package gov.nist.p25.issi.issiconfig;

import gov.nist.p25.issi.issiconfig.TopologyConfigParser.MyResolver;
import gov.nist.p25.issi.p25body.serviceprofile.user.UserServiceProfile;
import gov.nist.p25.issi.transctlmgr.ptt.TimerValues;

import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.text.ParseException;
import java.util.Hashtable;
import java.util.Iterator;

import javax.xml.parsers.SAXParserFactory;

import org.apache.log4j.Logger;
import org.xml.sax.Attributes;
import org.xml.sax.EntityResolver;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.DefaultHandler;

/**
 * A parser for the global topology file.
 * 
 * @author M. Ranganathan
 *
 */
public class GlobalTopologyParser extends DefaultHandler implements
		XMLTagsAndAttributes {
	private static Logger logger = Logger
			.getLogger("gov.nist.p25.issi.issiconfig");

	
	/**
	 * The following are used for parsing this structure.
	 */
	private boolean inuserProfile;

	private StringBuffer cdatabuffer;

	private SuConfig currentSuConfig;

	private boolean inGropProfile;

	/**
	 * File name from which this structure is generated.
	 */
	private String configFile;

	private GroupConfig currentGroupConfig;

	private TopologyConfig topologyConfig;

	private XMLReader saxParser;

	private boolean interactive;
	

	
	private static String LOOPBACK_ADDRESS = "127.0.0.1";
	
	
	
	// /////////////////////////////////////////////////////////////////////////////////////////
	// Entity resolver -- allows you to work offline
	// ////////////////////////////////////////////////////////////////////////////////////////
	class MyResolver implements EntityResolver {

		public InputSource resolveEntity(String publicId, String systemId)
				throws SAXException, IOException {
			if (systemId
					.equals("http://www-x.antd.nist.gov:8080/p25/issi-emulator/dtd/globaltopology.dtd")) {
				FileReader fileReader = new FileReader(new File(
						"dtd/globaltopology.dtd"));
				InputSource inputSource = new InputSource(fileReader);
				return inputSource;

			} else {
				return null;
			}
		}

	}

	// /////////////////////////////////////////////////////////////////////////////////////////
	// Constructor.
	// ////////////////////////////////////////////////////////////////////////////////////////
	/**
	 * Constructor. Takes a file name and constructs itself from that file.
	 * 
	 * @param interactive -- the flag that indicates whether to pick the real or emulated addresses.
	 * 
	 * @param fileName --
	 *            xml file from which the topology config constructs itself.
	 * 
	 */
	public GlobalTopologyParser(boolean interactive) throws Exception {

		try {

			this.interactive = interactive;
			SAXParserFactory saxParserFactory = SAXParserFactory.newInstance();
			this.saxParser = saxParserFactory.newSAXParser().getXMLReader();
			this.saxParser.setEntityResolver(new MyResolver());
			
			

			saxParser.setContentHandler(this);

			saxParser
					.setFeature("http://xml.org/sax/features/validation", true);

		} catch (SAXParseException spe) {
			// Error generated by the parser
			logger.error("\n** Parsing error" + ", line "
					+ spe.getLineNumber() + ", uri " + spe.getSystemId());

			logger.error("   " + spe.getMessage());
			// Use the contained exception, if any
			Exception x = spe;
			if (spe.getException() != null)
				x = spe.getException();
			x.printStackTrace();

			throw spe;
		} catch (SAXException sxe) {
			// Error generated by this application
			// (or a parser-initialization error)
			Exception x = sxe;
			if (sxe.getException() != null)
				x = sxe.getException();
			x.printStackTrace();
			throw sxe;

		} catch (Exception pce) {
			// Parser with specified options can't be built
			pce.printStackTrace();
			throw pce;
		}

	}

	// /////////////////////////////////////////////////////////////////////////////////
	// SAX Parsing methods that generate this structure from a XML File.
	// /////////////////////////////////////////////////////////////////////////////////

	public TopologyConfig parse(TopologyConfig systemTopology, String configFileName) throws IOException,
			SAXException {
		this.topologyConfig = systemTopology;
		topologyConfig.configFile = configFileName;
		this.configFile = configFileName;
		logger.debug("Parsing " + configFileName);
		saxParser.parse(new File(configFileName).toURI().toURL().toString());
		
		if ( ! this.interactive ) {
			for (RfssConfig rfssConfig : systemTopology.getRfssConfigurations()) {
				rfssConfig.ipAddress = LOOPBACK_ADDRESS;
				
				
			}
		}
		return this.topologyConfig;

	}

	@Override
	public void characters(char[] ch, int start, int length) {
		if (this.inuserProfile)
			this.cdatabuffer.append(ch, start, length);
		else if (this.inGropProfile)
			this.cdatabuffer.append(ch, start, length);

	}

	@Override
	public void startElement(String namespaceURI, String local, String name,
			Attributes attrs) throws SAXException {
		try {

			if (name.compareToIgnoreCase(GLOBAL_TOPOLOGY_CONFIG) == 0) {
				

			} else if (name.compareToIgnoreCase(RFSSCONFIG) == 0) {
				if (attrs.getValue(RFSS_NAME) == null) {
					throw new SAXException("Missing a required attribute "
							+ RFSS_NAME);
				}
				String rfssName = attrs.getValue(RFSS_NAME);
				RfssConfig rfssConfig = topologyConfig.getRfssConfigByName(rfssName);
				String selfTestPortString = attrs.getValue(LOOPBACK_PORT);
				if ( selfTestPortString == null ) throw new Exception("Missing a required attribute " + LOOPBACK_PORT);
				int selfTestPort = Integer.parseInt(selfTestPortString);
				
				
				rfssConfig.selfTestPort = selfTestPort;
				
				if ( ! interactive )  {
					String ipAddress = LOOPBACK_ADDRESS;
					
					rfssConfig.ipAddress = ipAddress;
					rfssConfig.sipPort = selfTestPort;
				
				}
				
				
			} else if (name.compareToIgnoreCase("suconfig") == 0) {
				if (attrs.getValue(SU_NAME) == null)
					throw new SAXException("Missing a required attribute "
							+ SU_NAME);
				String suName = attrs.getValue(SU_NAME);

				if (attrs.getValue(SU_ID) == null)
					throw new SAXException("Missing a required attribute "
							+ SU_ID);

				int suId = Integer.parseInt(attrs.getValue(SU_ID), 16);

				RfssConfig homeRfssConfig = topologyConfig
						.getRfssConfigByName(attrs.getValue(HOME_RFSS_NAME));

				if (homeRfssConfig == null)
					throw new SAXException(
							"Missing system configuration for home RFSS  = "
									+ attrs.getValue(HOME_RFSS_NAME));

				String rfssId = attrs.getValue(SERVING_RFSS_NAME);
				RfssConfig servingRfssConfig = topologyConfig
						.getRfssConfigByName(rfssId);
				if (servingRfssConfig == null)
					throw new SAXException(
							"missing configuration for rfssId = " + rfssId);

				SuState initialSuState = SuState.OFF;
				
				int homeRfssKnowsMeFor = -1;

				int inviteProcessingDelay = 0;

				int homeRfssRegistersMeFor = -1;

				boolean registerOnForce = true; 
				int servingRfssReferencesMeFor =  -1; 

				boolean isEmulated = servingRfssConfig.isEmulated() || !this.interactive ? true : false;
				
				boolean availability = true;
					
				SuConfig suConfig = new SuConfig(suId, suName,
						this.topologyConfig, homeRfssConfig, servingRfssConfig,
						initialSuState, homeRfssKnowsMeFor,
						servingRfssReferencesMeFor, inviteProcessingDelay,
						homeRfssRegistersMeFor, registerOnForce, 
						isEmulated, availability);

				this.currentSuConfig = suConfig;

				String radicalName = suConfig.getRadicalName();
				if (topologyConfig.suConfigTable.containsKey(radicalName))
					throw new SAXException("Duplicate SU" + radicalName);

				logger.debug("putting " + suConfig.getRadicalName()
						+ " into suConfig table");

				if (topologyConfig.suConfigTable.containsKey(suConfig
						.getRadicalName()))
					throw new SAXException("Duplicate SU detected. Check "
							+ this.configFile + " SU Name  = " + suName);

				topologyConfig.suConfigTable.put(suConfig.getRadicalName(),
						suConfig);

				this.topologyConfig.suNameTable.put(suName, suConfig);

			
		
			} else if (name.compareToIgnoreCase("sgconfig") == 0) {
				String groupName = attrs.getValue(GROUP_NAME);
				int sgid = Integer.parseInt(attrs.getValue(GROUP_ID), 16);

				String rfssId = attrs.getValue(HOME_RFSS_NAME);
				if (rfssId == null)
					throw new SAXException("Missing a required attribute "
							+ HOME_RFSS_NAME);

				RfssConfig homeRfssConfig = (RfssConfig) topologyConfig
						.getRfssConfigByName(rfssId);
				if (homeRfssConfig == null)
					throw new SAXException(
							"missing configuration for rfssId = " + rfssId);

				int homeRfssKnowsMeFor = attrs
						.getValue(HOME_RFSS_KNOWS_ABOUT_ME_FOR) == null
						|| attrs.getValue(HOME_RFSS_KNOWS_ABOUT_ME_FOR).equals(
								UNBOUNDED) ? -1 : Integer.parseInt(attrs
						.getValue(HOME_RFSS_KNOWS_ABOUT_ME_FOR));
				GroupConfig groupConfig = new GroupConfig(groupName, sgid,
						homeRfssConfig, homeRfssKnowsMeFor);

				String radicalName = groupConfig.getRadicalName();
				logger.debug("Adding group to table : name = " + radicalName);

				if (topologyConfig.groupTable.containsKey(radicalName))
					throw new SAXException(" Duplicate group detected in "
							+ this.configFile + " Check Group with name "
							+ groupName);

				topologyConfig.groupTable.put(radicalName, groupConfig);
				this.currentGroupConfig = groupConfig;
				topologyConfig.groupNameTable.put(groupName, groupConfig);

			} else if (name.compareToIgnoreCase("subscriber") == 0) {
				if (attrs.getValue(SU_NAME) == null)
					throw new SAXException("Missing a required attribute "
							+ SU_NAME);
				SuConfig suConfig = topologyConfig.getSuConfigByName(attrs
						.getValue(SU_NAME));

				if (suConfig == null)
					throw new SAXException("Could not find an rfss for "
							+ attrs.getValue(SU_NAME));
				else {
					this.currentGroupConfig.addSubscriber(suConfig);
					suConfig.addGroup(this.currentGroupConfig);
					suConfig.addSubscribedTo(this.currentGroupConfig);
				}

			}
		} catch (SAXException ex1) {
			throw ex1;
		} catch (Exception ex) {
			logger.error("Error parsing configuration file " + configFile, ex);
			throw new SAXException("Error parsing config file", ex);
		}

	}

	@Override
	public void endElement(String uri, String name, String qName)
			throws SAXException {

		 if ("suconfig".equalsIgnoreCase(qName)) {
			this.currentSuConfig = null;
		} else if ("sgconfig".equalsIgnoreCase(qName)) {
			this.currentGroupConfig = null;
		}

	}
}
