package gov.nist.p25.issi.issiconfig;

import gov.nist.p25.issi.p25body.serviceprofile.group.GroupServiceProfile;
import gov.nist.p25.issi.p25body.serviceprofile.user.UserServiceProfile;
import gov.nist.p25.issi.rfss.GroupHome;
import gov.nist.p25.issi.utils.ReadOnlyIterator;

import java.io.Serializable;
import java.text.ParseException;
import java.util.HashSet;
import java.util.Iterator;

/**
 * A configuration class for ISSI groups. A group has a number of subscriber
 * RFSS's. A group has a home RFSS that keeps track of its state and who is
 * subscribed to it. This class is constructed initially by reading the xml
 * configuration file for the group. Its function is to keep track of the
 * current group membership as well as the static information that is read from
 * the system configuration file. The current membership information is stored
 * at the home RFSS for the group.
 * 
 * @author M. Ranganathan
 * 
 */
public class GroupConfig implements Serializable {

	private String groupName; // The unique name that maps one to one with the
								// ID.

	private RfssConfig homeRfss;

	private SystemConfig sysConfig;

	private int groupId;

	private HashSet<SuConfig> subscribers;

	private GroupServiceProfile groupServiceProfile;

	private HashSet<RfssConfig> registerForbiddenFromRfss;

	private HashSet<RfssConfig> queryForbiddenFromRfss;

	// private boolean isKnownGroup;

	int homeRfssKnowsAboutMeFor;

	private CProtectedDisposition cProtectedDisposition;

	GroupConfig(String groupName, int sgId, RfssConfig homeRfss,
			int homeRfssKnowsAboutMeFor) {
		if (sgId > 0xffff)
			throw new IllegalArgumentException("Gid outside range");
		this.groupName = groupName;
		this.subscribers = new HashSet<SuConfig>();
		this.homeRfss = homeRfss;
		this.sysConfig = homeRfss.getSysConfig();
		this.registerForbiddenFromRfss = new HashSet<RfssConfig>();
		this.queryForbiddenFromRfss = new HashSet<RfssConfig>();
		if (sgId > 0xffff)
			throw new IllegalArgumentException(
					"SgId Outside allowable range : " + sgId);
		this.groupId = sgId;
		this.homeRfssKnowsAboutMeFor = homeRfssKnowsAboutMeFor;
		this.groupServiceProfile = new GroupServiceProfile();
		this.cProtectedDisposition = CProtectedDisposition.FORWARD_PROTECTED;

	}

	/**
	 * Get radical name for SU given its id and system id according to 3.4.1.2
	 * 
	 * @param sgId --
	 *            group id.
	 * @param systemConfig --
	 *            system config record.
	 */

	public static String getRadicalName(SystemConfig sysConfig, int sgId) {
		String wacnIdName = Integer.toHexString(sysConfig.getWacnId());

		// prepend 0's to the right length.
		int k = 5 - wacnIdName.length();

		for (int i = 0; i < k; i++) {
			wacnIdName = "0" + wacnIdName;
		}

		String systemIdName = Integer.toHexString(sysConfig.systemId);
		k = 3 - systemIdName.length();
		for (int i = 0; i < k; i++) {
			systemIdName = "0" + systemIdName;
		}
		String uidName = Integer.toHexString(sgId);
		k = 4 - uidName.length();
		for (int i = 0; i < k; i++) {
			uidName = "0" + uidName;
		}

		String radicalName = wacnIdName + systemIdName + uidName;

		return radicalName.toLowerCase();
	}

	/**
	 * @return Returns the homeRfss.
	 */
	public RfssConfig getHomeRfss() {
		return homeRfss;
	}

	/**
	 * @return Returns the sysConfig.
	 */
	public SystemConfig getSysConfig() {
		return sysConfig;
	}

	/**
	 * @return the group name
	 */
	public String getGroupName() {
		return this.groupName;
	}

	/**
	 * @return Returns the sgId.
	 */
	public int getGroupId() {
		return groupId;
	}

	public String getRadicalName() {
		return getRadicalName(this.sysConfig, this.groupId);
	}

	/**
	 * Add a SU that subscribes to the given group.
	 * 
	 * @param suConfig --
	 *            the Su to add.
	 */
	public void addSubscriber(SuConfig suConfig) {
		this.subscribers.add(suConfig);
	}

	/**
	 * Get the subscribers to the group.
	 * 
	 */
	public Iterator<SuConfig> getSubscribers() {
		return new ReadOnlyIterator<SuConfig>(this.subscribers.iterator());

	}

	/**
	 * Setter for the group service profile.
	 * 
	 * @param groupServiceProfile
	 */
	public void setGroupServiceProfie(GroupServiceProfile groupServiceProfile) {
		this.groupServiceProfile = groupServiceProfile;
	}

	/**
	 * Set the group service profile for this group
	 * 
	 * @param sgProfile;
	 */

	public void setGroupServiceProfile(String sgProfile) throws ParseException {
		this.groupServiceProfile = GroupServiceProfile
				.createGroupServiceProfile(sgProfile);

	}

	/**
	 * Get the group service profile.
	 */
	public GroupServiceProfile getGroupServiceProfile() {
		return this.groupServiceProfile;
	}

	/**
	 * Add a forbidden Rfss ( registrations from here are not accepted when a Su
	 * roams. )
	 * 
	 * @param rfssConfig
	 */
	public void addForbiddenRfss(RfssConfig rfssConfig) {
		this.registerForbiddenFromRfss.add(rfssConfig);

	}

	/**
	 * Return true if an SU is forbidden from registering from a given RFSS.
	 * 
	 * @param rfssConfig
	 * @return
	 */
	public boolean isRegisterForbiddenFromRfss(RfssConfig rfssConfig) {
		return this.registerForbiddenFromRfss.contains(rfssConfig);
	}

	/**
	 * Return true if a register query is forbidden from the given rfss.
	 * 
	 * @param rfssConfig
	 * @return
	 */
	public boolean isQueryForbiddenFromRfss(RfssConfig rfssConfig) {
		return this.queryForbiddenFromRfss.contains(rfssConfig);
	}

	/**
	 * The amount of time for which the Home RFSS keeps record of this group.
	 * 
	 * @return the amount of time that the home rfss keeps account of this
	 *         group.
	 */

	public int getHomeRfssKnowsAboutMeFor() {
		return this.homeRfssKnowsAboutMeFor;
	}

	/**
	 * Add a rfss config to the "query forbidden from" RFSS set.
	 * 
	 * @param rfssConfig --
	 *            rfss from which query is forbidden.
	 */
	public void addRegisterQueryForbiddenFromRfss(RfssConfig rfssConfig) {
		this.queryForbiddenFromRfss.add(rfssConfig);

	}

	public void setGroupId(int sgId) {
		if (sgId > 0xffff)
			throw new IllegalArgumentException(
					"SgId Outside allowable range : " + sgId);

		this.groupId = sgId;

	}

	public void setCProtectedDispsition(
			CProtectedDisposition forwardingDirective) {
		this.cProtectedDisposition = forwardingDirective;
	}

	public CProtectedDisposition getCProtectedDispsition() {
		return this.cProtectedDisposition;
	}

	public String toString() {
		StringBuffer sbuf = new StringBuffer();
		sbuf.append("<!-- Subscriber Group Definition -->\n");
		sbuf.append("<sgconfig\n");
		sbuf.append("\tgroupName = \"" + this.groupName + "\"\n");
		sbuf.append("\tsgId = \"" + Integer.toHexString(this.groupId) + "\"\n");
		sbuf.append("\thomeRfssName = \"" + this.getHomeRfss().getRfssName()
				+ "\"\n");
		sbuf.append("/>\n");
		return sbuf.toString();
	}

}