package gov.nist.p25.issi.issiconfig;

import gov.nist.p25.issi.rfss.RFSS;
import gov.nist.p25.issi.utils.ReadOnlyIterator;

import java.awt.Color;
import java.util.Collection;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.concurrent.ConcurrentHashMap;

import org.apache.log4j.Logger;

/**
 * This class represents the static configuration of an RFSS. This is
 * constructed by reading the configuraiton information of the emulator from the
 * XML topology configuration file.
 * 
 * @author M. Ranganathan
 */
public class RfssConfig  {
	
	private boolean advancedRtpResourceManagementSupported;
	
	private SystemConfig systemConfig;
	
	boolean pttRequestGranted;

	// Rfss id.
	private int rfssId;

	String ipAddress;

	// Port at which we get sip signaling.
	int sipPort;

	// The Rfss domain name.
	private String rfssDomainName;

	// Pointer to the runtime structure.
	private transient RFSS rfss;

	// Set this to true if the Rfss is emulated.
	boolean isEmulated;
	
	
	// List of SU for which this RFSS is home.
	private Hashtable<String, SuConfig> assignedSu= new Hashtable<String, SuConfig>() ;

	private ConcurrentHashMap<String, SuConfig> initialServedSubscriberUnits = new ConcurrentHashMap<String, SuConfig>();

	private Hashtable<String, GroupConfig> groups= new Hashtable<String, GroupConfig>();
	
	private HashSet<SuConfig> unknownSuTable = new HashSet<SuConfig>();


	// The symbolic name that we are using in the script.
	private String rfssName;
	
	// The time for which a RFSS is registered with us
	// Assuming we are the home rfss of a group.
	// after this time we will drop the registration.
	int    groupRegistrationExpiresTime;

	int servedGroupLeaseTime;
	
	int groupCallInviteProcessingTime;
	
	boolean rfResourcesAvailable;

    int maxRtpPorts;
		
	boolean suCredentialsQueriedBeforeRegister;

	boolean groupCredentialsQueriedBeforeRegister;

	int registerConfirmTime;
	
	boolean notInterestedInLosingAudio = true;
	
	Color color = Color.BLACK;

	private boolean isCallingSuInitialTransmitter;

	String tag;

	private TopologyConfig topologyConfig;

	private String colorString;

	int selfTestPort;

	private boolean checkSuPresenceOnRegister;
		

	private static Logger logger = Logger.getLogger(RfssConfig.class);
	
	
	/**
	 * Public constructor.
	 * 
	 * @param topologyConfig -- topology configuration.
	 * @param systemConfig  -- system configuration
	 * @param rfssId	-- RFSS ID
	 * @param systemConfig -- the system to which the Rfss belongs.
	 * @param topologyConfig -- the topology to which this Rfss belongs.
	 * @param isEmulated -- the emulated flag ( false means not emulated ).
	 * @param symbolicName -- the symbolic id for this RFSS.
	 * 
	 */
	public RfssConfig(TopologyConfig topologyConfig, SystemConfig systemConfig,
			int rfssId, String ipAddress, int port, String symbolicName, boolean isEmulated ) {
		
		if ( rfssId > 0xff ) throw new IllegalArgumentException("Bad RFSS ID  " + rfssId);
		
		this.ipAddress = ipAddress;
		this.sipPort = port;
		this.rfssName = symbolicName;
		this.systemConfig = systemConfig;
		this.topologyConfig = topologyConfig;
		this.isEmulated = isEmulated;
		this.rfssId = rfssId;
		systemConfig.setAssigned(true);
		
		// Default behaviors.
		this.groupRegistrationExpiresTime = -1;
		this.servedGroupLeaseTime = -1;
		this.groupCallInviteProcessingTime = 0;	
		this.rfResourcesAvailable = true;
		this.maxRtpPorts = -1;
		this.advancedRtpResourceManagementSupported = false;
		this.suCredentialsQueriedBeforeRegister = false;
		this.groupCredentialsQueriedBeforeRegister = false;
		this.pttRequestGranted = true;
		this.registerConfirmTime = 0;
		this.notInterestedInLosingAudio = true;
		this.isCallingSuInitialTransmitter = true;
		this.checkSuPresenceOnRegister = false;
		
	}
	

	

	public String toString() {
		StringBuffer retval =  new StringBuffer().append("<rfssconfig\n")
				.append("\trfssName = \"" + rfssName + "\"\n")
				.append("\trfssId = \"" + Integer.toHexString(rfssId) + "\"\n")
				.append("\tsystemName = \"" + systemConfig.getSymbolicName() + "\"\n");
		
		if ( ! this.isEmulated)
				retval.append("\tipAddress = \"" + ipAddress   + "\"\n") ;
				
		retval.append("\tport = \"" + sipPort + "\"\n")
				.append("\tcolor = \"" + this.colorString +"\"\n" )
				.append("\temulated = \"" + this.isEmulated +"\"\n")
				.append("/>\n");
		return retval
				.toString();
	}
	
	

	/**
	 * Get an iterator to the Su's for which this Rfss is the home subscriber
	 * unit.
	 * 
	 * @return
	 */
	public Iterator<SuConfig> getHomeSubsciberUnits() {
		return new ReadOnlyIterator<SuConfig>(this.assignedSu.values()
				.iterator());
	}

	/**
	 * Get an iterator to the groups assigned to the Rfss.
	 * 
	 * @return
	 */
	public Iterator<GroupConfig> getAssignedGroups() {
		return new ReadOnlyIterator<GroupConfig>(this.groups.values()
				.iterator());
	}

	/**
	 * Add a subscriber unit to the server set.
	 * 
	 * @param suConfig
	 */
	void addServedSubscriberUnit(SuConfig suConfig) {
		this.initialServedSubscriberUnits.put(suConfig.getRadicalName(),
				suConfig);
	}

	/**
	 * Add a subscriber unit to the home set.
	 * 
	 * @param suConfig
	 */
	void addAssignedSubscriberUnit(SuConfig suConfig) {
		String key = suConfig.getRadicalName();
		if (suConfig.getHomeRfssKnowsAboutMeFor() !=  0) {

			if (logger.isDebugEnabled()) {
				logger.debug("addAssignedSubsriberUnit : " + key);

			}
			assignedSu.put(key, suConfig);
		} else {
			if (logger.isDebugEnabled()) {
				logger
						.debug("SU is not a known SU -- not adding to the assigned Subscriber unit list");
			}
		}
	}

	/**
	 * Return a read-only iterator over the set of Subscriber units that are
	 * initially assigned to us at emulation startup.
	 * 
	 * @return -- iterator over subscriber units that are assigned to us.
	 * 
	 */
	public Collection<SuConfig> getInitialServedSubscriberUnits() {
		return this.initialServedSubscriberUnits
				.values();
	}

	public static String getDomainName(SystemConfig systemConfig, int rfssId) {
		return getDomainName(systemConfig.getWacnId(), systemConfig.getSystemId(), rfssId);
		
	}
	
	public static String getDomainName(int wacnId, int systemId, int rfssId)  {
		String systemName = SystemConfig.getSystemName(wacnId, systemId);
		String rfssIdString = Integer.toHexString(rfssId);
		int k = 2 - rfssIdString.length();
		for (int i = 0; i < k; i++) {
			rfssIdString = "0" + rfssIdString;
		}
		String rfssDomainName = rfssIdString + "." + systemName + "." + "p25dr";
		return rfssDomainName.toLowerCase();
	}

	public String getDomainName() {
		
		return  getDomainName(this.systemConfig, this.rfssId);
		
	}

	public String getIpAddress() {
		return this.ipAddress;
	}

	public int getSipPort() {
		return this.sipPort;
	}
	
	public void setSipPort(int port) {
		if (port < 0 ) throw new IllegalArgumentException("bad port value");
		this.sipPort = port;
	}

	public SuConfig getSubscriberUnit(int suId) {
		String key = SuConfig.getRadicalName(this.systemConfig, suId);
		logger.debug("RfssConfig: getSubscriberUnit() :" + key);
		return (SuConfig) this.assignedSu.get(key);
	}

	public SuConfig getSubscriberUnit(String radicalName) {
		return (SuConfig) this.assignedSu.get(radicalName);
	}

	/**
	 * Set a pointer to the dynamic emulation structure for RFSS.
	 * 
	 * @param rfss --
	 *            the dynamic emulation structure for an RFSS.
	 */
	public void setRFSS(RFSS rfss) {
		if (rfss == null)
			throw new NullPointerException("Null arg!");
		// if ( this.rfss != null ) throw new IllegalStateException("Already
		// Initialized");
		this.rfss = rfss;
	}

	/**
	 * Return the Runtime RFSS structure corresponding to this Configuration
	 * Structure.
	 * 
	 * @return
	 */
	public RFSS getRFSS() {
		return this.rfss;
	}

	public SystemConfig getSysConfig() {
		return this.systemConfig;
	}
	
	public String getSystemName() {
		return this.systemConfig.getSystemName();
	}
	
	
	public void setSystemName(String systemName)  {
		try {
		
		this.systemConfig = this.topologyConfig.getSystemConfigByName(systemName);
		} catch (Exception ex) {
			throw new IllegalArgumentException("System not found " + systemName,ex);
		}
	}
	
	

	public String getHostPort() {
		return this.ipAddress + ":" + sipPort;

	}

	public int getRfssId() {
		return this.rfssId;
	}

	public int getWacnId() {

		return this.systemConfig.getWacnId();
	}

	public int getSystemId() {
		return this.systemConfig.getSystemId();
	}

	/**
	 * Add a group for which I am the home RFSS.
	 * 
	 * @param groupConfig
	 */
	public void addAssignedGroup(GroupConfig groupConfig) {

		if (groupConfig.getHomeRfssKnowsAboutMeFor() != 0) {
			this.groups.put(groupConfig.getRadicalName(), groupConfig);
		}

	}

	/**
	 * Return true if a group is assigned to me.
	 * 
	 * @param groupId --
	 *            the group id.
	 * 
	 * @return true if a given group (identified by the group id ) is assigned
	 *         to me.
	 * 
	 */
	public boolean isAssignedGroup(String groupId) {

		return this.groups.containsKey(groupId);
	}

	/**
	 * Get a group from our assigned groups.
	 * 
	 * @param groupId
	 * @return -- the group configuration.
	 * 
	 */
	public GroupConfig getAssignedGroup(String groupId) {

		return this.groups.get(groupId);
	}



	public String getConsoleLog() {
		return "rfss" + this.getRfssId() + "consolelog.txt";
	}

	

	

	public boolean isSubscriberUnitMine(SuConfig suConfig) {
		return this.assignedSu.contains(suConfig);
	}

	/**
	 * Return true if a group belongs to us.
	 * 
	 * @param groupConfig
	 * @return
	 */
	public boolean isGroupMine(GroupConfig groupConfig) {
		return this.groups.containsKey(groupConfig.getRadicalName());
	}

	/**
	 * Return true if this RFSS is emulated.
	 * 
	 * @return
	 */
	public boolean isEmulated() {
		return this.isEmulated;
	}

	
	
	public void setEmulated(boolean emulatedFlag) {
		this.isEmulated = emulatedFlag;
	}

	public String getRfssName() {
		return rfssName;
	}



	/**
	 * @return Returns the groupRegistrationLeaseTime.
	 */
	public int getGroupRegistrationExpiresTime() {
		return groupRegistrationExpiresTime;
	}

	void setGroupRegistrationLeaseTime ( int leaseTime ) {
		this.groupRegistrationExpiresTime = leaseTime;
	}

	/**
	 * @return Returns the servedGroupLeaseTime.
	 */
	public int getServedGroupLeaseTime() {
		return servedGroupLeaseTime;
	}

	
   void setServedGroupLeaseTime ( int leaseTime ) {
	   this.servedGroupLeaseTime = leaseTime;
   }
	

	public int getGroupCallInviteProcessingTime() {
	
		return this.groupCallInviteProcessingTime;
	}
	
	void setGroupCallInviteProcessingTime(int processingTime ) {
		this.groupCallInviteProcessingTime = processingTime;
	}

	/**
	 * @param rfResourcesAvailable The rfResourcesAvailable to set.
	 */
	void setRfResourcesAvailable(boolean rfResourcesAvailable) {
		this.rfResourcesAvailable = rfResourcesAvailable;
	}

	/**
	 * @return Returns the rfResourcesAvailable.
	 */
	public boolean isRfResourcesAvailable() {
		return rfResourcesAvailable;
	}

	/**
	 * Limit on the RTP resources available on this machine.
	 * 
	 * @return
	 */
	public int  getMaxRtpPorts() {
			return this.maxRtpPorts;
	}

	void setMaxRtpPorts( int maxPort) {
		this.maxRtpPorts = maxPort;
	}
	

	/**
	 * @return the queryUnknownSu
	 */
	public boolean isSuCredentialsQueriedBeforeRegister() {
		return suCredentialsQueriedBeforeRegister;
	}

	
	/**
	 * @return the queryGroupCredentials
	 */
	public boolean isGroupCredentialsQueriedBeforeRegister() {
		return groupCredentialsQueriedBeforeRegister;
	}
	
	void setQueryGroupCredentials ( boolean groupCredentials) {
		this.groupCredentialsQueriedBeforeRegister = groupCredentials;
	}

	/**
	 * Default register confirm processing time (seconds).
	 * 
	 * @return
	 */
	public int getRegisterConfirmTime() {
		return this.registerConfirmTime;
	}

	void setRegisterConfirmTime( int registerConfirmTime) {
		this.registerConfirmTime = registerConfirmTime;
	}
	
	/**
	 * @return the advancedRtpResourceManagementSupported
	 */
	public boolean isAdvancedRtpResourceManagementSupported() {
		return advancedRtpResourceManagementSupported;
	}
	
	
    void setAdvancedRtpResourceManagementSupported( boolean advancedRtpResourceManagementSupported ) {
		this.advancedRtpResourceManagementSupported = advancedRtpResourceManagementSupported;
	}

	

	/**
	 * @return the pttRequestGranted
	 */
	public boolean isPttRequestGranted() {
		return pttRequestGranted;
	}

	public void addUnknownSu(SuConfig suConfig) {
		this.unknownSuTable.add(suConfig);
		
	}
	
	public boolean isSuUnknown(SuConfig suConfig) {
		return this.unknownSuTable.contains(suConfig);
	}

	public boolean isUnitKnown(int unitId) {
		for (SuConfig suConfig: this.unknownSuTable) {
			if (suConfig.getSuId() == unitId) return false;
		}
		return true;
	}
	
	void setColor ( String  color) {
		if ( color == null ) color  = "BLACK";
		if ( color.equals("BLACK")) {
			this.color = Color.BLACK;
		} else if ( color.equals("BLUE")) {
			this.color = Color.BLUE;
		} else if ( color.equals("MAGENTA")) {
			this.color = Color.MAGENTA;
		} else if ( color.equals("CYAN")) {
			this.color = Color.CYAN;
		} else if ( color.equals("GREEN")) {
			this.color = Color.GREEN;
		} else if (color.equals("RED")) {
			this.color = Color.RED;
		} 		
		this.colorString = color;
	}
	
	public Color getColor() {
		return this.color;
	}

	public boolean isNotInterestedInLosingAudio() {
		
		return this.notInterestedInLosingAudio;
	}

	public boolean isCallingSuInitialTransmitter() {
		
		return this.isCallingSuInitialTransmitter;
	}

	public void setCallingSuInitialTransmitter(boolean isCallingSuInitialTransmitter) {
		this.isCallingSuInitialTransmitter = isCallingSuInitialTransmitter;
	}
	
	static String generateTag(String name) {
		return  name  + ".daemon";
	}

	/**
	 * The tag to assign to the emulated DIETS daemons. This defines
	 * the edge between emulated RFSSs and the DIETS daemons where there
	 * are assigned to run.
	 * 
	 * @return
	 */
	public String getTag() {
		
		return generateTag(this.rfssName);
	}

	public void setIpAddress(String ipAddress) {
		this.ipAddress = ipAddress;
		
	}

	public void setTag(String newTag) {
		this.tag = newTag;
		
	}

	public int getSelfTestPort() {
	
		return this.selfTestPort;
	}

	public void setRfssId(int rfssId) {
		if ( rfssId > 0xff)
			throw new IllegalArgumentException("RFSS ID out of range should be <= 0xff was 0x" + Integer.toHexString(rfssId) );
		this.rfssId = rfssId;
		
	}




	




	





	

}
