package gov.nist.p25.issi.issiconfig;

import gov.nist.p25.issi.p25body.serviceprofile.user.UserServiceProfile;
import gov.nist.p25.issi.rfss.SuInterface;

import java.text.ParseException;
import java.util.HashSet;

import org.apache.log4j.Logger;

/**
 * This is the main test driver. This is the guy who emulates the SU. Eventually
 * this class will implement both SipListener and RTPListener.
 * 
 * @author M. Ranganathan
 * 
 */
public class SuConfig {
	
	private static Logger logger = Logger.getLogger(SuConfig.class);
	

	private TopologyConfig config;

	private SystemConfig sysConfig;

	private RfssConfig homeRfss;

	private RfssConfig initalServingRfss;
	
	private boolean isEmulated;
	
	
	// The user service Profile for the SU.
	private UserServiceProfile userServiceProfile;

	
	// The set of RFSSs that this SU is forbidden from 
	// registering with.
	private HashSet<RfssConfig> registrationForbiddenSet;
	
	private HashSet <RfssConfig> registerQueryForbiddenSet;

	// The set of groups for which the SU is a member
	private HashSet<GroupConfig> groups;

	// whether the SU is initially on or off.
    SuState initialState;
	
	int inviteProcessingDelay;

	/**
	 * These are specific to the SU.
	 */
	private int suId;
	
	// Temporary storage for radical name so we dont need to keep recomputing
	// it.

	private String radicalName;

	// The SU for this guy.
	private SuInterface su;
	
	// Time for which the Home keeps a record of this SU.
    int homeRfssKnowsAboutMeFor;
	
	// Time for which the Home Mobility manager keeps a registration record of this SU
    int homeRfssRegistersMeFor;

	private HashSet<GroupConfig> subscribedGroups;

	boolean registerOnForce;
	
	private String symbolicName;

	int servingRfssReferencesMeFor;
	
	boolean available;
	
	private boolean checkSuPresenceOnRegister  = true;
	
	private CProtectedDisposition cProtectedDisposition;
	
	SuConfig(int suId, String suName , TopologyConfig config, RfssConfig homeRfss,
			RfssConfig servingRfss, SuState initialSuState, 
			int homeRfssKnowsMeFor, 
			int servingRfssReferencesMeFor,
			int inviteProcessingDelay,int homeRfssRegistersMeFor, boolean registerOnForce,boolean isEmulated,
			boolean availability) {
		this.groups = new HashSet<GroupConfig>();
		this.config = config;
		this.homeRfss = homeRfss;
		this.initalServingRfss = servingRfss;
		this.sysConfig = homeRfss.getSysConfig();
		this.subscribedGroups = new HashSet<GroupConfig>();
		this.registrationForbiddenSet = new HashSet<RfssConfig>();
		this.registerQueryForbiddenSet = new HashSet<RfssConfig>();
		if (suId > 0xffffff)
			throw new IllegalArgumentException(
					"SuId Outside allowable range : " + suId);
		this.suId = suId;
		this.initialState = initialSuState; // The initial state of the SU ( on
											// or off ).
		this.homeRfssKnowsAboutMeFor = homeRfssKnowsMeFor;
		this.inviteProcessingDelay = inviteProcessingDelay;
		this.homeRfssRegistersMeFor = homeRfssRegistersMeFor;
		this.registerOnForce = registerOnForce;
		this.userServiceProfile = new UserServiceProfile();
		this.symbolicName = suName;
		this.servingRfssReferencesMeFor = servingRfssReferencesMeFor;
		this.isEmulated = isEmulated;
		this.available = availability;
		
	}

	public void setSU(SuInterface su) {
		// Set a pointer to the Dynamic SU structure for
		// this su configuration.
		if (su == null)
			throw new NullPointerException("Null arg!");
		if (this.su != null)
			throw new IllegalStateException("Illegal state -- already set");
		this.su = su;
	}

	public SuInterface getSU() {
		return this.su;
	}

	public RfssConfig getHomeRfss() {
		return this.homeRfss;
	}

	public RfssConfig getInitialServingRfss() {
		return this.initalServingRfss;
	}

	/**
	 * Get radical name for SU given its id and system id according to 3.4.1.2
	 * 
	 * @param suID --
	 *            su id.
	 * @param systemConfig --
	 *            system config record.
	 */

	public static String getRadicalName(SystemConfig sysConfig, int suId) {
		String wacnIdName = Integer.toHexString(sysConfig.getWacnId());

		// prepend 0's to the right length.
		int k = 5 - wacnIdName.length();

		for (int i = 0; i < k; i++) {
			wacnIdName = "0" + wacnIdName;
		}

		String systemIdName = Integer.toHexString(sysConfig.systemId);
		k = 3 - systemIdName.length();
		for (int i = 0; i < k; i++) {
			systemIdName = "0" + systemIdName;
		}
		String uidName = Integer.toHexString(suId);
		k = 6 - uidName.length();
		for (int i = 0; i < k; i++) {
			uidName = "0" + uidName;
		}

		String radicalName = wacnIdName + systemIdName + uidName;

		return radicalName.toLowerCase();
	}

	/**
	 * Compute the radical name according to 3.4.1.2
	 * 
	 * @return the radical name.
	 */
	public String getRadicalName() {
		if (this.radicalName == null)
			this.radicalName = getRadicalName(this.sysConfig, this.suId);
		return this.radicalName;

	}

	public String toString() {

		StringBuffer sb = new StringBuffer();
		sb.append("<!-- Subscriber Unit (SU) definition -->\n");
		sb.append("<suconfig\n").append(
				"\tsuId = \"" + Integer.toHexString(this.suId) + "\"\n").append (
				"\tsuName = \"" + this.getSuName() + "\"\n").append(
				"\thomeRfssName = \"" + this.homeRfss.getRfssName() + "\"\n").append(
				"\tservingRfssName = \"" + initalServingRfss.getRfssName() + "\"\n");
		sb.append("/>\n");
		
		return sb.toString();

	}

	public int hashCode() {
		return this.getRadicalName().hashCode();
	}

	public boolean equals(Object that) {

		return (that != null && that instanceof SuConfig && this
				.getRadicalName().equals(((SuConfig) that).getRadicalName()));
	}

	public short getSuId() {

		return (short) this.suId;
	}

	public int getWacnId() {
		return this.sysConfig.getWacnId();
	}

	public short getSystemId() {
		return (short) this.sysConfig.systemId;
	}

	public void setUserServiceProfile(String buffer) throws ParseException {
		this.userServiceProfile = UserServiceProfile
				.createUserServiceProfile(buffer);

	}

	public UserServiceProfile getUserServiceProfile() {
		
		return this.userServiceProfile;
	}

	public void setServingRfss(RfssConfig rfssConfig) {
		if (this.initalServingRfss != null)
			throw new IllegalStateException("Already set.");
		this.initalServingRfss = rfssConfig;

	}

	public void addSubscribedTo(GroupConfig groupConfig) {
		this.subscribedGroups.add(groupConfig);

	}

	public HashSet<GroupConfig> getSubscribedGroups() {

		return this.subscribedGroups;
	}

	/**
	 * Add a forbidden RFSS. The SU is not allowed to register from a
	 * "forbidden" rfss.
	 * 
	 * @param rfssConfig --
	 *            the forbidden RFSS to add.
	 * 
	 */
	public void addForbiddenRfss(RfssConfig rfssConfig) {
		this.registrationForbiddenSet.add(rfssConfig);
		this.registerQueryForbiddenSet.add(rfssConfig);
	}

	/**
	 * add a forbidden for registration only RFSS.
	 */
	public void addRegisterForbiddenRfss(RfssConfig rfssConfig) {
		this.registrationForbiddenSet.add(rfssConfig);
	}
	
	/**
	 * Add a forbidden for query RFSS.
	 */
	public void addRegisterQueryForbiddenRfss(RfssConfig rfssConfig) {
		this.registerQueryForbiddenSet.add(rfssConfig);
	}
	/**
	 * Return true if a given Rfss is forbidden.
	 * 
	 * @param rfssConfig --
	 *            the rfss to check.
	 */
	public boolean isRegistrationForbiddenFrom(RfssConfig rfssConfig) {
		return this.registrationForbiddenSet.contains(rfssConfig);
	}
	
	public boolean isQueryForbiddenFrom(RfssConfig rfssConfig) {
		return this.registerQueryForbiddenSet.contains(rfssConfig);
		
	}

	/**
	 * Add a group to the set of groups to which this su is subscrbed.
	 * 
	 * @param groupConfig
	 */
	void addGroup(GroupConfig groupConfig) {
		this.groups.add(groupConfig);

	}

	public HashSet<GroupConfig> getGroups() {
		
		return this.groups;
	}

	public SuState getInitialSuState() {
		return this.initialState;
	}

	

	/**
	 * The time for which the Home RFSS has knowledge of me ( allows 
	 * you to generate anamolies for testing ).
	 * 
	 * @return
	 */
	public int getHomeRfssKnowsAboutMeFor() {
		return this.homeRfssKnowsAboutMeFor;
	}

	

	/**
	 * @return Returns the inviteProcessingDelay.
	 */
	public int getInviteProcessingDelay() {
		return inviteProcessingDelay;
	}

	/**
	 * The time for which registration records are held. 0 means the 
	 * unit is not registered.
	 * 
	 * @return
	 */
	public int getHomeRfssRegistersMeFor() {
		
		return this.homeRfssRegistersMeFor;
	}



	public boolean forceRegister() {
			return this.registerOnForce;
	}

	public String getSuName() {
		return this.symbolicName;
	}



	/**
	 * @return the servingRfssRefreencesMeFor
	 */
	public int getServingRfssReferencesMeFor() {
		return servingRfssReferencesMeFor;
	}

	

	/**
	 * @return the isEmulated
	 */
	public boolean isEmulated() {
		return isEmulated;
	}

	public void setUserServiceProfile(UserServiceProfile userServiceProfile) {
		this.userServiceProfile = userServiceProfile;
		
	}

	/**
	 * get the availability state. The availability state
	 * governs whether a "busy here" is returned or not.
	 * 
	 * @return
	 */
	public boolean isAvailable() {
		return available;
	}

	
	/**
	 * Set the SU ID.
	 * 
	 * @param suId
	 */
	public void setSuId(int suId) {
		if (suId > 0xffffff)
			throw new IllegalArgumentException(
					"SuId Outside allowable range : " + suId);
	
		this.suId = suId;
		
	}

	public void setCProtectedDisposition(CProtectedDisposition cProtectedDisposition) {
		this.cProtectedDisposition = cProtectedDisposition;
	}

	public CProtectedDisposition getCProtectedDisposition() {
		return cProtectedDisposition;
	}

	
	/**
	 * @return the checkSuPresenceOnRegister
	 */
	public boolean isCheckSuPresenceOnRegister() {
		return checkSuPresenceOnRegister;
	}
	

	/**
	 * @param checkSuPresenceOnRegister the checkSuPresenceOnRegister to set
	 */
	public void setCheckSuPresenceOnRegister(boolean checkSuPresenceOnRegister) {
		this.checkSuPresenceOnRegister = checkSuPresenceOnRegister;
	}

	
	
}
