package gov.nist.p25.issi.issiconfig;

import gov.nist.p25.issi.rfss.tester.ISSITesterConfiguration;
import gov.nist.p25.issi.rfss.tester.TestScript;
import gov.nist.p25.issi.utils.ReadOnlyIterator;

import java.io.File;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.util.Collection;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.List;

import org.apache.log4j.Logger;

/**
 * This class is responsible for reading the topology of the test system. We
 * assume that the entire topology is known at the start of the test i.e. the
 * assignments of SUs to Home and Serving RFSSs etc. Note that the ISSI
 * specification does not specify how these mappings will be exported to the
 * world so we have invented our own mapping scheme -- i.e. a simple XML
 * document.
 * 
 * @author M. Ranganathan
 * 
 */
public class TopologyConfig implements XMLTagsAndAttributes,
		java.io.Serializable {

	private static Logger logger = Logger
			.getLogger("gov.nist.p25.issi.issiconfig");

	/**
	 * A table of SU configurations.
	 */
	Hashtable<String, SuConfig> suConfigTable;

	/**
	 * A table of RFSS configuration records indexed by radical name
	 */
	private Hashtable<String, RfssConfig> rfssConfigTable;

	/**
	 * A table of system confiuration records.
	 */
	private Hashtable<String, SystemConfig> systemTable;

	/**
	 * A table of group configuration records.
	 */
	Hashtable<String, GroupConfig> groupTable;

	/**
	 * File name from which this structure is generated.
	 */
	String configFile;

	/**
	 * Group Name to group ID mapping.
	 */

	Hashtable<String, GroupConfig> groupNameTable;

	/**
	 * SU Name to SU ID mapping
	 * 
	 */
	Hashtable<String, SuConfig> suNameTable;

	/**
	 * WACN Name to ID mapping
	 * 
	 */
	Hashtable<String, WacnConfig> wacnNameTable;

	/**
	 * RFSS Name to ID mapping
	 * 
	 */
	private Hashtable<String, RfssConfig> rfssNameTable;

	/**
	 * System Name to ID mapping
	 * 
	 */
	private Hashtable<String, SystemConfig> systemNameTable;
	
	/**
	 * List of names of the RFSS that appear in the trace 
	 * if such information is available.
	 */
	List<String> traceOrder;

	// /////////////////////////////////////////////////////////////////////////////
	// Constructor. (package local)
	// //////////////////////////////////////////////////////////////////////////////
	public TopologyConfig() {
		rfssConfigTable = new Hashtable<String, RfssConfig>();
		systemTable = new Hashtable<String, SystemConfig>();
		suConfigTable = new Hashtable<String, SuConfig>();
		groupTable = new Hashtable<String, GroupConfig>();
		groupNameTable = new Hashtable<String, GroupConfig>();
		wacnNameTable = new Hashtable<String, WacnConfig>();
		rfssNameTable = new Hashtable<String, RfssConfig>();
		systemNameTable = new Hashtable<String, SystemConfig>();
		suNameTable = new Hashtable<String, SuConfig>();
		traceOrder = new LinkedList<String>();

	}

	/**
	 * Print writer used to write out rtp records.
	 */
	private static PrintWriter rtpPrintWriter;

	private static File msgLogFile = null;

	private static FileWriter rtpMessageLogFileWriter;

	public synchronized static PrintWriter getRtpMessageLogStream() {
		try {
			if (rtpPrintWriter != null) {
				return rtpPrintWriter;
			} else if (TestScript.rtpMessagelog != null) {
				msgLogFile = new File(TestScript.rtpMessagelog);
				rtpMessageLogFileWriter = new FileWriter(msgLogFile);
				rtpPrintWriter = new PrintWriter(rtpMessageLogFileWriter);
				return rtpPrintWriter;
			} else
				return null;
		} catch (Exception ex) {
			return null;
		}
	}

	public static void closeRtpMessageLogStream() {
		try {
			if (rtpPrintWriter != null) {
				rtpMessageLogFileWriter.close();
				rtpPrintWriter.close();
				// msgLogFile.delete();
				rtpPrintWriter = null;
				rtpMessageLogFileWriter = null;
			}
		} catch (Exception ex) {
			logger.error("Unexpected exception", ex);
		}

	}
	
	public RfssConfig getRfssConfigByTag(String tag) { 
		for ( RfssConfig rfssConfig : this.getRfssConfigurations()) {
			if ( tag.equals(rfssConfig.getTag())) return rfssConfig;
			
		}
		return null;
	}
	
	/**
	 * Get the trace order.
	 * 
	 * @return a list of names of the trace ordering of this test.
	 */
	public List<String> getTraceOrder() {
		return this.traceOrder;
	}
	
	/**
	 * Set the trace order.
	 * 
	 * @param traceOrder -- trace order to set.
	 */
	public void setTraceOrder(List<String> traceOrder) {
		this.traceOrder = traceOrder;
	}

	/**
	 * Get the topology file name.
	 * 
	 * @return -- the topology file name.
	 */
	public String getTopologyFile() {
		return this.configFile;
	}

	/**
	 * Retrieve the system configuration given a WacnID and systemID
	 * 
	 * @param wacnId
	 * @param systemId
	 * @return
	 */
	public SystemConfig getSysConfig(int wacnId, int systemId) {
		String sysId = SystemConfig.getSystemName(wacnId, systemId);
		return (SystemConfig) this.systemTable.get(sysId);
	}

	/**
	 * Retrieve the SU config given wacnId, systemId and suId
	 * 
	 * @param wacnId
	 * @param systemId
	 * @param suId
	 * @return
	 */
	public SuConfig getSuConfig(int wacnId, int systemId, int suId) {
		SystemConfig sysConfig = getSysConfig(wacnId, systemId);
		String radicalName = SuConfig.getRadicalName(sysConfig, suId);
		return (SuConfig) this.suConfigTable.get(radicalName);
	}

	public RfssConfig getRfssConfig(int wacnId, int systemId, int rfssId) {
		String domainName = RfssConfig.getDomainName(wacnId, systemId, rfssId);
		return this.rfssConfigTable.get(domainName);
	}

	public Collection<RfssConfig> getRfssConfigurations() {
		return this.rfssConfigTable.values();
	}

	public RfssConfig getRfssConfig(String rfssDomainName) {

		logger.debug("getting Rfss Config " + rfssDomainName);
		return (RfssConfig) this.rfssConfigTable.get(rfssDomainName.toLowerCase());
	}

	public SuConfig getSuConfig(String suRadicalName) {
		return (SuConfig) this.suConfigTable.get(suRadicalName);
	}

	public Collection<SuConfig> getSuConfigurations() {
		return this.suConfigTable.values();
				
	}

	public RfssConfig getRfssConfig(String ipAddress, int port) {
		for (RfssConfig rfss : this.rfssConfigTable.values()) {
			if (rfss.getIpAddress().equals(ipAddress) && rfss.getSipPort() == port)
				return rfss;
		}

		logger.error("Could not find rfss corresponding to " + ipAddress
				+ " port = " + port);
		logger.debug("Here is what is in the table ");
		if (logger.isDebugEnabled()) {
			for (RfssConfig rfss : this.rfssConfigTable.values()) {
				logger.debug("stored ip address = " + rfss.getIpAddress()
						+ " stored port = " + rfss.getSipPort());
			}
		}
		return null;
	}

	public GroupConfig getGroupConfig(String groupRadicalName) {
		return this.groupTable.get(groupRadicalName);
	}

	public GroupConfig getGroupConfig(int wacnId, int systemId, int sgId) {
		logger.debug("getGroupConfig: wacnId = " + wacnId + "system Id = "
				+ systemId + "sgId = " + sgId);
		SystemConfig sysConfig = this.getSysConfig(wacnId, systemId);
		if (sysConfig == null)
			return null;
		String radicalName = GroupConfig.getRadicalName(sysConfig, sgId);
		logger.debug("radicalName = " + radicalName);
		GroupConfig retval = this.getGroupConfig(radicalName);
		return retval;
	}

	/**
	 * Get the RFSS ID given its name ( i.e. the symbolic name in the test ).
	 * 
	 * @param key
	 * @return
	 * @throws Exception
	 *             if the id is not found.
	 * 
	 */
	public int getRfssIdByName(String key) throws Exception {
		if (!rfssNameTable.containsKey(key))
			throw new Exception("Could not getRfss Id for " + key);
		else
			return this.rfssNameTable.get(key).getRfssId();
	}

	/**
	 * Return the system ID given its symbolic name.
	 * 
	 * @param key --
	 *            the symbolic name.
	 * @return -- the id if it exists.
	 * @throws Exception
	 */
	public int getSystemIdByName(String key) throws Exception {
		if (!this.systemNameTable.containsKey(key))
			throw new Exception("Could not get system ID for " + key);
		else
			return this.systemNameTable.get(key).getSystemId();
	}

	/**
	 * Get the wacn ID given its symbolic name.
	 * 
	 * @param key --
	 *            the wacn symbolic name.
	 * 
	 * @return -- the wacn id.
	 * @throws Exception
	 */

	public int getWacnIdByName(String key) throws Exception {
		if (!this.wacnNameTable.containsKey(key))
			throw new Exception("Could not find wacn ID for " + key);
		else
			return this.wacnNameTable.get(key).getWacnId();
	}
	
	/**
	 * Add a new WACN to the Topology
	 * 
	 * @param wacnConfig -- the wacn to add to the topology.
	 */
	public void addWacnConfig(WacnConfig wacnConfig) throws Exception {
		if (this.wacnNameTable.contains(wacnConfig) ) 
			throw new Exception("Already exists! " + wacnConfig.getWacnName());
		this.wacnNameTable.put(wacnConfig.getWacnName(), wacnConfig);
	}

	/**
	 * Get the SU ID given its symbolic name.
	 * 
	 * @param suName --
	 *            the symbolic name of the SU for which the SUId is desired.
	 * 
	 * @return -- the suId of the SU with that name
	 * 
	 * @throws Exception --
	 *             if no su by that name is configured.
	 * 
	 */
	public int getSuIdByName(String suName) throws Exception {
		if (!this.suNameTable.containsKey(suName))
			throw new Exception("Could not find SU ID for " + suName);
		else
			return this.suNameTable.get(suName).getSuId();
	}

	/**
	 * Get the configuration for a given RFSS given its symbolic name.
	 * 
	 * @param name --
	 *            the name for which to retrieve the configuration.
	 * 
	 * @return the configuration corresponding to the name.
	 * 
	 * @throws Exception --
	 *             if no rfss corresponding to the given symbolic name is
	 *             configured.
	 */
	public RfssConfig getRfssConfigByName(String name) throws Exception {
			return this.rfssNameTable.get(name);
	}
	
	/**
	 * Add an Rfss configuration.
	 * 
	 * @param rfssConfig -- rfss config to add.
	 */
	public void addRfssConfig(RfssConfig rfssConfig) {
		logger.debug("Adding rfss " + rfssConfig.getRfssName() + " domain name " + rfssConfig.getDomainName() );
		this.rfssNameTable.put(rfssConfig.getRfssName(), rfssConfig);
		this.rfssConfigTable.put(rfssConfig.getDomainName(), rfssConfig);
	}

	/**
	 * Get the configuration for a given SU given its name.
	 * 
	 * @param name --
	 *            the name of the SU for which the configuration is desired.
	 * 
	 * @return the Su configuration corresponding to the su name.
	 * 
	 * @throws Exception --
	 *             if no su by that name can be found
	 * 
	 */
	public SuConfig getSuConfigByName(String name) throws Exception {
		if (!suNameTable.containsKey(name))
			throw new Exception("SU configuration not found for " + name);
		return this.suNameTable.get(name);
	}

	/**
	 * Get the System configuration corresponding to a given System name.
	 * 
	 * @param name --
	 *            the name of the system for which the configuration is desired
	 * 
	 * @return -- the system configuration for the system named.
	 * 
	 * 
	 */
	public SystemConfig getSystemConfigByName(String name) {
			return this.systemNameTable.get(name);
	}
	
	/**
	 * Add a new system configuration node.
	 * 
	 * @param systemConfig -- configuration to add.
	 */
	public void addSystemConfiguration( SystemConfig systemConfig) {
		this.systemNameTable.put(systemConfig.getSymbolicName(), systemConfig);
		this.systemTable.put(systemConfig.getSystemName(), systemConfig);
	}

	/**
	 * Get the group configuraiton given its symbolic name.
	 * 
	 * @param name --
	 *            symbolic name of the group.
	 * 
	 * @return -- the group configuration for the group.
	 * 
	 */
	public GroupConfig getGroupConfigByName(String name)  {
		return this.groupNameTable.get(name);
	}

	/**
	 * Get all the groups that are configured for the system.
	 * 
	 * @return a collection of groups.
	 */
	public Collection<GroupConfig> getGroupConfigurations() {
		return this.groupTable.values();
	}

	/**
	 * Get the wacn configuration given its name.
	 * 
	 * @param wacnName
	 * @return
	 */
	public WacnConfig getWacnConfig(String wacnName) {
		return this.wacnNameTable.get(wacnName);
	}

	public Collection<WacnConfig> getWacnConfigurations() {
		return this.wacnNameTable.values();
	}

	public Collection<SystemConfig> getSystemConfigurations() {
		return this.systemNameTable.values();
	}
	
	public String exportSystemTopology() {
		String preamble = "<?xml version=\"1.0\" ?>" +
				"<!DOCTYPE system-topology SYSTEM \"http://www-x.antd.nist.gov:8080/p25/issi-emulator/dtd/systemtopology.dtd\">\n";
		StringBuffer sbuf = new StringBuffer();
		sbuf.append(preamble);
		sbuf.append("<system-topology> \n");
		for ( WacnConfig wacnConfig : this.getWacnConfigurations()) {
			sbuf.append(wacnConfig.toString());
		}
		for ( SystemConfig systemConfig : this.getSystemConfigurations() ) {
			sbuf.append(systemConfig.toString() );
		}
		for ( RfssConfig rfssConfig : this.getRfssConfigurations()) {
			sbuf.append(rfssConfig.toString() );
		}
		sbuf.append("</system-topology>\n");
		return sbuf.toString();
	}

	public String exportGlobalTopology() {
		
		String preamble = "<?xml version=\"1.0\" ?>\n" +
		"<!DOCTYPE global-topology SYSTEM \"http://www-x.antd.nist.gov:8080/p25/issi-emulator/dtd/globaltopology.dtd\">\n";
		StringBuffer sbuf = new StringBuffer();
		sbuf.append(preamble);
		sbuf.append("<global-topology> \n");
	
	
		
		for ( RfssConfig rfssConfig : this.getRfssConfigurations()) {
			sbuf.append("<rfssconfig\n");
			sbuf.append("\trfssName = \"" + rfssConfig.getRfssName() + "\"\n" );
			sbuf.append("\tselftestPort =\"" + rfssConfig.getSelfTestPort() +  "\"\n" );
			sbuf.append("/>\n");
		}
		
		for ( SuConfig suConfig : this.getSuConfigurations()) {
			sbuf.append(suConfig.toString() );
		}
		
		for (GroupConfig sgConfig : this.getGroupConfigurations()) {
			sbuf.append(sgConfig.toString());
		}
		
		sbuf.append("</global-topology>\n");
		return sbuf.toString();
	}

	/**
	 * @param systemNameTable the systemNameTable to set
	 */
	public void setSystemNameTable(Hashtable<String, SystemConfig> systemNameTable) {
		this.systemNameTable = systemNameTable;
	}

	/**
	 * @return the systemNameTable
	 */
	public Hashtable<String, SystemConfig> getSystemNameTable() {
		return systemNameTable;
	}

	/**
	 * @param systemTable the systemTable to set
	 */
	public void setSystemTable(Hashtable<String, SystemConfig> systemTable) {
		this.systemTable = systemTable;
	}

	/**
	 * @return the systemTable
	 */
	public Hashtable<String, SystemConfig> getSystemTable() {
		return systemTable;
	}

	/**
	 * Add a reference alias in the RFSS name table.
	 * 
	 * @param id
	 * @param rfssConfig
	 */
	public void addRfssAlias(String id, RfssConfig rfssConfig) {
		this.rfssNameTable.put(id, rfssConfig);
		
	}

	
	/**
	 * Rebind IP addresses for all the emulator nodes to conform to a new Tester config.
	 * This method is called to dynamically change a tester topology.
	 * 
	 * @param testerconfig -- the tester config.
	 */
	public void rebindIpAddresses(ISSITesterConfiguration testerconfig) {
		for ( RfssConfig rfssConfig: this.getRfssConfigurations()) {
			if ( rfssConfig.isEmulated()) {
				String tag = rfssConfig.getTag();
				if ( testerconfig.getDaemonByName(tag)!= null 	) {
					rfssConfig.setIpAddress(testerconfig.getDaemonByName(tag).getIpAddress());
				}
			}
		} 
		
	}

}
