package gov.nist.p25.issi.issiconfig;

import gov.nist.p25.issi.p25body.serviceprofile.group.GroupServiceProfile;
import gov.nist.p25.issi.p25body.serviceprofile.user.AvailabilityCheck;
import gov.nist.p25.issi.p25body.serviceprofile.user.UserServiceProfile;
import gov.nist.p25.issi.transctlmgr.ptt.TimerValues;

import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.text.ParseException;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;

import javax.xml.parsers.SAXParserFactory;

import org.apache.log4j.Logger;
import org.xml.sax.Attributes;
import org.xml.sax.EntityResolver;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.DefaultHandler;

/**
 * A parser for the topology configuration file.
 * 
 * @author M. Ranganathan
 * 
 */
public class TopologyConfigParser extends DefaultHandler implements
		XMLTagsAndAttributes {

	private static Logger logger = Logger
			.getLogger("gov.nist.p25.issi.issiconfig");

	/**
	 * The following are used for parsing this structure.
	 */
	private boolean inuserProfile;

	private StringBuffer cdatabuffer;

	private SuConfig currentSuConfig;

	private boolean inGropProfile;

	/**
	 * File name from which this structure is generated.
	 */
	private String configFile;

	private GroupConfig currentGroupConfig;

	private TopologyConfig topologyConfig;

	private XMLReader saxParser;

	private TopologyConfig globalTopology;

	private RfssConfig rfssConfig;

	private UserServiceProfile userServiceProfile;

	private GroupServiceProfile groupServiceProfile;

	// /////////////////////////////////////////////////////////////////////////////////////////
	// Entity resolver -- allows you to work offline
	// ////////////////////////////////////////////////////////////////////////////////////////
	class MyResolver implements EntityResolver {

		public InputSource resolveEntity(String publicId, String systemId)
				throws SAXException, IOException {
			if (systemId
					.equals("http://www-x.antd.nist.gov:8080/p25/issi-emulator/dtd/topologyconfig.dtd")) {
				FileReader fileReader = new FileReader(new File(
						"dtd/topologyconfig.dtd"));
				InputSource inputSource = new InputSource(fileReader);
				return inputSource;

			} else {
				return null;
			}
		}

	}

	// /////////////////////////////////////////////////////////////////////////////////////////
	// Constructor.
	// ////////////////////////////////////////////////////////////////////////////////////////
	/**
	 * Constructor. Takes a file name and constructs itself from that file.
	 * 
	 * @param fileName --
	 *            xml file from which the topology config constructs itself.
	 * 
	 */
	public TopologyConfigParser() throws Exception {

		try {

			SAXParserFactory saxParserFactory = SAXParserFactory.newInstance();
			this.saxParser = saxParserFactory.newSAXParser().getXMLReader();
			this.saxParser.setEntityResolver(new MyResolver());

			saxParser.setContentHandler(this);

			saxParser
					.setFeature("http://xml.org/sax/features/validation", true);

		} catch (SAXParseException spe) {
			// Error generated by the parser
			logger.error("\n** Parsing error" + ", line " + spe.getLineNumber()
					+ ", uri " + spe.getSystemId());

			logger.error("   " + spe.getMessage());
			// Use the contained exception, if any
			Exception x = spe;
			if (spe.getException() != null)
				x = spe.getException();
			x.printStackTrace();

			throw spe;
		} catch (SAXException sxe) {
			// Error generated by this application
			// (or a parser-initialization error)
			Exception x = sxe;
			if (sxe.getException() != null)
				x = sxe.getException();
			x.printStackTrace();
			throw sxe;

		} catch (Exception pce) {
			// Parser with specified options can't be built
			pce.printStackTrace();
			throw pce;
		}

	}

	// /////////////////////////////////////////////////////////////////////////////////
	// SAX Parsing methods that generate this structure from a XML File.
	// /////////////////////////////////////////////////////////////////////////////////

	public TopologyConfig parse(TopologyConfig globalTopology,
			String configFileName) throws IOException, SAXException {
		this.globalTopology = globalTopology;
		this.topologyConfig = new TopologyConfig();
		topologyConfig.configFile = configFileName;
		this.configFile = configFileName;

		// Inherit the values from global topology.
		if (globalTopology != null) {
			topologyConfig.wacnNameTable.putAll(globalTopology.wacnNameTable);
			topologyConfig.getSystemNameTable()
					.putAll(globalTopology.getSystemNameTable());
			topologyConfig.getSystemTable().putAll(globalTopology.getSystemTable());
			TimerValues.resetToDefaults();

		}

		saxParser.parse(configFileName);
		return this.topologyConfig;

	}

	@Override
	public void characters(char[] ch, int start, int length) {
		if (this.inuserProfile)
			this.cdatabuffer.append(ch, start, length);
		else if (this.inGropProfile)
			this.cdatabuffer.append(ch, start, length);

	}

	@Override
	public void startElement(String namespaceURI, String local, String name,
			Attributes attrs) throws SAXException {
		try {
			if (name.compareToIgnoreCase("conformance-test-config") == 0) {
				if (attrs.getValue("THEARTBEAT") != null) {
					int heartbeat = Integer.parseInt(attrs
							.getValue("THEARTBEAT"));
					TimerValues.setHeartbeat(heartbeat);
				}
				if ( attrs.getValue("traceOrder")!= null) {
					String[] rfssOrder = attrs.getValue("traceOrder").split(",");
					LinkedList<String> orders = new LinkedList<String>();
					for ( String s : rfssOrder) {
						orders.add(s.trim());
					}
					this.topologyConfig.setTraceOrder(orders);
				}
			} else if (name.compareToIgnoreCase("userprofile") == 0) {
				this.inuserProfile = true;
				this.cdatabuffer = new StringBuffer();
				this.userServiceProfile = new UserServiceProfile();
				
				this.userServiceProfile.setAvailabilityCheck(attrs
						.getValue("availCheckSupported"));
				this.userServiceProfile.setFullDuplex(Boolean
						.parseBoolean(attrs.getValue("isFullDuplex")));

				this.userServiceProfile.setAccessPermission(attrs
						.getValue("accessPermission"));

				this.userServiceProfile.setSecureCapable(Boolean
						.parseBoolean(attrs.getValue("isSecureCapable")));

				this.userServiceProfile.setGroupCallPermission(attrs
						.getValue("groupCallPermission"));

				this.userServiceProfile.setUnitToUnitCallPermission(attrs
						.getValue("unitToUnitCallPermission"));

				this.userServiceProfile.setCallPriority(Integer.parseInt(attrs
						.getValue("unitToUnitCallPriority")));

				this.userServiceProfile.setUnitToUnitCallSetupPreference(attrs
						.getValue("unitToUnitCallSetupPreference"));
				
				if ( attrs.getValue("interconnectCallPriority") != null) {
					int icp = Integer.parseInt(attrs.getValue("interconnectCallPriority"));
					this.userServiceProfile.setInterconnectCallPriority(icp);
				}

				// TBD -- setters for the other ones.

			} else if (name.compareToIgnoreCase("groupprofile") == 0) {
				this.inGropProfile = true;
				this.cdatabuffer = new StringBuffer();
				this.groupServiceProfile = new GroupServiceProfile();
				
				groupServiceProfile.setAccessPermission(attrs.getValue("accessPermission"));
				
				if (attrs.getValue("announcementGroup") != null ) {
					groupServiceProfile.setAnnouncementGroupId(Long.parseLong(attrs.getValue("announcementGroup")));
				}
				groupServiceProfile.setGroupPriority(Integer.parseInt(attrs.getValue("groupPriority")));
				
				if ( attrs.getValue("rfHangTime") != null) {
					groupServiceProfile.setRfHangTime(Integer.parseInt(attrs.getValue("rfHangTime")));
				}
				
				if ( attrs.getValue("confirmedGroupCallSetupTime") != null ) {
					groupServiceProfile.setConfirmedGroupCallSetupTime
						(Integer.parseInt(attrs.getValue("confirmedGroupCallSetupTime")));
				}
				
				groupServiceProfile.setInterruptMode(attrs.getValue("interruptMode"));
				
				groupServiceProfile.setGroupSecurityLevel(attrs.getValue("groupSecurityLevel"));
				
				groupServiceProfile.setInterconnectFlag(Boolean.parseBoolean(attrs.getValue("interconnectFlag")));
				
				groupServiceProfile.setInterconnectSecurity(attrs.getValue("interconnectSecurity"));

			} else if (name.compareToIgnoreCase("rfssconfig") == 0) {
				String rfssName = attrs.getValue(RFSS_NAME);
				if (attrs.getValue(RFSS_NAME) == null) {
					throw new SAXException("Missing a required attribute "
							+ RFSS_NAME);
				}
				if (rfssName == null)
					throw new SAXException("Missing a required attribute "
							+ RFSS_NAME);

				rfssConfig = this.globalTopology.getRfssConfigByName(rfssName);
				if (rfssConfig == null)
					throw new SAXException("Could not find global reference "
							+ rfssName);
				this.topologyConfig.addRfssConfig(rfssConfig);
				
				int groupRegistrationLeaseTime = attrs
						.getValue(GROUP_REGISTRATION_EXPIRES_TIME) == null
						|| attrs.getValue(GROUP_REGISTRATION_EXPIRES_TIME)
								.equals(UNBOUNDED) ? -1 : Integer
						.parseInt(attrs
								.getValue(GROUP_REGISTRATION_EXPIRES_TIME));
				if (groupRegistrationLeaseTime < -1)
					throw new SAXException(
							"Illegal value specified for groupRegistrationExpiresTime ");

				int servedGroupLeaseTime = attrs
						.getValue(SERVED_GROUP_LEASE_TIME) == null
						|| attrs.getValue(SERVED_GROUP_LEASE_TIME).equals(
								UNBOUNDED) ? -1 : Integer.parseInt(attrs
						.getValue(SERVED_GROUP_LEASE_TIME));

				int groupCallInviteProcessingTime = attrs
						.getValue(GROUP_CALL_INVITE_PROCESSING_TIME) == null ? 0
						: Integer.parseInt(attrs
								.getValue(GROUP_CALL_INVITE_PROCESSING_TIME));

				boolean rfResavail = attrs.getValue(RF_RESOURCES_AVAILABLE) == null ? true
						: attrs.getValue(RF_RESOURCES_AVAILABLE).equals("true");

				int maxRtpPorts = attrs.getValue(MAX_RTP_PORTS) == null
						|| attrs.getValue(MAX_RTP_PORTS).equals(UNBOUNDED) ? -1
						: Integer.parseInt(attrs.getValue(MAX_RTP_PORTS));

				boolean isAdvancedRtpResourceManagementSupported = attrs
						.getValue(IS_ADVANCED_RTP_RESOURCE_MANAGEMENT_SUPPORTED) == null ? false
						: "true"
								.equals(attrs
										.getValue(IS_ADVANCED_RTP_RESOURCE_MANAGEMENT_SUPPORTED));

				boolean queryUnknownSu = attrs.getValue(QUERY_SU_CREDENTIALS) == null ? false
						: attrs.getValue(QUERY_SU_CREDENTIALS).equals("true");

				boolean queryUnknownGroup = attrs
						.getValue(QUERY_GROUP_CREDENTIALS) == null ? false
						: attrs.getValue(QUERY_GROUP_CREDENTIALS)
								.equals("true");

				int registerConfirmTime = attrs.getValue(REGISTER_CONFIRM_TIME) == null ? 0
						: Integer.parseInt(attrs
								.getValue(REGISTER_CONFIRM_TIME));
				boolean grantPttRequest = attrs.getValue(GRANT_PTT_REQUEST) == null ? true
						: "true".equals(attrs.getValue(GRANT_PTT_REQUEST));

				boolean isInterestedInLosingAudio = attrs
						.getValue(IS_INTERESTED_IN_LOSING_AUDIO) == null ? false
						: "true".equals(attrs
								.getValue(IS_INTERESTED_IN_LOSING_AUDIO)) ? true
								: false;
				boolean isCallingSuInitialTransmitter = attrs
						.getValue(IS_CALLING_SU_INITRANS) == null ? true
						: "true".equals(attrs
						.getValue(IS_CALLING_SU_INITRANS));
				
				//rfssConfig = topologyConfig.getRfssConfigByName(rfssName);

				rfssConfig.notInterestedInLosingAudio = !isInterestedInLosingAudio;

				
				rfssConfig.groupRegistrationExpiresTime = groupRegistrationLeaseTime;
				rfssConfig.servedGroupLeaseTime = servedGroupLeaseTime;
				rfssConfig.groupCallInviteProcessingTime = groupCallInviteProcessingTime;
				rfssConfig.registerConfirmTime = registerConfirmTime;
				rfssConfig.maxRtpPorts = maxRtpPorts;
				rfssConfig.rfResourcesAvailable = rfResavail;
				rfssConfig.suCredentialsQueriedBeforeRegister = queryUnknownSu;
				rfssConfig.groupCredentialsQueriedBeforeRegister = queryUnknownGroup;
				rfssConfig
						.setAdvancedRtpResourceManagementSupported(isAdvancedRtpResourceManagementSupported);
				rfssConfig.pttRequestGranted = grantPttRequest;
				rfssConfig
						.setCallingSuInitialTransmitter(isCallingSuInitialTransmitter);
			
				String id = attrs.getValue(ID);
				if (id != null) {
					topologyConfig.addRfssAlias(id,rfssConfig);
				}
			} else if (name.compareToIgnoreCase("suconfig") == 0) {
				if (attrs.getValue(SU_NAME) == null)
					throw new SAXException("Missing a required attribute "
							+ SU_NAME);

				String suName = attrs.getValue(SU_NAME);

				if (suName == null)
					throw new SAXException("Missing a required attribute "
							+ SU_NAME);
				SuConfig suConfig = null;

				suConfig = globalTopology.getSuConfigByName(suName);
				if (suConfig == null)
					throw new SAXException("Could not find a global su name "
							+ suName);
				RfssConfig homeRfss = suConfig.getHomeRfss();
				
				this.topologyConfig.addRfssConfig(homeRfss);
				RfssConfig servingRfss = suConfig.getInitialServingRfss();
				this.topologyConfig.addRfssConfig(servingRfss);
				
				this.topologyConfig.suNameTable.put(suConfig.getSuName(),
						suConfig);
				this.topologyConfig.suConfigTable.put(
						suConfig.getRadicalName(), suConfig);

				SuState initialSuState = SuState.OFF;
				if (attrs.getValue(INITIAL_STATE) != null) {
					initialSuState = attrs.getValue(INITIAL_STATE).equals("on") ? SuState.ON
							: SuState.OFF;
				}
				int homeRfssKnowsMeFor = attrs
						.getValue(HOME_RFSS_KNOWS_ABOUT_ME_FOR) == null
						|| attrs.getValue(HOME_RFSS_KNOWS_ABOUT_ME_FOR).equals(
								UNBOUNDED) ? -1 : Integer.parseInt(attrs
						.getValue(HOME_RFSS_KNOWS_ABOUT_ME_FOR));

				int inviteProcessingDelay = attrs
						.getValue(INVITE_PROCESSING_DELAY) == null ? 0
						: Integer.parseInt(attrs
								.getValue(INVITE_PROCESSING_DELAY));

				int homeRfssRegistersMeFor = attrs
						.getValue(HOME_RFSS_REGISTERS_ME_FOR) == null
						|| attrs.getValue(HOME_RFSS_REGISTERS_ME_FOR).equals(
								UNBOUNDED) ? -1 : Integer.parseInt(attrs
						.getValue(HOME_RFSS_REGISTERS_ME_FOR));

				boolean registerOnForce = Boolean
						.parseBoolean(attrs.getValue(REGISTER_ON_FORCE) == null ? "true"
								: "false");
				int servingRfssReferencesMeFor = attrs
						.getValue(SERVING_RFSS_REFERENCES_ME_FOR) == null
						|| attrs.getValue(SERVING_RFSS_REFERENCES_ME_FOR)
								.equals(UNBOUNDED) ? -1 : Integer
						.parseInt(attrs
								.getValue(SERVING_RFSS_REFERENCES_ME_FOR));
				boolean availability = Boolean.parseBoolean(attrs
						.getValue(AVAILABILITY) == null ? "true" : attrs.getValue(AVAILABILITY));
				
				boolean isSuToSuCallAllowed = Boolean.parseBoolean(attrs
						.getValue(IS_SU_TO_SU_CALL_ALLOWED) == null ? "true" : attrs.getValue(IS_SU_TO_SU_CALL_ALLOWED));
				
				String prot = attrs.getValue(PROTECTION_DISPOSITION);
				CProtectedDisposition disp = CProtectedDisposition.getValue(prot);
				suConfig = topologyConfig.suNameTable.get(suName);
				suConfig.initialState = initialSuState;
				suConfig.homeRfssKnowsAboutMeFor = homeRfssKnowsMeFor;
				suConfig.servingRfssReferencesMeFor = servingRfssReferencesMeFor;
				suConfig.inviteProcessingDelay = inviteProcessingDelay;
				suConfig.homeRfssRegistersMeFor = homeRfssRegistersMeFor;
				suConfig.registerOnForce = registerOnForce;
				suConfig.available = availability;
				suConfig.setCProtectedDisposition(disp);
				boolean checkSuPresenceOnRegister = attrs.getValue(CHECK_SU_PRESENCE_ON_REGISTER) == null ? true : 
					"true".equals(attrs.getValue(CHECK_SU_PRESENCE_ON_REGISTER));
				
				suConfig.setCheckSuPresenceOnRegister(checkSuPresenceOnRegister);

				RfssConfig homeRfssConfig = suConfig.getHomeRfss();
				RfssConfig servingRfssConfig = suConfig.getInitialServingRfss();
				homeRfssConfig.addAssignedSubscriberUnit(suConfig);

				servingRfssConfig.addServedSubscriberUnit(suConfig);
				this.currentSuConfig = suConfig;
				if (attrs.getValue(ID) != null) {
					this.topologyConfig.suNameTable.put(attrs.getValue(ID),
							suConfig);
				}

			} else if (name.equalsIgnoreCase("forbidden")) {
				if (attrs.getValue(RFSS_NAME) == null)
					throw new SAXException("Missing a required value "
							+ RFSS_NAME);
				String rfssName = attrs.getValue(RFSS_NAME);

				RfssConfig rfssConfig = this.globalTopology
						.getRfssConfigByName(rfssName);
				if (rfssConfig == null) {
					throw new SAXException("could not find global rfss "
							+ attrs.getValue(RFSS_NAME));
				}
				
				this.topologyConfig.addRfssConfig(rfssConfig);
				String operation = attrs.getValue(OPERATION);
				if (this.currentSuConfig != null) {
					if (operation.equals("REGISTER"))
						this.currentSuConfig
								.addRegisterForbiddenRfss(rfssConfig);
					else if (operation.equals("REGISTER-QUERY"))
						this.currentSuConfig
								.addRegisterQueryForbiddenRfss(rfssConfig);
					else
						this.currentSuConfig.addForbiddenRfss(rfssConfig);
				} else if (this.currentGroupConfig != null) {
					if (currentGroupConfig.getHomeRfss() == rfssConfig)
						throw new SAXException(
								"Home RFSS of the group cannot be on the Forbidden set");
					else {
						for (Iterator<SuConfig> subscribers = this.currentGroupConfig
								.getSubscribers(); subscribers.hasNext();) {
							SuConfig subscriber = subscribers.next();
							if (subscriber.getInitialServingRfss() == rfssConfig
									&& subscriber.getInitialSuState() == SuState.ON)
								throw new SAXException(
										"Configuration error : "
												+ " Group Member Initial Serving RFSS cannot be in the forbidden set");
						}

					}
					if (operation.equals("REGISTER"))
						this.currentGroupConfig.addForbiddenRfss(rfssConfig);
					else if (operation.equals("REGISTER-QUERY"))
						this.currentGroupConfig
								.addRegisterQueryForbiddenFromRfss(rfssConfig);
					else
						this.currentGroupConfig.addForbiddenRfss(rfssConfig);
				} else
					throw new SAXException("Unexpected element : forbidden ");
			} else if (name.compareToIgnoreCase("unknown-unit") == 0) {
				if (attrs.getValue(SU_NAME) == null) {
					throw new SAXException("Missing a required value "
							+ SU_NAME);
				}

				String suName = attrs.getValue(SU_NAME);
				SuConfig suConfig = this.globalTopology
						.getSuConfigByName(suName);
				this.rfssConfig.addUnknownSu(suConfig);

			} else if (name.compareToIgnoreCase("sgconfig") == 0) {
				String groupName = attrs.getValue(GROUP_NAME);
				if (groupName == null)
					throw new SAXException("Missing a required attribute "
							+ GROUP_NAME);
				GroupConfig groupConfig = null;
				groupConfig = globalTopology.getGroupConfigByName(groupName);
				if (groupConfig == null)
					throw new SAXException(
							"Could not find a global group name " + groupName);
				RfssConfig homeRfss = groupConfig.getHomeRfss();
				
				this.topologyConfig.addRfssConfig(homeRfss);
				topologyConfig.groupNameTable.put(groupName, groupConfig);
				topologyConfig.groupTable.put(groupConfig.getRadicalName(),
						groupConfig);

				int homeRfssKnowsMeFor = attrs
						.getValue(HOME_RFSS_KNOWS_ABOUT_ME_FOR) == null
						|| attrs.getValue(HOME_RFSS_KNOWS_ABOUT_ME_FOR).equals(
								UNBOUNDED) ? -1 : Integer.parseInt(attrs
						.getValue(HOME_RFSS_KNOWS_ABOUT_ME_FOR));
				groupConfig.homeRfssKnowsAboutMeFor = homeRfssKnowsMeFor;
				
				String protectionDisposition = attrs.getValue(PROTECTION_DISPOSITION);
				
				CProtectedDisposition pdisp = CProtectedDisposition.getValue(protectionDisposition);
				
				groupConfig.setCProtectedDispsition(pdisp);

				this.currentGroupConfig = groupConfig;
				topologyConfig.groupNameTable.put(groupName, groupConfig);
				RfssConfig homeRfssConfig = groupConfig.getHomeRfss();
				homeRfssConfig.addAssignedGroup(groupConfig);
				String refId = attrs.getValue(ID);
				if (refId != null) {
					this.topologyConfig.groupNameTable.put(refId, groupConfig);
				}

			} else if (name.compareToIgnoreCase("subscriber") == 0) {
				String suName = attrs.getValue(SU_NAME);

				if (suName == null)
					throw new SAXException("Missing a required attribute "
							+ SU_NAME);

				SuConfig suConfig = null;

				if (globalTopology.getSuConfigByName(suName) == null) {
					throw new SAXException(
							"Could not find global defintion for su " + suName);
				}
				suConfig = globalTopology.getSuConfigByName(suName);
				RfssConfig homeRfss = suConfig.getHomeRfss();
				this.topologyConfig.addRfssConfig(homeRfss);
				
				RfssConfig servingRfss = suConfig.getInitialServingRfss();
				this.topologyConfig.addRfssConfig(servingRfss);
				this.topologyConfig.suNameTable.put(suConfig.getSuName(),
						suConfig);
				this.topologyConfig.suConfigTable.put(
						suConfig.getRadicalName(), suConfig);
				homeRfss.addAssignedSubscriberUnit(suConfig);

				servingRfss.addServedSubscriberUnit(suConfig);

				if (suConfig == null)
					throw new SAXException("Could not find an rfss for "
							+ attrs.getValue(SU_NAME));
				else {

					this.currentGroupConfig.addSubscriber(suConfig);
					suConfig.addGroup(this.currentGroupConfig);
					suConfig.addSubscribedTo(this.currentGroupConfig);

				}

			}
		} catch (SAXException ex1) {
			throw ex1;
		} catch (Exception ex) {
			logger.error("Error parsing configuration file " + configFile, ex);
			throw new SAXException("Error parsing config file", ex);
		}

	}

	@Override
	public void endElement(String uri, String name, String qName)
			throws SAXException {
		if ("userprofile".equalsIgnoreCase(qName)) {

			this.inuserProfile = false;
			logger.debug("cdataBuffer = "
					+ this.cdatabuffer.toString().trim().length());
			try {
				if (this.cdatabuffer.toString().trim().length() != 0) {
					this.currentSuConfig.setUserServiceProfile(this.cdatabuffer
							.toString());
				} else {
					this.currentSuConfig
							.setUserServiceProfile(this.userServiceProfile);
				}
			} catch (ParseException ex) {
				logger.error("error user profile in file " + this.configFile,
						ex);
				throw new SAXException(this.cdatabuffer.toString(), ex);
			}
		} else if ("groupprofile".equalsIgnoreCase(qName)) {
			this.inGropProfile = false;
			try {
				if (this.cdatabuffer.toString().trim().length() != 0) {
					this.currentGroupConfig
							.setGroupServiceProfile(this.cdatabuffer.toString());
				} else {
					this.currentGroupConfig
							.setGroupServiceProfie(this.groupServiceProfile);
				}
			} catch (ParseException ex) {
				logger.error("error user profile in file " + this.configFile,
						ex);
				throw new SAXException(this.cdatabuffer.toString(), ex);
			}
		} else if ("suconfig".equalsIgnoreCase(qName)) {
			this.currentSuConfig = null;
		} else if ("sgconfig".equalsIgnoreCase(qName)) {
			this.currentGroupConfig = null;
		}

	}
}
