/*
 *This software was developed by employees of the National Institute of
 *Standards and Technology (NIST), an agency of the Federal Government.
 *Pursuant to title 15 Untied States Code Section 105, works of NIST
 *employees are not subject to copyright protection in the United States
 *and are considered to be in the public domain.  As a result, a formal
 *license is not needed to use the software.
 *
 *
 *The ISSI Emulator software is provided by NIST as a service and is expressly
 *provided "AS IS."  NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED
 *OR STATUTORY, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF
 *MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT
 *AND DATA ACCURACY.  NIST does not warrant or make any representations
 *regarding the use of the software or the results thereof, including but
 *not limited to the correctness, accuracy, reliability or usefulness of
 *the software.
 *
 *
 *Permission to use this software is contingent upon your acceptance
 *of the terms of this agreement
 *.
 *
 */
package gov.nist.p25.issi.p25body.params;

import java.text.ParseException;

import org.apache.log4j.Logger;

/**
 * These attributes are used in conjunction with an SDP announcement. All these
 * parameters are optional. Each parameter is a boolean value represented by 0
 * or 1.
 * 
 * @author M. Ranganathan
 * 
 */
public class CallParam {
	
	/*
	 * A private flag that records whether the c-resavail flag was explicitly
	 * set.
	 */
	
	 private boolean resAvailIsSet;
	 
	/*
	 * This parameter indicates the first transmitter in a call. It is named
	 * c-initrans or c-i in its compact form. Two values are allowed: 0 if the
	 * calling SU is chosen to transmit first or 1 if it is the called SU. Note
	 * that if unset, the Intial Transmitter is null, then the field is unset.
	 */
	private Boolean callingSuInitialTransmitter;

	/*
	 * This parameter indicates whether RF resources are available in an RFSS
	 * area. A value of 0, which is the default value, indicates that no RF
	 * resource is available. This field SHALL be present in confirmed group
	 * calls: - in the first INVITE message from a serving RFSS to the Home
	 * RFSS. - in the re-INVITE messages from any serving RFSS to the Home RFSS
	 * when the RF resource allocation state changes. - in the 200 OK response
	 * from each serving RFSS that provides a Type 1 SDP answer to the Home
	 * RFSSs SIP (Re-)INVITE request.
	 * 
	 * This field SHOULD be present in SU-to-SU calls in all the INVITE, 200 OK
	 * and ACK messages.
	 * 
	 */
	private boolean rfResourceAvailable;

	/*
	 * This parameter indicates to the Called Serving RFSS whether it SHOULD
	 * perform an availability check with the called SU prior to granting the
	 * request and allocating a traffic channel, or to process the call as a
	 * direct call. The called SU is expected to respond to the availability
	 * check with a refusal or an acceptance.
	 */

	private boolean availabilityCheckRequested;

	/*
	 * This parameter indicates the desired duplexity.
	 * 
	 */
	private boolean fullDuplexRequested;
	
	
	/*
	 * Protected mode call setup requested.
	 */
	private boolean protectedMode;

	/*
	 * This parameter indicates the roaming of an SU in an SU-to-SU Call. It is
	 * named c-incallroaming with a compact form c-icr.
	 */
	private boolean incallRoaming;

	/*
	 * Grup call type.
	 */
	private boolean confirmedGroupCall;


	private static final String HEADER_START = "{\\insrsid15927085 ";

	private static final String IGNORED_FIELD = "}{\\i\\fs16\\insrsid15927085 ";

	private static final String HEADER_END = "\\par }\n";

	private static final String HEADER_RESTART = "} " + HEADER_START;

	private static Logger logger = Logger.getLogger(CallParam.class);

	private static boolean getBooleanFromString(String st)
			throws ParseException {
		String str = st.trim();
		if (str.equals("1"))
			return true;
		else if (str.equals("0"))
			return false;
		else
			throw new ParseException("bad token [" + str + "]", 0);
	}

	private static String getStringFromBoolean(boolean bool) {
		return (bool ? "1" : "0");
	}

	/**
	 * Default constructor (initialized with the default values from Table 25 BACA).
	 */
	public CallParam() {
		this.callingSuInitialTransmitter = null; // Not assigned.
		this.rfResourceAvailable = false;
		this.availabilityCheckRequested = true;
		this.fullDuplexRequested = false;
		this.incallRoaming = false;
		this.setProtectedMode(false);

	}

	
	/**
	 * @param callingSuInitialTransmitter
	 *            The callingSuInitialTransmitter to set.
	 */
	public void setCallingSuInitialTransmitter(
			Boolean callingSuInitialTransmitter) {
		this.callingSuInitialTransmitter = callingSuInitialTransmitter;
		
	}

	/**
	 * @return Returns true if calling su is set to be the initial transmitter.
	 * 
	 */
	public boolean isCallingSuInitialTransmitter() {
		return this.callingSuInitialTransmitter != null
				&& callingSuInitialTransmitter;
	}

	public boolean isCalledSuInitialTransmitter() {
		return this.callingSuInitialTransmitter != null
				&& !this.callingSuInitialTransmitter;
	}

	/**
	 * @param rfResourceAvailable
	 *            The rfResourceAvailable to set.
	 */
	public void setRfResourceAvailable(boolean rfResourceAvailable) {
		this.rfResourceAvailable = rfResourceAvailable;
		this.resAvailIsSet = true;
	}

	/**
	 * @return Returns the rfResourceAvailable.
	 */
	public boolean isRfResourceAvailable() {
		return rfResourceAvailable;
	}

	/**
	 * @param availabilityCheckRequested
	 *            The availabilityCheckRequested to set.
	 */
	public void setAvailabilityCheckRequested(boolean availabilityCheckRequested) {
		this.availabilityCheckRequested = availabilityCheckRequested;
	}

	/**
	 * @return Returns the availabilityCheckRequested.
	 */
	public boolean isAvailabilityCheckRequested() {
		return availabilityCheckRequested;
	}

	/**
	 * @param fullDuplexRequested
	 *            The fullDuplexRequested to set.
	 */
	public void setFullDuplexRequested(boolean fullDuplexRequested) {
		this.fullDuplexRequested = fullDuplexRequested;
	}

	/**
	 * @return Returns the fullDuplexRequested.
	 */
	public boolean isFullDuplexRequested() {
		return fullDuplexRequested;
	}

	/**
	 * @param incallRoaming
	 *            The incallRoaming to set.
	 */
	public void setIncallRoaming(boolean incallRoaming) {
		this.incallRoaming = incallRoaming;
	}

	/**
	 * @return Returns the incallRoaming.
	 */
	public boolean isIncallRoaming() {
		return incallRoaming;
	}

	/**
	 * @param confirmedGroupCall
	 *            The confirmedGroupCall to set.
	 */
	public void setConfirmedGroupCall(boolean confirmedGroupCall) {
		this.confirmedGroupCall = confirmedGroupCall;
	}

	/**
	 * @return Returns the confirmedGroupCall.
	 */
	public boolean isConfirmedGroupCall() {
		return confirmedGroupCall;
	}

	/**
	 * @param protectedMode the protectedMode to set
	 */
	public void setProtectedMode(boolean protectedMode) {
		this.protectedMode = protectedMode;
	}

	/**
	 * @return the protectedMode
	 */
	public boolean isProtectedMode() {
		return protectedMode;
	}

	/**
	 * Parse and return a call param string.
	 * 
	 * @param callParam
	 * @return
	 * @throws ParseException
	 */
	public static CallParam createCallParam(String cParam)
			throws ParseException {

		StringBuffer sbuf = new StringBuffer(cParam);
		for (int i = 0; i < cParam.length(); i++) {
			if (cParam.charAt(i) != '\r')
				sbuf.append(cParam.charAt(i));
		}
		String callParam = sbuf.toString();
		String[] lines = callParam.split("\n");
		// if ( lines.length == 0 ) lines = callParam.split("\n");
		if (lines.length == 0)
			throw new ParseException("Bad Param set [" + callParam + "]", 0);
		CallParam retval = new CallParam();
		for (String lin : lines) {
			if (lin.equals(""))
				continue;
			String li[] = lin.split(":");
			if (li.length != 2)
				throw new ParseException("Bad param [" + lin + "]", 0);
			String name = li[0].trim();
			String val = li[1].trim();
			logger.debug("createCallParam : " + lin + " name = " + name
					+ " val = " + val);
			if (name.equalsIgnoreCase("c-initrans")
					|| name.equalsIgnoreCase("c-i"))
				retval.callingSuInitialTransmitter = new Boolean(
						! getBooleanFromString(val));
			else if (name.equalsIgnoreCase("c-resavail")
					|| name.equalsIgnoreCase("c-r"))
				retval.rfResourceAvailable = getBooleanFromString(val);
			else if (name.equalsIgnoreCase("c-pref")
					|| name.equalsIgnoreCase("c-prf"))
				retval.availabilityCheckRequested = getBooleanFromString(val);
			else if (name.equalsIgnoreCase("c-duplex")
					|| name.equalsIgnoreCase("c-d"))
				retval.fullDuplexRequested = getBooleanFromString(val);
			else if (name.equalsIgnoreCase("c-incallroaming")
					|| name.equalsIgnoreCase("c-icr"))
				retval.incallRoaming = getBooleanFromString(val);
			else if (name.equalsIgnoreCase("c-groupcalltype")
					|| name.equalsIgnoreCase("c-gct")) {
				retval.confirmedGroupCall = getBooleanFromString(val);
			} else if (name.equalsIgnoreCase("c-protected") || name.equals("c-p")){
				retval.setProtectedMode(getBooleanFromString(val));
			} else {
				throw new ParseException("Unrecognised field [" + lin + "]", 0);
			}

		}

		logger.debug(" parsed value = [ " + retval + " ] ");
		return retval;
	}

	private static void appendField(String fieldName, boolean value,
			StringBuffer sb) {
		sb.append(fieldName).append(":").append(getStringFromBoolean(value))
				.append("\r\n");
	}

	private static void appendRtfField(String fieldName, boolean value,
			StringBuffer sb) {
		sb.append(HEADER_START).append(fieldName).append(": ").append(
				IGNORED_FIELD).append(getStringFromBoolean(value)).append(
				HEADER_END);
	}

	public String toRtfString() {
		StringBuffer sb = new StringBuffer();
		if (this.callingSuInitialTransmitter != null) {
			appendRtfField("c-initrans", ! this.callingSuInitialTransmitter, sb);
		}
		CallParam reference = new CallParam();
		if (reference.rfResourceAvailable != this.rfResourceAvailable || this.resAvailIsSet)
			appendRtfField("c-resavail", this.rfResourceAvailable, sb);
		if ( reference.availabilityCheckRequested != this.availabilityCheckRequested)
			appendRtfField("c-pref", this.availabilityCheckRequested, sb);
		if (reference.fullDuplexRequested != this.fullDuplexRequested)
			appendRtfField("c-duplex", this.fullDuplexRequested, sb);
		if ( reference.incallRoaming != this.incallRoaming)
			appendRtfField("c-incallroaming", this.incallRoaming, sb);
		if ( reference.confirmedGroupCall != this.confirmedGroupCall)
			appendRtfField("c-groupcalltype", this.confirmedGroupCall, sb);
		if (reference.isProtectedMode() != this.isProtectedMode())
			appendRtfField("c-protected",this.isProtectedMode(),sb);
		return sb.toString();
	}

	@Override
	public String toString() {
		StringBuffer sb = new StringBuffer();
		if (this.callingSuInitialTransmitter != null) {
			appendField("c-initrans", ! this.callingSuInitialTransmitter, sb);
		}
		CallParam reference = new CallParam();
		if (reference.rfResourceAvailable != this.rfResourceAvailable || this.resAvailIsSet)
			appendField("c-resavail", this.rfResourceAvailable, sb);
		if ( reference.availabilityCheckRequested != this.availabilityCheckRequested)			
			appendField("c-pref", this.availabilityCheckRequested, sb);
		if (reference.fullDuplexRequested != this.fullDuplexRequested)	
			appendField("c-duplex", this.fullDuplexRequested, sb);
		if ( reference.incallRoaming != this.incallRoaming)	
			appendField("c-incallroaming", this.incallRoaming, sb);
		if ( reference.confirmedGroupCall != this.confirmedGroupCall)
			appendField("c-groupcalltype", this.confirmedGroupCall, sb);
		if (reference.isProtectedMode() != this.isProtectedMode())
			appendField("c-protected",this.isProtectedMode(),sb);
	
		return sb.toString();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object other) {
		if (!(other instanceof CallParam))
			return false;
		CallParam template = (CallParam) other;
		if ( template.callingSuInitialTransmitter == null ^ this.callingSuInitialTransmitter == null )
			return false;
		return (( template.callingSuInitialTransmitter == null  && this.callingSuInitialTransmitter == null ||
					template.callingSuInitialTransmitter.equals(this.callingSuInitialTransmitter) ) &&
				template.availabilityCheckRequested == this.availabilityCheckRequested
				&& template.confirmedGroupCall == this.confirmedGroupCall
				&& template.fullDuplexRequested == this.fullDuplexRequested
				&& template.incallRoaming == this.incallRoaming
				&& template.rfResourceAvailable == this.rfResourceAvailable 
				&& template.protectedMode == this.protectedMode) ;

	}

	public boolean isDefault() {
		return (new CallParam().equals(this)) && ! this.resAvailIsSet;
	}
	
	/**
	 * Almost the same opertation as equals except that initialTransmitter is not compared ( except for
	 * existence).
	 * 
	 * @param other -- the template to compare with.
	 * 
	 * @return true if a match occured.
	 */
	
	public boolean match(CallParam other) {
			if (!(other instanceof CallParam))
			return false;
		CallParam template = (CallParam) other;
		if ( template.callingSuInitialTransmitter == null ^ this.callingSuInitialTransmitter == null )
			return false;
		return ( ( template.callingSuInitialTransmitter ==  this.callingSuInitialTransmitter  ||
				template.callingSuInitialTransmitter != null &&   this.callingSuInitialTransmitter != null ) &&
				template.availabilityCheckRequested == this.availabilityCheckRequested
				&& template.confirmedGroupCall == this.confirmedGroupCall
				&& template.fullDuplexRequested == this.fullDuplexRequested
				&& template.incallRoaming == this.incallRoaming
				&& template.rfResourceAvailable == this.rfResourceAvailable 
				&& template.protectedMode == this.protectedMode) ;
	}

}
