/*
 *This software was developed by employees of the National Institute of
 Standards and Technology (NIST), an agency of the Federal Government.
 Pursuant to title 15 Untied States Code Section 105, works of NIST
 employees are not subject to copyright protection in the United States
 and are considered to be in the public domain.  As a result, a formal
 license is not needed to use the software.


 The ISSI Emulator software is provided by NIST as a service and is expressly
 provided "AS IS."  NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED
 OR STATUTORY, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF
 MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT
 AND DATA ACCURACY.  NIST does not warrant or make any representations
 regarding the use of the software or the results thereof, including but
 not limited to the correctness, accuracy, reliability or usefulness of
 the software.


 Permission to use this software is contingent upon your acceptance
 of the terms of this agreement
 .

 */
package gov.nist.p25.issi.p25body.serviceprofile.group;

import gov.nist.p25.issi.issiconfig.XMLTagsAndAttributes;
import gov.nist.p25.issi.p25body.serviceprofile.ServiceProfile;
import gov.nist.p25.issi.p25body.serviceprofile.ServiceProfileLine;
import gov.nist.p25.issi.p25body.serviceprofile.user.UserServiceProfile;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.text.ParseException;
import java.util.Hashtable;

import org.apache.log4j.Logger;
import org.xml.sax.Attributes;

/**
 * Group Service profile structure. Implements the group service profile
 * attributes defined in section 3.8 of the ISSI Specification. Except for the
 * mandatory fields, each field has a default profile line assigned to it when
 * this structure is created.
 * 
 * @author M. Ranganathan
 * 
 */
public class GroupServiceProfile extends ServiceProfile {

	private static Logger logger = Logger.getLogger(GroupServiceProfile.class);

	static {

		profileLineClasses.put(AccessPermission.NAME, AccessPermission.class);
		profileLineClasses.put(AccessPermission.SHORTNAME,
				AccessPermission.class);
		profileLineClasses.put(AnnouncementGroup.NAME, AnnouncementGroup.class);
		profileLineClasses.put(AnnouncementGroup.SHORTNAME,
				AnnouncementGroup.class);

		profileLineClasses.put(ConfirmedCallSetupTime.NAME.toLowerCase(),
				ConfirmedCallSetupTime.class);
		profileLineClasses.put(ConfirmedCallSetupTime.SHORTNAME,
				ConfirmedCallSetupTime.class);

		profileLineClasses.put(EmergencyCapable.NAME, EmergencyCapable.class);
		profileLineClasses.put(EmergencyCapable.SHORTNAME,
				EmergencyCapable.class);

		profileLineClasses.put(EmergencyPreemption.NAME,
				EmergencyPreemption.class);
		profileLineClasses.put(EmergencyPreemption.SHORTNAME,
				EmergencyPreemption.class);

		profileLineClasses.put(GroupPriority.NAME, GroupPriority.class);
		profileLineClasses.put(GroupPriority.SHORTNAME, GroupPriority.class);

		profileLineClasses.put(GroupSecurityLevel.NAME,
				GroupSecurityLevel.class);
		profileLineClasses.put(GroupSecurityLevel.SHORTNAME,
				GroupSecurityLevel.class);

		profileLineClasses.put(InterconnectSecurity.NAME,
				InterconnectSecurity.class);
		profileLineClasses.put(InterconnectSecurity.SHORTNAME,
				InterconnectSecurity.class);

		profileLineClasses.put(InterruptMode.NAME, InterruptMode.class);
		profileLineClasses.put(InterruptMode.SHORTNAME, InterruptMode.class);

		profileLineClasses.put(InterconnectFlag.NAME, InterconnectFlag.class);
		// No short form for this.

		profileLineClasses.put(RfHangTime.NAME, RfHangTime.class);
		profileLineClasses.put(RfHangTime.SHORTNAME, RfHangTime.class);

	}

	/**
	 * Create a new group service profile with default values.
	 * 
	 */
	public GroupServiceProfile() {
		this.profileLines = new Hashtable<String, ServiceProfileLine>();
		/** Add default profile lines where ever applicable */
		this.addProfileLine(new AccessPermission());
		this.addProfileLine(new ConfirmedCallSetupTime());
		this.addProfileLine(new EmergencyCapable());
		this.addProfileLine(new GroupPriority());
		this.addProfileLine(new InterconnectSecurity());
		this.addProfileLine(new RfHangTime());
		this.addProfileLine(new GroupSecurityLevel());
		this.addProfileLine(new EmergencyPreemption());
		this.addProfileLine(new InterconnectFlag());
		this.addProfileLine(new InterruptMode());
	}

	/**
	 * Create a group service profile class given the unparsed line.
	 * 
	 * @param groupServiceProfileLine
	 * @return
	 * @throws ParseException --
	 *             if an error was reched parsing the line.
	 */
	protected ServiceProfileLine createServiceProfileLine(
			String groupServiceProfileLine) throws ParseException {
		if (groupServiceProfileLine == null)
			throw new NullPointerException("null argument");

		String[] tokens = groupServiceProfileLine.split(":");
		if (tokens == null || tokens.length != 2)
			throw new ParseException(groupServiceProfileLine, 0);
		return createServiceProfileLine(tokens[0], tokens[1]);

	}

	/**
	 * Create a group service profile line class given name and value.
	 * 
	 * @param name
	 * @param value
	 * @return
	 * @throws ParseException
	 */
	public static ServiceProfileLine createServiceProfileLine(String name,
			String value) throws ParseException {
		if (name == null || value == null)
			throw new NullPointerException("null argument");

		try {

			Class<? extends ServiceProfileLine> gspLineClass = getProfileLineClass(name);
			if (gspLineClass == null)
				throw new ParseException("Invalid field Name : " + name, 0);
			Constructor cons = gspLineClass
					.getDeclaredConstructor(new Class[] { String.class });
			ServiceProfileLine retval = (ServiceProfileLine) cons
					.newInstance((Object[]) (new String[] { value }));
			return retval;
		} catch (InvocationTargetException ex) {
			// The cause has the actual reason failure
			throw new RuntimeException(ex.getCause());
		} catch (NoSuchMethodException ex) {
			throw new RuntimeException(ex);
		} catch (IllegalArgumentException e) {
			throw new RuntimeException(e);
		} catch (InstantiationException e) {
			throw new RuntimeException(e);
		} catch (IllegalAccessException e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 * Create an announcement group field. Note there is no default value for
	 * this. This field is mandatory.
	 */
	public void setAnnouncementGroupId(long groupId) {
		AnnouncementGroup retval = new AnnouncementGroup(groupId);
		this.profileLines.put(AnnouncementGroup.NAME, retval);

	}

	/**
	 * Get the announcement group from this structure.
	 */
	public AnnouncementGroup getAnounementGroup() {
		return (AnnouncementGroup) getByName(AnnouncementGroup.NAME);
	}

	/**
	 * Get the access permission line.
	 */
	public AccessPermission getAccessPermission() {
		return (AccessPermission) profileLines.get(AccessPermission.NAME);
	}
	
	/**
	 * Set the access permission profile line.
	 * 
	 * @param accessPermission
	 */
	public void setAccessPermission(String accessPermission) {
		this.addProfileLine(AccessPermission.createFromXmlAttribute(accessPermission));
		
	}

	/**
	 * get the confirmed call setup time.
	 */
	public ConfirmedCallSetupTime getConfirmedCallSetupTime() {
		return (ConfirmedCallSetupTime) getByName(ConfirmedCallSetupTime.NAME);
	}
	
	/**
	 * Set the confirmed group call setup time
	 */
	public void setConfirmedGroupCallSetupTime ( int time ) {
		this.addProfileLine(new ConfirmedCallSetupTime(time));
	}
	
	

	/**
	 * get the emergency capable line
	 */
	public EmergencyCapable getEmergencyCapable() {
		return (EmergencyCapable) getByName(EmergencyCapable.NAME);
	}
	
	
	/**
	 * Set the emergency capable line
	 */
	public void setEmergencyCapable(String str) {
		this.addProfileLine(EmergencyCapable.createFromXmlAttribute(str));
	}
	

	/**
	 * Get the group priority line
	 */
	public GroupPriority getGroupPriority() {
		return (GroupPriority) this.getByName(GroupPriority.NAME);
	}
	/**
	 * Set the group priority
	 */
	public void setGroupPriority(int priority) {
		 this.addProfileLine(new GroupPriority(priority));
	}

	/**
	 * Get the interconnect security line.
	 */
	public InterconnectSecurity getInterconnectSecurity() {
		return (InterconnectSecurity) this.getByName(InterconnectSecurity.NAME);
	}
	
	public void setInterconnectSecurity(String security) {
		this.getInterconnectSecurity().setInterconnectCallsSecure(security.equals("Secure"));
	}

	/**
	 * Get the RF Hangtime line.
	 */
	public RfHangTime getRfHangTime() {
		return (RfHangTime) getByName(RfHangTime.NAME);
	}
	
	/**
	 * Set the rf hang timer
	 */
	public void setRfHangTime(int time) throws ParseException {
		RfHangTime  rfhangTime = new RfHangTime(time);
		this.addProfileLine(rfhangTime);
	}

	/**
	 * Get the GroupSecurityLevel
	 */
	public GroupSecurityLevel getGroupSecurityLevel() {
		return (GroupSecurityLevel) getByName(GroupSecurityLevel.NAME);
	}
	
	/**
	 * Set the group security level
	 */
	public void setGroupSecurityLevel(String level) {
		this.addProfileLine(GroupSecurityLevel.createFromXmlAttributes(level));
	}

	/**
	 * Get the emergency preemption line.
	 */
	public EmergencyPreemption getEmergencyPreemption() {
		return (EmergencyPreemption) getByName(EmergencyPreemption.NAME);
	}

	/**
	 * Set the emergency preemption string
	 */
	public void setEmergencyPreemption( String attr) {
		this.addProfileLine(EmergencyPreemption.createFromXmlAttribute(attr));
	}
	/**
	 * Get the interrupt mode.
	 * 
	 * @return
	 */
	public InterruptMode getInterruptMode() {
		return (InterruptMode) this.getByName(InterruptMode.NAME);
	}
	
	public void setInterruptMode(String interruptMode) {
		
		this.addProfileLine(InterruptMode.createFromXmlAttributes(interruptMode));
	}

	/**
	 * Get the interconnect flag line
	 */
	public InterconnectFlag getInterconnectFlag() {
		return (InterconnectFlag) getByName(InterconnectFlag.NAME);
	}
	
	/**
	 * Set the interconnect flag
	 */
	public void setInterconnectFlag(boolean flag) {
		this.getInterconnectFlag().setInterconnectAllowed(flag);
		
	}

	@Override
	public String toString() {
		StringBuffer sbuf = new StringBuffer();

		for (ServiceProfileLine gsLine : this.profileLines.values()) {

			sbuf.append(gsLine.toString()).append("\r\n");
		}
		return sbuf.toString();

	}

	@Override
	public String toRtfString() {

		StringBuffer sbuf = new StringBuffer();

		for (ServiceProfileLine gsLine : this.profileLines.values()) {

			sbuf.append(gsLine.toRtfString()).append("\r\n");
		}
		return sbuf.toString();
	}

	public String getXMLTag() {
		return XMLTagsAndAttributes.GROUP_SERVICE_PROFILE;
	}

	/**
	 * Parse and return a Group Service profile structure from an input buffer
	 * by parsing it.
	 * 
	 * @param buffer
	 * @return
	 */

	public static GroupServiceProfile createGroupServiceProfile(String buffer)
			throws ParseException {

		GroupServiceProfile retval = new GroupServiceProfile();

		parseProfileLines(retval, buffer);

		
		return retval;

	}

	@Override
	public boolean match(Object template) {
		if (!(template instanceof GroupServiceProfile))
			return false;
		return true;
	}

}
