package gov.nist.p25.issi.rfss;

import gov.nist.p25.issi.issiconfig.ISSIConstants;
import gov.nist.p25.issi.issiconfig.RfssConfig;
import gov.nist.p25.issi.transctlmgr.TransmissionControlManager;
import gov.nist.p25.issi.transctlmgr.ptt.PttSession;
import gov.nist.p25.issi.transctlmgr.ptt.PttSessionMultiplexer;

import javax.sdp.Connection;
import javax.sdp.Media;
import javax.sdp.MediaDescription;
import javax.sdp.SessionDescription;
import javax.sip.ClientTransaction;
import javax.sip.DialogTerminatedEvent;
import javax.sip.IOExceptionEvent;
import javax.sip.ListeningPoint;
import javax.sip.RequestEvent;
import javax.sip.ResponseEvent;
import javax.sip.ServerTransaction;
import javax.sip.SipListener;
import javax.sip.SipProvider;
import javax.sip.TimeoutEvent;
import javax.sip.TransactionTerminatedEvent;
import javax.sip.address.SipURI;
import javax.sip.header.ToHeader;
import javax.sip.message.Message;
import javax.sip.message.Request;
import javax.sip.message.Response;

import org.apache.log4j.Logger;

/**
 * The top level call control manager. This is the listener that fields the
 * INVITE. It looks at the user parameter of the To: address and sends the
 * INVITE off to the right place (i.e. UnitToUnitCallControlManager or
 * GroupCallControlManager)
 * 
 * @author M. Ranganathan
 * 
 */
public class CallControlManager implements SipListener {

	private UnitToUnitCallControlManager unitToUnitCallControlManager;

	private GroupCallControlManager groupCallControlManager;

	private SipProvider provider;

	private RfssConfig rfssConfig;

	public static int CC_SETUP_TEARDOWN_REQUEST = 2;

	public static int CC_SETUP_SEND_REQUEST = 1;

	private static Logger logger = Logger.getLogger(CallControlManager.class);

	/**
	 * A utility function that fixes up the SDP announce to match the port of the PttSession.
	 * 
	 * @param session --
	 *            RTPPoint to multipoint session from which to extract params to
	 *            stuff into SdpAnnounce
	 * @param sdpContent --
	 *            sdpContent to fixup.
	 * 
	 * @throws Exception
	 */
	public static void fixupSdpAnnounce(RfssConfig rfss, PttSession session,
			SessionDescription sdpAnnounce) {

		// Addres where the current RFSS
		// is expected to be receiving IP packets.
		try {
			String myIpAddress = rfss.getIpAddress();

			// TransmissionControlManager.addMediaLeg(session, sdpAnnounce) ;

			Connection connection = sdpAnnounce.getConnection();
			connection.setAddress(myIpAddress);

			MediaDescription outgoingMediaDescription = (MediaDescription) (sdpAnnounce
					.getMediaDescriptions(true).get(0));
			Media outgoingMedia = outgoingMediaDescription.getMedia();

			// This the place where this RFSS is goint to be recieving IP
			// PACKETS.

			outgoingMedia.setMediaPort(session.getMyRtpRecvPort());

			sdpAnnounce.getOrigin().setAddress(myIpAddress);

			session.setSessionDescription(sdpAnnounce);
		} catch (Exception ex) {
			logger.fatal("Unexpected exception ", ex);
		}

	}
	/**
	 * A utility function that fixes up the SDP announce to match the port of the PttSession.
	 * 
	 * @param session --
	 *           	ptt session mux from which to extract parameters.
	 *           
	 * @param sdpContent --
	 *            sdpContent to fixup.
	 * 
	 * @throws Exception
	 */
	public static void fixupSdpAnnounce(RfssConfig rfss, PttSessionMultiplexer session,
			SessionDescription sdpAnnounce) {

		// Addres where the current RFSS
		// is expected to be receiving IP packets.
		try {
			String myIpAddress = rfss.getIpAddress();

			// TransmissionControlManager.addMediaLeg(session, sdpAnnounce) ;

			Connection connection = sdpAnnounce.getConnection();
			connection.setAddress(myIpAddress);

			MediaDescription outgoingMediaDescription = (MediaDescription) (sdpAnnounce
					.getMediaDescriptions(true).get(0));
			Media outgoingMedia = outgoingMediaDescription.getMedia();

			// This the place where this RFSS is goint to be recieving IP
			// PACKETS.

			outgoingMedia.setMediaPort(session.getRtpSession().getMyRtpRecvPort());

			sdpAnnounce.getOrigin().setAddress(myIpAddress);

			session.setSessionDescription(sdpAnnounce);
		} catch (Exception ex) {
			logger.fatal("Unexpected exception ", ex);
		}

	}
	public CallControlManager(RFSS rfss) {
		try {
			this.provider = rfss.getProvider();
			this.rfssConfig = rfss.getRfssConfig();
			ListeningPoint listeningPoint = this.provider
					.getListeningPoint("udp");
			logger.info("setSentBy " + this.rfssConfig.getDomainName());
			listeningPoint.setSentBy(this.rfssConfig.getDomainName());

			this.unitToUnitCallControlManager = new UnitToUnitCallControlManager(
					rfss);
			this.groupCallControlManager = new GroupCallControlManager(
					provider, rfssConfig);
		} catch (Exception ex) {
			logger.fatal("Unexpected exception", ex);
			this.rfssConfig.getRFSS().getTestHarness().fail(
					"unexpected exception ", ex);
		}
	}

	private boolean isU2UCall(Message message) {
		ToHeader toHeader = (ToHeader) message.getHeader(ToHeader.NAME);
		String user = ((SipURI) toHeader.getAddress().getURI())
				.getParameter("user");
		return (user != null && user.equals(ISSIConstants.TIA_P25_SU));
	}

	private boolean isGroupCall(Message message) {
		ToHeader toHeader = (ToHeader) message.getHeader(ToHeader.NAME);
		String user = ((SipURI) toHeader.getAddress().getURI())
				.getParameter("user");
		return (user != null && user.equals(ISSIConstants.TIA_P25_SG));
	}

	public void processRequest(RequestEvent requestEvent) {
		Request request = requestEvent.getRequest();
		if (isU2UCall(request))
			this.unitToUnitCallControlManager.processRequest(requestEvent);
		else if (isGroupCall(request))
			this.groupCallControlManager.processRequest(requestEvent);
		else
			this.rfssConfig.getRFSS().getTestHarness()
					.fail("Unknown call type");

	}

	public void processResponse(ResponseEvent responseEvent) {
		Response response = responseEvent.getResponse();
		if (isU2UCall(response))
			this.unitToUnitCallControlManager.processResponse(responseEvent);
		else if (isGroupCall(response))
			this.groupCallControlManager.processResponse(responseEvent);
		else
			this.rfssConfig.getRFSS().getTestHarness()
					.fail("Unknown call type");

	}

	public void processTimeout(TimeoutEvent timeoutEvent) {
		if (timeoutEvent.isServerTransaction()) {
			ServerTransaction st = timeoutEvent.getServerTransaction();
			Request request = st.getRequest();
			if (isU2UCall(request)) {
				this.unitToUnitCallControlManager.processTimeout(timeoutEvent);
			} else if (isGroupCall(request)) {
				this.groupCallControlManager.processTimeout(timeoutEvent);
			} else {
				this.rfssConfig.getRFSS().getTestHarness().fail(
						"Unknown call type");
			}
		} else {
			ClientTransaction ct = timeoutEvent.getClientTransaction();
			Request request = ct.getRequest();
			if (isU2UCall(request)) {
				this.unitToUnitCallControlManager.processTimeout(timeoutEvent);
			} else if (isGroupCall(request)) {
				this.groupCallControlManager.processTimeout(timeoutEvent);
			} else {
				this.rfssConfig.getRFSS().getTestHarness().fail(
						"Unknown call type");
			}
		}

	}

	/*
	 * (non-Javadoc)
	 * @see javax.sip.SipListener#processIOException(javax.sip.IOExceptionEvent)
	 */
	public void processIOException(IOExceptionEvent arg0) {
		this.rfssConfig.getRFSS().getTestHarness().fail("unexpected event");

	}
	
	/*
	 * (non-Javadoc)
	 * @see javax.sip.SipListener#processTransactionTerminated(javax.sip.TransactionTerminatedEvent)
	 */

	public void processTransactionTerminated(
			TransactionTerminatedEvent transactionTerminated) {
		if (transactionTerminated.isServerTransaction()) {
			ServerTransaction st = transactionTerminated.getServerTransaction();
			Request request = st.getRequest();
			if (isU2UCall(request)) {
				this.unitToUnitCallControlManager
						.processTransactionTerminated(transactionTerminated);
			} else if (isGroupCall(request)) {
				this.groupCallControlManager
						.processTransactionTerminated(transactionTerminated);
			} else {
				this.rfssConfig.getRFSS().getTestHarness().fail(
						"Unknown call type");
			}
		} else {
			ClientTransaction ct = transactionTerminated.getClientTransaction();
			Request request = ct.getRequest();
			if (isU2UCall(request)) {
				this.unitToUnitCallControlManager
						.processTransactionTerminated(transactionTerminated);
			} else if (isGroupCall(request)) {
				this.groupCallControlManager
						.processTransactionTerminated(transactionTerminated);
			} else {
				this.rfssConfig.getRFSS().getTestHarness().fail(
						"Unknown call type");
			}
		}

	}

	/*
	 * (non-Javadoc)
	 * @see javax.sip.SipListener#processDialogTerminated(javax.sip.DialogTerminatedEvent)
	 */
	public void processDialogTerminated(DialogTerminatedEvent dte) {
		if (dte.getDialog().getApplicationData() instanceof UnitToUnitCall) {
			this.unitToUnitCallControlManager.processDialogTerminated(dte);
		} else {
			this.groupCallControlManager.processDialogTerminated(dte);
		}

	}

	public UnitToUnitCallControlManager getUnitToUnitCallControlManager() {

		return this.unitToUnitCallControlManager;
	}

	public GroupCallControlManager getGroupCallControlManager() {

		return this.groupCallControlManager;
	}

	

   void alertResourceChange(boolean resourceVal) throws Exception {
		this.unitToUnitCallControlManager.alertRfResourceChange(resourceVal);
		this.groupCallControlManager.alertRfResourceChange(resourceVal);
		
	}
	
	void alertGroupPriorityChange ( String groupRadicalName, int newPriority , boolean isEmergency )
		throws Exception{
		this.groupCallControlManager.groupPriorityChange(groupRadicalName, newPriority, isEmergency);
	}

}
