package gov.nist.p25.issi.rfss;

import gov.nist.javax.sdp.fields.OriginField;
import gov.nist.p25.issi.issiconfig.ISSIConstants;
import gov.nist.p25.issi.issiconfig.RfssConfig;
import gov.nist.p25.issi.issiconfig.SuConfig;
import gov.nist.p25.issi.issiconfig.TopologyConfig;
import gov.nist.p25.issi.p25body.CallParamContent;
import gov.nist.p25.issi.p25body.ContentList;
import gov.nist.p25.issi.p25body.SdpContent;
import gov.nist.p25.issi.p25body.params.CallParam;
import gov.nist.p25.issi.p25body.serviceprofile.AvailabilityCheckType;
import gov.nist.p25.issi.p25body.serviceprofile.CallSetupPreferenceType;
import gov.nist.p25.issi.p25body.serviceprofile.DuplexityType;
import gov.nist.p25.issi.p25body.serviceprofile.user.UserServiceProfile;
import gov.nist.p25.issi.p25payload.TransmitPriorityType;
import gov.nist.p25.issi.rfss.tester.GroupCallSetupScenario;
import gov.nist.p25.issi.rfss.tester.SuToSuCallSetupScenario;
import gov.nist.p25.issi.transctlmgr.PttPointToMultipointSession;
import gov.nist.p25.issi.transctlmgr.TransmissionControlManager;
import gov.nist.p25.issi.transctlmgr.ptt.LinkType;
import gov.nist.p25.issi.transctlmgr.ptt.PttSession;
import gov.nist.p25.issi.transctlmgr.ptt.SessionType;
import gov.nist.p25.issi.transctlmgr.ptt.SmfSession;
import gov.nist.p25.issi.utils.SipUtils;
import gov.nist.p25.issi.utils.WarningCodes;
import gov.nist.rtp.RtpException;

import java.util.Hashtable;

import javax.sdp.SessionDescription;
import javax.sip.ClientTransaction;
import javax.sip.Dialog;
import javax.sip.RequestEvent;
import javax.sip.ResponseEvent;
import javax.sip.ServerTransaction;
import javax.sip.SipException;
import javax.sip.SipProvider;
import javax.sip.TransactionState;
import javax.sip.address.Address;
import javax.sip.address.SipURI;
import javax.sip.header.CSeqHeader;
import javax.sip.header.CallIdHeader;
import javax.sip.header.ContactHeader;
import javax.sip.header.ContentTypeHeader;
import javax.sip.header.ToHeader;
import javax.sip.message.Request;
import javax.sip.message.Response;

import org.apache.log4j.Logger;

/**
 * Implementation of the Call Control SAP abstraction for
 * the emulated rfss.
 * 
 * @author M. Ranganathan
 *
 */
public class CallControlSAP {

	/* The subscriber unit to which we are attached */
	final SuConfig suConfig;

	final UnitToUnitCallControlManager unitToUnitCallControlManager;

	final GroupCallControlManager groupCallControlManager;

	final TransmissionControlManager transmissionControlManager;
	
	private ServiceAccessPoints saps;

	
	final RfssConfig rfssConfig;

	final TopologyConfig topologyConfig;

	private static Logger logger = Logger.getLogger(CallControlSAP.class);

	Hashtable<String, PttSession> pendingRtpSessions;

	CallControlSAP(RFSS rfss, SuConfig suConfig) {
		this.suConfig = suConfig;
		
		this.rfssConfig = rfss.getRfssConfig();
		this.topologyConfig = this.rfssConfig.getSysConfig()
				.getTopologyConfig();
		this.unitToUnitCallControlManager = rfss.getCallControlManager()
				.getUnitToUnitCallControlManager();
		this.groupCallControlManager = rfss.getCallControlManager()
				.getGroupCallControlManager();
		this.transmissionControlManager = rfss.getTransmissionControlManager();
		this.pendingRtpSessions = new Hashtable<String, PttSession>();

	}
	
	
	public void setSaps( ServiceAccessPoints saps) {
		this.saps = saps;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see gov.nist.p25.issi.rfss.CallControlSAP#ccSendCallSetupRequest(gov.nist.p25.issi.issiconfig.SuConfig)
	 */
	public UnitToUnitCall ccSendCallSetupRequest(
			SuToSuCallSetupScenario callSetupScenario) throws Exception {

		SuConfig peerSU = callSetupScenario.getCalledSuConfig();

		boolean isEmergency = callSetupScenario.isEmergency();
	
		logger.debug("ccSendCallSetupRequest: protection = " + callSetupScenario.isProtectedCall());

		SipProvider provider = this.unitToUnitCallControlManager.getProvider();

		CallIdHeader callId = provider.getNewCallId();

		PttPointToMultipointSession ptpSession = this.transmissionControlManager
				.getPttPointToMultipointSession(callId.getCallId(),
						SessionType.CALLING_SERVING);

		// the following should be random
		long sessId = (long) (Math.random() * 100000);
		long sessVersion = (long) (Math.random() * 100000);
//		 we are setting up the side of the RTP session
		// where we are going to be listening for incoming packets.
		// this is where we setup the initial SDP OFFER. We attach
		// it to the SIP INIVTE. Note in that we do not know
		// where the next hop is listening at this point. We will
		// know that only when the answer comes back.

		SmfSession smfSession = ptpSession.createSmfSession(
				LinkType.UNIT_TO_UNIT_CALLING_SERVING_TO_CALLING_HOME, new Long(sessId).toString());
		smfSession.setRemoteRfssDomainName(suConfig.getHomeRfss().getDomainName());
		
		// In this case the current test case becomes the listener
		// becuase we want to be able to trigger on message match.
		
		smfSession.getSmfReceiver().addListener(this.rfssConfig.getRFSS().getCurrentTestCase());

		int port = smfSession.getMyRtpRecvPort();

		this.pendingRtpSessions.put(new Long(sessId).toString(), smfSession);
	
		String ipAddress = suConfig.getHomeRfss().getIpAddress();
		
		String callIdSuffix = callSetupScenario.getId();
		
		Request inviteRequest = unitToUnitCallControlManager
			.createInviteForU2UCall( callSetupScenario) ;
		     
		String sdpBody = "v=0\r\n" + "o=- " + sessId + " " + sessVersion + " IN IP4 " + ipAddress
				+ "\r\n" + "s=TIA-P25-U2UCall\r\n" + "t=0 0\r\n"
				+ "c=IN IP4 " + ipAddress + "\r\n" + "m=audio " + port
				+ " RTP/AVP 100\r\n" + "a=rtpmap:100 X-TIA-P25-IMBE/8000\r\n";

		
		inviteRequest.setHeader(callId);

		ContentList contentList = new ContentList();

		CallParamContent callParamContent = CallParamContent
				.createCallParamContent();
		CallParam callParam = callParamContent.getCallParam();
		callParam.setFullDuplexRequested(callSetupScenario.isFullDuplexCall());
		callParam.setProtectedMode(callSetupScenario.isProtectedCall());
		
		SuConfig callingSuConfig = callSetupScenario.getCallingSuConfig();
		UserServiceProfile userProfile = callingSuConfig
				.getUserServiceProfile();
		CallSetupPreferenceType csetupPref = userProfile
				.getCallSetupPreference().getCallSetupPreferenceType();
		callParam
				.setAvailabilityCheckRequested(csetupPref
						.equals(CallSetupPreferenceType.CALLER_PREFERS_AVAILABILITY_CHECK));
		callParam.setRfResourceAvailable(this.rfssConfig.getRFSS()
				.isRfResourcesAvailable());
		SdpContent sdpContent = SdpContent.createSdpContent(sdpBody);
		smfSession.setSessionDescription(sdpContent.getSessionDescription());
		contentList.add(sdpContent);
		contentList.add(callParamContent);

		ContentTypeHeader contentTypeHeader = contentList.getContentTypeHeader();

		inviteRequest.setHeader(contentTypeHeader);
		inviteRequest.setContent(contentList.toString(), contentTypeHeader);

		UnitToUnitCall unitToUnitCall = this.unitToUnitCallControlManager
				.setupRequest(inviteRequest, this.suConfig.getSU(), peerSU);

		unitToUnitCall.setCallParam(callParamContent);
		unitToUnitCall.setEmergency(isEmergency);
		unitToUnitCall.setPttSession(smfSession);
		unitToUnitCall.setPointToMultipointSession(ptpSession);
		

		return unitToUnitCall;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see gov.nist.p25.issi.rfss.CallControlSAP#ccSetupIndicateResponse(gov.nist.p25.issi.rfss.CallSetupRequestEvent,
	 *      int)
	 */
	public void ccSetupIndicateResponse(
			CallSetupRequestEvent callSetupRequestEvent, int responseCode)
			throws Exception {
		ServerTransaction st = callSetupRequestEvent.getServerTransaction();
		if (st.getState() != TransactionState.PROCEEDING) {
			logger.debug("Too late to send the response... returning");
			return;
		}
		RequestEvent requestEvent = callSetupRequestEvent.getRequestEvent();
		Response response = UnitToUnitCallControlManager.messageFactory
				.createResponse(responseCode, requestEvent.getRequest());
		ContactHeader contactHeader = SipUtils.createContactHeaderForSU(
				this.rfssConfig, this.suConfig);
		response.setHeader(contactHeader);
		ToHeader toHeader = (ToHeader) response.getHeader(ToHeader.NAME);
		// Tag for to header is mandatory in 2xx response.
		if (response.getStatusCode() / 100 == 2) {
			String toTag = SipUtils.createTag();
			toHeader.setTag(toTag);
		}

		SipURI toUri = (SipURI) toHeader.getAddress().getURI();
		String radicalname = SipUtils.getRadicalName(toUri);
		SuConfig calledSu = this.topologyConfig.getSuConfig(radicalname);
		if (calledSu == null) {
			throw new Exception("Could not find called su");
		}

		UserServiceProfile usp = calledSu.getUserServiceProfile();

		// need to add additional information here.
		// TransmissionControlSAP tcSAP = this.mySubscriberUnit
		// .getTransmissionControlSAP();

		// If responding with a Success responseCode. Then create
		// a new RTP Session for this SU if needed.
		// Modify the response SDP Body to reflect the IP address
		// and port of the new session we just created and
		// send the response back.

		if (responseCode == Response.OK) {

			ContentList contentList = ContentList
					.getContentListFromMessage(requestEvent.getRequest());
			SdpContent sdpContent = (SdpContent) contentList.getContent(
					"application", "sdp");
			SessionDescription sdpAnnounce = (SessionDescription) sdpContent
					.getSessionDescription();

			// The SU is answering with a SIP OK. At this point,
			// we need to allocate a new RTP Session (i.e. a PTT session),
			// from which we extract the IP address and port and send it
			// in the OK.

			String sessionId = SipUtils.getCallIdFromMessage(response);
			PttPointToMultipointSession session = this.transmissionControlManager
					.getPttPointToMultipointSession(sessionId,SessionType.CALLED_SERVING);
			SmfSession smfSession = null;
			try {
			
				smfSession = session.createSmfSession(
						LinkType.UNIT_TO_UNIT_CALLED_SERVING_TO_CALLED_HOME,
						((OriginField)sdpAnnounce.getOrigin()).getSessIdAsString());
				smfSession.setRemoteRfssDomainName(calledSu.getHomeRfss().getDomainName());
				smfSession.setSessionDescription(sdpAnnounce);
				smfSession.getSmfReceiver().addListener(this.rfssConfig.getRFSS().getCurrentTestCase());
			} catch (RtpException ex) {
				Response errorResponse = SipUtils.createErrorResponse(
						this.rfssConfig, requestEvent.getRequest(),
						WarningCodes.NO_RTP_RESOURCE);
				st.sendResponse(errorResponse);
				return;

			}
			
			String toHost = SipUtils.getContactHost(callSetupRequestEvent.getServerTransaction().getRequest());
			transmissionControlManager.addMediaLeg(0, smfSession, sdpAnnounce,toHost);

			
			saps.getTransmissionControlSAP().setPttSession(smfSession);

			CallControlManager.fixupSdpAnnounce(this.rfssConfig, smfSession,
					sdpAnnounce);
			ContentTypeHeader contentType = (ContentTypeHeader) requestEvent
					.getRequest().getHeader(ContentTypeHeader.NAME);

			// logger.debug("contentList = " + contentList.toString());
			CallParamContent callparams = contentList.getCallParamContent();

			CallParam callparam = null;
			if (callparams != null)
				callparam = callparams.getCallParam();
			callparam.setCallingSuInitialTransmitter(null);
			if (callparam != null && callparam.isAvailabilityCheckRequested()) {
				logger.debug("availcheckrequested");
				logger
						.debug("availchektype = "
								+ usp.getAvailabilityCheck()
										.getAvailabilityCheckType());

				if ((!usp.getAvailabilityCheck().getAvailabilityCheckType()
						.equals(AvailabilityCheckType.AVAIL_CHECK_ONLY))
						&& !usp
								.getAvailabilityCheck()
								.getAvailabilityCheckType()
								.equals(
										AvailabilityCheckType.AVAIL_CHECK_AND_DIRECT_CALL)) {
					logger.debug("setAvailabilityCheckRequested: false");
					callparam.setAvailabilityCheckRequested(false);

				} else {
					callparam.setAvailabilityCheckRequested(true);

				}
			}

			if (callparam != null && callparam.isFullDuplexRequested()) {
				if (usp.getDuplexity().getDuplex().equals(DuplexityType.FULL)) {
					callparam.setFullDuplexRequested(true);
				} else {
					callparam.setFullDuplexRequested(false);
				}
			}

			String origin = ((OriginField) sdpAnnounce.getOrigin()).getSessIdAsString();

			this.pendingRtpSessions.put(origin, smfSession);

			// this.transmissionControlManager.addRtpListener(tcSap);

			response.setContent(contentList.toString(), contentList.getContentTypeHeader());

			Dialog dialog = st.getDialog();
			UnitToUnitCall unitToUnitCall = (UnitToUnitCall) dialog.getApplicationData();
			unitToUnitCall.setPointToMultipointSession(session);
			smfSession.getHeartbeatReceiver().setHeartbeatListener(unitToUnitCall);

		}
		this.unitToUnitCallControlManager.sendResponse(callSetupRequestEvent,
				response);

		if (logger.isDebugEnabled()) {
			logger
					.debug("Here are the sessions after response from called rfss: "
							+ this.transmissionControlManager.toString());
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see gov.nist.p25.issi.rfss.CallControlSAP#ccSetupConfirm(gov.nist.p25.issi.rfss.CallSetupResponseEvent)
	 */
	public void ccSetupConfirm(CallControlResponseEvent ccResponseEvent)
			throws Exception {
		ResponseEvent responseEvent = ccResponseEvent.getResponseEvent();
		ClientTransaction ct = responseEvent.getClientTransaction();
		Dialog dialog = ct.getDialog();
		Response response = responseEvent.getResponse();
		CSeqHeader cseqHeader = (CSeqHeader) response
				.getHeader(CSeqHeader.NAME);
		Request ackRequest = dialog.createAck(cseqHeader.getSeqNumber());
		dialog.sendAck(ackRequest);

		SessionDescription sdes = ContentList.getContentListFromMessage(
				response).getSdpContent().getSessionDescription();

		if (sdes == null) {

			logger
					.error("No session description in the response - dropping response");
			throw new Exception(
					"No session description in the response - dropping response");

		}
		// The sender preserves the origin field.

		// mranga -- got an OK from the invitee.
		// now extract the session that we originally stored away
		// when we sent out the invite. One half of the session
		// has already been set up. The OK Contains the SDP answer.
		// we extract the port from the SDP answer.
		String origin = ((OriginField)sdes.getOrigin()).getSessIdAsString();
		PttSession pendingSession = this.pendingRtpSessions.get(origin);

		// PttSession pendingSession = (PttSession) dialog.getApplicationData();

		if (pendingSession == null) {

			logger
					.error("Could not find pending session -- dropping the response");
			throw new Exception(
					"Could not find pending session - sdp corruption");

		}

		String  host = SipUtils.getContactHost(response);
		ToHeader toHeader = (ToHeader) response.getHeader(ToHeader.NAME);
		if (((SipURI) (toHeader.getAddress().getURI())).getUserParam().equals(
				ISSIConstants.TIA_P25_SG)) {
			String groupId = SipUtils.getGroupID((SipURI) toHeader.getAddress()
					.getURI());
			long groupIdLong = Long.parseLong(groupId, 16);
			transmissionControlManager.addMediaLeg(groupIdLong, pendingSession,
					sdes,host);
		} else {
			transmissionControlManager.addMediaLeg(0, pendingSession, sdes,host);
		}

		pendingSession.getHeartbeatTransmitter().start();
		
		saps.getTransmissionControlSAP().setPttSession(pendingSession);
		// this.transmissionControlManager.addRtpListener(this.tcSap);

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see gov.nist.p25.issi.rfss.CallControlSAP#ccTeardownCallSegment(gov.nist.p25.issi.rfss.CallSegment)
	 */
	public void ccTeardownCallSegment(HomeAgentInterface agent, UnitToUnitCall unitToUnitCall,
			String radicalName) throws Exception {

		this.unitToUnitCallControlManager.teardownRequest(agent, unitToUnitCall,
				radicalName);

	}

	
	/*
	 * (non-Javadoc)
	 * @see gov.nist.p25.issi.rfss.CallControlSAP#ccSendCallSetupInCallRoamingRequest(gov.nist.p25.issi.rfss.UnitToUnitCall, java.lang.String, javax.sip.address.Address)
	 */

	public void ccSendCallSetupInCallRoamingRequest(SuConfig roamingSu,
			UnitToUnitCall unitToUnitCall,
			String originalContact,
			Address newContact) throws Exception {

		if ( logger.isDebugEnabled()) {
			logger.debug("ccSendCallSetupInCallRoamingRequest: newContact  = "
				+ newContact + " originalContact  = " + originalContact);
		}
		
		PttPointToMultipointSession ptpSession = unitToUnitCall
				.getPointToMultipointSession();

		PttSession rtpSession = ptpSession.getPttSession(originalContact);
		
		if ( logger.isDebugEnabled()) {
			ptpSession.printPeerTable();
		}
		
		int port = rtpSession.getMyRtpRecvPort();
			// the following should be random
		long origin = (long) (Math.random() * 100000);

		this.pendingRtpSessions.put(new Long(origin).toString(), rtpSession);
		String ipAddress = suConfig.getHomeRfss().getIpAddress();

		String sdpBody = "v=0\r\n" + "o=- " + origin + " 0 IN IP4 " + ipAddress
				+ "\r\n" + "s=TIA-P25-U2UCall\r\n" + "t=0 0\r\n"
				+ "c=IN IP4 " + ipAddress + "\r\n" + "m=audio " + port
				+ " RTP/AVP 100\r\n" + "a=rtpmap:100 X-TIA-P25-IMBE/8000\r\n";
		
	
	
		
		Request inviteRequest = unitToUnitCallControlManager.createInviteForInCallRoaming(roamingSu,unitToUnitCall);
		SipURI sipUri = (SipURI) this.rfssConfig.getRFSS()
		.getUnitToUnitMobilityManager().getRegisteredContactURI(
				(SipURI)inviteRequest.getRequestURI());
		rtpSession.setRemoteRfssDomainName(sipUri.getHost());
		
		
	
		if (unitToUnitCall.getCallParamContent() == null) {
			SdpContent sdpContent = SdpContent.createSdpContent(sdpBody);
			CallParamContent callParamContent = CallParamContent
					.createCallParamContent();
			callParamContent.getCallParam().setIncallRoaming(true);
			boolean isCallingSuInitrans = unitToUnitCall.isCallingSuInitrans();
			callParamContent.getCallParam().setCallingSuInitialTransmitter(isCallingSuInitrans);
			ContentList contentList = new ContentList();
			contentList.add(sdpContent);
			contentList.add(callParamContent);
			inviteRequest.setContent(contentList.toString(),
					contentList.getContentTypeHeader());
		} else {
			ContentList contentList = new ContentList();
			contentList.add(SdpContent.createSdpContent(sdpBody));
			contentList.add(unitToUnitCall.getCallParamContent());
			contentList.getCallParamContent().getCallParam().setIncallRoaming(
					true);
			boolean isCallingSuInitrans = unitToUnitCall.isCallingSuInitrans();
			contentList.getCallParamContent().getCallParam().setCallingSuInitialTransmitter(isCallingSuInitrans);
			
			inviteRequest.setContent(contentList.toString(),
					contentList.getContentTypeHeader());
		}
		
		ContactHeader chdr = SipUtils.createContactHeaderForRfss(this.rfssConfig, ISSIConstants.TIA_P25_SU);
		inviteRequest.setHeader(chdr);

		unitToUnitCallControlManager.setupInCallRoamingRequest(inviteRequest, unitToUnitCall,
				this.suConfig.getSU());

	}

	/*
	 * (non-Javadoc)
	 * @see gov.nist.p25.issi.rfss.CallControlSAP#triggerGroupCallSetup(gov.nist.p25.issi.rfss.tester.GroupCallSetupScenario)
	 */
	public GroupServing triggerGroupCallSetup(GroupCallSetupScenario groupCallSetupScenario) throws Exception {
		
			return this.groupCallControlManager.subscribeToGroup(groupCallSetupScenario);
			
	}

	public RfssConfig getRfssConfig() {

		return rfssConfig;
	}

	public void ccCancelCallSegment(UnitToUnitCall unitToUnitCall) throws Exception {
		ClientTransaction ct = unitToUnitCall.getClientTransaction();
		if (ct.getState() == TransactionState.CONFIRMED
				|| ct.getState() == TransactionState.TERMINATED) {
			throw new Exception("Cannot cancel the request -- too late");

		}
		unitToUnitCallControlManager.sendCancel(ct);

	}
	
	public void ccCancelGroupCall( GroupServing groupServing ) throws Exception {
		ClientTransaction ct = groupServing.getClientTransaction();
		if ( groupServing.getGroupConfig().getHomeRfss() == this.getRfssConfig() ) {
			groupCallControlManager.sendCancel(groupServing);
		} else if ( ct.getState() == TransactionState.CONFIRMED ||
					ct.getState() == TransactionState.TERMINATED )  {
			throw new Exception ("Cannot cancel the request -- too late");
		} else {
			groupCallControlManager.sendCancel(groupServing);
		}
		
	}

	public void ccCancelResponse(ServerTransaction serverTransaction)
			throws Exception {
		if (serverTransaction.getState() == TransactionState.PROCEEDING) {
			unitToUnitCallControlManager.sendResponse(serverTransaction,
					Response.OK);
		} else {
			throw new SipException("Too late to cancel the request");
		}

	}

	public void ccSetupIndicateErrorResponse(
			CallSetupRequestEvent callSetupRequestEvent,
			WarningCodes warningCode) throws Exception {
		ServerTransaction st = callSetupRequestEvent.getServerTransaction();
		Response response = SipUtils.createErrorResponse(rfssConfig, st
				.getRequest(), warningCode);
		this.unitToUnitCallControlManager.sendResponse(callSetupRequestEvent,
				response);

	}

}
