package gov.nist.p25.issi.rfss;

import gov.nist.javax.sdp.fields.OriginField;
import gov.nist.p25.issi.ISSITimer;
import gov.nist.p25.issi.issiconfig.CProtectedDisposition;
import gov.nist.p25.issi.issiconfig.GroupConfig;
import gov.nist.p25.issi.issiconfig.ISSIConstants;
import gov.nist.p25.issi.issiconfig.RfssConfig;
import gov.nist.p25.issi.issiconfig.SuConfig;
import gov.nist.p25.issi.issiconfig.TopologyConfig;
import gov.nist.p25.issi.p25body.CallParamContent;
import gov.nist.p25.issi.p25body.ContentList;
import gov.nist.p25.issi.p25body.SdpContent;
import gov.nist.p25.issi.p25body.params.CallParam;
import gov.nist.p25.issi.p25body.serviceprofile.group.ConfirmedCallSetupTime;
import gov.nist.p25.issi.rfss.tester.AbstractSipSignalingTest;
import gov.nist.p25.issi.rfss.tester.GroupCallSetupScenario;
import gov.nist.p25.issi.rfss.tester.TestSU;
import gov.nist.p25.issi.testlauncher.TestHarness;
import gov.nist.p25.issi.transctlmgr.PttPointToMultipointSession;
import gov.nist.p25.issi.transctlmgr.TransmissionControlManager;
import gov.nist.p25.issi.transctlmgr.ptt.LinkType;
import gov.nist.p25.issi.transctlmgr.ptt.MmfSession;
import gov.nist.p25.issi.transctlmgr.ptt.PttManager;
import gov.nist.p25.issi.transctlmgr.ptt.PttSession;
import gov.nist.p25.issi.transctlmgr.ptt.PttSessionInterface;
import gov.nist.p25.issi.transctlmgr.ptt.PttSessionMultiplexer;
import gov.nist.p25.issi.transctlmgr.ptt.SessionType;
import gov.nist.p25.issi.transctlmgr.ptt.SmfSession;
import gov.nist.p25.issi.utils.SipUtils;
import gov.nist.p25.issi.utils.WarningCodes;
import gov.nist.rtp.RtpException;
import gov.nist.rtp.RtpSession;

import java.net.SocketException;
import java.util.Collection;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.ConcurrentHashMap;

import javax.sdp.MediaDescription;
import javax.sdp.SdpFactory;
import javax.sdp.SessionDescription;
import javax.sip.ClientTransaction;
import javax.sip.Dialog;
import javax.sip.DialogTerminatedEvent;
import javax.sip.IOExceptionEvent;
import javax.sip.InvalidArgumentException;
import javax.sip.RequestEvent;
import javax.sip.ResponseEvent;
import javax.sip.ServerTransaction;
import javax.sip.SipException;
import javax.sip.SipListener;
import javax.sip.SipProvider;
import javax.sip.TimeoutEvent;
import javax.sip.TransactionAlreadyExistsException;
import javax.sip.TransactionState;
import javax.sip.TransactionTerminatedEvent;
import javax.sip.TransactionUnavailableException;
import javax.sip.address.Address;
import javax.sip.address.AddressFactory;
import javax.sip.address.SipURI;
import javax.sip.header.AcceptHeader;
import javax.sip.header.CSeqHeader;
import javax.sip.header.CallIdHeader;
import javax.sip.header.ContactHeader;
import javax.sip.header.ContentDispositionHeader;
import javax.sip.header.ContentTypeHeader;
import javax.sip.header.FromHeader;
import javax.sip.header.HeaderFactory;
import javax.sip.header.MaxForwardsHeader;
import javax.sip.header.MimeVersionHeader;
import javax.sip.header.PriorityHeader;
import javax.sip.header.RetryAfterHeader;
import javax.sip.header.RouteHeader;
import javax.sip.header.TimeStampHeader;
import javax.sip.header.ToHeader;
import javax.sip.header.ViaHeader;
import javax.sip.header.WarningHeader;
import javax.sip.message.MessageFactory;
import javax.sip.message.Request;
import javax.sip.message.Response;

import org.apache.log4j.Logger;

/**
 * The group call control manager. This performs the Group Call Control function
 * for the emulator. The GCCF of the serving RFSS is responsible for the
 * following functions:
 * <ul>
 * <li> - Requesting a group call set up when an SU triggers a PTT request in
 * its area for a group for which no call currently exists
 * <li> - Participating in the establishment of new call segments in response to
 * requests from the home RFSS
 * <li> - Releasing a group call segment with the Home RFSS on loss of RTP
 * connectivity or other local policy condition
 * <li> - Once a group call segment has been set up, tearing down or setting up
 * RTP sessions upon preemption or Home RFSS request.
 * <li> - Communicating to the home RFSS the RTP session parameters
 * <li> - Notifying the home RFSS of the availability or lack of availability of
 * RF resources required for a confirmed call
 * <li> - Requesting a change of group call priority based on local policy
 * </ul>
 * 
 * The GCCF of the Home RFSS is responsible for the following functions:
 * 
 * <ul>
 * <li> - Establishing group call segments in response to group call setup
 * requests from serving RFSSs or SU PTT triggers within its own area,
 * including:
 * <li> - to get the list of registered serving RFSSs from the GMMF
 * <li> - to initiate group call set ups with the registered serving RFSSs
 * <li> - to bridge RTP sessions which are set up with the serving RFSSs
 * <li> - On-going management of group call segments during a call, including
 * <li> - To initiate group call segment set ups with any registered serving
 * RFSSs not already joined to the call (late call entry)
 * <li> - To handle a group call segment set up when it receives a group call
 * set up request from any serving RFSS
 * <li> - Once group call segment has been set up, tearing down or setting up
 * RTP sessions upon preemption or serving RFSS request.
 * <li> - Releasing the group call on ISSI hangtime expiry or other local policy
 * condition
 * <li> - Releasing a group call segment with a serving RFSS on loss of RTP
 * connectivity or other local policy condition
 * <li> - Communicating to each serving RFSS the RTP session parameters
 * associated with its call segment
 * <li> - Managing the establishment and maintenance of confirmed group calls
 * <li> - Changing the group call priority upon serving RFSS request or upon its
 * own local policy.
 * </ul>
 * 
 * 
 * @author M. Ranganathan
 * 
 */
public class GroupCallControlManager implements SipListener {

	private Hashtable<String, GroupInviteServingDelayTask> pendingTimers = new Hashtable<String, GroupInviteServingDelayTask>();

	/*
	 * Provider from where we will allocate transactions and get requests (this
	 * is our event source).
	 */
	private SipProvider provider;

	/*
	 * Configuration for our RFSS
	 */
	private RfssConfig rfssConfig;

	/*
	 * System topology configuration.
	 */
	private TopologyConfig topologyConfig;

	/*
	 * Factories for the SIP stack.
	 */
	private static AddressFactory addressFactory = ProtocolObjects.addressFactory;

	private static HeaderFactory headerFactory = ProtocolObjects.headerFactory;

	private static MessageFactory messageFactory = ProtocolObjects.messageFactory;

	/*
	 * The Currently active group calls we are fielding.
	 */
	private ConcurrentHashMap<String, GroupServing> activeGroupCalls;

	/*
	 * Sessions that are pending a respose (this is a temporary holding place)
	 */
	private static Logger logger = Logger
			.getLogger(GroupCallControlManager.class);

	private RFSS rfss;

	enum Operation { REINVITE, CALL_SETUP };
	
	class PendingResponse {

		Operation procedure;

		GroupServing callSegment;

		GroupCallSetupScenario scenario;

	}

	/**
	 * 
	 * The task that simulates a delay in invite processing so we can generate
	 * call collisions.
	 * 
	 */
	class GroupInviteServingDelayTask extends TimerTask {
		private ServerTransaction serverTransaction;

		private RequestEvent requestEvent;

		public GroupInviteServingDelayTask(RequestEvent requestEvent,
				ServerTransaction serverTransaction) {
			this.serverTransaction = serverTransaction;
			this.requestEvent = requestEvent;
		}

		@Override
		public void run() {
			try {
				pendingTimers.remove(serverTransaction.getBranchId());
				if (serverTransaction.getState() != TransactionState.TERMINATED)
					GroupCallControlManager.this.actuallyProcessRequest(
							requestEvent, serverTransaction);
			} catch (Exception ex) {
				rfss.getTestHarness().fail("Unexpected exception", ex);
			}

		}

		/**
		 * @return the serverTransaction
		 */
		public ServerTransaction getServerTransaction() {
			return serverTransaction;
		}

	}

	class GroupInviteHomeDelayTask extends TimerTask {

		private RequestEvent requestEvent;

		private GroupHome groupHome;

		private ServerTransaction st;

		public GroupInviteHomeDelayTask(RequestEvent requestEvent,
				GroupHome groupHome, ServerTransaction st) {
			this.requestEvent = requestEvent;
			this.groupHome = groupHome;
			this.st = st;
		}

		@Override
		public void run() {
			logger.debug("Done delaying the INVITE -- actually processing it!");
			Request request = requestEvent.getRequest();
			ToHeader toHeader = ((ToHeader) request.getHeader(ToHeader.NAME));
			SipURI toUri = ((SipURI) toHeader.getAddress().getURI());
			String groupId = SipUtils.getRadicalName(toUri);
			PriorityHeader priorityHeader = (PriorityHeader) request
					.getHeader(PriorityHeader.NAME);
			int priority = SipUtils.getPriorityValue(priorityHeader);
			boolean emergency = SipUtils.isEmergency(priorityHeader);
			GroupConfig groupConfig = topologyConfig.getGroupConfig(groupId);

			GroupMobilityManager gmm = getGroupMobilityManager();
			GroupHome groupHome = gmm.getGroup(groupConfig.getRadicalName());

			if (request.getMethod().equals(Request.INVITE)
					&& (!groupHome.getCallId().equals(
							((CallIdHeader) request
									.getHeader(CallIdHeader.NAME)).getCallId()))
					&& isCurrentCallHigheherPriority(emergency, priority,
							groupHome.isEmergency(), groupHome.getPriority())) {
				// There is a higher priority call queued.
				handleCollidingRequest(requestEvent, st);
				return;
			} else {
				actuallyProcessRequest(requestEvent, st);
			}
			groupHome.decrementPendingRequestCount();

		}

	}

	private GroupMobilityManager getGroupMobilityManager() {
		return this.rfss.getGroupMobilityManager();
	}

	public GroupCallControlManager(SipProvider provider, RfssConfig rfssConfig) {
		this.provider = provider;
		this.rfssConfig = rfssConfig;
		this.topologyConfig = rfssConfig.getSysConfig().getTopologyConfig();
		this.rfss = rfssConfig.getRFSS();
		this.activeGroupCalls = new ConcurrentHashMap<String, GroupServing>();

	}

	/**
	 * Return true if the current call is at a higher priority than the new
	 * call.
	 * 
	 * @param emergency
	 * @param priority
	 * @param currentCallIsEmergency
	 * @param currentCallPriority
	 * @return
	 */
	private boolean isCurrentCallHigheherPriority(boolean emergency,
			int priority, boolean currentCallIsEmergency,
			int currentCallPriority) {
		logger.debug("isHigherPriority : emergency = " + emergency
				+ " priority = " + priority + " currentCallIsEmergency =  "
				+ currentCallIsEmergency + " currentCallPriority = "
				+ currentCallPriority);

		if (emergency == currentCallIsEmergency) {
			return currentCallPriority >= priority;
		} else if (currentCallIsEmergency) {
			return true;
		} else {
			return false;
		}
	}

	/**
	 * Handle reINVITE at the group home. Essentially the same as processing an
	 * invite except re-use the previously allocated MmfSession.
	 * 
	 * @param requestEvent --
	 *            SIP Request Event.
	 * @param groupHome --
	 *            group home record for the group.
	 * @throws Exception
	 *             if an exception occured during processing.
	 */
	private void handleReInviteAtGroupHome(RequestEvent requestEvent,
			GroupHome groupHome) throws Exception {
		logger.debug("Processing a re-invite for the group " + groupHome);

		Request request = requestEvent.getRequest();

		ViaHeader viaHeader = (ViaHeader) request.getHeader(ViaHeader.NAME);
		String senderDomain = viaHeader.getHost();
		ServerTransaction st = requestEvent.getServerTransaction();
		groupHome.setServerTransaction(st);

		ClientTransaction ct = groupHome.getClientTransaction(senderDomain);

		if (logger.isDebugEnabled()) {
			logger.debug("retrieved client transaction " + ct);
			if (ct != null) {
				logger.debug("ct state = " + ct.getState());
			}
		}

		if (ct != null
				&& (ct.getState() == TransactionState.PROCEEDING || ct
						.getState() == TransactionState.CALLING)) {
			Response response = messageFactory.createResponse(
					Response.REQUEST_PENDING, st.getRequest());
			st.sendResponse(response);
			return;
		}
		PttSession pttSession = groupHome.getMmfSession(senderDomain);
		int mediaPort = SipUtils.getMediaPortFromMessage(request);
		int originalMediaPort = pttSession.getRtpSession()
				.getRemoteRtpRecvPort();
		pttSession.setRemoteRtpRecvPort(mediaPort);
		/*
		 * If he sends me a reinvite with a 0 port, I should also shutdown my
		 * port.
		 */
		if (mediaPort == 0) {
			pttSession.stopReceiver();
		}
		Response response = messageFactory.createResponse(Response.OK, st
				.getRequest());
		SessionDescription sessionDescription = SipUtils
				.getSessionDescriptionFromMessage(request);
		SessionDescription newSessionDescription = SdpFactory.getInstance()
				.createSessionDescription(sessionDescription.toString());
		CallControlManager.fixupSdpAnnounce(this.rfssConfig, pttSession,
				newSessionDescription);
		response.setContent(newSessionDescription.toString(), headerFactory
				.createContentTypeHeader("application", "sdp"));
		ContentDispositionHeader cdh = headerFactory
				.createContentDispositionHeader("session");
		cdh.setHandling("required");
		response.addHeader(cdh);

		MimeVersionHeader mv = headerFactory.createMimeVersionHeader(1, 0);
		response.addHeader(mv);

		ContactHeader contactHeader = SipUtils.createContactHeaderForGroup(
				groupHome.getGroupConfig(), rfssConfig);
		response.setHeader(contactHeader);
		st.sendResponse(response);
		// If this is a media port reassignment - just return.
		if (mediaPort != originalMediaPort) {
			logger.debug("media port reassignment -- returning ");
			return;
		}
		ContactHeader senderContactHeader = (ContactHeader) request
				.getHeader(ContactHeader.NAME);

		SipURI senderContactUri = (SipURI) senderContactHeader.getAddress()
				.getURI();
		// extract the sender domain name.
		String senderDomainName = SipUtils.getDomainName(senderContactUri);
		GroupConfig groupConfig = groupHome.getGroupConfig();
		for (Iterator<RfssConfig> it = groupHome.getSubscribers().iterator(); it
				.hasNext();) {
			RfssConfig rfssConfig = it.next();
			String domainName = rfssConfig.getDomainName();
			/*
			 * Forward the INVITE to all domains names excepting the sender.
			 */
			if (!domainName.equals(senderDomainName)
					&& !domainName.equals(this.rfssConfig.getDomainName())) {
				logger
						.debug("handleReInviteAtGroupHome : forwarding invite to "
								+ domainName);
				MmfSession mmfSession = groupHome.getMmfSession(domainName);
				sessionDescription = mmfSession.getSessionDescription();
				Dialog dialog = groupHome.getDialog(mmfSession);
				Request newRequest = dialog.createRequest(Request.INVITE);
				ViaHeader myViaHeader = SipUtils
						.createViaHeaderForRfss(this.rfssConfig);
				newRequest.setHeader(SipUtils.createContactHeaderForGroup(
						groupConfig, this.rfssConfig));
				TimeStampHeader tsHeader = headerFactory
						.createTimeStampHeader(0);
				long currentTime = System.currentTimeMillis();
				tsHeader.setTime(currentTime);
				newRequest.setHeader(myViaHeader);
				newRequest.setHeader(tsHeader);

				SipURI sipUri = addressFactory.createSipURI(groupConfig
						.getRadicalName(), domainName);
				sipUri.setUserParam(ISSIConstants.TIA_P25_SG);
				newRequest.removeHeader(RouteHeader.NAME);

				/*
				 * RouteHeader routeHeader = SipUtils
				 * .createRouteToRfss(rfssConfig); SipURI sipUri = (SipURI)
				 * routeHeader.getAddress().getURI();
				 * sipUri.setUser(ISSIConstants.P25_GROUP_CALL);
				 * newRequest.setHeader(routeHeader);
				 */
				newRequest.setRequestURI(sipUri);

				newSessionDescription = SdpFactory
						.getInstance()
						.createSessionDescription(sessionDescription.toString());

				// Get the old mmf and use the
				CallControlManager.fixupSdpAnnounce(this.rfssConfig,
						mmfSession, newSessionDescription);
				CallParamContent content = groupHome.getCallParamContent();
				if (content == null) {
					ContentTypeHeader ctHeader = headerFactory
							.createContentTypeHeader("application", "sdp");
					newRequest.setContent(newSessionDescription.toString(),
							ctHeader);
				} else {
					ContentList contentList = new ContentList();
					contentList.add(content);
					SdpContent sdpContent = SdpContent.createSdpContent();
					sdpContent.setSessionDescription(newSessionDescription);
					contentList.add(sdpContent);
					newRequest.setContent(contentList.toString(), contentList
							.getContentTypeHeader());
				}
				newRequest.setHeader(request.getHeader(PriorityHeader.NAME));

				ClientTransaction clientTransaction = this.provider
						.getNewClientTransaction(newRequest);
				groupHome.setClientTransaction(domainName, clientTransaction);
				groupHome.addPttSession(domainName, mmfSession);

				// Store away the MMF session in the application data
				// location.

				clientTransaction.setApplicationData(mmfSession);
				dialog.sendRequest(clientTransaction);

			}

		}

	}

	/**
	 * Handle re-invitation at the group serving.
	 * 
	 * @param requestEvent
	 * @param groupServing
	 * @throws Exception
	 */
	private void handleReInviteAtGroupServing(RequestEvent requestEvent,
			GroupServing groupServing) throws Exception {
		logger.debug("handleReInviteAtGroupServing for group "
				+ groupServing.getGroupConfig().getGroupId());
		Request request = requestEvent.getRequest();
		GroupConfig groupConfig = groupServing.getGroupConfig();
		ServerTransaction st = requestEvent.getServerTransaction();

		ClientTransaction ct = groupServing.getClientTransaction();
		if (ct != null
				&& (ct.getState() == TransactionState.CALLING || ct.getState() == TransactionState.PROCEEDING)) {
			Response response = messageFactory.createResponse(
					Response.REQUEST_PENDING, st.getRequest());
			RetryAfterHeader retryAfter = headerFactory
					.createRetryAfterHeader((int) (Math.random() * 100));
			response.addHeader(retryAfter);
			st.sendResponse(response);
			return;
		} else {
			if (ct != null)
				logger.debug("ct state at group serving is " + ct.getState());
		}
		ContentList clist = ContentList.getContentListFromMessage(request);
		if (clist == null)
			throw new Exception("Expecting SDP Content");
		SdpContent sdpContent = clist.getSdpContent();
		SessionDescription sdes = sdpContent.getSessionDescription();
		MediaDescription mdes = (MediaDescription) sdes.getMediaDescriptions(
				false).get(0);
		int remotePort = mdes.getMedia().getMediaPort();
		if (remotePort == 0) {
			groupServing.getMultiplexer().getRtpSession()
					.stopRtpPacketReceiver();
		}
		groupServing.getMultiplexer().setRemoteRtpRecvPort(remotePort);
		this.sendResponse(requestEvent, groupConfig, st, Response.OK);
	}

	/**
	 * Handle a colliding request at the Home RFSS.
	 * 
	 * @param requestEvent --
	 *            sip request event
	 * @param st --
	 *            server transaction
	 */
	private void handleCollidingRequest(RequestEvent requestEvent,
			ServerTransaction st) {

		Request request = requestEvent.getRequest();

		Response response = SipUtils.createErrorResponse(this.rfssConfig,
				request, WarningCodes.DECLINE_COLLIDING_CALL);
		if (st == null) {
			try {
				st = this.provider.getNewServerTransaction(request);
			} catch (TransactionAlreadyExistsException ex) {
				// Already seen this request so ignore the retransmission
				logger.debug("Ignoring retransmission of request ");
				return;
			} catch (TransactionUnavailableException e) {
				this.rfss.logError("unexpected exception", e);
				return;
			}
		}
		try {
			st.sendResponse(response);
		} catch (Exception ex) {
			this.rfss.getTestHarness().fail("Unexpected exception", ex);
		}

	}

	private void actuallyProcessRequest(RequestEvent requestEvent,
			ServerTransaction st) {
		if (requestEvent.getRequest().getMethod().equals(Request.INVITE)) {
			processInvite(requestEvent, st);
		} else if (requestEvent.getRequest().getMethod().equals(Request.ACK)) {
			processAck(requestEvent);
		} else if (requestEvent.getRequest().getMethod().equals(Request.BYE)) {
			processBye(requestEvent);
		} else if (requestEvent.getRequest().getMethod().equals(Request.CANCEL)) {
			processCancel(requestEvent, st);
		}
	}

	/**
	 * Process incoming BYE event.
	 * 
	 * @param requestEvent --
	 *            request event to process.
	 * 
	 */
	private void processBye(RequestEvent requestEvent) {
		try {
			ServerTransaction st = requestEvent.getServerTransaction();
			Request request = requestEvent.getRequest();
			ToHeader to = (ToHeader) requestEvent.getRequest().getHeader(
					ToHeader.NAME);
			String radicalName = SipUtils.getGroupID(to.getAddress().getURI());
			GroupConfig groupConfig = topologyConfig
					.getGroupConfig(radicalName);
			if (this.rfssConfig == groupConfig.getHomeRfss()) {
				GroupHome groupHome = this.getGroupMobilityManager().getGroup(
						radicalName);
				groupHome.removeDialog(st.getDialog());
			} else {
				logger.debug("Got a bye at group serving - tearing down call");
				GroupServing groupServing = this.getGroupCall(radicalName);
				groupServing.getMultiplexer().getRtpSession().shutDown();
			}
			Response response = messageFactory.createResponse(Response.OK,
					request);
			st.sendResponse(response);

		} catch (Exception ex) {
			logger.fatal("Unexpected exception -- failing test. ", ex);
			this.rfss.getTestHarness().fail("Unexpected exception", ex);
		}
	}

	/**
	 * Process an incoming ACK event.
	 * 
	 * @param requestEvent --
	 *            the event to process.
	 */
	private void processAck(RequestEvent requestEvent) {

		try {
			ToHeader to = (ToHeader) requestEvent.getRequest().getHeader(
					ToHeader.NAME);
			SipURI sipUri = (SipURI) to.getAddress().getURI();
			String groupId = SipUtils.getGroupID(sipUri);

			GroupConfig groupConfig = this.topologyConfig
					.getGroupConfig(groupId);
			this.rfss.getTestHarness().assertTrue(groupConfig != null);
			if (this.rfssConfig == groupConfig.getHomeRfss()) {
				GroupHome groupHome = this.getGroupMobilityManager().getGroup(
						groupId);
				ViaHeader viaHeader = (ViaHeader) requestEvent.getRequest()
						.getHeader(ViaHeader.NAME);
				String rfssDomainName = viaHeader.getHost();
				PttSession pttSession = groupHome
						.getPttPointToMultipointSession().getPttSession(
								rfssDomainName);
				if (pttSession.getRtpSession().getRemoteRtpRecvPort() > 0)
					pttSession.getHeartbeatTransmitter().start();

			} else {
				GroupServing groupServing = this.getGroupCall(groupId);
				PttSessionMultiplexer pttSessionMux = groupServing
						.getMultiplexer();
				if (pttSessionMux.getRtpSession().getRemoteRtpRecvPort() > 0) {
					pttSessionMux.getHeartbeatTransmitter().start();
				} else {
					logger
							.debug("no remote rtp port available -- not starting heartbeat transmitter");
				}
			}
		} catch (Exception ex) {
			this.rfss.getTestHarness().fail("Unexpected exception", ex);
		}

	}

	/**
	 * Process Cancel for the group call.
	 * 
	 * @param requestEvent
	 * @param st
	 */

	private void processCancel(RequestEvent requestEvent, ServerTransaction st) {
		try {
			Request request = requestEvent.getRequest();

			ToHeader toHeader = ((ToHeader) request.getHeader(ToHeader.NAME));
			SipURI toUri = ((SipURI) toHeader.getAddress().getURI());

			String groupId = SipUtils.getGroupID(toUri);
			GroupConfig groupConfig = topologyConfig.getGroupConfig(groupId);
			GroupMobilityManager gmm = this.getGroupMobilityManager();
			// I am at the home RFSS.
			if (groupConfig.getHomeRfss() == this.rfssConfig) {

				GroupHome groupHome = gmm
						.getGroup(groupConfig.getRadicalName());
				ServerTransaction inviteServerTransaction = groupHome
						.getServerTransaction();
				Response cancelResponse = messageFactory.createResponse(
						Response.OK, request);
				st.sendResponse(cancelResponse);
				if (inviteServerTransaction.getState() == TransactionState.PROCEEDING) {
					if (groupHome.getDelayTask() != null)
						groupHome.getDelayTask().cancel();
					Response response = messageFactory.createResponse(
							Response.REQUEST_TERMINATED,
							inviteServerTransaction.getRequest());
					inviteServerTransaction.sendResponse(response);

				}
				for (ClientTransaction ct : groupHome.getClientTransactions()) {
					if (ct.getState() == TransactionState.PROCEEDING) {
						Request cancelRequest = ct.createCancel();
						ClientTransaction cancelTx = provider
								.getNewClientTransaction(cancelRequest);
						cancelTx.sendRequest();
					}

				}
			} else {
				GroupInviteServingDelayTask delayTask = this.pendingTimers
						.get(st.getBranchId());
				if (delayTask == null) {
					st.sendResponse(messageFactory.createResponse(Response.OK,
							st.getRequest()));
					return;
				} else {
					ServerTransaction inviteTx = delayTask
							.getServerTransaction();
					if (inviteTx.getState() != TransactionState.TERMINATED) {
						delayTask.cancel();
						this.pendingTimers.remove(st.getBranchId());
						inviteTx.sendResponse(messageFactory.createResponse(
								Response.REQUEST_TERMINATED, inviteTx
										.getRequest()));

					}
					st.sendResponse(messageFactory.createResponse(Response.OK,
							st.getRequest()));

				}
			}

		} catch (Exception ex) {
			this.rfss.getTestHarness().fail("Unexpected exception", ex);

		}
	}

	/**
	 * Processes an Invite Request.
	 * 
	 * @param requestEvent --
	 *            the JAIN-SIP request event.
	 * @param st --
	 *            allocated SIP Server transaction.
	 */
	private void processInvite(RequestEvent requestEvent, ServerTransaction st) {
		try {

			Request request = requestEvent.getRequest();

			ToHeader toHeader = ((ToHeader) request.getHeader(ToHeader.NAME));
			SipURI toUri = ((SipURI) toHeader.getAddress().getURI());

			String groupId = SipUtils.getGroupID(toUri);
			long gid = Long.parseLong(groupId, 16);
			GroupConfig groupConfig = topologyConfig.getGroupConfig(groupId);

			// Get the sender.
			ContactHeader sender = (ContactHeader) request
					.getHeader(ContactHeader.NAME);
			if (sender == null) {
				Response response = messageFactory.createResponse(
						Response.BAD_REQUEST, request);
				st.sendResponse(response);
				return;
			}
			ViaHeader via = (ViaHeader) request.getHeader(ViaHeader.NAME);
			String callingRfssRadicalName = via.getHost();
			RfssConfig callingRfss = topologyConfig
					.getRfssConfig(callingRfssRadicalName);
			GroupMobilityManager gmm = this.getGroupMobilityManager();

			GroupHome groupHome = gmm.getGroup(groupConfig.getRadicalName());
			ContentList contentList = ContentList.getContentListFromMessage(request);
			CallParamContent callParamContent = contentList.getCallParamContent();
		
			if (rfssConfig.isAssignedGroup(groupId)) {
				if (groupHome.getClientTransaction(callingRfssRadicalName) != null) {
					ClientTransaction ct = groupHome
							.getClientTransaction(callingRfssRadicalName);
					if (ct.getState() == TransactionState.CALLING
							|| ct.getState() == TransactionState.PROCEEDING) {
						Response response = messageFactory.createResponse(
								Response.REQUEST_PENDING, st.getRequest());
						st.sendResponse(response);
						return;
					}
				}
				/*
				 * I got an INVITE from an RFSS which is not a group member.
				 * Reject the INVITE.
				 */
				if (!groupHome.isSubscribed(callingRfss)) {
					Response errorResponse = SipUtils.createErrorResponse(
							rfssConfig, request,
							WarningCodes.FORBIDDEN_SG_NOT_REGISTERED);
					st.sendResponse(errorResponse);
					return;
				}
				/*
				 * I am configured to reject protected calls for the group.
				 */
				if ( groupConfig.getCProtectedDispsition().equals(CProtectedDisposition.REJECT_PROTECTED) && callParamContent != null) {
					CallParam callParam = callParamContent.getCallParam();
					if ( callParam.isProtectedMode()) {
						
						Response response = SipUtils.createErrorResponse(
								rfssConfig, request,
								WarningCodes.FORBIDDEN_PROTECTED_MODE_AUDIO_CANNOT_BE_GRANTED);
					
						st.sendResponse(response);
						return;	
					}
						
				}
				
				/*
				 * I am configured to reject unprotected calls for the group.
				 */
				if ( groupConfig.getCProtectedDispsition().equals(CProtectedDisposition.REJECT_UNPROTECTED) && callParamContent != null) {
					CallParam callParam = callParamContent.getCallParam();
					if (! callParam.isProtectedMode()) {
						
						Response response = SipUtils.createErrorResponse(
								rfssConfig, request,
								WarningCodes.FORBIDDEN_UNPROTECTED_MODE_AUDIO_CANNOT_BE_GRANTED);
					
						st.sendResponse(response);
						return;	
					}
						
				}
				/*
				 * The Home group abstraction keeps a hash table of SIP Dialogs
				 * indexed by radical name of each of the memebers. It also
				 * keeps a hash table of all PTT MMF sessions hashed by the same
				 * key. If we see the Dialog already in the set of dialogs that
				 * are tracked by the group home, then this is a re-Invite.
				 * 
				 */
				if (groupHome.getDialogs().contains(st.getDialog())) {
					this.handleReInviteAtGroupHome(requestEvent, groupHome);
					return;
				}

				/*
				 * this is a group for which we are the home rfss. broadcast the
				 * request to all serving rfss's
				 */
				Response response = ProtocolObjects.getMessageFactory()
						.createResponse(Response.OK, request);
				ContactHeader contactHeader = SipUtils
						.createContactHeaderForGroup(groupConfig,
								this.rfssConfig);
				response.setHeader(contactHeader);
				ToHeader to = (ToHeader) response.getHeader(ToHeader.NAME);
				String tag = SipUtils.createTag();
				to.setTag(tag);

				groupHome
						.addDialog(callingRfss.getDomainName(), st.getDialog());
				if (request.getHeader(PriorityHeader.NAME) != null) {
					int priority = SipUtils
							.getPriorityValue((PriorityHeader) request
									.getHeader(PriorityHeader.NAME));

					boolean emergency = SipUtils
							.isEmergency((PriorityHeader) request
									.getHeader(PriorityHeader.NAME));

					String callId = ((CallIdHeader) request
							.getHeader(CallIdHeader.NAME)).getCallId();
					groupHome.setCallId(callId);
					groupHome.setPriority(priority);
					groupHome.setEmergency(emergency);
				}
				st.getDialog().setApplicationData(groupHome);
				PttPointToMultipointSession pttSession = groupHome
						.getPttPointToMultipointSession();
				MmfSession mmfSession = null;
				ContactHeader senderContactHeader = (ContactHeader) request
						.getHeader(ContactHeader.NAME);

				SipURI senderContactUri = (SipURI) senderContactHeader
						.getAddress().getURI();
				// extract the sender domain name.
				String senderDomainName = SipUtils
						.getDomainName(senderContactUri);
				int senderRfssId = SipUtils.getRfssId(senderDomainName);
				if (logger.isDebugEnabled()) {
					logger.debug("senderDomainName = " + senderDomainName
							+ " senderRfssId = " + senderRfssId);
				}

				SdpContent sdpContent = (SdpContent) ContentList
						.getContentByType(contentList, "application", "sdp");

				SessionDescription sessionDescription = sdpContent
						.getSessionDescription();
				String sessionId = ((OriginField) sessionDescription
						.getOrigin()).getSessIdAsString();

				try {
					mmfSession = pttSession.createMmfSession(
							LinkType.GROUP_HOME, sessionId);
					boolean muteFlag = rfssConfig
							.isNotInterestedInLosingAudio();
					mmfSession.setSetIsNotInterestedInLosingAudio(muteFlag);

					mmfSession.setRemoteRfssDomainName(callingRfssRadicalName);
				} catch (RtpException rtpException) {
					logger.debug("no rtp resources ", rtpException);
					Response errorResponse = SipUtils.createErrorResponse(
							rfssConfig, request, WarningCodes.NO_RTP_RESOURCE);
					st.sendResponse(errorResponse);
					return;
				}
				String sendingHost = SipUtils.getContactHost(request);
				rfss.getTransmissionControlManager().addMediaLeg(gid,
						mmfSession, sessionDescription, sendingHost);
				SessionDescription newSessionDescription = SdpFactory
						.getInstance().createSessionDescription(
								sessionDescription.toString());

				CallControlManager.fixupSdpAnnounce(this.rfssConfig,
						mmfSession, newSessionDescription);

				

				response.setContent(newSessionDescription.toString(),
						headerFactory.createContentTypeHeader("application",
								"sdp"));

				ContentDispositionHeader cdh = headerFactory
						.createContentDispositionHeader("session");
				cdh.setHandling("required");
				response.addHeader(cdh);

				MimeVersionHeader mv = headerFactory.createMimeVersionHeader(1,
						0);
				response.addHeader(mv);

				groupHome.addPttSession(senderDomainName, mmfSession);
				st.sendResponse(response);
				if (logger.isDebugEnabled()) {
					logger.debug("Sender domain name = " + senderDomainName);
				}
				for (Iterator<RfssConfig> it = groupHome.getSubscribers()
						.iterator(); it.hasNext();) {
					RfssConfig rfssConfig = it.next();
					String domainName = rfssConfig.getDomainName();
					// Forward the INVITE to all domains names excepting the
					// sender.
					if (!domainName.equals(senderDomainName)
							&& !domainName.equals(this.rfssConfig
									.getDomainName())
							&& !groupHome.hasPttSession(rfssConfig)) {
						if (logger.isDebugEnabled()) {
							logger.debug("forwarding invite to " + domainName);
						}
						Request newRequest = (Request) request.clone();
						ViaHeader viaHeader = SipUtils
								.createViaHeaderForRfss(this.rfssConfig);
						newRequest.setHeader(SipUtils
								.createContactHeaderForGroup(groupConfig,
										this.rfssConfig));
						TimeStampHeader tsHeader = headerFactory
								.createTimeStampHeader(0);
						long currentTime = System.currentTimeMillis();
						tsHeader.setTime(currentTime);
						newRequest.setHeader(viaHeader);
						newRequest.setHeader(tsHeader);
						SipURI sipUri = addressFactory.createSipURI(groupConfig
								.getRadicalName(), domainName);
						sipUri.setUserParam(ISSIConstants.TIA_P25_SG);
						newRequest.setRequestURI(sipUri);
						newRequest.removeHeader(RouteHeader.NAME);

						newRequest.setHeader(cdh);

						ToHeader newto = (ToHeader) newRequest
								.getHeader(ToHeader.NAME);
						newto.removeParameter("tag");
						FromHeader newfrom = (FromHeader) newRequest
								.getHeader(FromHeader.NAME);
						newfrom.setTag(SipUtils.createTag());
						String newSessionId = null;
						mmfSession = null;

						newSessionId = new Long((long) (Math.random() * 100000))
								.toString();

						logger
								.debug("groupCallControlManager: forwarding invite to using session Id : "
										+ newSessionId);
						try {

							mmfSession = pttSession.createMmfSession(
									LinkType.GROUP_HOME, newSessionId);
							mmfSession.setRemoteRfssDomainName(domainName);

						} catch (RtpException rtpException) {
							// The mmf session could not be created for some
							// reason. Just discard it. ??
							this.rfss.logError("Unexpected RtpException : ",
									rtpException);
							continue;

						}

						newSessionDescription = SdpFactory.getInstance()
								.createSessionDescription(
										sessionDescription.toString());

						CallControlManager.fixupSdpAnnounce(this.rfssConfig,
								mmfSession, newSessionDescription);
						((OriginField) newSessionDescription.getOrigin())
								.setSessionId(newSessionId);
						CallParamContent content = groupHome
								.getCallParamContent();
						
					
						if (content == null) {
							
							

							ContentTypeHeader ctHeader = headerFactory
									.createContentTypeHeader("application",
											"sdp");
							newRequest.setContent(newSessionDescription
									.toString(), ctHeader);
						} else {
							CallParam callParam = content.getCallParam(); 
						
							// Called home overrides the protected mode parameter settting if needed.
							if ( callParam.isProtectedMode() &&  
									groupConfig.getCProtectedDispsition().equals(CProtectedDisposition.CLEAR_PROTECTED) ) {
									callParam.setProtectedMode(false);
							} 
							callParam.setRfResourceAvailable(
									this.rfss.isRfResourcesAvailable());
							contentList = new ContentList();
							contentList.add(content);
							sdpContent = SdpContent.createSdpContent();
							sdpContent
									.setSessionDescription(newSessionDescription);
							contentList.add(sdpContent);
							newRequest.setContent(contentList.toString(),
									contentList.getContentTypeHeader());
						}

						ClientTransaction ct = this.provider
								.getNewClientTransaction(newRequest);
						groupHome.setClientTransaction(domainName, ct);
						groupHome.addPttSession(domainName, mmfSession);

						// Store away the MMF session in the application data
						// location.

						ct.setApplicationData(mmfSession);
						groupHome.addDialog(domainName, ct.getDialog());
						ct.getDialog().setApplicationData(groupHome);
						ct.sendRequest();

					}

				}
			} else {
				// I am at the serving RFSS.
				GroupMobilityManager mm = this.getGroupMobilityManager();
				if (!mm.isKnownGroup(groupConfig.getRadicalName())) {
					logger.debug("could not find " + groupId);
					Response response = SipUtils.createErrorResponse(
							this.rfssConfig, request,
							WarningCodes.NOT_FOUND_UNKNOWN_TARGET_GROUP);
					st.sendResponse(response);
					return;
				}
				// Check if any of our SUs are subscribed to the group
				boolean found = false;
				for (SuConfig suconfig : this.rfss.getServedSubscriberUnits()) {
					if (suconfig.getSubscribedGroups().contains(groupConfig)) {
						found = true;
					}
				}
				if (!found) {
					logger.debug("could not find " + groupId);
					Response response = SipUtils.createErrorResponse(
							this.rfssConfig, request,
							WarningCodes.NOT_FOUND_UNKNOWN_TARGET_GROUP);
					st.sendResponse(response);

				} else if (this.activeGroupCalls.containsKey(groupId)) {
					/*
					 * Look for the collinding case ( we have sent out an invite
					 * at the same time as the group has sent back an invite to
					 * us )
					 */
					if (requestEvent.getDialog() != this.activeGroupCalls.get(
							groupId).getDialog()) {
						logger
								.debug("Found an existing Active group call - declining invite");
						ClientTransaction ct = this.activeGroupCalls.get(
								groupId).getClientTransaction();
						PriorityHeader priorityHeader = (PriorityHeader) request
								.getHeader(PriorityHeader.NAME);
						boolean emergency = SipUtils
								.isEmergency(priorityHeader);
						int priority = SipUtils
								.getPriorityValue(priorityHeader);
						PriorityHeader currentCallPriorityHeader = (PriorityHeader) ct
								.getRequest().getHeader(PriorityHeader.NAME);
						int currentCallPriority = SipUtils
								.getPriorityValue(currentCallPriorityHeader);
						boolean currentCallIsEmergency = SipUtils
								.isEmergency(currentCallPriorityHeader);
						if (ct.getState() == TransactionState.CALLING
						/* || ct.getState() == TransactionState.PROCEEDING */) {

							Response response = messageFactory.createResponse(
									Response.REQUEST_PENDING, st.getRequest());
							st.sendResponse(response);
							return;

						} else if (!ct.getState().equals(
								TransactionState.COMPLETED)
								&& !ct.getState().equals(
										TransactionState.TERMINATED)) {
							if (isCurrentCallHigheherPriority(emergency,
									priority, currentCallIsEmergency,
									currentCallPriority)) {
								this.handleCollidingRequest(requestEvent, st);
							} else {
								// GroupServing(RFSS rfss, Dialog myDialog,
								// ServerTransaction st,
								// GroupConfig groupConfig, SipProvider
								// provider)
								this.activeGroupCalls.put(groupId,
										new GroupServing(this.rfss,
												requestEvent.getDialog(), st,
												groupConfig, this.provider));
								this.sendResponse(requestEvent, groupConfig,
										st, Response.OK);

							}
						} else {
							logger
									.debug("Current tx state is "
											+ ct.getState());

							this.handleCollidingRequest(requestEvent, st);
						}
					} else {
						GroupServing groupServing = this.activeGroupCalls
								.get(groupId);

						this.handleReInviteAtGroupServing(requestEvent,
								groupServing);

					}

				} else {
					// TODO check for resources etc.
					this.sendResponse(requestEvent, groupConfig, st,
							Response.OK);

				}
			}

		} catch (Exception ex) {
			this.rfss.logError("Unexpected exception", ex);
			this.rfss.getTestHarness().fail("unexpected exception", ex);
		}

	}

	/**
	 * Create and send a response back in response to the
	 * GroupCallSetupRequestEvent.
	 * 
	 * @param responseCode --
	 *            response code to send back.
	 * 
	 */
	private void sendResponse(RequestEvent requestEvent,
			GroupConfig groupConfig, ServerTransaction serverTransaction,
			int responseCode) throws Exception {

		Response response = messageFactory.createResponse(responseCode,
				requestEvent.getRequest());
		ContactHeader contactHeader = SipUtils.createContactHeaderForGroup(
				groupConfig, this.rfssConfig);
		response.setHeader(contactHeader);
		ToHeader toHeader = (ToHeader) response.getHeader(ToHeader.NAME);
		// Tag for to header is mandatory in 2xx response.
		boolean processingReInvite = true;
		if (response.getStatusCode() / 100 == 2 && toHeader.getTag() == null) {
			String toTag = SipUtils.createTag();
			toHeader.setTag(toTag);
			processingReInvite = false;
		}

		// If responding with a Success responseCode. Then create
		// a new RTP Session for this SU if needed.
		// Modify the response SDP Body to reflect the IP address
		// and port of the new session we just created and
		// send the response back.

		if (responseCode == Response.OK) {

			ContentList contentList = ContentList
					.getContentListFromMessage(requestEvent.getRequest());
			SdpContent sdpContent = (SdpContent) contentList.getContent(
					"application", "sdp");
			SessionDescription sessionDescription = (SessionDescription) sdpContent
					.getSessionDescription();

			// The SU is answering with a SIP OK. At this point,
			// we need to allocate a new RTP Session (i.e. a PTT session),
			// from which we extract the IP address and port and send it
			// in the OK.

			// String sessionId = SipUtils.getRtpSessionIdFromMessage(response);

			PttSessionMultiplexer mux = null;
			try {
				String sdpSessionId = ((OriginField) sessionDescription
						.getOrigin()).getSessIdAsString();
				mux = this.rfss.getTransmissionControlManager().getPttManager()
						.createPttSessionMultiplexer(sdpSessionId,
								groupConfig.getGroupId());
				mux.setOwningRfss(this.rfssConfig);
				mux.setRemoteRfssDomainName(groupConfig.getHomeRfss()
						.getDomainName());

				MediaDescription mediaDescription = (MediaDescription) sessionDescription
						.getMediaDescriptions(true).get(0);

				String ipAddress = sessionDescription.getConnection()
						.getAddress();

				if (mediaDescription.getConnection() != null) {

					ipAddress = mediaDescription.getConnection().getAddress();

				}

				int port = mediaDescription.getMedia().getMediaPort();

				mux.setRemoteIpAddress(ipAddress);
				logger.debug("original remote rtp recvport = "
						+ mux.getRtpSession().getRemoteRtpRecvPort());
				mux.setRemoteRtpRecvPort(port);

				Iterator<SuConfig> it = groupConfig.getSubscribers();
				while (it.hasNext()) {
					SuConfig groupMember = it.next();
					if (this.rfss.isSubscriberUnitServed(groupMember)) {
						SmfSession smfSession = this.rfss
								.getTransmissionControlManager()
								.getPttManager().createSmfSession(
										mux.getRtpSession(),
										LinkType.GROUP_SERVING);
						logger.debug("smf session 2 " + smfSession);
						smfSession.getSmfReceiver().addListener(
								this.rfss.getCurrentTestCase());
						smfSession.setGroupId(groupConfig.getGroupId());
						smfSession.setSessionDescription(sessionDescription);
						int tsn = this.rfss.getTransmissionControlManager()
								.getPttManager().getNewTsn(
										LinkType.GROUP_SERVING);
						smfSession.setTsn(groupMember.getSuId(), tsn);
						mux.addPttSession(tsn, smfSession);
						mux.addSu(groupMember.getRadicalName(), tsn);

					}
				}

			} catch (RtpException ex) {
				Response errorResponse = SipUtils.createErrorResponse(
						rfssConfig, requestEvent.getRequest(),
						WarningCodes.NO_RTP_RESOURCE);
				serverTransaction.sendResponse(errorResponse);
				return;
			}

			Request request = requestEvent.getRequest();
			SipURI requestUri = (SipURI) request.getRequestURI();
			String groupId = SipUtils.getGroupID(requestUri);
			long gid = Long.parseLong(groupId, 16);
			CallControlManager.fixupSdpAnnounce(this.rfssConfig, mux,
					sessionDescription);
			ContentTypeHeader contentType = (ContentTypeHeader) requestEvent
					.getRequest().getHeader(ContentTypeHeader.NAME);

			if (contentType.getContentType().equals("multipart")
					&& contentType.getContentSubType().equals("mixed")) {
				ContentList clist = new ContentList();
				SdpContent retContent = SdpContent.createSdpContent();
				retContent.setSessionDescription(sessionDescription);

				CallParamContent callParamContent = CallParamContent
						.createCallParamContent();

				callParamContent.getCallParam().setRfResourceAvailable(
						this.rfss.isRfResourcesAvailable());
				clist.add(retContent);
				clist.add(callParamContent);
				response.setContent(clist.toString(), clist
						.getContentTypeHeader());

			} else {
				if (!this.rfss.isRfResourcesAvailable()) {
					/* NO RF resources available */
					CallParamContent callParamContent = CallParamContent
							.createCallParamContent();
					callParamContent.getCallParam().setRfResourceAvailable(
							this.rfssConfig.isRfResourcesAvailable());
					ContentList cl = new ContentList();
					cl.add(callParamContent);
					SdpContent sc = SdpContent.createSdpContent();
					sc.setSessionDescription(sessionDescription);
					cl.add(sc);
					response.setContent(cl.toString(), cl
							.getContentTypeHeader());
				} else {
					response.setContent(sessionDescription.toString(),
							contentType);
				}

			}

			ContentDispositionHeader cdh = headerFactory
					.createContentDispositionHeader("session");
			cdh.setHandling("required");

			response.setHeader(cdh);

			MimeVersionHeader mvh = headerFactory.createMimeVersionHeader(1, 0);
			response.setHeader(mvh);

			// this.transmissionControlManager.addRtpListener(tcSap);

			serverTransaction.sendResponse(response);
			Dialog dialog = serverTransaction.getDialog();

			GroupServing groupServing = new GroupServing(this.rfss, dialog,
					serverTransaction, groupConfig, provider);
			groupServing.setMultiplexer(mux);
			Iterator<SuConfig> it = groupConfig.getSubscribers();
			while (it.hasNext()) {
				SuConfig groupMember = it.next();
				if (this.rfss.isSubscriberUnitServed(groupMember)) {
					groupServing.addSu(groupMember);

				}
			}

			int priority = SipUtils.getPriorityValue((PriorityHeader) request
					.getHeader(PriorityHeader.NAME));
			boolean emergency = SipUtils.isEmergency((PriorityHeader) request
					.getHeader(PriorityHeader.NAME));
			groupServing.setPriority(priority);
			groupServing.setEmergency(emergency);
			this.activeGroupCalls.put(groupConfig.getRadicalName(),
					groupServing);

		} else {
			serverTransaction.sendResponse(response);
		}

	}

	/**
	 * Create a re-invite for a group call.
	 * 
	 * @param calledGroup --
	 *            group to call.
	 * @param dialog --
	 *            original dialog for which the group call was originally
	 *            launched.
	 * @return -- the formatted request.
	 */
	Request createReInviteForGroupCall(GroupHome calledGroup, Dialog dialog) {
		try {
			Request sipRequest = dialog.createRequest(Request.INVITE);
			ServerTransaction st = calledGroup.getServerTransaction();
			Request originalRequest = st.getRequest();
			sipRequest.setHeader(originalRequest.getHeader(CallIdHeader.NAME));
			ContactHeader contactHeader = SipUtils.createContactHeaderForRfss(
					this.rfssConfig, ISSIConstants.TIA_P25_SG);
			SipURI contactUri = (SipURI) contactHeader.getAddress().getURI();
			contactUri.setUser(calledGroup.getGroupConfig().getRadicalName());
			sipRequest.setHeader(contactHeader);

			// sipRequest.setRequestURI(originalRequest.getRequestURI());
			TimeStampHeader timestamp = ProtocolObjects.headerFactory
					.createTimeStampHeader(0);
			timestamp.setTime(System.currentTimeMillis());
			sipRequest.setHeader(timestamp);
			sipRequest.addHeader(headerFactory
					.createAllowHeader(Request.REGISTER));
			sipRequest.addHeader(headerFactory
					.createAllowHeader(Request.INVITE));
			sipRequest.addHeader(headerFactory.createAllowHeader(Request.ACK));
			sipRequest.addHeader(headerFactory.createAllowHeader(Request.BYE));
			sipRequest.addHeader(headerFactory
					.createAllowHeader(Request.CANCEL));

			AcceptHeader acceptHeader = headerFactory.createAcceptHeader(
					ISSIConstants.APPLICATION, "sdp");
			acceptHeader.setParameter("level", "1");
			sipRequest.addHeader(acceptHeader);
			sipRequest.addHeader(headerFactory.createAcceptHeader(
					ISSIConstants.APPLICATION, ISSIConstants.X_TIA_P25_ISSI));

			sipRequest
					.addHeader(originalRequest.getHeader(PriorityHeader.NAME));

			ContentDispositionHeader cd = headerFactory
					.createContentDispositionHeader(ContentDispositionHeader.SESSION);
			cd.setHandling("required");
			sipRequest.addHeader(cd);
			MimeVersionHeader mvh = headerFactory.createMimeVersionHeader(1, 0);
			sipRequest.setHeader(mvh);

			if (logger.isDebugEnabled())
				logger
						.debug("createdReInviteForGroupCall : (GroupHome) Created a request = "
								+ sipRequest);

			return sipRequest;
		} catch (Exception ex) {
			logger.fatal("unexpected exception ", ex);
			return null;

		}
	}

	/**
	 * Create a re-invite for a group call.
	 * 
	 * @param calledGroup --
	 *            group to call.
	 * @param dialog --
	 *            original dialog for which the group call was originally
	 *            launched.
	 * @return -- the formatted request.
	 */
	private Request createReInviteForGroupCall(GroupServing calledGroup,
			Dialog dialog) {
		try {
			Request sipRequest = dialog.createRequest(Request.INVITE);
			ClientTransaction ct = calledGroup.getClientTransaction();
			Request originalRequest = calledGroup.getLastTransaction()
					.getRequest();
			sipRequest.setHeader(originalRequest.getHeader(CallIdHeader.NAME));
			sipRequest.setHeader(SipUtils.createContactHeaderForRfss(
					this.rfssConfig, ISSIConstants.TIA_P25_SG));
			// sipRequest.setRequestURI(originalRequest.getRequestURI());
			TimeStampHeader timestamp = ProtocolObjects.headerFactory
					.createTimeStampHeader(0);
			timestamp.setTime(System.currentTimeMillis());
			sipRequest.setHeader(timestamp);
			sipRequest.addHeader(headerFactory
					.createAllowHeader(Request.REGISTER));
			sipRequest.addHeader(headerFactory
					.createAllowHeader(Request.INVITE));
			sipRequest.addHeader(headerFactory.createAllowHeader(Request.ACK));
			sipRequest.addHeader(headerFactory.createAllowHeader(Request.BYE));
			sipRequest.addHeader(headerFactory
					.createAllowHeader(Request.CANCEL));

			AcceptHeader acceptHeader = headerFactory.createAcceptHeader(
					ISSIConstants.APPLICATION, "sdp");
			acceptHeader.setParameter("level", "1");
			sipRequest.addHeader(acceptHeader);
			MimeVersionHeader mvh = headerFactory.createMimeVersionHeader(1, 0);
			sipRequest.addHeader(mvh);

			sipRequest.addHeader(headerFactory.createAcceptHeader(
					ISSIConstants.APPLICATION, ISSIConstants.X_TIA_P25_ISSI));

			sipRequest
					.addHeader(originalRequest.getHeader(PriorityHeader.NAME));

			ContentDispositionHeader cd = headerFactory
					.createContentDispositionHeader(ContentDispositionHeader.SESSION);
			cd.setHandling("required");
			sipRequest.addHeader(cd);

			if (logger.isDebugEnabled())
				logger
						.debug("createReInviteForGroupCall: (GroupServing) Created a request = "
								+ sipRequest);

			return sipRequest;
		} catch (Exception ex) {
			logger.fatal("unexpected exception ", ex);
			return null;

		}
	}

	/**
	 * Set up an outgoing invite
	 * 
	 * @param callingSu
	 * @param calledGroup
	 * @return
	 * @throws Exception
	 */
	private Request createInviteForGroupCall(GroupConfig calledGroup,
			int priority, boolean isEmergency)
			throws Exception {
		// create my SIP URI. This goes into the From Header.

		String calledRadicalName = calledGroup.getRadicalName();
		SipURI requestUri = addressFactory.createSipURI(calledRadicalName,
				ISSIConstants.P25DR);
		requestUri.setUserParam(ISSIConstants.TIA_P25_SG);
		SipURI callingSipURI = SipUtils.createSipURI(calledRadicalName,
				ISSIConstants.P25DR);
		callingSipURI.setParameter("user", ISSIConstants.TIA_P25_SG);
		String tag = SipUtils.createTag();
		Address fromAddress = addressFactory.createAddress(callingSipURI);
		FromHeader fromHeader = headerFactory
				.createFromHeader(fromAddress, tag);

		// Create the SIP URI for the Called RFSS.
		SipURI calledSipURI = SipUtils.createSipURI(calledRadicalName,
				ISSIConstants.P25DR);
		calledSipURI.setParameter("user", ISSIConstants.TIA_P25_SG);
		Address toAddress = addressFactory.createAddress(calledSipURI);

		ToHeader toHeader = headerFactory.createToHeader(toAddress, null);

		ContactHeader contactHeader = SipUtils.createContactHeaderForGroup(
				calledGroup, this.rfssConfig);

		ViaHeader viaHeader = SipUtils.createViaHeaderForRfss(this.rfssConfig);
		LinkedList<ViaHeader> viaHeaders = new LinkedList<ViaHeader>();
		viaHeaders.add(viaHeader);

		// Get a new call ID header for the outgoing invite.
		String newCallId = System.currentTimeMillis() + ".1@"
				+ this.rfssConfig.getDomainName();
		CallIdHeader callIdHeader = headerFactory.createCallIdHeader(newCallId); // BUGBUG

		CSeqHeader cseqHeader = headerFactory.createCSeqHeader(1L,
				Request.INVITE);

		MaxForwardsHeader maxForwardsHeader = headerFactory
				.createMaxForwardsHeader(70);

		Request sipRequest = messageFactory.createRequest(requestUri,
				Request.INVITE, callIdHeader, cseqHeader, fromHeader, toHeader,
				viaHeaders, maxForwardsHeader);

		if (calledGroup.getHomeRfss() != this.rfssConfig) {
			RouteHeader routeHeader = SipUtils.createRouteToRfss(calledGroup
					.getHomeRfss());
			SipURI sipUri = (SipURI) routeHeader.getAddress().getURI();
			sipUri.setUser(ISSIConstants.P25_GROUP_CALL);
			sipRequest.addHeader(routeHeader);
		}

		sipRequest.addHeader(contactHeader);

		TimeStampHeader timestamp = ProtocolObjects.headerFactory
				.createTimeStampHeader(0);
		timestamp.setTime(System.currentTimeMillis());
		sipRequest.setHeader(timestamp);
		sipRequest.addHeader(headerFactory.createAllowHeader(Request.REGISTER));
		sipRequest.addHeader(headerFactory.createAllowHeader(Request.INVITE));
		sipRequest.addHeader(headerFactory.createAllowHeader(Request.ACK));
		sipRequest.addHeader(headerFactory.createAllowHeader(Request.BYE));
		sipRequest.addHeader(headerFactory.createAllowHeader(Request.CANCEL));

		AcceptHeader acceptHeader = headerFactory.createAcceptHeader(
				ISSIConstants.APPLICATION, "sdp");
		acceptHeader.setParameter("level", "1");
		sipRequest.addHeader(acceptHeader);
		sipRequest.addHeader(headerFactory.createAcceptHeader(
				ISSIConstants.APPLICATION, ISSIConstants.X_TIA_P25_ISSI));

		PriorityHeader priorityHeader = headerFactory
				.createPriorityHeader(priority + ";"
						+ (isEmergency ? "e" : "a"));
		sipRequest.addHeader(priorityHeader);

		ContentDispositionHeader cd = headerFactory
				.createContentDispositionHeader(ContentDispositionHeader.SESSION);
		cd.setHandling("required");
		sipRequest.addHeader(cd);
		MimeVersionHeader mvh = headerFactory.createMimeVersionHeader(1, 0);
		sipRequest.addHeader(mvh);

		if (logger.isDebugEnabled())
			logger.debug("createInviteForGroupCall: created a request = "
					+ sipRequest);

		return sipRequest;
	}

	private void groupCallConfirm(ResponseEvent responseEvent) throws Exception {
		ClientTransaction ct = responseEvent.getClientTransaction();
		Dialog dialog = ct.getDialog();
		Response response = responseEvent.getResponse();

		SessionDescription sdes = ContentList.getContentListFromMessage(
				response).getSdpContent().getSessionDescription();

		if (sdes == null) {

			this.rfss
					.logError("No session description in the response - dropping response "
							+ response);
			return;
		}
		// The sender preserves the origin field.

		// got an OK from the invitee.
		// now extract the session that we originally stored away
		// when we sent out the invite. One half of the session
		// has already been set up. The OK Contains the SDP answer.
		// we extract the port from the SDP answer.
		// long origin = sdes.getOrigin().getSessionId();
		// PttSession pendingSession = this.pendingSessions.get(origin);

		String groupId = SipUtils.getGroupIdFromMessage(response);

		GroupServing groupServing = this.getGroupCall(groupId);

		if (groupServing == null) {
			this.rfss
					.logError("Could not find the serving group record for group "
							+ groupId);
			return;
		}

		// GroupServing groupServing = (GroupServing)
		// dialog.getApplicationData();
		PttSessionMultiplexer pendingSession = groupServing.getMultiplexer();
		pendingSession.setRemoteRfssDomainName(groupServing.getGroupConfig()
				.getHomeRfss().getDomainName());

		if (pendingSession == null) {

			this.rfss
					.logError("Could not find pending session while processing  "
							+ response);
			return;

		}
		String ipAddress = sdes.getConnection().getAddress();

		MediaDescription mediaDescription = (MediaDescription) sdes
				.getMediaDescriptions(true).get(0);

		if (mediaDescription.getConnection() != null) {

			ipAddress = mediaDescription.getConnection().getAddress();

		}

		int port = mediaDescription.getMedia().getMediaPort();

		if (logger.isDebugEnabled()) {

			logger.debug("Setting ipAddress : " + ipAddress);
			logger.debug("Setting port " + port);

		}

		// The last four hex digits of the radical name are the
		// group ID.
		pendingSession.setRemoteIpAddress(ipAddress);

		pendingSession.setRemoteRtpRecvPort(port);

		if (pendingSession.getRtpSession().getRemoteRtpRecvPort() > 0) {
			pendingSession.getHeartbeatTransmitter().start();
		}

		CSeqHeader cseqHeader = (CSeqHeader) response
				.getHeader(CSeqHeader.NAME);
		Request ackRequest = dialog.createAck(cseqHeader.getSeqNumber());
		dialog.sendAck(ackRequest);

	}

	// //////////////////////////////////////////////////////////////////////////
	// Public methods
	// //////////////////////////////////////////////////////////////////////////

	/**
	 * Processes incoming requests for GROUP calls ( both at group home and at
	 * serving rfss).
	 * 
	 * @param requestEvent -
	 *            the JAIN-SIP request event.
	 */
	public void processRequest(RequestEvent requestEvent) {
		Request request = requestEvent.getRequest();
		ToHeader toHeader = ((ToHeader) request.getHeader(ToHeader.NAME));
		SipURI toUri = ((SipURI) toHeader.getAddress().getURI());
		String groupId = SipUtils.getRadicalName(toUri);
		PriorityHeader priorityHeader = (PriorityHeader) request
				.getHeader(PriorityHeader.NAME);
		String callId = ((CallIdHeader) request.getHeader(CallIdHeader.NAME))
				.getCallId();
		int priority = SipUtils.getPriorityValue(priorityHeader);
		boolean emergency = SipUtils.isEmergency(priorityHeader);
		GroupConfig groupConfig = this.topologyConfig.getGroupConfig(groupId);

		GroupMobilityManager gmm = this.getGroupMobilityManager();
		GroupHome groupHome = gmm.getGroup(groupConfig.getRadicalName());

		ServerTransaction st = null;
		try {
			if (request.getMethod().equals(Request.INVITE)) {
				// We are processing the INVITE for the group call.
				// Look for whether there are colliding invite calls.

				st = requestEvent.getServerTransaction();
				if (st == null) {
					try {
						st = provider.getNewServerTransaction(request);
					} catch (Exception ex) {
						logger.error("Request already processed!!", ex);
						return;
					}
				}

				if (this.rfssConfig == groupConfig.getHomeRfss()) {
					// Check to see if we know about this group. If not, return
					// an error.
					if (groupHome == null) {
						Response response = SipUtils.createErrorResponse(
								rfssConfig, request,
								WarningCodes.NOT_FOUND_UNKNOWN_TARGET_GROUP);
						try {
							st.sendResponse(response);
						} catch (Exception ex) {
							logger.error("Unexpected error sending response",
									ex);
						}
						return;
					}

					// If we are at the home RFSS check if there are any ongoing
					// delayed requests.
					// if so, then block the call nd return error.

					if (groupHome.getPendingRequestCount() > 0) {
						// This is not a reinvitation.
						if (isCurrentCallHigheherPriority(emergency, priority,
								groupHome.isEmergency(), groupHome
										.getPriority())) {
							// There is a higher priority call queued.
							logger
									.debug("collision detected -- higher priority call in progress");
							this.handleCollidingRequest(requestEvent, st);

						} else {
							// If we are a higher priority call, we are at the
							// top of the queue.
							groupHome.setPriority(priority);
							groupHome.setEmergency(emergency);
							// groupHome.incrementPendingRequestCount();
							groupHome.setCallId(callId);

							this.actuallyProcessRequest(requestEvent, st);

						}
					} else {
						groupHome.setPriority(priority);
						groupHome.setEmergency(emergency);
						groupHome.setCallId(callId);
						groupHome.setServerTransaction(st);
						ViaHeader viaHeader = (ViaHeader) request
								.getHeader(ViaHeader.NAME);
						logger.debug("myRfssId = "
								+ this.rfssConfig.getDomainName());
						logger.debug("request is sent from = "
								+ viaHeader.getHost());
						if (rfssConfig.getGroupCallInviteProcessingTime() != 0
								&& !viaHeader.getHost().equals(
										this.rfssConfig.getDomainName())) {
							logger
									.debug("Delaying processing of the INVITE at Group Home");
							groupHome.incrementPendingRequestCount();

							TimerTask delayTask = new GroupInviteHomeDelayTask(
									requestEvent, groupHome, st);
							groupHome.setDelayTask(delayTask);
							ISSITimer
									.getTimer()
									.schedule(
											delayTask,
											rfssConfig
													.getGroupCallInviteProcessingTime() * 1000);

						} else {
							this.actuallyProcessRequest(requestEvent, st);
						}

					}

				} else {
					if (rfssConfig.getGroupCallInviteProcessingTime() != 0) {
						logger
								.debug("Creating Group Invite serving delay task "
										+ rfssConfig
												.getGroupCallInviteProcessingTime());
						GroupInviteServingDelayTask delayTask = new GroupInviteServingDelayTask(
								requestEvent, st);
						this.pendingTimers.put(st.getBranchId(), delayTask);
						ISSITimer
								.getTimer()
								.schedule(
										delayTask,
										rfssConfig
												.getGroupCallInviteProcessingTime() * 1000);

					} else {
						logger
								.debug("GroupInvite serving delay is 0 -- processing request");
						this.actuallyProcessRequest(requestEvent, st);
					}
				}
			} else {
				st = requestEvent.getServerTransaction();
				this.actuallyProcessRequest(requestEvent, st);

			}

		} catch (Exception ex) {
			logger.error("Unexpected exception", ex);
			rfss.getTestHarness().fail("Unexpected exception ");
		} finally {
			if (groupHome != null)
				logger.debug("Pending count = "
						+ groupHome.getPendingRequestCount());
		}

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see javax.sip.SipListener#processResponse(javax.sip.ResponseEvent)
	 */
	public void processResponse(ResponseEvent responseEvent) {
		logger.debug("Got a response");
		ClientTransaction ct = responseEvent.getClientTransaction();

		if (ct == null) {
			logger.debug("response retransmission -- dropping");
			return;
		}

		Response response = responseEvent.getResponse();
		ToHeader toHeader = (ToHeader) response.getHeader(ToHeader.NAME);
		CSeqHeader cseq = (CSeqHeader) response.getHeader(CSeqHeader.NAME);
		if (!cseq.getMethod().equals(Request.INVITE))
			return;
		SipURI toUri = (SipURI) toHeader.getAddress().getURI();
		String groupId = SipUtils.getRadicalName(toUri);
		GroupConfig groupConfig = this.topologyConfig.getGroupConfig(groupId);

		if (groupConfig == null) {
			logger.error("Unepxected error, group not found ");
			return;
		}

		if (response.getStatusCode() == Response.OK) {
			try {
				// I am the home RFSS of this Group.
				if (ct.getApplicationData() != null
						&& ct.getApplicationData() instanceof MmfSession) {

					// This is the controller that got the response.
					MmfSession mmfSession = (MmfSession) ct
							.getApplicationData();
					ContentList clist = ContentList
							.getContentListFromMessage(response);
					SessionDescription sessionDescription = clist
							.getSdpContent().getSessionDescription();
					// Got a response.
					long gid = Long.parseLong(groupId, 16);
					String senderDomainName = SipUtils.getContactHost(response);
					this.rfss.getTransmissionControlManager().addMediaLeg(gid,
							mmfSession, sessionDescription, senderDomainName);
					Dialog dialog = ct.getDialog();
					long seqno = cseq.getSeqNumber();
					// Ack the 200 OK.
					GroupHome groupHome = (GroupHome) ct.getDialog()
							.getApplicationData();
					groupHome.addPttSession(senderDomainName.toLowerCase(),
							mmfSession);

					Request ack = dialog.createAck(seqno);
					dialog.sendAck(ack);
					if (mmfSession.getRtpSession().getRemoteRtpRecvPort() > 0) {
						mmfSession.getHeartbeatTransmitter().start();
					}
				} else {

					PendingResponse pendingResponse = (PendingResponse) ct
							.getApplicationData();

					GroupServing calledGroup = this.getGroupCall(groupId);
					if (calledGroup == null) {
						this.rfss
								.getTestHarness()
								.fail(
										"Could not find group from the Response Check To heade");
						return;
					}

					int rc = responseEvent.getResponse().getStatusCode();
					if (rc / 100 == 2) {

						this.groupCallConfirm(responseEvent);

					} else {
						logger.info("Ignoring response code " + rc);

					}
					assert calledGroup != null;
					this.activeGroupCalls.put(calledGroup.getGroupConfig()
							.getRadicalName(), calledGroup);
					GroupCallSetupScenario scenario = pendingResponse.scenario;
					if (scenario.isTalkSpurtSentAfterCallSetup()) {
						SuConfig suConfig = scenario.getCallingSuConfig();
						if (this.rfss.isSubscriberUnitServed(suConfig)) {
							((TestSU) suConfig.getSU())
									.doSendTalkSpurt(scenario);
						}
					}

				}
			} catch (Exception ex) {
				this.rfss.getTestHarness().fail("Unexpected exception ", ex);
			}
		} else if (response.getStatusCode() / 100 > 2) {

			// handle the error cases here.
			if (ct.getApplicationData() != null
					&& ct.getApplicationData() instanceof MmfSession) {
				PttSessionInterface mmfSession = (PttSessionInterface) ct
						.getApplicationData();
				mmfSession.shutDown();
			} else {
				logger.debug("application data = "
						+ ct.getDialog().getApplicationData());
				GroupServing calledGroup = (GroupServing) ct.getDialog()
						.getApplicationData();
				calledGroup.getMultiplexer().shutDown();
			}
			this.activeGroupCalls.remove(groupConfig.getRadicalName());

		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see javax.sip.SipListener#processTimeout(javax.sip.TimeoutEvent)
	 */
	public void processTimeout(TimeoutEvent timeoutEvent) {
		logger.error("Client tx = " + timeoutEvent.getClientTransaction());
		Request request = null;
		if (timeoutEvent.getClientTransaction() != null) {
			logger.error("Ct request = "
					+ timeoutEvent.getClientTransaction().getRequest());
			request = timeoutEvent.getClientTransaction().getRequest();
		}

		if (timeoutEvent.getServerTransaction() != null) {
			logger.error("St request = "
					+ timeoutEvent.getServerTransaction().getRequest());
			request = timeoutEvent.getServerTransaction().getRequest();
		}
		this.rfss.logError("Timeout occured for the following request :\n"
				+ request);

		if (!timeoutEvent.isServerTransaction()) {
			// Client transaction timed out. S
			ClientTransaction ct = timeoutEvent.getClientTransaction();

			ToHeader toHeader = ((ToHeader) request.getHeader(ToHeader.NAME));
			SipURI toUri = ((SipURI) toHeader.getAddress().getURI());
			String groupId = SipUtils.getRadicalName(toUri);
			GroupConfig groupConfig = this.topologyConfig
					.getGroupConfig(groupId);
			if (groupConfig.getHomeRfss() == this.rfssConfig
					&& request.getMethod().equals(Request.INVITE)) {
				GroupHome groupHome = this.getGroupMobilityManager().getGroup(
						groupId);
				Dialog dialog = ct.getDialog();
				groupHome.removeDialog(dialog);
			}
		}

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see javax.sip.SipListener#processIOException(javax.sip.IOExceptionEvent)
	 */
	public void processIOException(IOExceptionEvent ioex) {
		this.rfss.getTestHarness().fail("Unexpected event -- IOException ");
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see javax.sip.SipListener#processTransactionTerminated(javax.sip.TransactionTerminatedEvent)
	 */
	public void processTransactionTerminated(
			TransactionTerminatedEvent transactionTerminatedEvent) {
		ClientTransaction ct = transactionTerminatedEvent
				.getClientTransaction();
		if (ct != null) {
			if (ct.getDialog().getApplicationData() instanceof GroupHome) {
				GroupHome groupHome = (GroupHome) ct.getDialog()
						.getApplicationData();
				groupHome.removeClientTransaction(ct);
			}
		}

	}

	public void processDialogTerminated(
			DialogTerminatedEvent dialogTerminatedEvent) {
		Dialog dialog = dialogTerminatedEvent.getDialog();

		if (dialog.getApplicationData() instanceof GroupHome) {
			GroupHome groupHome = (GroupHome) dialog.getApplicationData();

			groupHome.removeDialog(dialog);
		} else {
			GroupServing gc = (GroupServing) dialog.getApplicationData();
			if (gc != null) {
				String gid = gc.getGroupConfig().getRadicalName();
				this.activeGroupCalls.remove(gid);
				gc.getMultiplexer().shutDown();

			}
		}

	}

	/**
	 * Join an arriving SU to an existing group.
	 * 
	 * @param suConfig
	 * @param groupRadicalName
	 * @return
	 * @throws Exception
	 */
	public GroupServing joinExistingGroup(SuConfig suConfig,
			String groupRadicalName) throws Exception {
		GroupConfig groupConfig = this.topologyConfig
				.getGroupConfig(groupRadicalName);
		logger.debug(String.format("joinExistingGroup: %s %s ", suConfig
				.getSuName(), groupConfig.getGroupName()));
		if (this.activeGroupCalls.containsKey(groupRadicalName)) {
			GroupServing gc = this.activeGroupCalls.get(groupRadicalName);
			logger.debug("Found an active group call at rfss "
					+ this.rfssConfig.getRfssName());

			assert !gc.containsSu(suConfig);

			RtpSession rtpSession = gc.getMultiplexer().getRtpSession();
			SmfSession smfSession = this.rfss.getTransmissionControlManager()
					.getPttManager().createSmfSession(rtpSession,
							LinkType.GROUP_SERVING);
			int tsn = this.rfss.getTransmissionControlManager().getPttManager()
					.getNewTsn(LinkType.GROUP_SERVING);

			smfSession.setTsn(suConfig.getSuId(), tsn);

			gc.getMultiplexer().addPttSession(tsn, smfSession);
			gc.getMultiplexer().addSu(suConfig.getRadicalName(), tsn);
			gc.addSu(suConfig);
			// Need to add an smf session to the smf mux for the group.

			return gc;
		} else
			return null;

	}

	/**
	 * Send out a request to establish a group call.
	 * 
	 * @param groupConfig
	 * @return
	 * @throws Exception
	 */
	public GroupServing subscribeToGroup(
			GroupCallSetupScenario groupCallSetupScenario) throws Exception {
		logger.debug("subscribeToGroup "
				+ groupCallSetupScenario.getCallingSuConfig().getSuName());
		GroupConfig calledGroup = groupCallSetupScenario.getCalledGroupConfig();
		ConfirmedCallSetupTime ccSetp = calledGroup.getGroupServiceProfile()
				.getConfirmedCallSetupTime();

		SuConfig suConfig = groupCallSetupScenario.getCallingSuConfig();
		int priority = groupCallSetupScenario.getPriority();
		boolean isEmergency = groupCallSetupScenario.isEmergency();
		boolean isConfirmed = ccSetp.getSetupTime() > 0;
		if (groupCallSetupScenario.isConfirmed()) {
			isConfirmed = groupCallSetupScenario.isConfirmed();
		}
		
		boolean isProtected = groupCallSetupScenario.isProtected();

		GroupServing groupServing = joinExistingGroup(suConfig, calledGroup
				.getRadicalName());
		if (groupServing != null)
			return groupServing;
		try {

			String contentType = "application/sdp";

			CallIdHeader callId = headerFactory.createCallIdHeader(System
					.currentTimeMillis()
					+ ".1@" + this.rfssConfig.getDomainName());

			PttManager pttManager = this.rfss.getTransmissionControlManager()
					.getPttManager();

			// the following should be random
			long sessId = (long) (Math.random() * 100000);

			// we are setting up the side of the RTP session
			// where we are going to be listening for incoming packets.
			// this is where we setup the initial SDP OFFER. We attach
			// it to the SIP INIVTE. Note in that we do not know
			// where the next hop is listening at this point. We will
			// know that only when the answer comes back.

			String ipAddress = this.rfss.getIpAddress();
			long sessVersion = (long) (Math.random() * 100000);
			int gid = calledGroup.getGroupId();
			PttSessionMultiplexer pttSessionMultiplexer = pttManager
					.createPttSessionMultiplexer(Long.toString(sessId), gid);
			pttSessionMultiplexer.setRemoteRfssDomainName(calledGroup
					.getHomeRfss().getDomainName());
			SmfSession smfSession = pttManager.createSmfSession(
					pttSessionMultiplexer.getRtpSession(),
					LinkType.GROUP_SERVING);
			smfSession.setRemoteRfssDomainName(calledGroup.getHomeRfss()
					.getDomainName());
			smfSession.getSmfReceiver().addListener(
					this.rfss.getCurrentTestCase());

			pttSessionMultiplexer.setOwningRfss(this.rfssConfig);
			int tsn = this.rfss.getTransmissionControlManager().getPttManager()
					.getNewTsn(LinkType.GROUP_SERVING);
			smfSession.setTsn(suConfig.getSuId(), tsn);
			pttSessionMultiplexer.addPttSession(tsn, smfSession);

			if (logger.isDebugEnabled()) {
				logger.debug("smf session : " + smfSession);
			}
			int port = pttSessionMultiplexer.getRtpSession().getMyRtpRecvPort();

			String sdpBody = "v=0\r\n" + "o=- " + Long.toString(sessId) + " "
					+ Long.toString(sessVersion) + " IN IP4 " + ipAddress
					+ "\r\n" + "s=" + ISSIConstants.P25_GROUP_CALL + "\r\n"
					+ "c=IN IP4 " + ipAddress + "\r\n" + "t=0 0\r\n"
					+ "m=audio " + port + " RTP/AVP 100\r\n"
					+ "a=rtpmap:100 X-TIA-P25-IMBE/8000\r\n";

			Request inviteRequest = this.createInviteForGroupCall(calledGroup,
					priority, isEmergency);

			inviteRequest.setHeader(callId);
			inviteRequest.removeHeader(ContentDispositionHeader.NAME);
			boolean rfResourceAvailable = this.rfss.isRfResourcesAvailable();

			if (isConfirmed || isProtected) {
				ContentList contentList = new ContentList();

				SdpContent sdpContent = SdpContent.createSdpContent(sdpBody);
				CallParamContent callParams = CallParamContent
						.createCallParamContent();

				callParams.getCallParam().setConfirmedGroupCall(isConfirmed);
				callParams.getCallParam().setProtectedMode(isProtected);
				if (this.rfss.isRfResourcesAvailable()) {
					callParams.getCallParam().setRfResourceAvailable(
							rfResourceAvailable);

				}
				contentList.add(sdpContent);
				contentList.add(callParams);

				inviteRequest.setContent(contentList.toString(), contentList
						.getContentTypeHeader());
			} else {

				ContentTypeHeader contentTypeHeader = (ContentTypeHeader) ProtocolObjects.headerFactory
						.createHeader(ContentTypeHeader.NAME, contentType);

				inviteRequest.setContent(sdpBody, contentTypeHeader);

			}
			SessionDescription sd = SdpFactory.getInstance()
					.createSessionDescription(sdpBody);
			smfSession.setSessionDescription(sd);
			PendingResponse pendingResponse = new PendingResponse();
			pendingResponse.procedure = Operation.CALL_SETUP;
			ClientTransaction ct = provider
					.getNewClientTransaction(inviteRequest);

			pendingResponse.scenario = groupCallSetupScenario;
			ct.setApplicationData(pendingResponse);

			Dialog dialog = ct.getDialog();
			groupServing = new GroupServing(this.rfss, dialog, ct, calledGroup,
					provider);
			pendingResponse.callSegment = groupServing;
			groupServing.setMultiplexer(pttSessionMultiplexer);
			groupServing.setPriority(priority);
			groupServing.setEmergency(isEmergency);
			this.activeGroupCalls.put(calledGroup.getRadicalName(),
					groupServing);
			dialog.setApplicationData(groupServing);
			ct.sendRequest();
			groupServing.addSu(suConfig);
			pttSessionMultiplexer.addSu(suConfig.getRadicalName(), tsn);
			// this.activeGroupCalls.put(calledGroup.getRadicalName(),
			// callSegment);
			return groupServing;
		} catch (Exception ex) {
			this.rfss.getTestHarness().fail("unexpected exception", ex);
			logger.error("Unexpected exception", ex);
			throw ex;

		}
	}

	/**
	 * RTP resources have become available at the home RFSS. Re-invite group
	 * members who have established half sessions with me. Reinvite myself to
	 * groups that I am serving.
	 * 
	 * 
	 * @param nports - #
	 *            of ports to add.
	 * 
	 * @throws Exception
	 */
	void rtpResourcesAvailable(int nports) throws Exception {
		int portsUsed = 0;
		logger.debug("activeGroupCalls : " + this.activeGroupCalls);
		logger.debug("GroupCallControlManager : Adding an rtp port ");
		Collection<GroupHome> groupHomes = this.getGroupMobilityManager()
				.getGroups();
		for (GroupHome groupHome : groupHomes) {
			for (MmfSession mmfSession : groupHome.getMmfSessions()) {
				if (mmfSession.getMyRtpRecvPort() == 0) {

					mmfSession.assignRtpRecvPort();
					Dialog dialog = groupHome.getDialog(mmfSession);
					String domainName = ((SipURI) dialog.getRemoteTarget()
							.getURI()).getHost();
					if (logger.isDebugEnabled())
						logger.debug("domain name for remote target = "
								+ domainName);

					ClientTransaction clientTransaction = groupHome
							.getClientTransaction(domainName);
					Request originalRequest = clientTransaction != null ? clientTransaction
							.getRequest()
							: groupHome.getServerTransaction().getRequest();
					ContentList contentList = ContentList
							.getContentListFromMessage(originalRequest);
					CallParamContent callParam = null;
					if (contentList.getCallParamContent() != null) {
						callParam = contentList.getCallParamContent();
					}
					Request request = this.createReInviteForGroupCall(
							groupHome, dialog);
					SessionDescription sd = mmfSession.getSessionDescription();
					CallControlManager.fixupSdpAnnounce(rfssConfig, mmfSession,
							sd);
					if (callParam == null) {
						request.setContent(sd.toString(), headerFactory
								.createContentTypeHeader("application", "sdp"));
					} else {
						ContentList newContentList = new ContentList();
						SdpContent sdpContent = SdpContent.createSdpContent();
						sdpContent.setSessionDescription(sd);
						newContentList.add(sdpContent);
						callParam.getCallParam().setRfResourceAvailable(
								this.rfss.isRfResourcesAvailable());
						newContentList.add(callParam);
						request.setContent(newContentList.toString(),
								newContentList.getContentTypeHeader());

					}
					ClientTransaction ct = groupHome.getProvider()
							.getNewClientTransaction(request);
					ct.setApplicationData(mmfSession);
					groupHome.setClientTransaction(domainName, ct);
					dialog.sendRequest(ct);
					portsUsed++;
					if (portsUsed >= nports)
						return;

				}
			}
		}

		// look for groups for which we are serving and for which we dont
		// have
		// rtp resources;
		for (GroupServing groupServing : this.activeGroupCalls.values()) {
			if (groupServing.getMultiplexer().getMyRtpRecvPort() == 0) {
				groupServing.getMultiplexer().assignRtpRecvPort();
				Dialog dialog = groupServing.getDialog();
				// The SIP dialog which we are using to talk to our home
				Request request = this.createReInviteForGroupCall(groupServing,
						dialog);
				Request originalRequest = groupServing.getLastTransaction()
						.getRequest();
				ContentList contentList = ContentList
						.getContentListFromMessage(originalRequest);
				if (contentList.getCallParamContent() != null) {
					contentList.getCallParamContent().getCallParam()
							.setRfResourceAvailable(
									this.rfss.isRfResourcesAvailable());
				}
				SessionDescription sd = contentList.getSdpContent()
						.getSessionDescription();

				CallControlManager.fixupSdpAnnounce(rfssConfig, groupServing
						.getMultiplexer(), sd);

				request.setContent(contentList.toString(),
						(ContentTypeHeader) originalRequest
								.getHeader(ContentTypeHeader.NAME));
				ClientTransaction ct = groupServing.getProvider()
						.getNewClientTransaction(request);
				groupServing.setClientTransaction(ct);
				PendingResponse pendingResponse = new PendingResponse();
				pendingResponse.callSegment = groupServing;
				pendingResponse.procedure = Operation.REINVITE;
				dialog.sendRequest(ct);
				portsUsed++;
				if (portsUsed >= nports)
					return;
			}
		}

		this.rfss.getTransmissionControlManager().getPttManager()
				.incrementPortLimit(nports - portsUsed);
	}

	/**
	 * Remove rtp resources.
	 * 
	 */
	void removeRtpPorts(int nports) throws Exception {
		int portsReleased = 0;
		logger.debug("activeGroupCalls : " + this.activeGroupCalls);
		logger.debug("GroupCallControlManager : removing rtp ports ");
		Collection<GroupHome> groupHomes = this.getGroupMobilityManager()
				.getGroups();
		for (GroupHome groupHome : groupHomes) {
			for (MmfSession mmfSession : groupHome.getMmfSessions()) {
				if (mmfSession.getMyRtpRecvPort() != 0) {

					mmfSession.stopReceiver();
					Dialog dialog = groupHome.getDialog(mmfSession);
					String domainName = ((SipURI) dialog.getRemoteTarget()
							.getURI()).getHost();
					if (logger.isDebugEnabled())
						logger.debug("domain name for remote target = "
								+ domainName);
					ClientTransaction clientTransaction = groupHome
							.getClientTransaction(domainName);
					logger
							.debug("client tx "
									+ groupHome
											.getClientTransaction(domainName)
									+ " server tx  "
									+ groupHome.getServerTransaction());
					Request originalRequest = clientTransaction != null ? clientTransaction
							.getRequest()
							: groupHome.getServerTransaction().getRequest();
					ContentList contentList = ContentList
							.getContentListFromMessage(originalRequest);
					CallParamContent callParam = null;
					if (contentList.getCallParamContent() != null) {
						callParam = contentList.getCallParamContent();
					}
					Request request = this.createReInviteForGroupCall(
							groupHome, dialog);
					SessionDescription sd = mmfSession.getSessionDescription();
					CallControlManager.fixupSdpAnnounce(rfssConfig, mmfSession,
							sd);
					if (callParam == null) {
						request.setContent(sd.toString(), headerFactory
								.createContentTypeHeader("application", "sdp"));
					} else {
						ContentList newContentList = new ContentList();
						SdpContent sdpContent = SdpContent.createSdpContent();
						sdpContent.setSessionDescription(sd);
						newContentList.add(sdpContent);
						callParam.getCallParam().setRfResourceAvailable(
								this.rfss.isRfResourcesAvailable());
						newContentList.add(callParam);
						request.setContent(newContentList.toString(),
								newContentList.getContentTypeHeader());

					}
					ClientTransaction ct = groupHome.getProvider()
							.getNewClientTransaction(request);
					ct.setApplicationData(mmfSession);
					groupHome.setClientTransaction(domainName, ct);
					dialog.sendRequest(ct);
					portsReleased++;
					if (portsReleased > nports)
						return;

				}
			}
		}
		// look for groups for which we are serving and for which we dont
		// have
		// rtp resources;
		for (GroupServing groupServing : this.activeGroupCalls.values()) {
			if (groupServing.getMultiplexer().getMyRtpRecvPort() != 0) {
				groupServing.getMultiplexer().stopReceiver();
				Dialog dialog = groupServing.getDialog();
				// The SIP dialog which we are using to talk to our home
				Request request = this.createReInviteForGroupCall(groupServing,
						dialog);
				Request originalRequest = groupServing.getLastTransaction()
						.getRequest();
				ContentList contentList = ContentList
						.getContentListFromMessage(originalRequest);
				
				ContentList newContentList = new ContentList();
		
				if (contentList.getCallParamContent() != null && 
						contentList.getCallParamContent().getCallParam().isConfirmedGroupCall()) {
					CallParamContent newCallParamContent = CallParamContent.createCallParamContent();
					newCallParamContent.getCallParam().setRfResourceAvailable(this.rfss.isRfResourcesAvailable());
					newContentList.add(newCallParamContent);
					newContentList.add(newCallParamContent);
				}
				SessionDescription sd = contentList.getSdpContent()
						.getSessionDescription();

				CallControlManager.fixupSdpAnnounce(rfssConfig, groupServing
						.getMultiplexer(), sd);
				SdpContent newSdpContent = SdpContent.createSdpContent();
				newSdpContent.setSessionDescription(sd);
				newContentList.add(newSdpContent);

				request.setContent(newContentList.toString(), newContentList.getContentTypeHeader());
						
				ClientTransaction ct = groupServing.getProvider()
						.getNewClientTransaction(request);
				groupServing.setClientTransaction(ct);
				dialog.sendRequest(ct);
				portsReleased++;
				if (portsReleased > nports)
					return;
			}
		}

	}

	/**
	 * Get the current group call corresponding to a group radical name.
	 * 
	 * @return -- the group call if it exists or null if not.
	 */
	GroupServing getGroupCall(String groupRadicalName) {
		return this.activeGroupCalls.get(groupRadicalName);
	}

	/**
	 * For the all groups for which I am HOME RFSS, send bye on all call legs.
	 * For all groups for which I am a member RFSS send bye.
	 * 
	 */
	void dropAllCalls() {
		try {
			logger
					.debug("Active group calls "
							+ this.activeGroupCalls.values());
			for (GroupConfig groupConfig : this.topologyConfig
					.getGroupConfigurations()) {
				RfssConfig home = groupConfig.getHomeRfss();
				if (home == this.rfssConfig) {
					GroupMobilityManager gmm = this.getGroupMobilityManager();
					GroupHome groupHome = gmm.getGroup(groupConfig
							.getRadicalName());

					for (Dialog dialog : groupHome.getDialogs()) {

						Request bye = dialog.createRequest(Request.BYE);
						ClientTransaction ct = this.provider
								.getNewClientTransaction(bye);
						dialog.sendRequest(ct);
					}
				}

			}

			/*
			 * For all groups for which I am a member, drop the call.
			 */
			for (GroupServing groupServing : this.activeGroupCalls.values()) {
				Dialog dialog = groupServing.getDialog();
				Request request = dialog.createRequest(Request.BYE);
				ClientTransaction ctx = this.provider
						.getNewClientTransaction(request);
				dialog.sendRequest(ctx);
			}
		} catch (Exception ex) {
			logger.fatal("Error sending or creating bye ", ex);

		}

	}

	/**
	 * Re-invite a group member to a call.
	 * 
	 * @param groupId -
	 *            Group Id to reinvite
	 * @param rfssConfig -
	 *            Rfss to reinvite.
	 */

	void reInviteRfssToGroup(String radicalName, RfssConfig rfssConfig)
			throws Exception {
		GroupHome groupHome = this.getGroupMobilityManager().getGroup(
				radicalName);
		if (groupHome == null)
			throw new Exception("Group not found " + radicalName);
		String rfssDomainName = rfssConfig.getDomainName();
		if (logger.isDebugEnabled())
			logger.debug("domain name for remote target = " + rfssDomainName);

		MmfSession mmfSession = groupHome.getMmfSession(rfssDomainName);
		if (mmfSession == null)
			throw new Exception("MmfSession not found for this Rfss "
					+ rfssConfig.getRfssName());

		Dialog dialog = groupHome.getDialog(mmfSession);
		if (dialog == null)
			throw new Exception("Could not find dialog for re-inviting");

		Request request = this.createReInviteForGroupCall(groupHome, dialog);
		SessionDescription sd = mmfSession.getSessionDescription();
		CallControlManager.fixupSdpAnnounce(rfssConfig, mmfSession, sd);
		request.setContent(sd.toString(), headerFactory
				.createContentTypeHeader("application", "sdp"));
		ClientTransaction ct = groupHome.getProvider().getNewClientTransaction(
				request);
		ct.setApplicationData(mmfSession);
		groupHome.setClientTransaction(rfssDomainName, ct);
		dialog.sendRequest(ct);

	}

	/**
	 * Send a cancel for an outstanding invite of a group call.
	 * 
	 * @param groupServing -
	 *            group serving to send cancel to.
	 * 
	 * @throws Exception
	 */
	void sendCancel(GroupServing groupServing) throws Exception {

		ClientTransaction ct = groupServing.getClientTransaction();
		if (ct.getState() == TransactionState.PROCEEDING) {
			Request cancelRequest = ct.createCancel();
			ClientTransaction cancelTx = provider
					.getNewClientTransaction(cancelRequest);
			cancelTx.setApplicationData(groupServing);
			cancelTx.sendRequest();
		} else if (rfssConfig.isAssignedGroup(groupServing.getGroupConfig()
				.getRadicalName())) {
			logger.debug("Sending CANCEL from group home!");
			// We are home for this group.
			// We send CANCEL to the other legs.
			GroupHome groupHome = this.getGroupMobilityManager().getGroup(
					groupServing.getGroupConfig().getRadicalName());
			for (ClientTransaction inviteTx : groupHome.getClientTransactions()) {
				if (inviteTx.getState() == TransactionState.PROCEEDING) {
					Request cancelRequest = inviteTx.createCancel();
					ClientTransaction cancelTx = provider
							.getNewClientTransaction(cancelRequest);
					cancelTx.sendRequest();
				}

			}

		}
	}

	/**
	 * Re-invite to the group home or group serving when RF resources become
	 * available.
	 * 
	 * @param resourceVal --
	 *            true / false indicates whether rf resources became available
	 *            or not
	 * 
	 * @throws Exception
	 */
	void alertRfResourceChange(boolean resourceVal) throws Exception {

		for (GroupServing groupServing : this.activeGroupCalls.values()) {
			Dialog dialog = groupServing.getDialog();
			// The SIP dialog which we are using to talk to our home
			Request request = this.createReInviteForGroupCall(groupServing,
					dialog);
			PttSessionMultiplexer pttSession = groupServing.getMultiplexer();
			Request originalRequest = groupServing.getLastTransaction()
					.getRequest();
			ContentList contentList = ContentList
					.getContentListFromMessage(originalRequest);
			if (contentList.getCallParamContent() != null) {
				contentList.getCallParamContent().getCallParam()
						.setRfResourceAvailable(
								this.rfss.isRfResourcesAvailable());
			}
			SessionDescription sd = contentList.getSdpContent()
					.getSessionDescription();
			CallControlManager.fixupSdpAnnounce(rfssConfig, pttSession, sd);

			request.setContent(contentList.toString(), contentList
					.getContentTypeHeader());
			ClientTransaction ct = groupServing.getProvider()
					.getNewClientTransaction(request);
			groupServing.setClientTransaction(ct);
			dialog.sendRequest(ct);

		}

		for (GroupHome groupHome : this.getGroupMobilityManager().getGroups()) {
			for (MmfSession mmfSession : groupHome.getMmfSessions()) {
				Dialog dialog = groupHome.getDialog(mmfSession);
				String domainName = ((SipURI) dialog.getRemoteTarget().getURI())
						.getHost();
				if (logger.isDebugEnabled())
					logger.debug("domain name for remote target = "
							+ domainName);

				Request originalRequest = groupHome.getClientTransaction(
						domainName).getRequest();
				ContentList contentList = ContentList
						.getContentListFromMessage(originalRequest);
				CallParamContent callParam = null;
				if (contentList.getCallParamContent() != null) {
					callParam = contentList.getCallParamContent();
				}
				Request request = this.createReInviteForGroupCall(groupHome,
						dialog);
				SessionDescription sd = mmfSession.getSessionDescription();
				CallControlManager.fixupSdpAnnounce(rfssConfig, mmfSession, sd);
				if (callParam == null) {
					request.setContent(sd.toString(), headerFactory
							.createContentTypeHeader("application", "sdp"));
				} else {
					ContentList newContentList = new ContentList();
					SdpContent sdpContent = SdpContent.createSdpContent();
					sdpContent.setSessionDescription(sd);
					newContentList.add(sdpContent);
					callParam.getCallParam().setRfResourceAvailable(
							this.rfss.isRfResourcesAvailable());
					newContentList.add(callParam);
					request.setContent(newContentList.toString(),
							newContentList.getContentTypeHeader());

				}
				ClientTransaction ct = groupHome.getProvider()
						.getNewClientTransaction(request);
				ct.setApplicationData(mmfSession);
				groupHome.setClientTransaction(domainName, ct);
				dialog.sendRequest(ct);

			}

		}

	}

	/**
	 * Alert for group priority change.
	 * 
	 * @param radicalName --
	 *            group radical name.
	 */
	void groupPriorityChange(String groupRadicalName, int newPriority,
			boolean isEmergency) throws Exception {

		logger.debug("groupPriorityChange : " + groupRadicalName
				+ " : newPiroirity " + newPriority + " isEmergency "
				+ isEmergency);
		GroupServing groupServing = this.getGroupCall(groupRadicalName);

		logger.debug("groupServing = " + groupServing);
		if (groupServing != null
				&& (groupServing.getPriority() != newPriority || groupServing
						.isEmergency() != isEmergency)) {

			Dialog dialog = groupServing.getDialog();
			// The SIP dialog which we are using to talk to our home
			Request request = this.createReInviteForGroupCall(groupServing,
					dialog);
			Request originalRequest = groupServing.getLastTransaction()
					.getRequest();
			ContentList contentList = ContentList
					.getContentListFromMessage(originalRequest);
			PriorityHeader priorityHeader = (PriorityHeader) request
					.getHeader(PriorityHeader.NAME);
			if (isEmergency) {
				priorityHeader.setPriority(new StringBuffer().append(
						Integer.toString(newPriority)).append(";").append(
						isEmergency ? "e" : "a").toString());
			}
			request.setHeader(priorityHeader);
			if (contentList.getCallParamContent() != null) {
				contentList.getCallParamContent().getCallParam()
						.setRfResourceAvailable(
								this.rfss.isRfResourcesAvailable());
			}
			SessionDescription sd = contentList.getSdpContent()
					.getSessionDescription();
			PttSessionMultiplexer pttSessionMux = groupServing.getMultiplexer();

			CallControlManager.fixupSdpAnnounce(rfssConfig, pttSessionMux, sd);

			request.setContent(contentList.toString(), contentList
					.getContentTypeHeader());
			ClientTransaction ct = groupServing.getProvider()
					.getNewClientTransaction(request);
			groupServing.setClientTransaction(ct);
			dialog.sendRequest(ct);

		} else {
			GroupHome groupHome = this.getGroupMobilityManager().getGroup(
					groupRadicalName);
			logger.debug("groupHome = " + groupHome);

			if (groupHome == null
					|| (groupHome.getPriority() == newPriority && groupHome
							.isEmergency() == isEmergency)) {
				logger.debug("could not find group home record ... returning ");
				return;
			}

			for (MmfSession mmfSession : groupHome.getMmfSessions()) {
				Dialog dialog = groupHome.getDialog(mmfSession);
				String domainName = ((SipURI) dialog.getRemoteTarget().getURI())
						.getHost();
				if (logger.isDebugEnabled())
					logger.debug("domain name for remote target = "
							+ domainName);

				Request originalRequest = groupHome.getClientTransaction(
						domainName).getRequest();
				ContentList contentList = ContentList
						.getContentListFromMessage(originalRequest);
				CallParamContent callParam = null;
				if (contentList.getCallParamContent() != null) {
					callParam = contentList.getCallParamContent();
				}
				Request request = this.createReInviteForGroupCall(groupHome,
						dialog);
				PriorityHeader priorityHeader = (PriorityHeader) request
						.getHeader(PriorityHeader.NAME);
				if (isEmergency) {
					priorityHeader.setPriority(new StringBuffer().append(
							Integer.toString(newPriority)).append(";").append(
							isEmergency ? "e" : "a").toString());
				}

				request.setHeader(priorityHeader);
				SessionDescription sd = mmfSession.getSessionDescription();
				CallControlManager.fixupSdpAnnounce(rfssConfig, mmfSession, sd);
				if (callParam == null) {
					request.setContent(sd.toString(), headerFactory
							.createContentTypeHeader("application", "sdp"));
				} else {
					ContentList newContentList = new ContentList();
					SdpContent sdpContent = SdpContent.createSdpContent();
					sdpContent.setSessionDescription(sd);
					newContentList.add(sdpContent);
					callParam.getCallParam().setRfResourceAvailable(
							this.rfss.isRfResourcesAvailable());
					newContentList.add(callParam);
					request.setContent(newContentList.toString(),
							newContentList.getContentTypeHeader());

				}
				ClientTransaction ct = groupHome.getProvider()
						.getNewClientTransaction(request);
				ct.setApplicationData(mmfSession);
				groupHome.setClientTransaction(domainName, ct);
				dialog.sendRequest(ct);

			}
		}

	}

}
