package gov.nist.p25.issi.rfss;

import gov.nist.p25.issi.ISSITimer;
import gov.nist.p25.issi.issiconfig.GroupConfig;
import gov.nist.p25.issi.issiconfig.ISSIConstants;
import gov.nist.p25.issi.issiconfig.RfssConfig;
import gov.nist.p25.issi.issiconfig.SuConfig;
import gov.nist.p25.issi.issiconfig.SuState;
import gov.nist.p25.issi.issiconfig.TopologyConfig;
import gov.nist.p25.issi.p25body.ContentList;
import gov.nist.p25.issi.p25body.params.RegisterParam;
import gov.nist.p25.issi.p25body.serviceprofile.ServiceProfile;
import gov.nist.p25.issi.rfss.UnitToUnitMobilityManager.PendingResponse;
import gov.nist.p25.issi.rfss.UnitToUnitMobilityManager.Proceedure;
import gov.nist.p25.issi.rfss.tester.AbstractSipSignalingTest;
import gov.nist.p25.issi.rfss.tester.TestSU;
import gov.nist.p25.issi.testlauncher.TestHarness;
import gov.nist.p25.issi.transctlmgr.PttPointToMultipointSession;
import gov.nist.p25.issi.transctlmgr.ptt.SessionType;
import gov.nist.p25.issi.utils.SipUtils;
import gov.nist.p25.issi.utils.WarningCodes;

import java.util.Collection;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.TimerTask;

import javax.sip.ClientTransaction;
import javax.sip.DialogTerminatedEvent;
import javax.sip.IOExceptionEvent;
import javax.sip.RequestEvent;
import javax.sip.ResponseEvent;
import javax.sip.SipListener;
import javax.sip.SipProvider;
import javax.sip.TimeoutEvent;
import javax.sip.TransactionState;
import javax.sip.TransactionTerminatedEvent;
import javax.sip.address.Address;
import javax.sip.address.AddressFactory;
import javax.sip.address.SipURI;
import javax.sip.header.AcceptHeader;
import javax.sip.header.AllowHeader;
import javax.sip.header.CSeqHeader;
import javax.sip.header.CallIdHeader;
import javax.sip.header.ContactHeader;
import javax.sip.header.ContentDispositionHeader;
import javax.sip.header.ContentTypeHeader;
import javax.sip.header.ExpiresHeader;
import javax.sip.header.FromHeader;
import javax.sip.header.HeaderFactory;
import javax.sip.header.MaxForwardsHeader;
import javax.sip.header.RouteHeader;
import javax.sip.header.TimeStampHeader;
import javax.sip.header.ToHeader;
import javax.sip.header.ViaHeader;
import javax.sip.message.MessageFactory;
import javax.sip.message.Request;
import javax.sip.message.Response;

import org.apache.log4j.Logger;

/**
 * Implements the functionality of the Group Mobility Manager gor groups. From
 * the ISSI spec, the function is described as follows: Within each RFSS, the
 * Group Mobility Management Function (GMMF) is responsible for group mobility
 * management. The GMMF supports different functions depending on the role of
 * the RFSS with respect to the group in question.
 * 
 * When the RFSS is the serving RFSS relative to the group, the GMMF is
 * responsible for the following functions:
 * 
 * <ul>
 * <li> - Registration of the group with the Home RFSS when the RFSS detects it
 * has SUs currently affiliated with the group in its serving area,
 * <li> - De-registration of the group with the home RFSS when the RFSS
 * determines that it no longer has affiliated members for the group,
 * <li> - Responding to the home RFSS when the Home RFSS requests confirmation
 * that the RFSS still has an interest in being registered for the group,
 * <li> - Requesting service profile information for the group on the ISSI only
 * when the RFSS is currently registered to the SG, or is in the process of
 * registering with the group.
 * </ul>
 * 
 * 
 * When the RFSS is the home RFSS relative to the group, the GMMF is responsible
 * for the following function:
 * <ul>
 * <li> - Providing the group service profile information to suitably authorized
 * serving RFSSs that request the information.
 * <li> - Prohibiting registration of the group in serving domains for which it
 * is not authorized,
 * <li> - Maintaining a list of serving RFSSs of the SG
 * <li> - Checking whether a serving RFSS still has an interest in being
 * registered for the SG, in accordance with local policy.
 * </ul>
 * 
 * <b>
 * 
 * NOTE The list of affiliated P25 subscriber units to a group in a serving RFSS
 * is not known at the SGs home RFSS. As a consequence, the SGs home RFSS might
 * only control the right of a serving RFSS to be in a group call coverage or
 * not.
 * 
 * </b>
 * 
 * Implementation Notes: This class is structured as a SipListener although it
 * is not directly registered with the SIP Stack. Instead it is registered with
 * the Mobility manager which directs SIP Events to an instance of this class by
 * looking at the From and To headers of the incoming SIP request/response to
 * determine whether the Request/Response belongs to a Group or to an SU to SU
 * call. Although SU's belong to Groups, during runtime, group memebership is
 * tracked on the basis of RFSS's. This is because a given Serving RFSS keeps at
 * most one active session with a group and multiplexes all signialing and media
 * to all the SU's in that group which it is currently serving. Thus, at the
 * RFSS, we maintain a dynamic structure called "Group" which tracks RFSS
 * membership of the Group.
 * 
 * @see gov.nist.p25.rfss.GroupHome
 * 
 * @author M. Ranganathan
 */

public class GroupMobilityManager implements SipListener {

	private RfssConfig rfssConfig;

	private RFSS rfss;

	private SipProvider provider;

	private TopologyConfig topologyConfig;

	private static HeaderFactory headerFactory = ProtocolObjects.headerFactory;

	private static MessageFactory messageFactory = ProtocolObjects.messageFactory;

	private static AddressFactory addressFactory = ProtocolObjects.addressFactory;

	private static Logger logger = Logger.getLogger(GroupMobilityManager.class);

	/*
	 * The Groups for which we are the home RFSS
	 */
	private Hashtable<String, GroupHome> myGroups;

	/*
	 * The Groups for which there are currently served SUs.
	 */
	private Hashtable<String, Integer> groupRefcountTable;
	
	enum Operation {
		REGISTER, QUERY, DEREGISTER, HOME_QUERY
	}
	class PendingResponse {
		Operation op;
		ClientTransaction ct;
		Response response;
	}
	
	/**
	 * Create a home query request.
	 * 
	 * SIP Register-Query is sent by a serving RFSS to query a home RFSS for
	 * service profile information, and by a home RFSS to confirm the interest
	 * of a serving RFSS in a mobility object. It is distinguished by the
	 * absence of both the expires parameter and the contact parameter. The SIP
	 * Register-Query message is used in two procedures described in ?5.5, the
	 * Home Query procedure and the Serving Query procedure. The content of the
	 * message is consistent in each case, and the particular procedure is
	 * determined by the roles of the participating RFSSs pursuant to the
	 * mobility object involved.
	 */
	private Request createHomeQueryRequest(GroupConfig groupConfig,
			RfssConfig targetRfss, boolean force, boolean confirm) {
		try {
			String groupName = groupConfig.getRadicalName();
			SipURI fromURI = (SipURI) SipUtils.createGroupSipURI(groupName);
			String tag = SipUtils.createTag();
			Address fromAddress = addressFactory.createAddress(fromURI);
	
			FromHeader fromHeader = headerFactory.createFromHeader(fromAddress,
					tag);
	
			SipURI toURI = fromURI;
			Address toAddress = addressFactory.createAddress(toURI);
	
			ToHeader toHeader = headerFactory.createToHeader(toAddress, null);
	
			String domainName = targetRfss.getDomainName();
	
			// The request URI for Register requests has the form
			// sip:domain-name
			SipURI registrationURI = SipUtils.createDomainSipURI(domainName);
	
			// Create the via header for the outgoing request.
			ViaHeader viaHeader = SipUtils
					.createViaHeaderForRfss(this.rfssConfig);
			LinkedList<ViaHeader> viaHeaders = new LinkedList<ViaHeader>();
			viaHeaders.add(viaHeader);
	
			// Create a maxforwards header for the outgoing request.
			MaxForwardsHeader maxForwards = headerFactory
					.createMaxForwardsHeader(70);
	
			CallIdHeader callIdHeader = provider.getNewCallId();
	
			CSeqHeader cseqHeader = headerFactory.createCSeqHeader(1L,
					Request.REGISTER);
	
			Request sipRequest = messageFactory.createRequest(registrationURI,
					Request.REGISTER, callIdHeader, cseqHeader, fromHeader,
					toHeader, viaHeaders, maxForwards);
	
			SipUtils.addAllowHeaders(sipRequest);
			AcceptHeader acceptHeader = headerFactory.createAcceptHeader(
					ISSIConstants.APPLICATION, ISSIConstants.SDP);
			acceptHeader.setParameter("level", "1");
			sipRequest.addHeader(acceptHeader);
			acceptHeader = headerFactory.createAcceptHeader(
					ISSIConstants.APPLICATION, ISSIConstants.X_TIA_P25_ISSI);
			sipRequest.addHeader(acceptHeader);
	
			TimeStampHeader tsHeader = headerFactory.createTimeStampHeader(0);
			tsHeader.setTime(System.currentTimeMillis());
			sipRequest.setHeader(tsHeader);
			RegisterParam registerParams = RegisterParam.createRegisterParam();
			ContentTypeHeader ctHdr = headerFactory.createContentTypeHeader(
					ISSIConstants.APPLICATION, ISSIConstants.X_TIA_P25_ISSI);
			if (force) {
				registerParams.setForceRegister(force);
			}
	
			if (confirm) {
				registerParams.setConfirmRegister(confirm);
			}
			sipRequest.setContent(registerParams.toString(), ctHdr);
	
			return sipRequest;
		} catch (Exception ex) {
			this.rfss.getTestHarness().fail("unexpected exception", ex);
			logger.fatal("unexpected exception ", ex);
			return null;
		}
	
	}

	/**
	 * Create a register request and point it to the home rfss of the SU who
	 * sent us a register.
	 * 
	 * SIP Register-Register is used as part of the Registration procedure to
	 * provide a means for a serving RFSS to express interest in a mobility
	 * object. The Register message is distinguished by the presence of a
	 * contact parameter and a non-zero expires parameter.
	 * 
	 * 
	 * SIP Register-Deregister is used by a serving RFSS to deregister a
	 * mobility object or by a home RFSS to request deregistration of an SU. The
	 * SIP Register-Deregister command is distinguished by the presence of a
	 * contact parameter and a zero expires parameter. As with the SIP
	 * Register-Query message, the SIP Register-Deregister message is used in
	 * two separate procedures, the De-registration procedure and the Roamed
	 * procedure. Again, the choice of procedure is dependant on the roles of
	 * the RFSSs involved.
	 * 
	 * SIP Register-Query is sent by a serving RFSS to query a home RFSS for
	 * service profile information, and by a home RFSS to confirm the interest
	 * of a serving RFSS in a mobility object. It is distinguished by the
	 * absence of both the expires parameter and the contact parameter. The SIP
	 * Register-Query message is used in two procedures described in ?5.5, the
	 * Home Query procedure and the Serving Query procedure. The content of the
	 * message is consistent in each case, and the particular procedure is
	 * determined by the roles of the participating RFSSs pursuant to the
	 * mobility object involved.
	 * 
	 * 
	 * 
	 * @param groupConfig -
	 *            the group to register.
	 * @param expiresTime
	 * @return -- the formatted Register request
	 * 
	 */
	private Request createRegisterRequest(GroupConfig groupConfig,
			int expiresTime) {
	
		try {
			String groupName = groupConfig.getRadicalName();
			SipURI fromURI = (SipURI) SipUtils.createGroupSipURI(groupName);
			String tag = SipUtils.createTag();
			Address fromAddress = addressFactory.createAddress(fromURI);
	
			FromHeader fromHeader = headerFactory.createFromHeader(fromAddress,
					tag);
	
			SipURI toURI = fromURI;
			Address toAddress = addressFactory.createAddress(toURI);
	
			ToHeader toHeader = headerFactory.createToHeader(toAddress, null);
	
			RfssConfig homeRfssConfig = groupConfig.getHomeRfss();
	
			String domainName = homeRfssConfig.getDomainName();
	
			// The request URI for Register requests has the form
			// sip:domain-name
			SipURI registrationURI = SipUtils.createDomainSipURI(domainName);
	
			// Create the via header for the outgoing request.
			ViaHeader viaHeader = SipUtils
					.createViaHeaderForRfss(this.rfssConfig);
			LinkedList<ViaHeader> viaHeaders = new LinkedList<ViaHeader>();
			viaHeaders.add(viaHeader);
	
			// Create a maxforwards header for the outgoing request.
			MaxForwardsHeader maxForwards = headerFactory
					.createMaxForwardsHeader(70);
	
			CallIdHeader callIdHeader = provider.getNewCallId();
	
			CSeqHeader cseqHeader = headerFactory.createCSeqHeader(1L,
					Request.REGISTER);
	
			Request sipRequest = messageFactory.createRequest(registrationURI,
					Request.REGISTER, callIdHeader, cseqHeader, fromHeader,
					toHeader, viaHeaders, maxForwards);
	
			SipUtils.addAllowHeaders(sipRequest);
			
			AcceptHeader acceptHeader = headerFactory.createAcceptHeader(
					ISSIConstants.APPLICATION, ISSIConstants.SDP);
			acceptHeader.setParameter("level", "1");
			sipRequest.addHeader(acceptHeader);
			acceptHeader = headerFactory.createAcceptHeader(
					ISSIConstants.APPLICATION, ISSIConstants.X_TIA_P25_ISSI);
			sipRequest.addHeader(acceptHeader);
	
			ContactHeader contactHeader = SipUtils.createContactHeaderForGroup(
					groupConfig, rfssConfig);
			contactHeader.removeParameter("expires");
			sipRequest.addHeader(contactHeader);
	
			// RouteHeader routeHeader = SipUtils
			// .createRouteToRfss(homeRfssConfig);
			// sipRequest.addHeader(routeHeader);
	
			ExpiresHeader expiresHeader = headerFactory
					.createExpiresHeader(expiresTime);
			sipRequest.addHeader(expiresHeader);
			TimeStampHeader tsHeader = headerFactory.createTimeStampHeader(0);
			tsHeader.setTime(System.currentTimeMillis());
			sipRequest.setHeader(tsHeader);
	
			return sipRequest;
		} catch (Exception ex) {
			this.rfss.getTestHarness().fail("unexpected exception", ex);
			logger.fatal("unexpected exception ", ex);
			return null;
		}
	}

	/*******************************************
	 * Private Methods
	 *******************************************/
	
	private void processRegister(RequestEvent requestEvent, Request request, GroupConfig groupConfig, GroupHome groupHome) throws Exception {
//		 All the error handling is done at this point we can go ahead and
		// process the request.
		if (MobilityManager.isRegisterRegister(request)
				|| MobilityManager.isRegisterDeregister(request)) {

			ContactHeader contactHeader = (ContactHeader) request
					.getHeader(ContactHeader.NAME);

			Address address = contactHeader.getAddress();
			if (contactHeader.isWildCard()) {
				Response response = messageFactory.createResponse(
						Response.BAD_REQUEST, request);
				provider.sendResponse(response);

				this.rfss.logError("Expected a valid address -- got wild card");
				return;
			}

			SipURI sipUri = (SipURI) address.getURI();
			String rfssId = SipUtils.getDomainName(sipUri);
			RfssConfig registeringRfss = topologyConfig
					.getRfssConfig(rfssId);
			if (registeringRfss == null) {
				Response response = messageFactory.createResponse(
						Response.NOT_FOUND, request);
				SipUtils.addAllowHeaders(response);
				
				provider.sendResponse(response);
				logger.debug("RFSS not found");
				return;

			}

			if (!groupHome.getSubscribers().contains(registeringRfss)
					&& MobilityManager.isRegisterDeregister(request)) {
				Response response = messageFactory.createResponse(
						Response.METHOD_NOT_ALLOWED, request);
				SipUtils.addAllowHeaders(response);
				provider.sendResponse(response);
				return;
			}
			// ???
			if (MobilityManager.isRegisterRegister(request)) {

				// Add a subscriber to the group iff the registration lease time is non zero.
				if ( this.rfssConfig.getGroupRegistrationExpiresTime()!= 0)
					groupHome.addSubscriber(registeringRfss);

			} else if (MobilityManager.isRegisterDeregister(request)) {
				groupHome.removeSubscriber(registeringRfss);

			}
			// /
			Response response = messageFactory.createResponse(Response.OK,
					request);
			ContactHeader myContactHeader = SipUtils.createContactHeaderForRfss(this.rfssConfig, ISSIConstants.TIA_P25_SG);
			response.setHeader(myContactHeader);
			ContentDispositionHeader contentDisposition = headerFactory
					.createContentDispositionHeader(ContentDispositionHeader.SESSION);
			contentDisposition.setHandling("required");
			response.addHeader(contentDisposition);
			ServiceProfile sp = groupConfig.getGroupServiceProfile();

			if (MobilityManager.isRegisterRegister(request)) {
				ContentTypeHeader ctHdr = ProtocolObjects.headerFactory
						.createContentTypeHeader(ISSIConstants.APPLICATION,
								ISSIConstants.X_TIA_P25_ISSI);
				response.setContent(sp.toString(), ctHdr);
				if ( groupHome.isSubscribed(registeringRfss)) {
					ExpiresHeader expiresHeader = headerFactory.createExpiresHeader(81400); // some number.
					response.setHeader(expiresHeader);
				} else {
					ExpiresHeader expiresHeader = headerFactory.createExpiresHeader(0); // no registration record.
					response.setHeader(expiresHeader);
				}
			}
			SipUtils.addAllowHeaders(response);
			provider.sendResponse(response);

		} else {

			ContentList contentList = ContentList
					.getContentListFromMessage(request);
			
			Response response = messageFactory.createResponse(Response.OK,
					request);
			ContactHeader myContactHeader = SipUtils.createContactHeaderForRfss(this.rfssConfig, ISSIConstants.TIA_P25_SG);
			response.setHeader(myContactHeader);
			
			ContentDispositionHeader contentDisposition = headerFactory
					.createContentDispositionHeader(ContentDispositionHeader.SESSION);
			contentDisposition.setHandling("required");
			response.addHeader(contentDisposition);
			SipUtils.addAllowHeaders(response);

			ServiceProfile sp = groupConfig.getGroupServiceProfile();
			if (sp != null) {
				if (MobilityManager.isRegisterQuery(request) && 
						rfssConfig == groupConfig.getHomeRfss()) {
					ContentTypeHeader ctHdr = ProtocolObjects.headerFactory
							.createContentTypeHeader(
									ISSIConstants.APPLICATION,
									ISSIConstants.X_TIA_P25_ISSI);
					response.setContent(sp.toString(), ctHdr);
				}
			}
			provider.sendResponse(response);
			RegisterParam registerParam = null;

			if (contentList != null) {
				registerParam = contentList.getRegisterParamContent()
						.getRegisterParam();
			}
			if ( registerParam != null && rfssConfig != groupConfig.getHomeRfss() &&  registerParam.isForceRegister()) {
				// See if we have any SU's that are registered with the group.
				if ( this.groupRefcountTable.get(groupConfig.getRadicalName()) != null &&
						this.groupRefcountTable.get(groupConfig.getRadicalName()).intValue() > 0 ) {
					Request registerRequest = this.createRegisterRequest(groupConfig, 86400);
					ClientTransaction ct = this.provider.getNewClientTransaction(registerRequest);
					PendingResponse pendingResponse = new PendingResponse ();
					pendingResponse.op = Operation.REGISTER;
					pendingResponse.ct = ct;
					ct.setApplicationData(pendingResponse);
					ct.sendRequest();
				}
			}
		}

	}
	
	/*************************************************************
	 * Public Constructor
	 *************************************************************/

	/**
	 * Constructor for group mobility manager.
	 * 
	 * @undocumented
	 * @param rfss --
	 *            the RFSS to which I am assigned.
	 */
	public GroupMobilityManager(RFSS rfss) {
		this.rfss = rfss;
		this.provider = rfss.getProvider();
		this.rfssConfig = rfss.getRfssConfig();
		this.topologyConfig = rfssConfig.getSysConfig().getTopologyConfig();
		this.myGroups = new Hashtable<String, GroupHome>();
		// Populate the tables for which we are responsible.
		for (Iterator<GroupConfig> groupIterator = this.rfssConfig
				.getAssignedGroups(); groupIterator.hasNext();) {
			GroupConfig gc = groupIterator.next();
			String sessionId = gc.getRadicalName();
			PttPointToMultipointSession pttPointToMultipointSession = this.rfss
					.getTransmissionControlManager()
					.getPttPointToMultipointSession(sessionId,
							SessionType.GROUP_HOME);
			GroupHome groupHome = new GroupHome(this.rfss,pttPointToMultipointSession, this.provider,gc);
			this.myGroups.put(gc.getRadicalName(), groupHome);
			// Load the initial configuration into the group
			// This is the starting scenario for the test system.
			for (Iterator<SuConfig> it = gc.getSubscribers(); it.hasNext();) {
				SuConfig su = it.next();
				if (su.getInitialSuState() == SuState.ON) {
					RfssConfig servingRfss = su.getInitialServingRfss();
					if (servingRfss.getGroupRegistrationExpiresTime() != 0) {
						groupHome.addSubscriber(servingRfss);
					}
				}
			}

		}

		// Populate the group refcount table.
		// The group refcount table tracks how many subscribers belong to a
		// given
		// group in a serving RFSS.
		this.groupRefcountTable = new Hashtable<String, Integer>();

		for (SuConfig suConfig : this.topologyConfig
				.getSuConfigurations()) {
			if (suConfig.getInitialSuState() == SuState.ON) {
				// For all the RFSSs that we are minding, increment our
				// group reference count table for the groups for which this
				// rfss is a member.
				if (suConfig.getInitialServingRfss() == this.rfssConfig) {
					HashSet<GroupConfig> gcset = suConfig.getGroups();
					if (gcset != null) {
						for (GroupConfig gc : gcset) {
							String radicalName = gc.getRadicalName();
							if (this.rfssConfig.getServedGroupLeaseTime() != 0) {
								if (!this.groupRefcountTable
										.containsKey(radicalName)) {
									this.groupRefcountTable.put(radicalName,
											new Integer(1));

								} else {
									Integer rc = this.groupRefcountTable
											.get(radicalName);
									Integer newrc = new Integer(
											rc.intValue() + 1);
									this.groupRefcountTable.put(radicalName,
											newrc);
								}
							} else {
								logger
										.debug(rfssConfig.getRfssName()
												+ " : Group with 0 lease time - not registering group");
							}
						}

					}
				}

			} else {
				logger
						.debug(rfssConfig.getRfssName()
								+ " : Not inserting into the group configuration table -- su is off");
			}
		}

	}

	/********************************************************************************
	 * Public methods.																*
	 *******************************************************************************/
	/**
	 * Return true if this is a group which is known to the Serving Rfss.
	 */
	public boolean isKnownGroup(String groupRadicalName) {
		logger.debug(rfssConfig.getRfssName() + " isKnownGroup : "
				+ groupRadicalName);
		return this.groupRefcountTable.containsKey(groupRadicalName);
	}

	
	/**
	 * Handles group registration requests. This method updates the registered
	 * RFSSs for a given group.
	 * 
	 * @undocumented
	 * 
	 */
	public void processRequest(RequestEvent requestEvent) {
		Request request = requestEvent.getRequest();
		assert request.getMethod().equals(Request.REGISTER);
		ToHeader toHeader = (ToHeader) request.getHeader(ToHeader.NAME);
		String groupName = ((SipURI) (toHeader.getAddress().getURI()))
				.getUser();

		try {

			GroupConfig groupConfig = topologyConfig.getGroupConfig(groupName);
			if (groupConfig == null) {
				Response response = SipUtils.createErrorResponse(
						this.rfssConfig, request,
						WarningCodes.NOT_FOUND_UNKNOWN_TARGET_GROUP);
				provider.sendResponse(response);
				logger.error("cannot find target group in topology config");
				return;
			}
			GroupHome group = this.myGroups.get(groupName);
			// Request is aimed at a group that I know nothing about.
			// Cannot deregister a group that I know nothing about.
			// This condition could happen when the RFSS is deliberately
			// misconfigured.

			if (group == null && groupConfig.getHomeRfss() == this.rfssConfig) {
				if (MobilityManager.isRegisterDeregister(request)) {
					Response response = messageFactory.createResponse(
							Response.METHOD_NOT_ALLOWED, request);
					provider.sendResponse(response);

				} else {
					Response response = SipUtils.createErrorResponse(
							this.rfssConfig, request,
							WarningCodes.NOT_FOUND_UNKNOWN_TARGET_GROUP);
					provider.sendResponse(response);
					logger.error("cannot find target group");
				}
				return;
			}

			String sender = ((ViaHeader) request.getHeader(ViaHeader.NAME))
					.getHost();

			RfssConfig sendingRfss = this.topologyConfig.getRfssConfig(sender);

			// Register query can only be issued by home to a group member.
			if (MobilityManager.isRegisterQuery(request)) {
				if (groupConfig.getHomeRfss() != this.rfssConfig
						&& sendingRfss != groupConfig.getHomeRfss()) {
					Response response = SipUtils.createErrorResponse(
							this.rfssConfig, request,
							WarningCodes.FORBIDDEN_SG_NOT_AUTHORIZED);
					provider.sendResponse(response);
					return;

				} else if (groupConfig.isQueryForbiddenFromRfss(sendingRfss)
						&& MobilityManager.isRegisterQuery(request)) {
					Response response = SipUtils.createErrorResponse(
							this.rfssConfig, request,
							WarningCodes.FORBIDDEN_SG_NOT_AUTHORIZED);
					provider.sendResponse(response);
					return;
				} else if (groupConfig.getHomeRfss() != this.rfssConfig &&
						!this.groupRefcountTable.containsKey(groupConfig
						.getRadicalName())) {
					// I am a serving Rfss for this group but I have
					// no knowledge of the group.
					Response response = SipUtils.createErrorResponse(
							this.rfssConfig, request,
							WarningCodes.NOT_FOUND_UNKNOWN_TARGET_GROUP);
					provider.sendResponse(response);
					return;
				}
			} else if (MobilityManager.isRegisterRegister(request)
					|| MobilityManager.isRegisterDeregister(request)) {
				if (groupConfig.isRegisterForbiddenFromRfss(sendingRfss)
						&& MobilityManager.isRegisterRegister(request)) {
					Response response = SipUtils.createErrorResponse(
							this.rfssConfig, request,
							WarningCodes.FORBIDDEN_SG_NOT_AUTHORIZED);
					provider.sendResponse(response);
					return;
				}
			} else {
				Response response = messageFactory.createResponse(
						Response.BAD_REQUEST, request);
				provider.sendResponse(response);
				logger.error("Could not classify request");
				return;

			}
			
			//If we are configured to do so send a session progress and then send the response.
			if ( this.rfssConfig.getRegisterConfirmTime() != 0 ) {
				Response response = messageFactory.createResponse(Response.SESSION_PROGRESS,request);
				provider.sendResponse(response);
				SipUtils.addAllowHeaders(response);
				ISSITimer.getTimer().schedule( new  TimerTask () {
					private RequestEvent requestEvent;
					private Request request;
					private GroupHome groupHome;
					private GroupConfig groupConfig;
					public TimerTask setParameters( RequestEvent requestEvent, Request request, GroupConfig groupConfig,GroupHome groupHome)	{
						this.requestEvent  = requestEvent;
						this.request = request;
						this.groupHome = groupHome;
						this.groupConfig = groupConfig;
						return this;
					}
					@Override
					public void run() {
						try {
							GroupMobilityManager.this.processRegister(requestEvent, request, groupConfig, groupHome);
						} catch (Exception e) {
							logger.error("Unexpected exception", e);
							GroupMobilityManager.this.rfss.getTestHarness().fail("Unexpected exception",e);
						}
						
					}
					
				}.setParameters(requestEvent, request, groupConfig, group), this.rfssConfig.getRegisterConfirmTime() * 1000);
				
			} else {
				this.processRegister(requestEvent, request, groupConfig, group);
			}
		} catch (Exception ex) {
			this.rfss.getTestHarness().fail("Unexpected exception ", ex);
			logger.fatal("Unexpected exception ", ex);
		}

	}

	/*
	 * @see javax.sip.SipListener#processResponse(javax.sip.ResponseEvent)
	 */
	public void processResponse(ResponseEvent responseEvent) {
		assert responseEvent.getClientTransaction() != null;
		PendingResponse pendingResponse = (PendingResponse) responseEvent.getClientTransaction().getApplicationData();
		pendingResponse.response = responseEvent.getResponse();

	}

	/*
	 * @see javax.sip.SipListener#processDialogTerminated(javax.sip.DialogTerminatedEvent)
	 */
	public void processTimeout(TimeoutEvent timeoutEvent) {
		Request request = timeoutEvent.getClientTransaction() != null ? timeoutEvent.getClientTransaction().getRequest() :
				timeoutEvent.getServerTransaction().getRequest(); 
		this.rfss.logError("Unexpected timeout event for request :\n" + request );

	}

	/*
	 * @see javax.sip.SipListener#processIOException(javax.sip.IOExceptionEvent)
	 */
	public void processIOException(IOExceptionEvent ioe) {
		this.rfss.logError("Unexpected IOException event");
	}

	public void processTransactionTerminated(TransactionTerminatedEvent arg0) {

	}

	/**
	 * @see javax.sip.SipListener#processDialogTerminated(javax.sip.DialogTerminatedEvent)
	 */
	public void processDialogTerminated(DialogTerminatedEvent arg0) {
		this.rfss.logError("Unexpected Dialog terminated event");
	}

	/**
	 * The mobility manager invokes this method when an SU departs from a given
	 * RFSS or when the Mobility manager detects that a SU has departed. Check
	 * if the departure of the SU has resulted in any groups becoming empty. If
	 * so then, send a deregistration request to the home RFSS of the given
	 * group to deregister this RFSS from the home RFSS.
	 * 
	 * 
	 * @param suConfig --
	 *            the configuration file for the SU that is departing this RFSS
	 */

	public synchronized void suDeparted(SuConfig suConfig) {
		logger.info("GroupMobilityManager: suDeparted : " + suConfig.getSuId());

		try {
			for (GroupConfig gc : suConfig.getGroups()) {
				Integer rc = this.groupRefcountTable.get(gc.getRadicalName());
				Integer newRc = new Integer(rc.intValue() - 1);
				if (newRc == 0) {
					this.groupRefcountTable.remove(gc.getRadicalName());
					// Generate a Register-Deregister message from the old
					// location.
					Request request = this.createRegisterRequest(gc, 0);
					ClientTransaction ct = this.provider
							.getNewClientTransaction(request);
					PendingResponse pendingResponse = new PendingResponse();
					pendingResponse.op = Operation.DEREGISTER;
					pendingResponse.ct = ct;
					ct.setApplicationData(pendingResponse);
					ct.sendRequest();

				}
			}
		} catch (Exception ex) {
			logger.error("unexpected exception ", ex);
			this.rfss.getTestHarness().fail("Unexpected exception ", ex);
		}

	}

	/**
	 * This methiod is invoked when an SU arrives at an RFSS or when a Mobility
	 * manager detects that the SU has come alive. Check if there are any active
	 * calls in progress for the given RFSS. If so do nothing. If not, send a
	 * registration request to the home RFSS of the group to register this RFSS
	 * with the group. Note that in this case the group home sends out the Re-INVITE
	 * when we register for the group call.
	 * 
	 * @param suConfig --
	 *            the config file for the SU that has arrived.
	 */
	public synchronized void suArrived(SuConfig suConfig) {

		logger.info("GroupMobilityManager: SuArrived ");
		if (suConfig.getSU().getState() == gov.nist.p25.issi.issiconfig.SuState.ON) {
			try {
				for (GroupConfig gc : suConfig.getSubscribedGroups()) {
					if (!this.groupRefcountTable.containsKey(gc
							.getRadicalName())) {
						if (rfssConfig.isGroupCredentialsQueriedBeforeRegister()) {
							ClientTransaction ct = this.sendRegisterQuery(gc);
							while ( ct.getState() != TransactionState.TERMINATED) {
								Thread.sleep(1000); // Let the OK come back for the query.
							}
							PendingResponse pendingResponse = (PendingResponse) ct.getApplicationData();
							assert pendingResponse != null;
							if ( pendingResponse.response == null) {
								logger.error("Group Mobility Manager : Tx timed out maybe the peer died -- returning silently");
								return;
							}
							if ( pendingResponse.response.getStatusCode()/100 > 2 ) {
								logger.debug("Error in query -- returning without registering!");
								return;
							}
						}
						Request request = this.createRegisterRequest(gc, 84600);
						ClientTransaction ct = this.provider
								.getNewClientTransaction(request);
						PendingResponse pendingResponse = new PendingResponse();
						pendingResponse.ct = ct;
						pendingResponse.op = Operation.REGISTER;
						ct.sendRequest();
						ct.setApplicationData(pendingResponse);
						Integer rc = new Integer(1);
						if ( this.rfssConfig.getServedGroupLeaseTime() != 0)
							this.groupRefcountTable.put(gc.getRadicalName(), rc);

					} else {
						logger.info("Not sending register refcount = "
								+ this.groupRefcountTable.get(gc
										.getRadicalName()));
					}

				}
			} catch (Exception ex) {
				logger.error("Unexpected exception ", ex);
				this.rfss.getTestHarness().fail("Unexpected exception", ex);
			}
		}

	}

	/**
	 * Get a Group given the group id.
	 * 
	 * @undocumented
	 * @param radicalName
	 * @return
	 */
	public GroupHome getGroup(String radicalName) {
		return this.myGroups.get(radicalName);
	}

	public ClientTransaction sendRegisterQuery(GroupConfig sgConfig) throws Exception {

		Request request = this.createRegisterRequest(sgConfig, 0);
		request.removeHeader(ExpiresHeader.NAME);
		request.removeHeader(ContactHeader.NAME);
		ClientTransaction ct = this.provider.getNewClientTransaction(request);
		PendingResponse pendingResponse = new PendingResponse();
		pendingResponse.op = Operation.QUERY;
		pendingResponse.ct = ct;
		ct.setApplicationData(pendingResponse);
		ct.sendRequest();
		return ct;

	}

	/**
	 * Send off a query from a group to a member. This allows you to send
	 * queries to any RFSS at all ( it should do some parameter checking but it
	 * does not ).
	 * 
	 * @param groupConfig
	 * @param targetRfss
	 */
	public void homeQueryGroup(GroupConfig sgConfig, RfssConfig targetRfss,
			boolean force, boolean confirm) throws Exception {

		Request request = this.createHomeQueryRequest(sgConfig, targetRfss,
				force, confirm);

		ClientTransaction ct = this.provider.getNewClientTransaction(request);
		PendingResponse pendingResponse = new PendingResponse();
		pendingResponse.op = Operation.HOME_QUERY;
		pendingResponse.ct = ct;
		ct.setApplicationData(pendingResponse);
		ct.sendRequest();

	}

	public Collection<GroupHome> getGroups() {
		return this.myGroups.values();
	}

}