package gov.nist.p25.issi.rfss;

import gov.nist.p25.issi.ISSITimer;
import gov.nist.p25.issi.issiconfig.ISSIConstants;
import gov.nist.p25.issi.issiconfig.RfssConfig;
import gov.nist.p25.issi.issiconfig.SuConfig;
import gov.nist.p25.issi.issiconfig.TopologyConfig;
import gov.nist.p25.issi.rfss.tester.TestSU;
import gov.nist.p25.issi.utils.SipUtils;
import gov.nist.p25.issi.utils.WarningCodes;

import java.util.HashSet;
import java.util.Timer;
import java.util.TimerTask;

import javax.sip.ClientTransaction;
import javax.sip.DialogTerminatedEvent;
import javax.sip.IOExceptionEvent;
import javax.sip.RequestEvent;
import javax.sip.ResponseEvent;
import javax.sip.ServerTransaction;
import javax.sip.SipException;
import javax.sip.SipListener;
import javax.sip.SipProvider;
import javax.sip.TimeoutEvent;
import javax.sip.TransactionTerminatedEvent;
import javax.sip.address.SipURI;
import javax.sip.header.ContactHeader;
import javax.sip.header.ExpiresHeader;
import javax.sip.header.FromHeader;
import javax.sip.header.ToHeader;
import javax.sip.message.Message;
import javax.sip.message.Request;
import javax.sip.message.Response;

import org.apache.log4j.Logger;

/**
 * Top level Mobility Manager. This multiplexes requests to either a SU to SU
 * mobility manager or a Group Mobilty Manager depending upon whether the
 * request belongs to a Unit to Unit call or to a Group call.
 * 
 * @author M. Ranganathan
 * 
 */
public class MobilityManager implements SipListener {

	private UnitToUnitMobilityManager unitToUnitMobilityManager;

	private GroupMobilityManager groupMobilityManager;

	private RFSS rfss;

	MembershipScanner membershipScanner;

	private static int SCAN_INTERVAL = 500; // milisec between membership change

	// scans.

	private static Logger logger = Logger.getLogger(MobilityManager.class);


	enum MessageType {
		U2U, GROUP
	};

	// //////////////////////////////////////////////////////////////////////////
	// Inner classes.
	// /////////////////////////////////////////////////////////////////////////

	/**
	 * This timer task scans the membership of the Rfss and looks for changes in
	 * membership. If it detects a change in membership, it notifies the unit to
	 * unit and group mobility managers. this.servedUnits contains the
	 * Subscriber units served by the RFSS in this scan.
	 * 
	 * @undocumented
	 */
	class MembershipScanner extends TimerTask {
		HashSet<SuConfig> servedUnits;

		RfssConfig rfssConfig;

		MembershipScanner() {
			this.rfssConfig = rfss.getRfssConfig();
			this.servedUnits = new HashSet<SuConfig>(rfss
					.getServedSubscriberUnits());
		}

		public synchronized void removeServedSubscriberUnit(SuConfig suConfig) {
			this.servedUnits.remove(suConfig);
			this.rfssConfig.getRFSS().removeServedSubscriberUnit(suConfig);
		}
		public synchronized void run() {
			HashSet<SuConfig> currentServed;
			currentServed = new HashSet<SuConfig>(rfss
					.getServedSubscriberUnits());
			// We dont need to sycnrhonize the entire scan because we will catch
			// it on the next scan.
			// this.servedUnits is what we saw in the last scan.
			for (SuConfig suConfig : this.servedUnits) {
				// check for departures.
				if (!currentServed.contains(suConfig)) {
					unitToUnitMobilityManager.suDeparted(suConfig);
					groupMobilityManager.suDeparted(suConfig);
				}

			}
			for (SuConfig suConfig : currentServed) {
				if (!this.servedUnits.contains(suConfig)) {
					unitToUnitMobilityManager.suArrived(suConfig);
					groupMobilityManager.suArrived(suConfig);

				}
			}
			this.servedUnits = currentServed;

		}

	}

	// ///////////////////////////////////////////////////////////////////////////////////////
	// Constructor
	// ////////////////////////////////////////////////////////////////////////////////////////

	/**
	 * Constructor
	 * 
	 * @undocumented
	 */
	MobilityManager(RFSS rfss) {
		assert rfss != null;
		this.rfss = rfss;
		this.unitToUnitMobilityManager = new UnitToUnitMobilityManager(rfss);
		this.groupMobilityManager = new GroupMobilityManager(rfss);

		this.membershipScanner = new MembershipScanner();
		ISSITimer.getTimer().schedule(membershipScanner, SCAN_INTERVAL, SCAN_INTERVAL);

	}

	// ////////////////////////////////////////////////////////////////////////////////////
	// Static methods
	// ////////////////////////////////////////////////////////////////////////////////////

	private MessageType getMessageType(Message message) {
		FromHeader fromHeader = (FromHeader) message.getHeader(FromHeader.NAME);
		ToHeader toHeader = (ToHeader) message.getHeader(ToHeader.NAME);
		SipURI sipUri = (SipURI) fromHeader.getAddress().getURI();
		if (sipUri.getUser() == null)
			this.rfss.getTestHarness().fail(
					"Null UserParameter in the From Header of Request");
		if (fromHeader.getTag() == null)
			this.rfss.getTestHarness().fail(
					"Missing tag Parameter in the From Header");
		/* if (toHeader.getTag() != null)
			this.rfss.getTestHarness().fail(
					"To Header contains a Tag Parameter");*/

		if (sipUri.getUserParam().equals(ISSIConstants.TIA_P25_SU)) {
			return MessageType.U2U;
		} else if (sipUri.getUserParam().equals(ISSIConstants.TIA_P25_SG)) {
			return MessageType.GROUP;
		} else {
			this.rfss.logError("Unknown type of request : " + message);
			this.rfss.getTestHarness().fail(
					"Unknown User Parameter in the FromHeader of request ");
			return null;
		}

	}

	/**
	 * SIP Register-Register is used as part of the Registration procedure to
	 * provide a means for a serving RFSS to express interest in a mobility
	 * object. The Register message is distinguished by the presence of a
	 * contact parameter and a non-zero expires parameter. The SIP header fields
	 * SHALL feature an "expires" parameter indicating its registration
	 * expiration interval chosen by the home RFSS, and the body SHALL include
	 * service profile information for the requested object.
	 * 
	 */

	public static boolean isRegisterRegister(Request request) {

		ExpiresHeader expires = request.getExpires();

		return request.getMethod().equals(Request.REGISTER)
				&& request.getHeader(ContactHeader.NAME) != null
				&& expires != null && expires.getExpires() != 0;
	}

	/**
	 * 
	 * SIP Register-Deregister is used by a serving RFSS to deregister a
	 * mobility object or by a home RFSS to request deregistration of an SU. The
	 * SIP Register-Deregister command is distinguished by the presence of a
	 * contact parameter and a zero expires parameter. As with the SIP
	 * Register-Query message, the SIP Register-Deregister message is used in
	 * two separate procedures, the De-registration procedure and the Roamed
	 * procedure. Again, the choice of procedure is dependant on the roles of
	 * the RFSSs involved.
	 * 
	 */

	public static boolean isRegisterDeregister(Request request) {
		ExpiresHeader expires = request.getExpires();
		return request.getMethod().equals(Request.REGISTER)
				&& request.getHeader(ContactHeader.NAME) != null
				&& expires != null && expires.getExpires() == 0;

	}

	/**
	 * SIP Register-Query is sent by a serving RFSS to query a home RFSS for
	 * service profile information, and by a home RFSS to confirm the interest
	 * of a serving RFSS in a mobility object. It is distinguished by the
	 * absence of both the expires parameter and the contact parameter. The SIP
	 * Register-Query message is used in two procedures described in ?5.5, the
	 * Home Query procedure and the Serving Query procedure. The content of the
	 * message is consistent in each case, and the particular procedure is
	 * determined by the roles of the participating RFSSs pursuant to the
	 * mobility object involved.
	 */

	public static boolean isRegisterQuery(Request request) {
		return request.getMethod().equals(Request.REGISTER)
				&& request.getHeader(ContactHeader.NAME) == null
				&& request.getHeader(ExpiresHeader.NAME) == null;
	}

	// ////////////////////////////////////////////////////////////////////////////
	// SIP Listener methods
	// ////////////////////////////////////////////////////////////////////////////

	/**
	 * @inheritDoc
	 * @undocumented
	 */
	public void processRequest(RequestEvent requestEvent) {
		Request request = requestEvent.getRequest();
		MessageType requestType = getMessageType(request);

		if (requestType == MessageType.GROUP)
			groupMobilityManager.processRequest(requestEvent);
		else if (requestType == MessageType.U2U)
			unitToUnitMobilityManager.processRequest(requestEvent);

	}

	/**
	 * @inheritDoc
	 * @undocumented
	 */

	public void processResponse(ResponseEvent responseEvent) {
		Response response = responseEvent.getResponse();
		MessageType requestType = getMessageType(response);
		if (requestType == MessageType.GROUP)
			groupMobilityManager.processResponse(responseEvent);
		else if (requestType == MessageType.U2U)
			unitToUnitMobilityManager.processResponse(responseEvent);
	}

	/**
	 * @inheritDoc
	 * @undocumented
	 */

	public void processTimeout(TimeoutEvent timeoutEvent) {
		if (timeoutEvent.isServerTransaction()) {
			ServerTransaction serverTransaction = timeoutEvent
					.getServerTransaction();
			Request request = serverTransaction.getRequest();
			MessageType type = getMessageType(request);
			if (type == MessageType.GROUP)
				groupMobilityManager.processTimeout(timeoutEvent);
			else
				unitToUnitMobilityManager.processTimeout(timeoutEvent);
		} else {
			ClientTransaction clientTransaction = timeoutEvent
					.getClientTransaction();
			Request request = clientTransaction.getRequest();
			MessageType type = getMessageType(request);
			if (type == MessageType.GROUP)
				groupMobilityManager.processTimeout(timeoutEvent);
			else
				unitToUnitMobilityManager.processTimeout(timeoutEvent);

		}

	}

	/**
	 * @inheritDoc
	 * @undocumented
	 */

	public void processIOException(IOExceptionEvent ioex) {
		this.rfss.logError("Unexpected event -- IO Exception : " + ioex.getHost() + ":" + ioex.getPort() + "/" + ioex.getTransport());
	}

	/**
	 * @inheritDoc
	 * @undocumented
	 */

	public void processTransactionTerminated(
			TransactionTerminatedEvent terminatedEvent) {
		if (terminatedEvent.isServerTransaction()) {
			ServerTransaction serverTransaction = terminatedEvent
					.getServerTransaction();
			Request request = serverTransaction.getRequest();
			MessageType type = getMessageType(request);
			if (type == MessageType.GROUP)
				groupMobilityManager
						.processTransactionTerminated(terminatedEvent);
			else
				unitToUnitMobilityManager
						.processTransactionTerminated(terminatedEvent);
		} else {
			ClientTransaction clientTransaction = terminatedEvent
					.getClientTransaction();
			Request request = clientTransaction.getRequest();
			MessageType type = getMessageType(request);
			if (type == MessageType.GROUP)
				groupMobilityManager
						.processTransactionTerminated(terminatedEvent);
			else
				unitToUnitMobilityManager
						.processTransactionTerminated(terminatedEvent);

		}

	}

	/**
	 * @inheritDoc
	 * @undocumented
	 */

	public void processDialogTerminated(DialogTerminatedEvent arg0) {
		this.rfss.getTestHarness().fail("unexpected event");
	}

	// ////////////////////////////////////////////////////////////////
	// Instance methods
	// ////////////////////////////////////////////////////////////////

	/**
	 * @inheritDoc
	 * @undocumented
	 */

	public UnitToUnitMobilityManager getUnitToUnitMobilityManager() {
		return this.unitToUnitMobilityManager;
	}

	/**
	 * @inheritDoc
	 * @undocumented
	 */

	public GroupMobilityManager getGroupMobilityManager() {
		return this.groupMobilityManager;

	}

	/**
	 * Stop the membership scanner timer task.
	 * 
	 */
	public void stop() {
		this.membershipScanner.cancel();
	}

}