package gov.nist.p25.issi.rfss;

import gov.nist.javax.sip.SipStackImpl;
import gov.nist.javax.sip.message.SIPRequest;
import gov.nist.javax.sip.message.SIPResponse;
import gov.nist.p25.issi.ISSITimer;
import gov.nist.p25.issi.issiconfig.GlobalTopologyParser;
import gov.nist.p25.issi.issiconfig.GroupConfig;
import gov.nist.p25.issi.issiconfig.RfssConfig;
import gov.nist.p25.issi.issiconfig.SuConfig;
import gov.nist.p25.issi.issiconfig.SuState;
import gov.nist.p25.issi.issiconfig.SystemTopologyParser;
import gov.nist.p25.issi.issiconfig.TopologyConfig;
import gov.nist.p25.issi.issiconfig.TopologyConfigParser;
import gov.nist.p25.issi.p25body.ContentList;
import gov.nist.p25.issi.p25payload.IMBEVoiceBlock;
import gov.nist.p25.issi.p25payload.IMBEVoiceDataGenerator;
import gov.nist.p25.issi.p25payload.P25Payload;
import gov.nist.p25.issi.rfss.tester.AbstractSipSignalingTest;
import gov.nist.p25.issi.rfss.tester.CallSetupTest;
import gov.nist.p25.issi.rfss.tester.ISSITesterConfiguration;
import gov.nist.p25.issi.rfss.tester.TestMessages;
import gov.nist.p25.issi.rfss.tester.TestMessagesParser;
import gov.nist.p25.issi.rfss.tester.TestScript;
import gov.nist.p25.issi.testlauncher.TestHarness;
import gov.nist.p25.issi.traceviewer.MessageData;
import gov.nist.p25.issi.traceviewer.SipTraceFileParser;
import gov.nist.p25.issi.transctlmgr.TransmissionControlManager;
import gov.nist.p25.issi.transctlmgr.TransmissionControlSAP;
import gov.nist.p25.issi.transctlmgr.ptt.PttSession;
import gov.nist.p25.issi.transctlmgr.ptt.PttSessionInterface;
import gov.nist.p25.issi.utils.SipUtils;
import gov.nist.rtp.RtpPacket;
import gov.nist.rtp.RtpSession;

import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.ListIterator;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.Semaphore;

import javax.sip.ClientTransaction;
import javax.sip.DialogTerminatedEvent;
import javax.sip.IOExceptionEvent;
import javax.sip.RequestEvent;
import javax.sip.ResponseEvent;
import javax.sip.SipListener;
import javax.sip.SipProvider;
import javax.sip.SipStack;
import javax.sip.TimeoutEvent;
import javax.sip.TransactionTerminatedEvent;
import javax.sip.address.SipURI;
import javax.sip.header.CSeqHeader;
import javax.sip.header.CallIdHeader;
import javax.sip.header.ContentLengthHeader;
import javax.sip.header.ExpiresHeader;
import javax.sip.header.FromHeader;
import javax.sip.header.MaxForwardsHeader;
import javax.sip.header.RouteHeader;
import javax.sip.header.TimeStampHeader;
import javax.sip.header.ToHeader;
import javax.sip.header.ViaHeader;
import javax.sip.header.WarningHeader;
import javax.sip.message.Message;
import javax.sip.message.Request;
import javax.sip.message.Response;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;

/**
 * This is the RFSS emulation class. Its function is mainly a management
 * interface. It registers a SIP listener and recieves SIP Events via the SIP
 * listener. It looks at the request/response event type of the incoming SIP
 * request and directs these to appropriate handlers. It creates one instance of
 * each of the following upon startup:
 * 
 * <ul>
 * <li> MobilityManager
 * <li> CallControlManager
 * <li> TransmissionControlManager
 * </ul>
 * 
 * <p>
 * All incoming REGISTER requests and responses are directed to the Mobility
 * Manager and all INVITE, ACK, BYE are directed to the Call Control Manager.
 * 
 * <p>
 * Each RFSS has an IP Address and port which is specificed in the
 * configuration.xml file for the system. It is provided as an argument to the
 * RFSS on startup. An RFSS configuration is contained in the class RfssConfig.
 * 
 * <p>
 * This class may be invoked in listener thread if we are emulating the system
 * or in its own JVM ( as a test tool ).
 * 
 * 
 * @author M. Ranganathan
 */

public class RFSS implements SipListener {

	// Logs all the reuqests and responses seen at a RFSS for
	// testing purposes.

	/*
	 * A table of RFSSId to RFSS
	 */

	private AbstractSipSignalingTest currentTestCase;

	private boolean isRfResourcesAvailable;

	private static Hashtable<String, RFSS> rfssTable = new Hashtable<String, RFSS>();

	private ArrayList<Request> requests;

	private ArrayList<Response> responses;

	private RfssConfig rfssConfig;

	private int port;

	private String ipAddress;

	private SipProvider provider;

	private TestHarness testHarness;

	private TransmissionControlManager transmissionControlManager;

	private TestScript testScript;

	/*
	 * Manages user and group service profiles. This should really be managed by
	 * a SLEE Profile in a reference implementation of this standards.
	 */

	private MobilityManager mobilityManager;

	private Hashtable<String, SipListener> methodHash;

	private HashSet<SuConfig> servedSubscriberUnits;

	private Hashtable<String, HomeAgent> homeAgents;

	private Hashtable<PttSessionInterface, CapturedPttPacket> lastCapturedPttPacketTable = new Hashtable<PttSessionInterface, CapturedPttPacket>();

	private CallControlManager callControlManager;

	private TopologyConfig topologyConfig;

	private ArrayList<CapturedPttPacket> capturedPackets = new ArrayList<CapturedPttPacket>();

	private String failureReason;

	private TestMessages testMessages;

	private TestMessages refMessages;

	static long lastRecordedTime;

	public static final ArrayList<IMBEVoiceBlock> imbeVoiceBlocks = IMBEVoiceDataGenerator
			.getIMBEVoiceData("voice/imbe-hex-test.txt");

	private static Logger logger = Logger.getLogger(RFSS.class);

	private static Logger errorLogger = Logger
			.getLogger("gov.nist.p25.ERRORLOG");

	// ////////////////////////////////////////////////////////////////////////////////
	// Private methods
	// ////////////////////////////////////////////////////////////////////////////////

	private static boolean matchSipRequest(SIPRequest request,
			SIPRequest template) {
		SIPRequest testSipRequest = (SIPRequest) request.clone();

		testSipRequest.removeHeader(CallIdHeader.NAME);

		template.removeHeader(CallIdHeader.NAME);

		testSipRequest.removeHeader(ExpiresHeader.NAME);

		template.removeHeader(ExpiresHeader.NAME);

		return testSipRequest.match(template);

	}

	private static boolean matchSipResponse(SIPResponse response,
			SIPResponse template) {
		SIPResponse testSipResponse = (SIPResponse) response.clone();

		/*
		 * We dont care about the call ID header in making the comparison.
		 * However call id header is mandatory and if the request did not have a
		 * call id header it would fail in transaction matching anyways.
		 */
		testSipResponse.removeHeader(CallIdHeader.NAME);

		template.removeHeader(CallIdHeader.NAME);

		testSipResponse.removeHeader(ExpiresHeader.NAME);

		template.removeHeader(ExpiresHeader.NAME);

		return testSipResponse.match(template);

	}

	/**
	 * Create an RFSS. Return one if its already created.
	 * 
	 * @param provider
	 * @param rfssConfig
	 * @return
	 */
	public static RFSS create(SipProvider provider, RfssConfig rfssConfig,
			TestScript testScript) throws Exception {
		if (rfssTable.get(rfssConfig.getDomainName()) != null) {
			return rfssTable.get(rfssConfig.getDomainName());

		} else {
			if (testScript.isInteractive()) {
				// load the traces for this test.
				String fname = rfssConfig.getRfssName() + ".xml";
				String dirName = testScript.getTestDirectory();

				String traceFileName = testScript.getReferenceMessagesDirName()
						+ "/" + fname;
				if (!new File(traceFileName).exists()) {
					if (logger.isDebugEnabled()) {
						logger
								.error("Reference Trace file not found for the emulated RFSS "
										+ rfssConfig.getRfssName());
						logger
								.error("Configured to receive SIP requests at the following address "
										+ rfssConfig.getHostPort());

					}
					ProtocolObjects.stop();
					RFSS.reset();
					throw new Exception(
							"Reference trace for RFSS is missing -- generate trace first!");

				}
			}

			RFSS rfss = new RFSS(provider, rfssConfig);
			rfssTable.put(rfssConfig.getDomainName(), rfss);
			rfss.testScript = testScript;
			((RfssRouter) provider.getSipStack().getRouter()).setRfss(rfss);
			if (testScript.isInteractive()) {
				// load the traces for this test.
				String fname = rfssConfig.getRfssName() + ".xml";
				String dirName = testScript.getTestDirectory();

				String traceFileName = testScript.getReferenceMessagesDirName()
						+ "/" + fname;

				rfss.testMessages = new TestMessagesParser(traceFileName)
						.parse();
				rfss.refMessages = new TestMessagesParser(traceFileName)
						.parse();

			}
			return rfss;
		}
	}

	/**
	 * Reset the table after the test is over.
	 */
	public static void reset() {

		rfssTable.clear();
		rfssTable = new Hashtable<String, RFSS>();
	}

	/**
	 * Capture a ptt packet for later analysis. Note that this just creates an
	 * object and stores it in memory for later retrieval. This is because
	 * otherwise we cannot handle a large volume of packets
	 * 
	 * @param payload
	 * @param isSender -
	 *            true if I am sender.
	 * @param session
	 */
	public void capturePttPacket(RtpPacket rtpPacket, P25Payload payload,
			boolean isSender, PttSessionInterface session) {
		RtpSession rtpSession = session.getRtpSession();

		/*
		 * Dont record packets that we route within an RFSS (internal to the
		 * RFSS).
		 */
		if (this.getTransmissionControlManager().getPttManager()
				.ownsRemoteRtpSession(rtpSession)) {
			logger.debug("Not logging a packet that we send to ourselves");
			return;
		}
		CapturedPttPacket pttPacket = new CapturedPttPacket(rtpPacket, payload,
				isSender, session);
		/*
		 * I would like to store the packets in memory at runtime and delay
		 * flushing.
		 */

		CapturedPttPacket lastPacket = this.lastCapturedPttPacketTable
				.get(session);

		if (lastPacket != null
				&& lastPacket.getRemotePort() == pttPacket.getRemotePort()
				&& lastPacket.getRemoteIpAddress().equals(
						pttPacket.getRemoteIpAddress())
				&& lastPacket.match(pttPacket)) {
			logger.debug("not recording packet -- matched previous one ");
			return;
		} else
			this.lastCapturedPttPacketTable.put(session, pttPacket);

		this.capturedPackets.add(pttPacket);
		logger.debug("Capture packets -- captured a ptt packet");

	}

	/**
	 * Flush all the captured messages to disk.
	 * 
	 */
	public void flushCapturedPttMessages() {
		logger.debug("flushing " + this.capturedPackets.size() + " packets ! ");
		for (CapturedPttPacket pttMessage : this.capturedPackets) {
			pttMessage.flush();

		}

		// this.capturedPackets.clear();

	}

	/**
	 * Save the expected messages as a file that can be later used to test
	 * against the messages that the RFSS actually sees.
	 * 
	 * @throws IOException
	 */
	public void saveTrace() throws IOException {
		try {
			logger.debug("saving trace for " + this.rfssConfig.getRfssName());
			String fname = this.rfssConfig.getRfssName() + ".xml";
			File dir = new File(testScript.getReferenceMessagesDirName());

			dir.mkdirs();
			File file = new File(testScript.getReferenceMessagesDirName() + "/"
					+ fname);
			file.delete();
			file.createNewFile();
			FileOutputStream fos = new FileOutputStream(file);
			PrintStream ps = new PrintStream(fos);
			ps.append("<?xml version=\"1.0\" ?>\n");
			ps
					.append("<!DOCTYPE expected-messages SYSTEM \"http://www-x.antd.nist.gov:8080/p25/issi-emulator/dtd/expected-messages.dtd\">\n");
			ps.append("<expected-messages>\n");
			ps.append("<siprequests>\n");
			for (Request sipRequest : this.requests) {
				ViaHeader viaHeader = (ViaHeader) sipRequest
						.getHeader(ViaHeader.NAME);
				String host = viaHeader.getHost();
				if (host.equals(this.rfssConfig.getDomainName())) {
					// We do not care about self routed requests.
					continue;
				}
				Request newRequest = (Request) sipRequest.clone();

				ps.append("<siprequest>\n");
				ps.append("<![CDATA[\n");
				// Timestamp header is not mandatory.
				newRequest.removeHeader(TimeStampHeader.NAME);
				newRequest.removeHeader(MaxForwardsHeader.NAME); // Maxforwards
				newRequest.removeHeader(WarningHeader.NAME);
				/*
				 * Remove the warning header (not compared)
				 */
				ListIterator li = newRequest.getHeaders(RouteHeader.NAME);
				// Only check the first route header.
				if (li != null && li.hasNext()) {
					RouteHeader route = (RouteHeader) li.next();
					newRequest.setHeader(route);
				}
				ListIterator it = newRequest.getHeaders(ViaHeader.NAME);
				newRequest.removeHeader(ViaHeader.NAME);
				while (it.hasNext()) {
					// Get rid of the branch in the request header
					ViaHeader via = (ViaHeader) it.next();
					via.removeParameter("branch");
					newRequest.addHeader(via);
				}
				ToHeader to = (ToHeader) newRequest.getHeader(ToHeader.NAME);
				// The tag is a random quanity -- get rid of it.
				to.removeParameter("tag");
				FromHeader from = (FromHeader) newRequest
						.getHeader(FromHeader.NAME);
				from.removeParameter("tag");
				ps.append(newRequest.toString());
				ps.append("]]>\n");
				ps.append("</siprequest>\n");
			}
			ps.append("</siprequests>\n");

			/**
			 * For each of the captured messages - save only the parts that we
			 * want
			 */
			ps.append("<sipresponses>\n");
			for (Response response : this.responses) {
				Response newResponse = (Response) response.clone();
				FromHeader from = (FromHeader) newResponse
						.getHeader(FromHeader.NAME);
				from.removeParameter("tag");
				ToHeader to = (ToHeader) newResponse.getHeader(ToHeader.NAME);
				to.removeParameter("tag");
				newResponse.removeHeader(TimeStampHeader.NAME);
				newResponse.removeHeader(MaxForwardsHeader.NAME);

				for (Iterator it = newResponse.getHeaders(ViaHeader.NAME); it
						.hasNext();) {
					ViaHeader viaHeader = (ViaHeader) it.next();
					viaHeader.removeParameter("branch");
				}

				if (response.getStatusCode() != 100) {
					ps.append("<sipresponse>\n");
					ps.append("<![CDATA[\n");
					ps.append(newResponse.toString());
					ps.append("]]>\n");
					ps.append("</sipresponse>\n");
				}
			}
			ps.append("</sipresponses>\n");
			// TODO -- store the PTT messages here.
			ps.append("</expected-messages>\n");
			fos.close();
			if (this.capturedPackets.size() > 0) {
				fname = RFSS.this.rfssConfig.getRfssName() + "_ptt.bin";
				file = new File(this.testScript.getReferenceMessagesDirName()
						+ "/" + fname);
				file.delete();
				file.createNewFile();
				fos = new FileOutputStream(file);

				ObjectOutputStream oos = new ObjectOutputStream(fos);

				for (CapturedPttPacket pttMessage : this.capturedPackets) {
					oos.writeObject(pttMessage);
				}
				oos.close();
				fos.close();
			}
		} catch (IOException ex) {
			logger.error("Error capturing trace", ex);
			throw ex;
		}

	}

	/**
	 * This predicate tests to see if we find all the messages that are captured
	 * by the RFSS.
	 * 
	 * @return true -- if each of the messages in the trace is seen in the
	 *         captured messages. False otherwise.
	 */
	public boolean testCapturedMessages() {
		try {

			// load the traces for this test.
			String fname = this.rfssConfig.getRfssName() + ".xml";
			
			String traceFileName = this.testScript
					.getReferenceMessagesDirName()
					+ "/" + fname;
			if (!new File(traceFileName).exists()) {
				if (logger.isDebugEnabled()) {
					logger
							.error("Reference Trace file not found for the emulated RFSS "
									+ this.rfssConfig.getRfssName());
					logger
							.error("Configured to receive SIP requests at the following address "
									+ rfssConfig.getHostPort());

				}
				this.testMessages = new TestMessages();
			} else {
				this.testMessages = new TestMessagesParser(traceFileName)
						.parse();
			}

			for (Request request : testMessages.getRequests()) {
				SIPRequest testSipRequest = (SIPRequest) request.clone();
				// Get the template sip request content list.
				ContentList contentList = ContentList
						.getContentListFromMessage(testSipRequest);
				/*
				 * We dont care about the call ID header in making the
				 * comparison. However call id header is mandatory and if the
				 * request did not have a call id header it would fail in
				 * transaction matching anyways.
				 */
				testSipRequest.removeHeader(CallIdHeader.NAME);
				/*
				 * The contents of the expires header are arbitrary so remove it
				 * from the match template.
				 */
				boolean expiresRemoved = (ExpiresHeader) request
						.getHeader(ExpiresHeader.NAME) != null;
				testSipRequest.removeHeader(ExpiresHeader.NAME);
				boolean matchFound = false;

				for (Request capturedRequest : this.requests) {
					ContentList contentList1 = ContentList
							.getContentListFromMessage(capturedRequest);

					if (((SIPRequest) capturedRequest).match(testSipRequest)
							&& ((!expiresRemoved) || capturedRequest
									.getExpires() != null)
							&& contentList.match(contentList1)) {
						matchFound = true;
						break;
					}
				}

				if (!matchFound) {
					logError("Could not find a match for this message :\n"
							+ testSipRequest);
					logger.debug("test databse =  "
							+ testMessages.getRequests());
					this.failureReason = "Could not find a match for this message "
							+ testSipRequest;
					return false;
				}

			}
			for (Response response : testMessages.getResponses()) {
				SIPResponse testSipResponse = (SIPResponse) response.clone();
				testSipResponse.removeHeader(CallIdHeader.NAME);

				ContentList testContentList = ContentList
						.getContentListFromMessage(testSipResponse);

				boolean expiresRemoved = testSipResponse
						.getHeader(ExpiresHeader.NAME) != null;
				testSipResponse.removeHeader(ExpiresHeader.NAME);
				response.removeHeader(ExpiresHeader.NAME);
				boolean matchFound = false;
				for (Response capturedResponse : this.responses) {
					SIPResponse testResponse = (SIPResponse) capturedResponse;
					ContentList contentList = ContentList
							.getContentListFromMessage(capturedResponse);
					testResponse.removeHeader(CallIdHeader.NAME);

					if (((SIPResponse) capturedResponse).match(testSipResponse)
							&& ((!expiresRemoved) || capturedResponse
									.getExpires() != null)
							&& testContentList.match(contentList)) {
						matchFound = true;
						break;
					}
				}
				if (!matchFound) {
					this.failureReason = "Could not find a match for this message:\n "
							+ testSipResponse;
					logError(failureReason);
					return false;
				}
			}
			String dirName = testScript.getTestDirectory();
			fname = RFSS.this.rfssConfig.getRfssName() + "_ptt.bin";
			File file = new File(this.testScript.getReferenceMessagesDirName()
					+ "/" + fname);
			if (file.exists()) {
				FileInputStream fis = new FileInputStream(file);

				ObjectInputStream oos = new ObjectInputStream(fis);

				int counter = 0;
				while (oos.available() != 0) {
					CapturedPttPacket pttPacket = (CapturedPttPacket) oos
							.readObject();
					while (this.capturedPackets.size() != 0) {
						CapturedPttPacket packetToCheck = this.capturedPackets
								.get(counter++);
						if (packetToCheck.match(pttPacket))
							break;
					}
					this.failureReason = "Could not find a packet to match this packet : \n"
							+ pttPacket.toString();
					logError(failureReason);
					return false;

				}
			}
			return true;
		} catch (Exception ex) {
			this.failureReason = "A parse error occured while trying to test messages";
			logError(failureReason, ex);
			return false;

		}

	}

	

	public static void shutDownAll() {
		for (RFSS rfss : rfssTable.values()) {
			rfss.shutDown();
		}
		rfssTable.clear();
	}

	public static Collection<RFSS> getAllRfss() {
		return rfssTable.values();
	}

	public String getRfssSessionInfo() {
		StringBuffer retval = new StringBuffer();
		retval.append("\n<rfss-session-info").append(
				"\n\trfssId = \"" + this.rfssConfig.getDomainName() + "\"")
				.append("\n>\n");
		retval.append(this.getTransmissionControlManager().getPttManager()
				.getSessionDescriptions());
		retval.append("\n</rfss-session-info>");
		return retval.toString();
	}

	/**
	 * Shut down an emulated RFSS>
	 * 
	 */
	public void shutDown() {
		logger.debug("Shutting down emulated rfss "
				+ this.rfssConfig.getDomainName());
		this.transmissionControlManager.shutDown();
	}

	/**
	 * The RFSS constructor.
	 * 
	 * @unpublished
	 * @paream protocolObjects -- encapsuated SIP Stack.
	 * @param provider -
	 *            the provider - created by the caller.
	 * @param rfssConfig -
	 *            rfss configuration for this rfss
	 */
	private RFSS(SipProvider provider, RfssConfig rfssConfig) {
		try {
			if (rfssConfig == null)
				throw new NullPointerException("null rfssconfig");
			this.rfssConfig = rfssConfig;
			this.isRfResourcesAvailable = rfssConfig.isRfResourcesAvailable();
			this.topologyConfig = rfssConfig.getSysConfig().getTopologyConfig();
			SipStack sipStack = provider.getSipStack();
			this.requests = new ArrayList<Request>();
			this.responses = new ArrayList<Response>();

			this.servedSubscriberUnits = new HashSet<SuConfig>();

			logger.debug("InitialServedSubscriberUnits : "
					+ rfssConfig.getRfssName() + " = "
					+ rfssConfig.getInitialServedSubscriberUnits());
			for (SuConfig suConfig : rfssConfig
					.getInitialServedSubscriberUnits()) {
				// Only add those SU's that are initially turned on.
				if (suConfig.getInitialSuState() == SuState.ON
						&& suConfig.getServingRfssReferencesMeFor() != 0) {
					this.servedSubscriberUnits.add(suConfig);
					/*
					 * start a timer to remove the record if the time is not
					 * unbounded.
					 */
					if (suConfig.getServingRfssReferencesMeFor() != -1) {
						ISSITimer.getTimer()
								.schedule(
										new TimerTask() {
											private SuConfig suConfig;

											public TimerTask setSuConfig(
													SuConfig suConfig) {
												this.suConfig = suConfig;
												return this;
											}

											public void run() {
												synchronized (RFSS.this) {
													RFSS.this.servedSubscriberUnits
															.remove(suConfig);
												}
											}
										}.setSuConfig(suConfig),
										suConfig
												.getServingRfssReferencesMeFor() * 1000);
					}

				} else {
					logger.debug("not inserting into served set "
							+ suConfig.getInitialSuState());
				}
			}

			/*
			 * This is a NIST-SIP only feature - add an address resolver to the
			 * SIP stack which knows how to convert from funky Via headers to
			 * actual IP address and port.
			 */
			((SipStackImpl) sipStack)
					.setAddressResolver(new RfssAddressResolver(topologyConfig));

			/*
			 * Another NIST-SIP only feature. We register our own log record
			 * factory to log the ISSI specific information into the log stream.
			 */
			((SipStackImpl) sipStack)
					.setLogRecordFactory(new LogRecordFactoryImpl(
							topologyConfig));

			this.ipAddress = rfssConfig.getIpAddress();
			this.port = rfssConfig.getSipPort();

			this.rfssConfig.setRFSS(this);

			this.provider = provider;
			// This manages the RTP sessions.
			this.transmissionControlManager = new TransmissionControlManager(
					this);

			// Manages registrations etc.
			this.mobilityManager = new MobilityManager(this);
			this.callControlManager = new CallControlManager(this);
			this.methodHash = new Hashtable<String, SipListener>();
			methodHash.put(Request.INVITE, callControlManager);
			methodHash.put(Request.ACK, callControlManager);
			methodHash.put(Request.BYE, callControlManager);
			methodHash.put(Request.CANCEL, callControlManager);
			methodHash.put(Request.REGISTER, mobilityManager);
			int counter = 0;
			this.homeAgents = new Hashtable<String, HomeAgent>();
			for (Iterator<SuConfig> it = this.rfssConfig
					.getHomeSubsciberUnits(); it.hasNext();) {
				counter++;
				SuConfig su = it.next();
				HomeAgent homeAgent = new HomeAgent(su, this.topologyConfig,
						this);
				ServiceAccessPoints saps = this.getServiceAccessPoints(su);
				assert su.getHomeRfss() == this.rfssConfig;
				logger.debug("adding Home su " + su.getRadicalName());
				homeAgent.setServiceAccessPoints(saps);
				this.homeAgents.put(su.getRadicalName(), homeAgent);

			}

			if (counter == 0)
				logger.debug("I am not HOME to any SU ");

			logger.info("Emulated RFSS initialized!");
			logger.info("RFSS ID = " + this.rfssConfig.getRfssId());
			logger.info("Served Set = " + this.servedSubscriberUnits);

		} catch (Exception ex) {
			logger.error("exception in initializing RFS", ex);
			throw new RuntimeException("Error Intializing the RFSS", ex);
		}

	}

	// //////////////////////////////////////////////////////////////////////////
	// Package local methods.
	// //////////////////////////////////////////////////////////////////////////
	/**
	 * @unpublished
	 * 
	 */
	UnitToUnitMobilityManager getUnitToUnitMobilityManager() {
		return this.mobilityManager.getUnitToUnitMobilityManager();

	}

	/**
	 * @unpublished
	 */
	GroupMobilityManager getGroupMobilityManager() {
		return this.mobilityManager.getGroupMobilityManager();
	}

	/**
	 * @unpublished
	 */
	CallControlManager getCallControlManager() {
		return this.callControlManager;
	}

	// //////////////////////////////////////////////////////////////////////
	// JAIN-SIP listener methods.
	// //////////////////////////////////////////////////////////////////////

	/**
	 * @see javax.sip.SipListener#processRequest(javax.sip.RequestEvent)
	 * @unpublished
	 */

	public void processRequest(RequestEvent requestEvent) {

		Request request = requestEvent.getRequest();
		this.requests.add(request);
		String method = request.getMethod();
		logger.debug("RFSS: " + this.getRfssConfig().getRfssId()
				+ " processRequest: " + method);
		SipListener listener = methodHash.get(method);

		if (this.testScript.isInteractive()) {
			if (this.refMessages.hasMoreRequests()) {
				SIPRequest template = (SIPRequest) this.refMessages
						.getFirstRequest();
				if (matchSipRequest((SIPRequest) request, (SIPRequest) template)) {
					refMessages.removeRequest(template);
				}

			}

			if (!this.refMessages.hasMoreRequests()) {
				logger.debug("Request COMPLETION seen on "
						+ this.rfssConfig.getRfssName());

			} else {
				logger.debug("REQUEST COMPLETION not yet seen "
						+ this.refMessages.getRequests().size());
			}
		}
		listener.processRequest(requestEvent);

	}

	/**
	 * @see javax.sip.SipListener#processResponse(javax.sip.ResponseEvent)
	 * 
	 * @unpublished
	 */

	public void processResponse(ResponseEvent responseEvent) {
		Response response = responseEvent.getResponse();
		ClientTransaction ct = responseEvent.getClientTransaction();

		if (ct == null) {
			logger.debug("RFSS: Dropping stray response! "
					+ responseEvent.getResponse());
			return;
		}

		ViaHeader viaHeader = (ViaHeader) ct.getRequest().getHeader(
				ViaHeader.NAME);

		String host = viaHeader.getHost();

		if (!host.equals(this.rfssConfig.getDomainName())) {
			// We do not care about self routed requests.
			this.responses.add(response);
		}

		CSeqHeader cseqHeader = (CSeqHeader) response
				.getHeader(CSeqHeader.NAME);
		String method = cseqHeader.getMethod();
		SipListener listener = methodHash.get(method);

		if (this.testScript.isInteractive()) {
			if (this.refMessages.hasMoreResponses()) {
				SIPResponse template = (SIPResponse) this.refMessages
						.getFirstResponse();
				if (matchSipResponse((SIPResponse) response,
						(SIPResponse) template)) {
					refMessages.removeResponse(template);
				}

			}
			if (!this.refMessages.hasMoreResponses()) {
				logger.debug("Response COMPLETION seen on "
						+ this.rfssConfig.getRfssName());

			} else {
				logger.debug("RESPONSE COMPLETION not yet seen "
						+ this.refMessages.getResponses().size());
			}
		}

		listener.processResponse(responseEvent);

	}

	/**
	 * @see javax.sip.SipListener#processTimeout(javax.sip.TimeoutEvent)
	 * 
	 * @unpublished
	 */
	public void processTimeout(TimeoutEvent timeoutEvent) {
		logger.debug("timeoutEvent ");
		Request request;
		if (timeoutEvent.isServerTransaction()) {
			request = timeoutEvent.getServerTransaction().getRequest();
		} else {
			request = timeoutEvent.getClientTransaction().getRequest();
		}
		SipListener listener = methodHash.get(request.getMethod());
		listener.processTimeout(timeoutEvent);

	}

	/**
	 * @unpubished
	 */
	public void processIOException(IOExceptionEvent ioExceptionEvent) {
		logger.debug("ioExceptionEvent");
		this.testHarness.fail("Unexpected IO Exception");

	}

	/**
	 * @see javax.sip.SipListener#processTransactionTerminated(javax.sip.TransactionTerminatedEvent)
	 * @unpublished
	 */
	public void processTransactionTerminated(
			TransactionTerminatedEvent terminatedEvent) {
		logger.debug("txTerminatedEvent : ");
		Request request;
		if (terminatedEvent.isServerTransaction()) {
			request = terminatedEvent.getServerTransaction().getRequest();
		} else {
			request = terminatedEvent.getClientTransaction().getRequest();
		}
		SipListener listener = methodHash.get(request.getMethod());
		listener.processTransactionTerminated(terminatedEvent);

	}

	/**
	 * @see javax.sip.SipListener#processDialogTerminated(javax.sip.DialogTerminatedEvent)
	 * @unpublished
	 */
	public void processDialogTerminated(DialogTerminatedEvent dte) {
		logger.debug("dialgTerminatedEvent ");

		this.callControlManager.processDialogTerminated(dte);
	}

	// /////////////////////////////////////////////////////////////////////////////////////
	// Public methods
	// /////////////////////////////////////////////////////////////////////////////////////

	/**
	 * @unpublished
	 */
	public TransmissionControlManager getTransmissionControlManager() {
		return this.transmissionControlManager;
	}

	/**
	 * @param currentTestCase
	 *            the currentTestCase to set
	 */
	public void setCurrentTestCase(AbstractSipSignalingTest currentTestCase) {
		this.currentTestCase = currentTestCase;
	}

	/**
	 * @return the currentTestCase
	 */
	public AbstractSipSignalingTest getCurrentTestCase() {
		return currentTestCase;
	}

	/**
	 * Each Subscriber Unit abstraction ( i.e. implementation of SU) talks to
	 * the individual services offered by the RFSS via Service Access points.
	 * 
	 * @param suConfig --
	 *            the subscriber unit configuration file.
	 * @param su --
	 *            the Subscriber unit instance.
	 * @return -- a set of service access points to talk to the RFSS.
	 */
	public ServiceAccessPoints getServiceAccessPoints(SuConfig suConfig) {
		assert suConfig != null;
		MobilityManagementSAP mmSap = new MobilityManagementSAP(this, suConfig);
		CallControlSAP ccSap = new CallControlSAP(this, suConfig);
		TransmissionControlSAP tcSap = new TransmissionControlSAP(this,
				suConfig);
		ServiceAccessPoints saps = new ServiceAccessPoints(ccSap, mmSap, tcSap);
		ccSap.setSaps(saps);

		return saps;
	}

	/**
	 * Get the IP Address of this RFSS.
	 * 
	 * @return -- the IP address of the RFSS
	 */
	public String getIpAddress() {
		return this.ipAddress;
	}

	/**
	 * Get the port of this RFSS.
	 * 
	 * @return -- the port of the RFSS
	 */
	public int getPort() {
		return this.port;
	}

	/**
	 * Get the configuration of this RFSS. The configuration of the RFSS is read
	 * from the topology configuration file when the system is started up. Each
	 * RFSS has an RFSS ID. The topology configuration file contains the IP
	 * address, port and other relevant information for the RFSS.
	 * 
	 * @return -- the RFSS configuration for this RFSS.
	 */

	public RfssConfig getRfssConfig() {
		return this.rfssConfig;
	}

	/**
	 * Get the Mobility Management comopnent for the RFSS.
	 * 
	 * @return -- the mobility manager.
	 */
	public MobilityManager getMobilityManager() {
		return this.mobilityManager;

	}

	/**
	 * Return a collection of all the subscriber units that we are currently
	 * serving.
	 * 
	 * @return -- a collection of the currently served subscriber units.
	 */
	public HashSet<SuConfig> getServedSubscriberUnits() {

		return this.servedSubscriberUnits;
	}

	/**
	 * Remove a served subscriber unit ( for roaming emulation)
	 * 
	 * @param suConfig
	 */
	public synchronized void removeServedSubscriberUnit(SuConfig suConfig) {
		this.servedSubscriberUnits.remove(suConfig);
	}

	/**
	 * Add a served subscriber unit ( for roaming emulation ).
	 * 
	 * @param suConfig --
	 *            suConfig to add
	 * @param suConfig
	 */
	public synchronized void addServedSubscriberUnit(final SuConfig suConfig) {
		this.servedSubscriberUnits.add(suConfig);
		if (suConfig.getServingRfssReferencesMeFor() != -1) {
			ISSITimer.getTimer().schedule(new TimerTask() {
				public void run() {
					synchronized (RFSS.this) {
						RFSS.this.servedSubscriberUnits.remove(suConfig);
					}
				}
			}, suConfig.getServingRfssReferencesMeFor() * 1000);
		}
	}

	public HomeAgent getHomeAgent(String radicalName) {
		return this.homeAgents.get(radicalName);
	}

	/**
	 * Boolean check for if rf resources are available.
	 * 
	 * @return
	 */
	public boolean isRfResourcesAvailable() {
		return this.isRfResourcesAvailable;
	}

	/**
	 * Set the resource flag.
	 * 
	 * @param resourceVal
	 */
	public void doSetRfResourceAvailable(boolean resourceVal) throws Exception {
		boolean alert = this.isRfResourcesAvailable ^ resourceVal;
		this.isRfResourcesAvailable = resourceVal;
		if (alert) {
			this.callControlManager.alertResourceChange(resourceVal);
		}

	}

	/**
	 * Return true if I am serving this SU.
	 * 
	 * @param suConfig
	 * @return
	 */
	public boolean isSubscriberUnitServed(SuConfig suConfig) {

		boolean retval = this.servedSubscriberUnits.contains(suConfig);

		if (retval) {
			logger.debug("RfssName " + rfssConfig.getRfssName()
					+ " Served Subscriber Units " + this.servedSubscriberUnits);
		}
		return retval;
	}

	/**
	 * @return Returns the provider.
	 */
	public SipProvider getProvider() {
		return provider;
	}

	/**
	 * @param testHarness
	 *            The testHarness to set.
	 */
	public void setTestHarness(TestHarness testHarness) {
		this.testHarness = testHarness;
	}

	/**
	 * @return Returns the testHarness.
	 */
	public TestHarness getTestHarness() {
		if (this.testHarness == null)
			this.testHarness = new TestHarness();
		return testHarness;
	}

	/**
	 * @return Returns the requests.
	 */
	public ArrayList<Request> getRequests() {
		return requests;
	}

	/**
	 * @return Returns the responses.
	 */
	public ArrayList<Response> getResponses() {
		return responses;
	}

	/**
	 * Send a query to the home rfss.
	 * 
	 * @param suConfig --
	 *            su to query.
	 */
	public void doServingQuerySu(SuConfig suConfig) throws Exception {
		this.mobilityManager.getUnitToUnitMobilityManager().sendRegisterQuery(
				suConfig);

	}

	/**
	 * Sends a query operation for the SU.
	 * 
	 * @param suConfig
	 * @throws Exception
	 */
	public void doHomeQuerySu(SuConfig suConfig, boolean force, boolean confirm)
			throws Exception {
		this.mobilityManager.getUnitToUnitMobilityManager().sendRegisterQuery(
				suConfig, force, confirm);

	}

	/**
	 * Sends a query operation for the SG.
	 * 
	 * @param sgConfig
	 * @throws Exception
	 */
	public void doQueryGroup(GroupConfig sgConfig) throws Exception {
		this.mobilityManager.getGroupMobilityManager().sendRegisterQuery(
				sgConfig);

	}

	/**
	 * Get the current GroupCall for the given radical name or null if nothing
	 * exists.
	 * 
	 * @param groupRadicalName
	 */
	public GroupServing getServedGroupByRadicalName(String groupRadicalName) {
		return this.callControlManager.getGroupCallControlManager()
				.getGroupCall(groupRadicalName);

	}

	/**
	 * Return the group structure from the home RFSS.
	 * 
	 * @param groupName
	 * @return -- the Group structure.
	 * @throws Exception --
	 *             if group not found.
	 */
	public GroupHome getHomeGroupByRadicalName(String radicalName)
			throws Exception {
		return this.getGroupMobilityManager().getGroup(radicalName);
	}

	/**
	 * Issue a query from the home Rfss of a group to the given target RFSS>
	 * 
	 * @param groupConfig --
	 *            the group configuration. This should be homed at this rfss
	 *            configuration but for misconfiguration emulation, this can be
	 *            different.
	 * @param targetRfss --
	 *            the target RFSS to query from the group home
	 * @param force -
	 *            the force parameter
	 * @param confirm --
	 *            the confirm parameter.
	 */
	public void doHomeQueryGroup(GroupConfig groupConfig,
			RfssConfig targetRfss, boolean force, boolean confirm)
			throws Exception {
		if (groupConfig.getHomeRfss() != this.rfssConfig) {
			logger
					.warn("Sending Home query from RFSS other than group home -- check your configuration! ");
		}
		this.getGroupMobilityManager().homeQueryGroup(groupConfig, targetRfss,
				force, confirm);

	}

	/**
	 * Add an RTP port to the RTP resources for this RFSS.
	 * 
	 * @param nports - #
	 *            of ports to add.
	 * 
	 */
	public void doAddRtpPort(int nports) {
		try {
			logger.debug("Adding rtp ports " + nports);
			this.callControlManager.getGroupCallControlManager()
					.rtpResourcesAvailable(nports);
		} catch (Exception e) {
			logger.error("Unexpected exception ", e);
			this.getTestHarness().fail("Unexpected exception occured ");

		}

	}

	/**
	 * Remove active RTP sessions for this RFSS.
	 * 
	 * @param nports - #
	 *            of active sessions to stop
	 */
	public void removeRtpPort(int nports) {
		try {
			this.callControlManager.getGroupCallControlManager()
					.removeRtpPorts(nports);
		} catch (Exception ex) {
			logger.error("Unexpected exception ", ex);
			this.getTestHarness().fail("Unexpected exception occured", ex);
		}
	}

	/**
	 * Reinvite an Rfss to a group call from the group call home.
	 * 
	 * @param radicalName --
	 *            group radical name.
	 * @param rfssConfig --
	 *            rfss config of the rfss to reinvite.
	 * 
	 * @throws Exception
	 */
	public void doReInviteGroup(String radicalName, RfssConfig rfssConfig)
			throws Exception {
		this.callControlManager.getGroupCallControlManager()
				.reInviteRfssToGroup(radicalName, rfssConfig);

	}

	/**
	 * Change the group call priority.
	 * 
	 * @param radicalName --
	 *            groupRadical name.
	 * @param newPriority --
	 *            new priority.
	 * @param isEmergency --
	 *            isEmergency.
	 */
	public void doChangeGroupCallPriority(String groupRadicalName,
			int newPriority, boolean isEmergency) throws Exception {
		this.callControlManager.alertGroupPriorityChange(groupRadicalName,
				newPriority, isEmergency);

	}

	/**
	 * Subscribe this SU to the given group.
	 * 
	 * @param groupRadicalName
	 * @param suConfig
	 */
	public void subscribeSuToServedGroup(String groupRadicalName,
			SuConfig suConfig) throws Exception {
		this.callControlManager.getGroupCallControlManager().joinExistingGroup(
				suConfig, groupRadicalName);
	}

	/**
	 * The method that gets called when an SU arrives at an RFSS.
	 * 
	 * @param suConfig
	 */
	public void suArrived(SuConfig suConfig) {
		getMobilityManager().getUnitToUnitMobilityManager().suArrived(suConfig);
		getMobilityManager().getGroupMobilityManager().suArrived(suConfig);

	}

	public void logError(String message) {

		logger.error(message);
		String newMessageStr = "[" + this.rfssConfig.getDomainName() + "] "
				+ message;
		errorLogger.error(newMessageStr);

	}

	public void logError(String message, Exception ex) {

		this.testHarness.fail(message, ex);
		logger.error(message, ex);
		String newMessageStr = "[" + this.rfssConfig.getDomainName() + "] "
				+ message;
		errorLogger.error(newMessageStr, ex);

	}

	public boolean isSignalingCompleted() {
		boolean retval = (!this.refMessages.hasMoreRequests())
				&& (!this.refMessages.hasMoreResponses());

		return retval;
	}

	public void doRegisterRegister(SuConfig suConfig, boolean booleanValue) throws Exception {
		this.mobilityManager.getUnitToUnitMobilityManager().sendRegisterRegister(suConfig,booleanValue);
		
	}

}
