package gov.nist.p25.issi.rfss.tester;

import gov.nist.p25.issi.issiconfig.RfssConfig;
import gov.nist.p25.issi.issiconfig.TopologyConfig;
import gov.nist.p25.issi.rfss.ISSITesterConstants;
import gov.nist.p25.issi.rfss.RFSS;
import gov.nist.p25.issi.rfss.servlets.ControlServlet;
import gov.nist.p25.issi.testlauncher.TestControllerInterface;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.FilenameFilter;
import java.io.PrintWriter;
import java.util.Calendar;
import java.util.Collection;
import java.util.HashSet;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServletResponse;

import org.apache.log4j.Logger;

/**
 * 
 * The server side of the remote test controller. The methods here may be either
 * directly invoked from the gui or via http.
 * 
 * @author M. Ranganathan
 * 
 */
public class RfssController implements TestControllerInterface {

	private static Logger logger = Logger.getLogger(RfssController.class);

	private HashSet<TestRFSS> testRfssCollection = new HashSet<TestRFSS>();

	private Collection<String> ipAddress;

	private String testSuiteName;

	public RfssController(Collection<String> ipAddress) {
		this.ipAddress = ipAddress;
	}

	public void setTestSuite(String testSuiteName) {
		this.testSuiteName = testSuiteName;

	}

	/**
	 * Load the test case and intialize all RFSSs.
	 * 
	 * @param testCase --
	 *            the test case to load.
	 * @param interactive --
	 *            whether interactive or not.
	 * 
	 * @throws Exception
	 */
	public void loadTest(String scenarioDir, String testCase,
			String testNumber,
			String testTopology, String globalTopologyName,
			String systemTopologyName, boolean interactive,
			boolean evaluatePassFail) throws Exception {
		this.testRfssCollection.clear();
		RFSS.reset();

		File logDir = new File(ISSITesterConstants.DEFAULT_LOGS_DIR);

		if (!logDir.exists()) {
			logDir.mkdirs();
		}

		File[] logFiles = logDir.listFiles(new FilenameFilter() {

			public boolean accept(File dir, String name) {
				return (name.endsWith(".sip") || name.endsWith(".ptt"));

			}

		});

		// Delete all the SIP log files
		for (File file : logFiles) {
			file.delete();
			logger.info("sipTraceFileLength  = " + file.length());
		}

		String pttTrace = TestScript.getRtpMessageLog();
		File pttTraceFile = new File(pttTrace);
		if (pttTraceFile.delete()) {
			logger.debug("successfully deleted ptt trace file");
		}
		/*
		 * File debugLogFile = new File(ISSITesterConstants.DEBUG_LOG); if
		 * (debugLogFile.delete()) { System.out.println("Successfully deleted
		 * debug log file"); }
		 */

		File errorLogFile = new File(ISSITesterConstants.ERROR_LOG);
		if (errorLogFile.delete()) {
			logger.debug("Successfully deleted errorlog.txt");
		}

		try {

			ISSITesterConstants.setScenarioDir(scenarioDir);

			testSuiteName = new File(scenarioDir).getParentFile().getName();

			logger.info("loadScenario" + testCase);
			String testDir = new File(testCase).getParent();
			testCase = scenarioDir + "/" + testCase;
			String localTopology = scenarioDir + "/" + testTopology;
			// Record the test case in the logs directory.

			File testCaseLog = new File(ISSITesterConstants.DEFAULT_LOGS_DIR
					+ "/" + ISSITesterConstants.CURRENT_TEST_CASE_XML);
			testCaseLog.delete();
			PrintWriter printWriter = new PrintWriter(new FileWriter(
					testCaseLog));
			printWriter.println("<testcase");
			printWriter.println("date = \"" + Calendar.getInstance().getTime()
					+ "\"\n");
			printWriter.println("testNumber = \""+testNumber + "\"\n");
			printWriter.println("interactive = \"" + interactive + "\"\n");
			printWriter.println("scenarioDir = \"" + scenarioDir + "\"\n");
			printWriter.println("testDirectory = \"" + testDir + "\"\n");
			printWriter.println("topology = \"" + testTopology + "\"\n");
			printWriter.println("/>");
			printWriter.close();

			File globalTopologyLog = new File(
					ISSITesterConstants.DEFAULT_LOGS_DIR + "/"
							+ ISSITesterConstants.GLOBAL_TOPOLOGY_XML);
			if (globalTopologyLog.exists())
				globalTopologyLog.delete();

			globalTopologyLog.createNewFile();

			// TODO read this from the manifest.

			File testGlobalTopology = new File(new File(scenarioDir) + "/"
					+ ISSITesterConstants.GLOBAL_TOPOLOGY_XML);

			FileReader reader = new FileReader(testGlobalTopology);

			char[] buffer = new char[(int) testGlobalTopology.length()];

			reader.read(buffer);

			FileWriter writer = new FileWriter(globalTopologyLog);
			writer.write(buffer);

			writer.close();
			reader.close();

			File systemTopologyLog = new File(
					ISSITesterConstants.DEFAULT_LOGS_DIR + "/"
							+ ISSITesterConstants.SYSTEM_TOPOLOGY_XML);

			if (systemTopologyLog.exists())
				systemTopologyLog.delete();

			systemTopologyLog.createNewFile();

			File testSystemTopology = new File(testGlobalTopology
					.getParentFile().getParentFile().getAbsolutePath()
					+ "/" + ISSITesterConstants.SYSTEM_TOPOLOGY_XML);

			reader = new FileReader(testSystemTopology);

			buffer = new char[(int) testSystemTopology.length()];

			reader.read(buffer);

			writer = new FileWriter(systemTopologyLog);
			writer.write(buffer);

			writer.close();
			reader.close();

			logger.debug("globalTopologyName = " + globalTopologyName);
			TestScript tscript = new TestScriptParser(testCase, localTopology,
					globalTopologyName, systemTopologyName, interactive)
					.parse();
			tscript.setInteractive(interactive);
			if (!interactive) {
				tscript.setTraceCaptureEnabled(true);
			} else {
				tscript.setTraceCaptureEnabled(false);
			}
			tscript.setEvaluatePassFail(evaluatePassFail);
			TopologyConfig topology = tscript.getTopologyConfig();

			for (String ipAddress : this.ipAddress) {
				for (RfssConfig rfssConfig : topology.getRfssConfigurations()) {

					if ((rfssConfig.getIpAddress().equals(ipAddress) && rfssConfig
							.isEmulated())
							|| !interactive) {
						TestRFSS testRfss = new TestRFSS(rfssConfig, topology);
						testRfss.loadScenario(tscript);
						testRfssCollection.add(testRfss);
					}
				}
			}
		} catch (Exception ex) {
			logger.error("Unexpected error loading test case", ex);
			throw ex;

		}

	}

	/**
	 * Run the loaded test case.
	 * 
	 */
	public void runTest() {
		for (TestRFSS testRfss : testRfssCollection) {
			testRfss.runScenario();
		}
	}

	public boolean isTestCompleted() throws Exception {
		try {

			boolean retval = true;
			for (TestRFSS rfss : this.testRfssCollection) {
				retval &= rfss.getRfss().isSignalingCompleted();
			}
			logger.debug("RfssController: isTestCompleted " + retval);
			return retval;
		} catch (Exception ex) {
			logger.error("Unexpected exception occured", ex);
			return false;
		}
	}

	/**
	 * Get the session information.
	 */
	public String getRfssPttSessionInfo() {
		StringBuffer retval = new StringBuffer();
		for (TestRFSS testRfss : this.testRfssCollection) {
			retval.append(testRfss.getPttSessionInfo());

		}
		return retval.toString();
	}

	/**
	 * Run the next scenario in prompting mode.
	 * 
	 * @param scenario
	 * @throws Exception
	 */
	public void signalNextScenario(String scenario) throws Exception {
		for (TestRFSS rfss : this.testRfssCollection) {
			rfss.getCurrentCallSetupTest().executeNextScenario(scenario);
		}

	}

	/**
	 * Tear down the test.
	 */
	public void tearDownCurrentTest() {

		for (TestRFSS testRfss : this.testRfssCollection) {
			testRfss.getCurrentCallSetupTest().abort();
			testRfss.getCurrentCallSetupTest().tearDown();
		}
	}

	/**
	 * Do the completion tasks
	 */
	public void signalTestCompletion() {
		for (TestRFSS testRfss : this.testRfssCollection) {
			testRfss.getCurrentCallSetupTest().startCompletionTimer();
		}
	}

	/**
	 * Get the test results from our RFSSs.
	 * 
	 * @return
	 */
	public boolean getTestResults() {
		boolean passed = true;

		for (TestRFSS testRfss : this.testRfssCollection) {
			if (testRfss.getCurrentCallSetupTest().isFailed()) {
				passed = false;
			}
		}
		return passed;

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see gov.nist.p25.issi.testlauncher.TestControllerInterface#getStackSipLog()
	 */
	public String getStackSipLog() throws Exception {
		try {
			StringBuffer retval = new StringBuffer();
			for (TestRFSS testRfss : this.testRfssCollection) {
				String msglog = testRfss.getSipMessageLog();
				retval.append(msglog);
			}
			return retval.toString();
		} catch (Exception ex) {
			return null;
		}
	}

	public String getStackSipLogs() throws Exception {
		return this.getStackSipLog();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see gov.nist.p25.issi.testlauncher.TestControllerInterface#getStackPttLog()
	 */
	public String getStackPttLog() throws Exception {
		try {
			for (TestRFSS testRfss : this.testRfssCollection) {
				testRfss.flushPttMessageLog();

			}
			TopologyConfig.closeRtpMessageLogStream();

			String messageLogFile = TestScript.getRtpMessageLog();
			BufferedReader bufferdReader = new BufferedReader(new FileReader(
					new File(messageLogFile)));
			StringBuffer stringBuffer = new StringBuffer();
			String line;
			while ((line = bufferdReader.readLine()) != null) {
				stringBuffer.append(line + "\n");
			}
			bufferdReader.close();
			return stringBuffer.toString();
		} catch (Exception ex) {
			return null;
		}

	}

	public String getStackPttLogs() throws Exception {
		return this.getStackPttLog();
	}

	public String getPttTraces() throws Exception {

		throw new Exception("Method not supported");
	}

	public String getSipTraces() throws Exception {

		throw new Exception("Method not supported");
	}

	public void reset() {
		this.testRfssCollection = new HashSet<TestRFSS>();
		RFSS.reset();

	}

	public String getErrorLog() throws Exception {
		return TestRFSS.getErrorLog();
	}

}
