package gov.nist.p25.issi.rfss.tester;

import gov.nist.p25.issi.issiconfig.GroupConfig;
import gov.nist.p25.issi.issiconfig.SuConfig;
import gov.nist.p25.issi.issiconfig.SuState;
import gov.nist.p25.issi.issiconfig.TopologyConfig;
import gov.nist.p25.issi.p25payload.TransmitPriorityType;

import org.apache.log4j.Logger;
import org.python.util.PythonInterpreter;

/**
 * This defines an event for an SU. An SU can specify a script to be executed at
 * a certain time during the run. It gets access to all the methods that are
 * defined in the TestSuInterface.
 * 
 * @author M. Ranganathan
 * 
 */
public class SuScript extends AbstractScenario implements SuScriptInterface {
	private PythonInterpreter interpreter;

	private static Logger logger = Logger.getLogger(SuScript.class);

	SuConfig suConfig;

	String scriptingEngine;

	String script;

	String method;

	String description;

	private TestScript testScript;

	private TopologyConfig topologyConfig;

	SuScript(TestScript testScript) {
		this.testScript = testScript;

	}

	private TestSU getSU() {
		return (TestSU) this.suConfig.getSU();
	}

	void init() throws Exception {
		if (scriptingEngine == null)
			this.scriptingEngine = "jython";
		if ((!"jython".equals(scriptingEngine))
				&& (!"python".equals(scriptingEngine))) {
			throw new Exception("Interpreter not supported " + scriptingEngine);
		}

		
		// logger.info("script = " + this.script);
		this.interpreter = new PythonInterpreter();
		interpreter.exec("from java.lang import *");
		interpreter.exec("from java.util import *");
		interpreter.exec("from gov.nist.p25.issi.issiconfig import *");
		interpreter.exec("from gov.nist.p25.issi.rfss.tester import *");
		interpreter.exec("global su");
		interpreter.set("su", this);
		interpreter.set("testScript", testScript);
		this.interpreter.exec(script);
	}

	@Override
	public void setDescription(String description) {
		super.setDescription("Perform the following action on the following SU : \n"
				+ "Su ID = %x" +  Integer.toHexString(this.suConfig.getSuId()) + "\nisEmulated : "+suConfig.isEmulated() +"\n" + description + "\n");
	}

	
	/* (non-Javadoc)
	 * @see gov.nist.p25.issi.rfss.tester.SuScriptInterface#terminateCallSegment()
	 */
	public void terminateCallSegment() {
		getSU().terminateUnitToUnitCall();
	}

	public void sendHeartbeatQuery() {
		getSU().getCurrentUnitToUnitCall().getPttSession().getHeartbeatTransmitter().sendHeartbeatQuery();
	}
	

	/* (non-Javadoc)
	 * @see gov.nist.p25.issi.rfss.tester.SuScriptInterface#sendTalkSpurt(int)
	 */
	public void sendTalkSpurt(int nblocks) {
		getSU().sendTalkSpurt(nblocks);

	}

	/* (non-Javadoc)
	 * @see gov.nist.p25.issi.rfss.tester.SuScriptInterface#sendTalkSpurt()
	 */
	public void sendTalkSpurt() {
		getSU().sendTalkSpurt();
	}

	/* (non-Javadoc)
	 * @see gov.nist.p25.issi.rfss.tester.SuScriptInterface#forceTalkSpurt(int)
	 */
	public void forceTalkSpurt(int nblocks) {
		getSU().forceTalkSpurt(nblocks);
	}

	/* (non-Javadoc)
	 * @see gov.nist.p25.issi.rfss.tester.SuScriptInterface#setState(gov.nist.p25.issi.issiconfig.SuState)
	 */

	public void setState(SuState suState) {
		this.getSU().setState(suState);

	}
	
	
	
	

	public void runScript(TopologyConfig topologyConfig) {

		/*
		 * We run the script for an emulated Su instance.
		 */
		try {
			if (this.suConfig.getSU() != null &&  this.suConfig.isEmulated() ) {
				this.topologyConfig = topologyConfig;
				this.interpreter.set("topologyConfig", topologyConfig);
				this.interpreter.eval(this.method + "()");
			} else {
				logger.info("nothing to run");
			}
		} catch (Throwable th) {
			System.err.println("Unexpected error caught , check the SU Script "
					+ th);
		}
	}

	/* (non-Javadoc)
	 * @see gov.nist.p25.issi.rfss.tester.SuScriptInterface#sendTalkSpurt(java.lang.String, int)
	 */
	public void sendTalkSpurt(String groupName, int nblocks) throws Exception {
		GroupConfig groupConfig = this.topologyConfig
				.getGroupConfigByName(groupName);
		if (groupConfig == null)
			throw new Exception("Group not found " + groupName);
		getSU().sendTalkSpurt(groupConfig.getRadicalName(), nblocks);

	}

	/* (non-Javadoc)
	 * @see gov.nist.p25.issi.rfss.tester.SuScriptInterface#sendTalkSpurt(java.lang.String)
	 */
	public void sendTalkSpurt(String groupName) throws Exception {
		GroupConfig groupConfig = this.topologyConfig
				.getGroupConfigByName(groupName);
		if (groupConfig == null)
			throw new Exception("Group not found " + groupName);
		getSU().sendTalkSpurt(groupConfig.getRadicalName());

	}
	
	/* (non-Javadoc)
	 * @see gov.nist.p25.issi.rfss.tester.SuScriptInterface#forceTalkSpurt(java.lang.String, int)
	 */
	public void forceTalkSpurt(String groupName, int nblocks ) throws Exception {
		GroupConfig groupConfig = this.topologyConfig.getGroupConfigByName(groupName);
		if (groupConfig == null)
			throw new Exception("Group not found " + groupName);
		getSU().forceTalkSpurt(groupConfig.getRadicalName(), nblocks);
	}
	
	/* (non-Javadoc)
	 * @see gov.nist.p25.issi.rfss.tester.SuScriptInterface#forceTalkSpurt(java.lang.String)
	 */
	public void forceTalkSpurt(String groupName ) throws Exception {
		GroupConfig groupConfig = this.topologyConfig.getGroupConfigByName(groupName);
		if (groupConfig == null)
			throw new Exception("Group not found " + groupName);
		getSU().forceTalkSpurt(groupConfig.getRadicalName());
	}
	
	/* (non-Javadoc)
	 * @see gov.nist.p25.issi.rfss.tester.SuScriptInterface#setTransmitPriority(java.lang.String, int)
	 */
	public void setTransmitPriority(String type,
			int level) throws Exception {
		TransmitPriorityType transmitPriority = null;
		if ( type.equals("NORMAL")) transmitPriority = TransmitPriorityType.NORMAL;
		else if ( type.equals("PRIORITY")) transmitPriority = TransmitPriorityType.PRIORITY;
		else if ( type.equals("PREEMPTIVE_PRIORITY")) transmitPriority = TransmitPriorityType.PREEMPTIVE_PRIORITY;
		if (level > (2<<4) -1 ) throw new Exception ("Bad argument - only values from 0 to 15");
		getSU().setTransmitPriorityType(transmitPriority);
		getSU().setTransmissionPriorityLevel(level);
	}

	public SuConfig getSuConfig() {
			return this.suConfig;
	}

}
