package gov.nist.p25.issi.rfss.tester;

import gov.nist.p25.issi.issiconfig.RfssConfig;
import gov.nist.p25.issi.issiconfig.TopologyConfig;
import gov.nist.p25.issi.issiconfig.WebServerAddress;
import gov.nist.p25.issi.rfss.ISSITesterConstants;
import gov.nist.p25.issi.rfss.ProtocolObjects;
import gov.nist.p25.issi.rfss.RFSS;
import gov.nist.p25.issi.rfss.RemoteAppender;
import gov.nist.p25.issi.rfss.servlets.ControlServlet;
import gov.nist.p25.issi.rfss.servlets.WelcomeServlet;
import gov.nist.p25.issi.startup.DietsConfigProperties;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileFilter;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.InputStream;
import java.util.HashSet;
import java.util.Properties;

import javax.sip.ListeningPoint;
import javax.sip.SipProvider;
import javax.sip.SipStack;
import javax.swing.JFrame;
import javax.swing.JOptionPane;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.log4j.PropertyConfigurator;
import org.apache.log4j.SimpleLayout;
import org.mortbay.http.HttpContext;
import org.mortbay.http.SocketListener;
import org.mortbay.http.handler.ResourceHandler;
import org.mortbay.jetty.Server;
import org.mortbay.jetty.servlet.ServletHandler;
import org.mortbay.util.FileResource;

/**
 * This is the test proxy wrapper. A test case is first configured by starting
 * as many of these test proxy servers as are necessary on multiple test
 * machines. A test is launched by invoking the methods loadScenario and
 * runScenario with the <b>same</b> arguments on each emulator.
 * 
 * @author M. Ranganathan
 * 
 */
public class TestRFSS  {
	/** The HTTP Server to query information and manage the RFSS */
	private static Server httpServer;

	private transient RFSS rfss;

	private transient SipProvider provider;

	private String ipAddress;

	private RfssConfig rfssConfig;

	private CallSetupTest currentCallSetupTest;

	public static Logger logger = Logger.getLogger(TestRFSS.class);

	private static boolean remoteAppenderInitialized;

	private static ISSITesterConfiguration testerConfiguration = null;

	private static String testerConfigFile;
	
	private static boolean configured;

	private static HttpContext httpContext;

	private static HttpContext logContext;

	// ////////////////////////////////////////////////////////////
	// Inner classes - this one is just a simple file filter.
	// ///////////////////////////////////////////////////////////
	static class TestDirectoryNameFilter implements FileFilter {

		public boolean accept(File pathname) {

			return pathname.isDirectory()
					&& !pathname.getPath().endsWith("dtd");

		}

	}

	// /////////////////////////////////////////////////////////////
	// Private methods
	// ////////////////////////////////////////////////////////////

	private void registerRemoteMonitor(String ipAddress, int httpPort)
			throws Exception {
		String key = ipAddress + ":" + httpPort;

		if (!TestRFSS.remoteAppenderInitialized) {
			Logger logger = Logger.getLogger("gov.nist.p25");
			logger.addAppender(new RemoteAppender(this.rfssConfig, ipAddress,
					httpPort, new SimpleLayout()));
			TestRFSS.remoteAppenderInitialized = true;
		}

	}

	

	/**
	 * Configure the web server to talk to the outside world.
	 * 
	 * 
	 * @param configPort --
	 *            port through which to talk to the outside world
	 * 
	 */
	private static void startWebServer()
			throws Exception {

		

		TestRFSS.httpContext = new HttpContext();
		httpContext.setContextPath("/rfss/*");
		ServletHandler servletHandler = new ServletHandler();
		// Register a welcome servlet with the Servlet Handler.
		servletHandler.addServlet("welcome", "/*", WelcomeServlet.class
				.getName());
			servletHandler.addServlet("control", "/control/*", ControlServlet.class
				.getName());
		// "gov.nist.p25.issi.servlets.ControlServlet");
		httpContext.addHandler(servletHandler);
		httpServer.addContext(httpContext);

		// Set up a handler to export the log files.
		TestRFSS.logContext = new HttpContext();
		logContext.setContextPath("/logs/*");
		FileResource logFilesDir = (FileResource) FileResource
				.newResource("./logs/");
		ResourceHandler rhandler = new ResourceHandler();
		rhandler.setName("logs");
		rhandler.setAllowedMethods(new String[] { "GET", "CONNECT", "HEAD" });
		logContext.setBaseResource(logFilesDir);
		rhandler.setDirAllowed(true);
		logContext.addHandler(rhandler);
		httpServer.addContext(logContext);
		rhandler.start();

		httpContext.start();
		
	}

	// ////////////////////////////////////////////////////////
	// PUBLIC METHODS
	// ////////////////////////////////////////////////////////

	/**
	 * Flush the PTT messsage log to the disk.
	 * 
	 * @throws Exception
	 */
	public void flushPttMessageLog() throws Exception {
		this.getRfss().flushCapturedPttMessages();

	}

	/**
	 * Get the SIP Message log.
	 * 
	 * @return
	 */
	public String getSipMessageLog() {
		try {

			File messageLogFile = new File(ISSITesterConstants.DEFAULT_LOGS_DIR
					+ "/" + this.rfssConfig.getRfssName() + ".messagelog.sip");

			StringBuffer stringBuffer = new StringBuffer();

			BufferedReader bufferdReader = new BufferedReader(new FileReader(
					messageLogFile));
			String line;
			while ((line = bufferdReader.readLine()) != null) {
				stringBuffer.append(line + "\r\n");
			}
			bufferdReader.close();

			return stringBuffer.toString();
		} catch (Exception ex) {
			logger.error("Error reading log file", ex);
			return "";
		}
	}

	/**
	 * Get the error log.
	 * 
	 */
	public static String getErrorLog() {
		try {

			File messageLogFile = new File(ISSITesterConstants.ERROR_LOG);

			StringBuffer stringBuffer = new StringBuffer();

			BufferedReader bufferdReader = new BufferedReader(new FileReader(
					messageLogFile));
			String line;
			while ((line = bufferdReader.readLine()) != null) {
				stringBuffer.append(line + "\r\n");
			}
			bufferdReader.close();

			return stringBuffer.toString();
		} catch (Exception ex) {
			logger.error("Error reading log file", ex);
			return "";
		}
	}

	/**
	 * Run the previously loaded scenario.
	 * 
	 */

	public void runScenario() {
		logger.debug("TestRFSS:runScenario() + " + this);
		this.currentCallSetupTest.doCallSetup();

	}

	/*
	 * Get the current call setup test.
	 */
	public CallSetupTest getCurrentCallSetupTest() {
		return this.currentCallSetupTest;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see gov.nist.p25.issi.rfss.RFSSTestProxyIF#loadScenario(java.lang.String)
	 */
	public void loadScenario(TestScript testCase) throws Exception {
		logger.info("Loading testCase : " + testCase);

		long currentTime = System.currentTimeMillis();
		if (this.currentCallSetupTest != null
				&& (currentTime - this.currentCallSetupTest.getStartTime() < ISSITesterConstants.TEST_RUNS_FOR)) {
			throw new Exception("Test currently in progress");
		}
		try {
			if (this.currentCallSetupTest != null) {
				provider.removeSipListener(this.currentCallSetupTest);
			}
			this.rfss = RFSS.create(provider, rfssConfig, testCase);

			this.currentCallSetupTest = new CallSetupTest(rfssConfig, testCase,
					rfss);
			this.rfss.setCurrentTestCase(this.currentCallSetupTest);

		} catch (Exception ex) {
			// ex.printStackTrace();
			throw new Exception(
					"Exception occured while launching the test case", ex);
		}
	}

	/**
	 * Get the RFSS object that this TestRFSS wraps.
	 * 
	 * @return the rfss object
	 */

	public RFSS getRfss() {
		return rfss;
	}

	/**
	 * Get the RFSS communication objects.
	 */
	public String getPttSessionInfo() {
		return this.rfss.getRfssSessionInfo();
	}

	/**
	 * The constructor.
	 * 
	 * @param rfssConfig -
	 *            the static configuration of the RFSS for which we want to
	 *            start this emulated instance.
	 * 
	 * @param topologyConfig --
	 *            the topology configuration
	 */
	public TestRFSS(RfssConfig rfssConfig, TopologyConfig topologyConfig)
			throws Exception {
		if (rfssConfig == null)
			throw new NullPointerException("Null arg!");
		this.rfssConfig = rfssConfig;

		this.ipAddress = rfssConfig.getIpAddress();
		int port = rfssConfig.getSipPort();
		SipStack sipStack = ProtocolObjects.getSipStack(rfssConfig
				.getRfssName(), topologyConfig);
		ListeningPoint listeningPoint = sipStack.createListeningPoint(
				ipAddress, port, "udp");

		this.provider = sipStack.createSipProvider(listeningPoint);

	}

	public static void setTesterConfiguration(
			ISSITesterConfiguration testerConfig) {
		testerConfiguration = testerConfig;
	}

	public static ISSITesterConfiguration getTesterConfiguration() {
		return testerConfiguration;
	}
	
	public static int configure(Server server, String configFileName ) 
		throws Exception {
		/*
		 * Start web severs for all addresses we own.
		 */
		int startCount = 0;
		int port = -1;
		testerConfigFile = configFileName;
		httpServer = server;
		TestRFSS.testerConfiguration = new ISSITesterConfigurationParser(testerConfigFile).parse();

		for (String ipAddress : testerConfiguration.getLocalAddresses()) {
			HashSet<String> testers = new HashSet<String> ();
			
			for (WebServerAddress config : testerConfiguration
					.getEmulatorConfigurations()) {
				if (config.getIpAddress().equals(ipAddress)) {
					port = config.getHttpPort();
				}
				testers.add(ipAddress);
			}

			ControlServlet.setIpAddress(testers);
			WelcomeServlet.setISSITestLauncherConfig(testerConfiguration);
			startWebServer();
			startCount ++;
			
		}
		logger.info("Start count " + startCount);
		configured = true;
		return startCount;
	}
	
	public static boolean isConfigured() {
		return configured;
	}
	
	public static void unconfigure() throws Exception {
		httpContext.stop();
		httpServer.removeContext(httpContext);
		logContext.stop();
		httpServer.removeContext(logContext);
		
	}

	// //////////////////////////////////////////////////////////////////
	// Main entry point for the RFSS when running in distributed mode.
	// /////////////////////////////////////////////////////////////////

	public static void main(String[] args) throws Exception {
		
		Properties props = new Properties();

		String fileName = null;
		
		for (int i = 0; i < args.length; i++) {
			 if (args[i].equals("-startup")) {
				fileName = args[++i];
				InputStream inStream = new FileInputStream(new File(fileName));

				props.load(inStream);
			} 

		}
		if (fileName == null ) fileName = System.getProperty("diets.startup");
		
		if ( fileName == null) {
			throw new Exception("Missing startup properties");
		}
		logger.info("Using startup property " + fileName);

		testerConfigFile = props.getProperty(DietsConfigProperties.DAEMON_CONFIG_PROPERTY,
				testerConfigFile);
		
		if (testerConfigFile == null)
			throw new Exception("Missing a required parameter -testerConfig");

		Server httpServer = new Server();
		TestRFSS.configure(httpServer, testerConfigFile);
		
		httpServer.start();

		

	}



	

}
