package gov.nist.p25.issi.rfss.tester;

import gov.nist.p25.issi.ISSITimer;
import gov.nist.p25.issi.issiconfig.GroupConfig;
import gov.nist.p25.issi.issiconfig.RfssConfig;
import gov.nist.p25.issi.issiconfig.SuConfig;
import gov.nist.p25.issi.issiconfig.SuState;
import gov.nist.p25.issi.p25body.CallParamContent;
import gov.nist.p25.issi.p25body.ContentList;
import gov.nist.p25.issi.p25payload.TransmitPriorityType;
import gov.nist.p25.issi.rfss.CallControlSAP;
import gov.nist.p25.issi.rfss.CallSetupConfirmEvent;
import gov.nist.p25.issi.rfss.CallSetupRequestEvent;
import gov.nist.p25.issi.rfss.CallTeardownIndicateEvent;
import gov.nist.p25.issi.rfss.CallTeardownResponseEvent;
import gov.nist.p25.issi.rfss.DeregistrationResponseEvent;
import gov.nist.p25.issi.rfss.GroupServing;
import gov.nist.p25.issi.rfss.MobilityManagementSAP;
import gov.nist.p25.issi.rfss.RFSS;
import gov.nist.p25.issi.rfss.RegisterEvent;
import gov.nist.p25.issi.rfss.RegistrationResponseEvent;
import gov.nist.p25.issi.rfss.SuInterface;
import gov.nist.p25.issi.rfss.ServiceAccessPoints;
import gov.nist.p25.issi.rfss.UnitToUnitCall;
import gov.nist.p25.issi.rfss.UnitToUnitCallControlResponseEvent;
import gov.nist.p25.issi.testlauncher.TestHarness;
import gov.nist.p25.issi.transctlmgr.TransmissionControlSAP;
import gov.nist.p25.issi.utils.ThreadedExecutor;
import gov.nist.p25.issi.utils.WarningCodes;

import java.util.TimerTask;
import java.util.concurrent.ThreadPoolExecutor;

import javax.sip.DialogState;
import javax.sip.ServerTransaction;
import javax.sip.address.SipURI;
import javax.sip.header.CallIdHeader;
import javax.sip.header.FromHeader;
import javax.sip.header.ToHeader;
import javax.sip.message.Request;
import javax.sip.message.Response;

import org.apache.log4j.Logger;

/**
 * This is an abstraction for a Test Subscriber Unit ( SU ). The Test SU is
 * called by the test driver to send requests and expect defined responses for
 * test cases.
 * 
 * @author M. Ranganathan
 * 
 */
public class TestSU implements SuInterface {

	private CallControlSAP ccSap;

	private SuConfig peerSuConfig;

	private TransmissionControlSAP tcSap;

	private MobilityManagementSAP mmSap;

	private boolean gotCCSetupResponse;

	private UnitToUnitCall callSegment;

	private GroupServing groupServing;

	private boolean byeOKReceived;

	private boolean byeSent;

	private boolean isCaller;

	private boolean okSent;

	private boolean ackReceived;

	private ServiceAccessPoints saps;

	private SuConfig suConfig;

	protected boolean gotRegisterDergister;

	protected boolean gotGroupCallSetupResponseEvent;

	private TransmitPriorityType transmitPriorityType = TransmitPriorityType.NORMAL;

	private int transmissionPriorityLevel = 0;

	private GroupConfig groupToCall;

	private TestHarness testHarness;

	private SuState suState;

	private AbstractScenario scenario;

	private boolean cancelSent;

	private static Logger logger = Logger.getLogger(TestSU.class);

	class SuToSuCallCancelTimerTask extends TimerTask {

		private UnitToUnitCall callSegment;

		SuToSuCallCancelTimerTask(UnitToUnitCall callSegment) {
			this.callSegment = callSegment;
		}

		@Override
		public void run() {

			try {
				TestSU.this.ccSap.ccCancelCallSegment(this.callSegment);
			} catch (Exception ex) {
				testHarness.fail("Unexpected exception", ex);
			}

		}

	}

	class DelayTask implements Runnable {

		private CallSetupRequestEvent ccRequestEvent;

		public DelayTask(CallSetupRequestEvent ccRequestEvent) {
			this.ccRequestEvent = ccRequestEvent;

		}

		public void run() {
			// Auto-generated method stub
			// send redundant provsional responses (one every minute -- see RFC
			// 3216 section 13.2.1.1)
			int count = suConfig.getInviteProcessingDelay() / 60;
			if (count == 0 && suConfig.getInviteProcessingDelay() > 0)
				count = 1;
			SuConfig calledSu = ccRequestEvent.getCallSegment()
					.getCalledSuConfig();
			CallParamContent callParamContent = ccRequestEvent.getCallSegment()
					.getCallParamContent();

			for (int niterations = 0; niterations < count; niterations++) {
				try {
					if (getCurrentRfss().getRFSS().isRfResourcesAvailable()) {
						ccSap.ccSetupIndicateResponse(ccRequestEvent,
								Response.RINGING);

					} else {
						if (calledSu.getUserServiceProfile()
								.isAvailabilityCheckSupported()) {
							if (callParamContent.getCallParam()
									.isAvailabilityCheckRequested()) {
								ccSap.ccSetupIndicateResponse(ccRequestEvent,
										Response.SESSION_PROGRESS);
							} else {
								break;
							}
						} else {
							break;
						}
					}
				} catch (Exception e1) {
					logger.error("unexpected exception", e1);
				}

				int timeToSleep = suConfig.getInviteProcessingDelay()
						- niterations * 60;

				try {
					if (timeToSleep < 60)
						Thread.sleep(timeToSleep * 1000);
					else
						Thread.sleep(60 * 1000);
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}

			}
			actuallyHandleSetupIndicate(ccRequestEvent);
		}

	}

	// ///////////////////////////////////////////////////////////////
	// Constructor
	// ///////////////////////////////////////////////////////////////
	public TestSU(SuConfig suConfig, TestHarness testHarness) {
		this.suConfig = suConfig;
		this.testHarness = testHarness;
		this.suState = suConfig.getInitialSuState();
		logger.debug("TestSU: InitialState = " + this.suState);
		logger.debug("TestSU: " + suConfig.getSuName());

	}

	// //////////////////////////////////////////////////////////////////////////////////////
	// get the current RFSS where this TestSU is residing.
	// /////////////////////////////////////////////////////////////////////////////////////

	private RfssConfig getCurrentRfss() {
		return this.ccSap.getRfssConfig();
	}

	private void doSendTalkSpurt(SuToSuCallSetupScenario callSetupScenario) {
		if (callSetupScenario.isForce()) {
			ISSITimer.getTimer().schedule(new TimerTask() {

				@Override
				public void run() {

					TestSU.this.forceTalkSpurt();
				}

			}, 1000);

		} else {
			ISSITimer.getTimer().schedule(new TimerTask() {

				@Override
				public void run() {

					TestSU.this.sendTalkSpurt();
				}

			}, 1000);

		}
	}

	

	// /////////////////////////////////////////////////////////////////////////////////////
	// class methods
	// /////////////////////////////////////////////////////////////////////////////////////
	public String getCurrentUnitToUnitCallCallId() {
		return this.callSegment == null ? null : this.callSegment.getCallID();
	}
	
	/**
	 * Send a talk spurt after you get an OK.
	 * 
	 * @param callSetupScenario
	 */
	public void doSendTalkSpurt(GroupCallSetupScenario callSetupScenario) {
		if (callSetupScenario.isTalkSpurtForced()) {
			ISSITimer.getTimer().schedule(new TimerTask() {

				private GroupCallSetupScenario groupCallSetupScenario;

				TimerTask setScenario(GroupCallSetupScenario scenario) {
					this.groupCallSetupScenario = scenario;
					return this;
				}

				@Override
				public void run() {

					TestSU.this.forceTalkSpurt(groupCallSetupScenario
							.getCalledGroupConfig().getRadicalName());
				}

			}.setScenario(callSetupScenario), 1000);

		} else {
			ISSITimer.getTimer().schedule(new TimerTask() {

				private GroupCallSetupScenario groupCallSetupScenario;

				TimerTask setScenario(GroupCallSetupScenario scenario) {
					this.groupCallSetupScenario = scenario;
					return this;
				}

				@Override
				public void run() {

					TestSU.this.sendTalkSpurt(groupCallSetupScenario
							.getCalledGroupConfig().getRadicalName());
				}

			}.setScenario(callSetupScenario), 1000);

		}
	}
	public void handleRegistrationResponseEvent(
			RegistrationResponseEvent registrationResponseEvent) {
		Response response = registrationResponseEvent.getResponse();

		if (response.getExpires() == null
				&& response.getStatusCode() == Response.OK) {
			testHarness.fail("No expires header in response");
		}
		FromHeader from = (FromHeader) response.getHeader(FromHeader.NAME);
		ToHeader toHeader = (ToHeader) response.getHeader(ToHeader.NAME);
		testHarness.assertEquals(
				"From and To Header Must have Identical Addresses!", from
						.getAddress(), toHeader.getAddress());

		logger.info("handleRegistrationResponseEvent");

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see gov.nist.p25.issi.rfss.tester.TestSUInterface#setState(gov.nist.p25.issi.issiconfig.SuState)
	 */

	public void setState(SuState suState) {
		logger.debug("suName = " + this.getSuConfig().getSuName());
		logger.debug("setting state of the SU to " + suState);
		logger.debug("current state is " + this.suState);
		try {
			SuState oldSuState = this.suState;
			this.suState = suState;
			if (oldSuState == SuState.OFF && suState == SuState.ON) {
				this.mmSap.mmSuArrived();

			} else if (oldSuState == SuState.ON && suState == SuState.OFF) {
				this.mmSap.mmSuDeparted();
			}
		} catch (Exception ex) {
			this.testHarness.fail("Unexpected exception ", ex);
		}
	}

	public SuState getState() {
		return this.suState;
	}

	public void handleMMRoamedResponseEvent(
			DeregistrationResponseEvent deregistrationResponseEvent) {
		logger.info("TestSU: gotDeregisterationResponseEvent ");
	}

	public void handleCallSetupResponseEvent(
			UnitToUnitCallControlResponseEvent ccResponseEvent) {
		try {
			int rc = ccResponseEvent.getStatusCode();
			if (rc / 100 == 2 && this.suState == SuState.ON) {
				this.gotCCSetupResponse = true;

				ccSap.ccSetupConfirm(ccResponseEvent);
				logger.debug("got a response event for "
						+ ccResponseEvent.getCallSegment());

			}
			if (logger.isDebugEnabled()) {
				logger.debug(this.getCurrentRfss().getRfssName()
						+ " Here is the state of the rfss at the caller"
						+ this.ccSap.getRfssConfig().getRFSS()
								.getTransmissionControlManager().toString());
			}
			SuToSuCallSetupScenario callSetupScenario = (SuToSuCallSetupScenario) this.scenario;
			if (callSetupScenario.getTalkSpurtSender() != null
					&& callSetupScenario.getTalkSpurtSender() == this.suConfig) {
				this.doSendTalkSpurt(callSetupScenario);

			}

		} catch (Exception ex) {
			logger.error("Unexpected exception ! ", ex);
			testHarness.fail("unexpected exception occured");
		}

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see gov.nist.p25.issi.su.SU#handeSetupIndicate(gov.nist.p25.issi.rfss.CallSetupRequestEvent)
	 */
	public void handeSetupIndicate(CallSetupRequestEvent ccRequestEvent) {

		try {
			if (this.suState == SuState.ON && this.callSegment == null) {
				this.callSegment = ccRequestEvent.getCallSegment();

			}
			/*
			 * Check if another call is in progress.
			 */
			Request request = ccRequestEvent.getServerTransaction()
					.getRequest();
			String requestCallId = ((CallIdHeader) request
					.getHeader(CallIdHeader.NAME)).getCallId();

			// If we are in a confirmed state and another call (with different
			// call Id
			// comes in, signal a busy here.
			FromHeader from = (FromHeader) ccRequestEvent
					.getServerTransaction().getRequest().getHeader(
							FromHeader.NAME);
			String radicalName = ((SipURI) from.getAddress().getURI())
					.getUser();
			if (callSegment.getDialog(radicalName).getState() == DialogState.CONFIRMED
					&& !requestCallId.equals(this.callSegment.getCallID())) {
				ccSap.ccSetupIndicateErrorResponse(ccRequestEvent,
						WarningCodes.BUSY_HERE);
				return;
			}

			ContentList clist = ContentList.getContentListFromMessage(request);

			boolean ac = clist.getCallParamContent().getCallParam()
					.isAvailabilityCheckRequested();

			int delay = ac && suConfig.getInviteProcessingDelay() == 0 ? 10
					: suConfig.getInviteProcessingDelay();

			int count = delay / 60;

			if (count == 0 && delay > 0)
				count = 1;
			if (count >= 1) {
				new ThreadedExecutor().execute(new DelayTask(ccRequestEvent));
			} else {
				this.actuallyHandleSetupIndicate(ccRequestEvent);
			}
		} catch (Exception ex) {
			logger.fatal("Unexpected exception", ex);
			System.exit(0);
		}

	}

	private void actuallyHandleSetupIndicate(
			CallSetupRequestEvent ccRequestEvent) {
		logger.info("got a call setup request");
		logger.debug("actuallyHandleSetupIndicate: callSegment = "
				+ ccRequestEvent.getCallSegment());
		try {
			if (suState == SuState.ON) {
				String callSegmentId = this.callSegment.getCallID();
				ServerTransaction st = ccRequestEvent.getServerTransaction();
				String callerCallId = ((CallIdHeader) st.getRequest()
						.getHeader(CallIdHeader.NAME)).getCallId();
				int incomingPriority = ccRequestEvent.getCallSegment()
						.getPriority();

				int currentCallPriority = this.callSegment.getPriority();

				logger.debug("TestSU: incomingPriority = " + incomingPriority
						+ " currentCallPriority = " + currentCallPriority);

				/*
				 * Check if another call is in progress.
				 */
				FromHeader from = (FromHeader) ccRequestEvent
						.getServerTransaction().getRequest().getHeader(
								FromHeader.NAME);
				String radicalName = ((SipURI) from.getAddress().getURI())
						.getUser();

				if (callSegment.getDialog(radicalName).getState() == DialogState.CONFIRMED) {
					ccSap.ccSetupIndicateErrorResponse(ccRequestEvent,
							WarningCodes.BUSY_HERE);
					return;
				}
				if ((!callSegmentId.equals(callerCallId))
						&& incomingPriority <= currentCallPriority) {
					ccSap.ccSetupIndicateErrorResponse(ccRequestEvent,
							WarningCodes.DECLINE_COLLIDING_CALL);
					return;
				}
				if (!this.suConfig.isAvailable()) {
					ccSap.ccSetupIndicateErrorResponse(ccRequestEvent,
							WarningCodes.BUSY_HERE);
					return;
				}
				if (!this.getCurrentRfss().getRFSS().isRfResourcesAvailable()) {
					ccSap.ccSetupIndicateErrorResponse(ccRequestEvent,
							WarningCodes.NO_RF_RESOURCE);
					return;
				}
				this.callSegment = ccRequestEvent.getCallSegment();
				ccSap.ccSetupIndicateResponse(ccRequestEvent, 200);
				this.okSent = true;

			}
		} catch (Exception ex) {
			logger.error("unexpected error ", ex);
			testHarness.fail("unexpected excetption ");
			throw new RuntimeException("unexpected error", ex);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see gov.nist.p25.issi.su.SU#handleTeardownResponse(gov.nist.p25.issi.rfss.CallTeardownResponseEvent)
	 */
	public void handleTeardownResponse(CallTeardownResponseEvent ccTeardownEvent) {
		this.byeOKReceived = true;
		logger.info("handle teardown response");

	}

	/**
	 * method is invoked when I get an ACK.
	 * 
	 * @param callSetupConfirmEvent --
	 *            encapsulation of the ACK event.
	 * 
	 */
	public void handleCallSetupConfirmEvent(
			CallSetupConfirmEvent callSetupConfirmEvent) {
		try {
			if (!okSent)
				testHarness
						.fail("This is not the callee! I should not see an ACK !");
			this.ackReceived = true;
			this.callSegment = callSetupConfirmEvent.getCallSegment();
			SuToSuCallSetupScenario callSetupScenario = (SuToSuCallSetupScenario) this.scenario;
			if (callSetupScenario.getTalkSpurtSender() != null
					&& callSetupScenario.getTalkSpurtSender() == this.suConfig) {
				this.doSendTalkSpurt(callSetupScenario);

			}

		} catch (Exception ex) {
			ex.printStackTrace();
			testHarness.fail("unexpected exception !");
		}

	}

	public void handleTeardownIndicate(
			CallTeardownIndicateEvent callTeardownIndicateEvent) {
		if (this.callSegment != null) {
			this.callSegment.getPointToMultipointSession().shutDown();
			this.callSegment = null;
		}

	}

	/**
	 * Send out the invite for an su to su call.
	 * 
	 */
	public void doCallSetup(SuToSuCallSetupScenario callsetupScenario) {

		try {

			logger.info("testCallSetup()");

			UnitToUnitCall callSegment = this.ccSap
					.ccSendCallSetupRequest(callsetupScenario);
			this.isCaller = true;
			if (callsetupScenario.getCancelAfter() != -1) {

				ISSITimer.getTimer().schedule(
						new SuToSuCallCancelTimerTask(callSegment),
						callsetupScenario.getCancelAfter() * 1000);
			}

			logger.debug("doCallSetup:CallSegment = " + callSegment
					+ " dialog = "
					+ callSegment.getDialog(this.suConfig.getRadicalName()));
			this.callSegment = callSegment;

		} catch (Exception ex) {
			logger.error("call setup this.rfss.getTestHarness().failed", ex);
			testHarness.fail("Call setup this.rfss.getTestHarness().failed");
		}

	}

	public void setServiceAccessPoints(ServiceAccessPoints saps) {
		this.saps = saps;
		this.ccSap = saps.getCallControlSAP();
		this.tcSap = saps.getTransmissionControlSAP();
		this.mmSap = saps.getMobilityManagementSAP();

	}

	public ServiceAccessPoints getServiceAccessPoints() {
		return saps;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see gov.nist.p25.issi.rfss.tests.TestSUInterface#getSuConfig()
	 */
	public SuConfig getSuConfig() {
		return this.suConfig;
	}

	public void setScenario(SuToSuCallSetupScenario scenario) {
		this.peerSuConfig = scenario.getCalledSuConfig();
		this.scenario = scenario;
	}

	public void setScenario(GroupCallSetupScenario scenario) {
		this.groupToCall = scenario.getCalledGroupConfig();
		this.scenario = scenario;
	}

	public void handleRegistrationEvent(RegisterEvent registerDeregister) {
		testHarness.fail("unexpected event");

	}

	public void handleMMRoamedEvent(RegisterEvent registerEvent) {
		try {
			assert registerEvent.eventType == RegisterEvent.REGISTER_DEREGISTER;
			this.gotRegisterDergister = true;

			//

			/*
			 * if (this.callSegment != null) { String radicalName =
			 * this.suConfig.getRadicalName();
			 * this.ccSap.ccTeardownCallSegment(callSegment, radicalName);
			 * this.callSegment = null; }
			 */

		} catch (Exception ex) {
			logger.error("Unexpected exception", ex);
			testHarness.fail("unexpected exception");
		}

	}

	public void handleDeRegistrationEvent(RegisterEvent event) {
		logger.info("Got a DeRegistration Event!");

	}

	public void handleDeRegistrationResponseEvent(
			DeregistrationResponseEvent registrationResponseEvent) {

	}

	public void doGroupCallSetup(GroupCallSetupScenario groupCallSetupScenario)
			throws Exception {
		logger.debug("TestSu: doGroupCallSetup()");
		this.groupServing = this.ccSap
				.triggerGroupCallSetup(groupCallSetupScenario);
		if (groupCallSetupScenario.getCancelAfter() != -1) {
			ISSITimer.getTimer().schedule(new TimerTask() {

				@Override
				public void run() {
					try {
						TestSU.this.ccSap
								.ccCancelGroupCall(TestSU.this.groupServing);
					} catch (Exception ex) {
						testHarness.fail("Unexpected exception", ex);
					}

				}

			}, groupCallSetupScenario.getCancelAfter() * 1000);
		}
		if (this.groupServing.getDialog().getState() == DialogState.CONFIRMED) {
			if (groupCallSetupScenario.isTalkSpurtSentAfterCallSetup()) {
				this.doSendTalkSpurt(groupCallSetupScenario);
			}
		}

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see gov.nist.p25.issi.rfss.tests.TestSUInterface#sendTalkSpurt()
	 */
	public void sendTalkSpurt(int nblocks) {
		try {
			if (this.suState == SuState.OFF)
				testHarness.fail("Cannot send talk spurt -- unit is off ");
			else if (this.callSegment == null) {
				testHarness.fail("No Call segment -- cannot send talk spurt");
			} else {
				logger.debug("sending talk spurt");

				this.tcSap.sendTalkSpurt(RFSS.imbeVoiceBlocks, this.suConfig
						.getSystemId(), this.suConfig.getSuId(),
						this.transmitPriorityType,
						this.transmissionPriorityLevel, nblocks);

			}
		} catch (Exception ex) {
			logger.error("unexpected exception", ex);
			testHarness.fail("Unexpected exception", ex);
		}

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see gov.nist.p25.issi.rfss.tests.TestSUInterface#sendTalkSpurt()
	 */
	public void sendTalkSpurt() {
		try {
			if (this.suState == SuState.OFF)
				testHarness.fail("Cannot send talk spurt -- unit is off ");
			else if (this.callSegment == null) {
				testHarness.fail("No Call segment -- cannot send talk spurt");
			} else {
				logger.debug("sending talk spurt");

				this.tcSap.sendTalkSpurt(RFSS.imbeVoiceBlocks, this.suConfig
						.getSystemId(), this.suConfig.getSuId(),
						this.transmitPriorityType,
						this.transmissionPriorityLevel, RFSS.imbeVoiceBlocks
								.size());

			}
		} catch (Exception ex) {
			logger.error("unexpected exception", ex);
			testHarness.fail("Unexpected exception", ex);
		}

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see gov.nist.p25.issi.rfss.tester.TestSUInterface#forceTalkSpurt(int)
	 */
	public void forceTalkSpurt(int nblocks) {

		try {
			if (this.suState == SuState.OFF)
				testHarness.fail("Cannot send talk spurt -- unit is off ");
			else if (this.callSegment == null) {
				testHarness.fail("No Call segment -- cannot send talk spurt");
			} else {
				logger.debug("sending talk spurt");

				this.tcSap.forceTalkSpurt(RFSS.imbeVoiceBlocks, this.suConfig
						.getSystemId(), this.suConfig.getSuId(),
						this.transmitPriorityType,
						this.transmissionPriorityLevel, nblocks);

			}
		} catch (Exception ex) {
			logger.error("unexpected exception", ex);
			testHarness.fail("Unexpected exception", ex);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see gov.nist.p25.issi.rfss.tester.TestSUInterface#forceTalkSpurt()
	 */
	public void forceTalkSpurt() {
		try {
			if (this.suState == SuState.OFF)
				testHarness.fail("Cannot send talk spurt -- unit is off ");
			else if (this.callSegment == null) {
				testHarness.fail("No Call segment -- cannot send talk spurt");
			} else {
				logger.debug("sending talk spurt");

				this.tcSap.forceTalkSpurt(RFSS.imbeVoiceBlocks, this.suConfig
						.getSystemId(), this.suConfig.getSuId(),
						this.transmitPriorityType,
						this.transmissionPriorityLevel, RFSS.imbeVoiceBlocks
								.size());

			}
		} catch (Exception ex) {
			logger.error("unexpected exception", ex);
			testHarness.fail("Unexpected exception", ex);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see gov.nist.p25.issi.rfss.tester.TestSUInterface#sendTalkSpurt(java.lang.String,
	 *      int)
	 */
	public void sendTalkSpurt(String groupRadicalName, int nblocks) {

		try {
			if (this.suState == SuState.OFF) {
				testHarness.fail("Cannot send talk spurt -- unit is off ");

			} else {
				logger.debug("sending talk spurt");

				this.tcSap.sendTalkSpurt(RFSS.imbeVoiceBlocks,
						groupRadicalName, this.suConfig,
						this.transmitPriorityType,
						this.transmissionPriorityLevel, nblocks);

			}
		} catch (Exception ex) {
			logger.error("Unexpected exception", ex);
			testHarness.fail("Unexpected exception", ex);
		}

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see gov.nist.p25.issi.rfss.tester.TestSUInterface#sendTalkSpurt(java.lang.String)
	 */
	public void sendTalkSpurt(String groupRadicalName) {

		try {
			if (this.suState == SuState.OFF) {
				testHarness.fail("Cannot send talk spurt -- unit is off ");

			} else {
				logger.debug("sending talk spurt");

				this.tcSap.sendTalkSpurt(RFSS.imbeVoiceBlocks,
						groupRadicalName, this.suConfig,
						this.transmitPriorityType,
						this.transmissionPriorityLevel, RFSS.imbeVoiceBlocks
								.size());

			}
		} catch (Exception ex) {
			logger.error("Unexpected exception", ex);
			testHarness.fail("Unexpected exception", ex);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see gov.nist.p25.issi.rfss.tests.TestSUInterface#terminateUnitToUnitCall()
	 */
	public void terminateUnitToUnitCall() {
		try {
			if (this.callSegment != null) {
				SuToSuCallSetupScenario scenario = (SuToSuCallSetupScenario) this.scenario;
				SuConfig suConfig = this.suConfig == scenario
						.getCallingSuConfig() ? scenario.getCalledSuConfig()
						: scenario.getCallingSuConfig();
				ccSap.ccTeardownCallSegment(this,this.callSegment, suConfig
						.getRadicalName());
				byeSent = true;
				this.callSegment = null;
			}
		} catch (Exception ex) {
			testHarness.fail("Unexpected exception tearing down call", ex);
		}
	}

	public void cancelUnitToUnitCall() {
		try {
			if (this.callSegment != null && this.isCaller) {
				ccSap.ccCancelCallSegment(this.callSegment);
				cancelSent = true;
				callSegment.setCanceled();
			} else if (this.callSegment != null) {
				this.callSegment.setCanceled();
			}
		} catch (Exception ex) {
			testHarness.fail("Unexpected exception in canceling call", ex);
		}

	}

	public void handleCancelIndicate(UnitToUnitCall callSegment) {
		try {
			ccSap.ccCancelResponse(callSegment.getServerTransaction());
		} catch (Exception ex) {
			testHarness.fail(
					"UnexpectedException in sending response to cancel ", ex);
		}

	}

	public void clearCallSegment() {
		this.callSegment = null;

	}

	/**
	 * Set the transmit priority type
	 * 
	 * @param transmitPriorityType
	 */
	public void setTransmitPriorityType(
			TransmitPriorityType transmitPriorityType) {
		this.transmitPriorityType = transmitPriorityType;

	}

	/**
	 * @param transmissionPriorityLevel
	 *            the transmissionPriorityLevel to set
	 */
	public void setTransmissionPriorityLevel(int transmissionPriorityLevel) {
		this.transmissionPriorityLevel = transmissionPriorityLevel;
	}

	/**
	 * Force a talk spur to the group.
	 * 
	 * @param radicalName
	 */
	public void forceTalkSpurt(String groupRadicalName, int nblocks) {
		try {
			if (this.suState == SuState.OFF) {
				testHarness.fail("Cannot send talk spurt -- unit is off ");

			} else {
				logger.debug("sending talk spurt");

				this.tcSap.forceTalkSpurt(RFSS.imbeVoiceBlocks,
						groupRadicalName, this.suConfig,
						this.transmitPriorityType,
						this.transmissionPriorityLevel, nblocks);

			}
		} catch (Exception ex) {
			logger.error("Unexpected exception", ex);
			testHarness.fail("Unexpected exception", ex);
		}

	}

	/**
	 * Force the entire talk spurt ( the whole file ) to the group.
	 * 
	 * @param groupRadicalNamem --
	 *            radical name.
	 */
	public void forceTalkSpurt(String groupRadicalName) {
		try {
			if (this.suState == SuState.OFF) {
				testHarness.fail("Cannot send talk spurt -- unit is off ");

			} else {
				logger.debug("sending talk spurt");

				this.tcSap.forceTalkSpurt(RFSS.imbeVoiceBlocks,
						groupRadicalName, this.suConfig,
						this.transmitPriorityType,
						this.transmissionPriorityLevel, RFSS.imbeVoiceBlocks
								.size());

			}
		} catch (Exception ex) {
			logger.error("Unexpected exception", ex);
			testHarness.fail("Unexpected exception", ex);
		}
	}

	@Override
	public UnitToUnitCall getCurrentUnitToUnitCall() {
		return this.callSegment;
	}

}
