package gov.nist.p25.issi.rfss.tester;

import java.io.IOException;
import java.util.Collection;
import java.util.HashSet;
import java.util.Hashtable;

import javax.xml.parsers.SAXParserFactory;

import gov.nist.p25.issi.issiconfig.GroupConfig;
import gov.nist.p25.issi.issiconfig.RfssConfig;
import gov.nist.p25.issi.issiconfig.SuConfig;
import gov.nist.p25.issi.issiconfig.SystemConfig;
import gov.nist.p25.issi.issiconfig.TopologyConfig;
import gov.nist.p25.issi.issiconfig.XMLTagsAndAttributes;
import gov.nist.p25.issi.rfss.ISSITesterConstants;

import org.apache.log4j.ConsoleAppender;
import org.apache.log4j.Logger;
import org.apache.log4j.SimpleLayout;
import org.apache.log4j.helpers.*;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.DefaultHandler;

/**
 * This is a the Java class that represents a test script. A test script
 * consists of several Scenarios that are fired at particular times during the
 * execution.
 * 
 * @author M. Ranganathan
 * 
 */
public class TestScript {
	
	private String description;
	
	
	int testCompletionDelay ;
	
	int traceGenerationTime;

		
	HashSet<PostCondition> postConditions;

	Hashtable<String, AbstractScenario> scenarios;

	private TopologyConfig topologyConfig;
	
	private String testName ;

	private String testDirectory;

	private boolean isInteractive;

	// Set this flag to capture trace.
	private boolean isTraceCaptureEnabled;

    
	private boolean evaluatePassFail;

	/**
	 * Where XML formatted rtp messages go.
	 */
	public static String rtpMessagelog = "logs/messagelog.ptt";

	/**
	 * Where RTP debug messages go
	 */
	
	public static String rtpDebuglog = "logs/rtpdebuglog.txt";

	/**
	 * Where SIP Debug messages go.
	 */
	public static String sipDebuglog = ISSITesterConstants.DEBUG_LOG;
	

	private static Logger logger = Logger.getLogger(TestScript.class);
	
	/*
	 * Whether or not to generate the SIP trace.
	 */

	public static boolean  generateSipTrace;

	/*
	 * Wheter or not a meaningful ptt trace is captured.
	 */
	public static boolean generatePttTrace;

	public TestScript() {
		this.scenarios = new Hashtable<String, AbstractScenario>();
		this.postConditions = new HashSet<PostCondition>();
		
		this.testCompletionDelay = 0;
		this.traceGenerationTime = 25000;

	}
	
	void setTopologyConfig (TopologyConfig topologyConfig) {
		this.topologyConfig = topologyConfig;
	}
	
	

	public AbstractScenario getScenario(String scenarioName) {

		return this.scenarios.get(scenarioName);
	}

	public Collection<AbstractScenario> getScenarios() {
		return this.scenarios.values();
	}

	public void printScenarios() {
		for (AbstractScenario sc : this.scenarios.values()) {
			logger.info("id = " + sc.getId() + " method = "
					+ sc.getMethodName() + " time = " + sc.getTrigger());
		}
	}

	public TopologyConfig getTopologyConfig() {
		return topologyConfig;
	}

	public String getReferenceMessagesDirName() {
		String retval = 
		this.getTopologyConfig().getTopologyFile().substring(0, this.topologyConfig.getTopologyFile().indexOf(".xml"))
		+ ".testmessages";
		String dirName = retval.replaceFirst(ISSITesterConstants.getScenarioDir(), 
				ISSITesterConstants.DEFAULT_TRACES_DIR + "/" + ISSITesterConstants.getTestSuite());
		logger.debug("referenceMessagesDirName = " + dirName);
		return dirName;
	}
	
	
	public void addPostCondition(PostCondition postCondition) {
		logger.info("AddPostCondition");
		this.postConditions.add(postCondition);
	}

	public HashSet<PostCondition> getPostConditions() {
		return this.postConditions;
	}

	/**
	 * A wordy description for the test.
	 * 
	 * @param description
	 */
	public void setDescription(String description) {
		this.description = description;
	}
	
	public String getDescription() {
		return this.description;
	}

	/**
	 * An optional descriptive name for the test.
	 * 
	 * @param testName
	 */
	public void setTestName(String testName) {
		this.testName = testName;
		
	}
	
	public String getTestName() {
		return this.testName;
	}

	/**
	 * The test directory is the subidrectory name where the tests
	 * and the captured traces reside.
	 * 
	 * @param dirName
	 */
	public void setTestDirectory(String dirName) {
		this.testDirectory = dirName;
		
	}
	
	public String getTestDirectory() {
		return this.testDirectory;
	}

	/**
	 * The interactive flag determines whether the test is run with prompting.
	 * 
	 * @return -- the interactive flag.
	 */
	public boolean isInteractive() {
		return this.isInteractive;
	}
	
	public void setInteractive(boolean interactive) {
		this.isInteractive = interactive;
	}
	
	/**
	 * @param isTraceCaptureEnabled The isTraceCaptureEnabled to set.
	 */
	public void setTraceCaptureEnabled(boolean isTraceCaptureEnabled) {
		this.isTraceCaptureEnabled = isTraceCaptureEnabled;
	}

	/**
	 * @return Returns the isTraceCaptureEnabled.
	 */
	public boolean isTraceCaptureEnabled() {
		return isTraceCaptureEnabled;
	}
	
	/**
	 * Evaluate predicate at the end of the test?
	 * 
	 * @param evaluatePassFail
	 */
	public void setEvaluatePassFail(boolean evaluatePassFail) {
		this.evaluatePassFail = evaluatePassFail;
		
	}
	
	public boolean isEvaluatePassFail() {
		return this.evaluatePassFail;
	}
	

	public int getTestCompletionDelay() {
		return testCompletionDelay;
	}
	
	public int getTraceGenerationTime() {
		return this.traceGenerationTime;
	}
	
	// /////////////////////////////////////////////////////////////////////
	// Static methods 
	// /////////////////////////////////////////////////////////////////////
	
	public static String getSipDebugLog() {
		return sipDebuglog;
	}

	public static String getRtpMessageLog() {
		return rtpMessagelog;
	}

	public static String getRtpDebugLog() {
		return rtpDebuglog;
	}

	

	

	

	
	

}
