/**
 * The client side of the application that prompts the user to do the next test action.
 * This implements a barrier synchronization with prompting so that each test action on
 * the server will wait for the user to prompt it to go to the next action.
 * 
 * @author mranga
 * 
 */
package gov.nist.p25.issi.testlauncher;

import gov.nist.p25.issi.issiconfig.SuConfig;
import gov.nist.p25.issi.rfss.ISSITesterConstants;
import gov.nist.p25.issi.rfss.ProtocolObjects;
import gov.nist.p25.issi.rfss.tester.AbstractScenario;
import gov.nist.p25.issi.rfss.tester.EmulatorWebServerAddress;
import gov.nist.p25.issi.rfss.tester.TestScript;

import java.util.Comparator;
import java.util.Iterator;
import java.util.PriorityQueue;
import java.util.TimerTask;
import java.util.TreeSet;
import java.util.concurrent.atomic.AtomicInteger;

import javax.swing.JOptionPane;

import org.apache.commons.httpclient.NameValuePair;
import org.apache.log4j.Logger;

/**
 * Put all the test actions in a priority queue and prompt until the priority
 * queue is empty. This is for interactive testing. The priority queue is
 * ordered by time. That is the head of the queue is all the ready tasks.
 * 
 * @author M. Ranganathan
 * 
 */
public class ScenarioPrompter {

	private DietsGUI dietsGUI;

	private PriorityQueue<AbstractScenario> priorityQueue;

	private TestScript testScript;

	private TestControllerInterface remoteTestController;

	private AtomicInteger counter;

	private static Logger logger = Logger.getLogger(ScenarioPrompter.class);

	ProgressMonitor progressMonitor;

	class MyPrompterTask implements Runnable {

		private AbstractScenario scenario;

		MyPrompterTask(AbstractScenario scenario) {
			this.scenario = scenario;
		}

		public void run() {
			logger.debug("Waiting..." + scenario);
			this.scenario.doWait();

			logger.debug("Prompting for next action.");
			int answer = JOptionPane
					.showConfirmDialog(dietsGUI,
							this.scenario.getDescription(), "Do this action",
							JOptionPane.YES_OPTION);
			try {
				if (answer == JOptionPane.YES_OPTION) {
					remoteTestController.signalNextScenario(scenario.getId());
				} else {
					remoteTestController.tearDownCurrentTest();
					dietsGUI.getTestExecutionPanel().enableAllButtons();
				}
			} catch (Exception e) {
				e.printStackTrace();
				logger.error("Error communicating with the remote tester", e);
				dietsGUI.getTestExecutionPanel().logError(
						"Error in communicating with the remote tester");
				priorityQueue.clear();
				return;

			}
			// The counter insures that the start prompting call is
			// implemented only once.
			if (ScenarioPrompter.this.counter.decrementAndGet() == 0) {
				startPrompting();
			}

		}

	}

	/**
	 * A class that waits for test execution to complete. This is basically an
	 * open loop timer. It is used in non-interactive execution.
	 * 
	 */
	class TestCompletionTimerTask extends TimerTask {

		int secondsRemaining;

		DietsGUI gui;

		boolean completed;

		private int completionSeconds;
		
		private boolean flag;

		public TestCompletionTimerTask(DietsGUI gui, int seconds,
				int completionSeconds) {
			gui.getStatusProgressBar().setIndeterminate(true);
			this.gui = gui;
			gui.getTestExecutionPanel().setProgressMonitor(new ProgressMonitor(gui));
			this.secondsRemaining = seconds;
			this.completionSeconds = completionSeconds;

		}

		public void run() {
			try {
				secondsRemaining = secondsRemaining - 1000;
				gui.getTestExecutionPanel().setProgressScreenMessage(
						"Time remaining: "
								+ (secondsRemaining / 1000) + " seconds");
				if (!completed)
					completed = remoteTestController.isTestCompleted();

				if (secondsRemaining <= 0) {
					if (DietsGUI.evaluatePassFail) {

						boolean pass = remoteTestController.getTestResults();

						JOptionPane
								.showMessageDialog(
										null,
										pass ? ISSITesterConstants.PASSED
												: ISSITesterConstants.FAILED
														+ "\nSee Error Diagnostic tab for more info");
						if (!pass) {
							String errorLog = remoteTestController
									.getErrorLog();
							gui.getTestExecutionPanel().logError(errorLog);
						}
					} else {

					}
					this.gui.getTestExecutionPanel().setProgressScreenMessage(
							"Getting SIP and PTT traces");
					this.gui.getTestExecutionPanel().getSipPttTraces(true);

					
					this.gui.getTestExecutionPanel().enableAllButtons();
					
					this.cancel();
					
				} else if (completed) {
					if ( ! flag ) secondsRemaining = completionSeconds;
					flag = true;
				}
			} catch (Exception ex) {
				this.cancel();
			}

		}

	}

	/**
	 * Signal that the test is complete and that the server should evaluate the
	 * predicates.
	 * 
	 */
	public void signalTestCompletion() throws Exception {
		this.remoteTestController.signalTestCompletion();

		TestExecutionPanel.timer.schedule(new TestCompletionTimerTask(
				this.dietsGUI, ISSITesterConstants.TEST_RUNS_FOR + 5000, testScript
						.getTestCompletionDelay() * 1000),0, 1000);

		
	}

	public ScenarioPrompter(DietsGUI dietsGui,
			TestControllerInterface remoteTestController, TestScript testScript) throws Exception {
		this.testScript = testScript;
		this.dietsGUI = dietsGui;
		this.priorityQueue = new PriorityQueue<AbstractScenario>();
		for (SuConfig suConfig : testScript.getTopologyConfig()
				.getSuConfigurations()) {
			int answer = JOptionPane.showConfirmDialog(dietsGUI,
					"Turn on the SU : %x"
							+ Integer.toHexString(suConfig.getSuId()) + "\n",
					"Do this action", JOptionPane.YES_OPTION);
			if (answer == JOptionPane.NO_OPTION) {
				if ( remoteTestController != null ) remoteTestController.tearDownCurrentTest();
				
				return;
			}

		}

		this.remoteTestController = remoteTestController;
		this.counter = new AtomicInteger(0);

		if (testScript.getScenarios().size() != 0) {
			for (AbstractScenario scenario : testScript.getScenarios()) {
				if (scenario.isTimeTriggered()) {
					priorityQueue.offer(scenario);
					new Thread(new MyPrompterTask(scenario)).start();
				}

			}
		} else {

		}

	}

	private boolean evaluatePostCondition() throws Exception {
		// No more events to schedule.
		logger
				.debug("Test is completed -- signal the server to complete the test ");
		if (DietsGUI.evaluatePassFail) {
			int answer = JOptionPane
					.showConfirmDialog(
							null,
							"Test setup is completed. Complete and evaluate PASS/FAIL?",
							"Test Completion Dialog", JOptionPane.YES_OPTION);
			if (answer == JOptionPane.YES_OPTION) {
				this.dietsGUI.getTestExecutionPanel().setProgressMonitor(
						new ProgressMonitor(this.dietsGUI));
				this.dietsGUI.getTestExecutionPanel().setProgressScreenMessage(
						"Completing test and evaluating predicate");
				this.signalTestCompletion();

			} else {
				if ( remoteTestController != null ) remoteTestController.tearDownCurrentTest();
				dietsGUI.getTestExecutionPanel().enableAllButtons();
			}
		} else {
			// Just using DIETS as a call generator.
			this.dietsGUI.getTestExecutionPanel().setProgressMonitor(
					new ProgressMonitor(this.dietsGUI));
			this.dietsGUI.getTestExecutionPanel().setProgressScreenMessage(
					"Completing test...");
			this.signalTestCompletion();
			

		}
		return false;
	}

	public boolean startPrompting() {
		try {
			if (this.priorityQueue.isEmpty()) {
				return this.evaluatePostCondition();
			}

			if (this.priorityQueue.isEmpty()) {
				logger.debug("scenario queue is empty -- returning");
				return false;
			}
			AbstractScenario scenario = this.priorityQueue.poll();
			logger.info("notifying : " + scenario);
			this.counter.incrementAndGet();
			logger.info("Time = "
					+ scenario.getTriggerAsTimeTrigger().getTime());
			scenario.doNotify();
			int time = scenario.getTriggerAsTimeTrigger().getTime();
			while (!this.priorityQueue.isEmpty()
					&& this.priorityQueue.peek().getTriggerAsTimeTrigger()
							.getTime() == time) {
				scenario = this.priorityQueue.poll();
				logger.info("notifying : " + scenario);
				this.counter.incrementAndGet();
				logger.info("Time = "
						+ scenario.getTriggerAsTimeTrigger().getTime());
				scenario.doNotify();
				break;

			}

			return !this.priorityQueue.isEmpty();
		} catch (Exception ex) {
			ex.printStackTrace();
			return false;
		}
	}

}
