/* The software was developed by employees of the National Institute of
 * Standards and Technology (NIST), an agency of the Federal Government.
 * Pursuant to title 15 United States Code Section 105, works of NIST
 * employees are not subject to copyright protection in the United States
 * and are considered to be in the public domain.  As a result, a formal
 * license is not needed to use the software.
 * 
 * The ISSI Emulator software is provided by NIST as a service and is expressly
 * provided "AS IS".  NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED
 * OR STATUTORY, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT
 * AND DATA ACCURACY.  NIST does not warrant or make any representations
 * regarding the use of the software or the results thereof including, but
 * not limited to, the correctness, accuracy, reliability or usefulness of
 * the software.
 * 
 * Permission to use this software is contingent upon your acceptance
 * of the terms of this agreement and upon your providing appropriate
 * acknowledgments of NIST's ownership of the software.
 */
package gov.nist.p25.issi.traceviewer;

import gov.nist.p25.issi.issiconfig.GroupConfig;
import gov.nist.p25.issi.issiconfig.SuConfig;

import java.util.*;

import javax.swing.*;

import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.awt.print.*;
import java.awt.image.*;
import java.io.*;
import javax.imageio.*;

import org.apache.log4j.Logger;

/**
 * This class implements 2D graphics for the trace viewer sequence diagram.
 * 
 * @author steveq@nist.gov
 * @version $Revision: 1.5 $, $Date: 2007/10/05 18:53:05 $
 * @since 1.5
 */
public class DiagramPanel extends JPanel implements MouseListener,
		MouseMotionListener, ActionListener {

	private static Logger logger = Logger.getLogger("gov.nist.p25.issi");

	// /////////////////////////////////////////////////////////////////////
	// Constants
	// /////////////////////////////////////////////////////////////////////

	private static final long serialVersionUID = 0;

	private static final int TOP_PAD = 50;

	private static final int NODE_SEPARATOR_PAD = 60;

	private static final int NODE_WIDTH = 130;

	private static final int NODE_HEIGHT = 60;

	private static final int FONT_PT_SIZE = 11;

	private static final int MESSAGE_SEPARATOR_PAD = 40;

	private static final int MESSAGE_LABEL_SEPARATOR_PAD = 10;

	private static final int MESSAGE_SELECT_HEIGHT = 20;

	private static final int ARROW_WIDTH = 10;

	private static final int ARROW_TOP_HEIGHT = 5;

	private static final int ARROW_BOTTOM_HEIGHT = 5;

	private static final double ARC_HEIGHT = 60;

	private static final double ARC_WIDTH = 60;

	private static final int RFSS_Y_PAD = 10;

	private static final int RFSS_X_PAD = 10;

	private final static BasicStroke stroke = new BasicStroke(1.0f);

	private final static int NUM_PRINTED_MESSAGES_PER_PAGE = 20;

	// /////////////////////////////////////////////////////////////////////
	// Variables
	// /////////////////////////////////////////////////////////////////////

	private TracePanel tracePanel = null;

	private JPopupMenu rightClickPopupMenu;

	private ZoomLevel currentZoomLevel = ZoomLevel.ZOOM_100;

	private ZoomLevel printingZoomLevel = ZoomLevel.ZOOM_50;

	private int mousePressedXPos = 0;

	private int mousePressedYPos = 0;

	private int numPrintPages = 0;

	private JMenuItem saveAsClickMenuItem;

	static private final String newline = System.getProperty("line.separator");

	// /////////////////////////////////////////////////////////////////////
	// Constructor
	// /////////////////////////////////////////////////////////////////////

	DiagramPanel(TracePanel tracePanel) {

		this.tracePanel = tracePanel;
		addMouseListener(this);
		addMouseMotionListener(this);

		rightClickPopupMenu = new JPopupMenu();

		JMenuItem newWindowClickMenuItem = new JMenuItem("New Window");
		newWindowClickMenuItem.setActionCommand("New Window");
		newWindowClickMenuItem.addActionListener(this);

		this.saveAsClickMenuItem = new JMenuItem("Save As");
		saveAsClickMenuItem.setActionCommand("Save As");
		saveAsClickMenuItem.addActionListener(this);

		JMenuItem closeRightClickMenuItem = new JMenuItem("Close");
		closeRightClickMenuItem.setActionCommand("Close");
		closeRightClickMenuItem.addActionListener(this);

		rightClickPopupMenu.add(newWindowClickMenuItem);
		rightClickPopupMenu.addSeparator();

		rightClickPopupMenu.add(saveAsClickMenuItem);
		rightClickPopupMenu.addSeparator();
		rightClickPopupMenu.add(closeRightClickMenuItem);

		this.setBackground(Color.white);

	}

	// /////////////////////////////////////////////////////////////////////
	// Methods
	// /////////////////////////////////////////////////////////////////////

	public void setData(Collection<RfssData> rfssList,
			ArrayList<MessageData> messageList) {

		double zoomMax = ZoomLevel.ZOOM_MAX.value();

		int rfssSize = rfssList.size();
		int totalWidth = (int) (rfssSize
				* ((NODE_SEPARATOR_PAD * zoomMax) + (NODE_WIDTH * zoomMax)) + (NODE_SEPARATOR_PAD * zoomMax));

		int messageListSize = messageList.size();
		int totalHeight = (int) (messageListSize
				* (MESSAGE_SEPARATOR_PAD * zoomMax) + MESSAGE_SEPARATOR_PAD);
		this.setPreferredSize(new Dimension(totalWidth, totalHeight));
		this.revalidate();

		// Set number of print pages
		int numMessages = messageList.size();
		numPrintPages = (int) (numMessages / NUM_PRINTED_MESSAGES_PER_PAGE);

	}

	private void disableMessageSelected() {

		ArrayList messageList = tracePanel.getMessageList();

		for (int i = 0; i < messageList.size(); i++) {

			MessageData messageData = (MessageData) messageList.get(i);

			if (messageData.isSelected()) {

				messageData.setSelected(false);
				break;
			}

		}
	}

	private void disableRfssSelected() {

		Collection<RfssData> rfssList = tracePanel.getRfssList();

		for (RfssData rfssData : rfssList) {
			if (rfssData.isSelected()) {
				rfssData.setSelected(false);
				break;
			}
		}

	}

	private RfssData getRfssById(String seekPort) {

		// logger.debug("getRfssByPort " + seekPort);
		RfssData foundRfss = null;
		Collection<RfssData> rfssList = tracePanel.getRfssList();

		for (RfssData rfssData : rfssList) {

			if ((rfssData.getRfssConfig().getDomainName()).equals(seekPort)) {

				foundRfss = rfssData;

			}

			if (foundRfss != null)
				break;

		}

		return foundRfss;

	}

	private RfssData getRfssByPort(String seekPort) {
		logger.debug("getRfssByPort " + seekPort);
		RfssData foundRfss = null;
		Collection<RfssData> rfssList = tracePanel.getRfssList();

		for (RfssData rfssData : rfssList) {

			for (int j = 0; j < rfssData.getPorts().size(); j++) {

				String port = rfssData.getPorts().get(j);

				if (port.equals(seekPort)) {

					foundRfss = rfssData;
					break;

				}

			}

			if (foundRfss != null)
				break;

		}
		if (foundRfss == null) {
			logger.debug("Here is the list of rfss " + rfssList);
			for (RfssData rfssData : rfssList) {
				logger.debug("ports = " + rfssData.getPorts());
			}
		}

		return foundRfss;

	}

	private RfssData getRfssByLocation(int mouseX, int mouseY) {

		RfssData foundRfss = null;
		Collection<RfssData> rfssList = tracePanel.getRfssList();

		for (RfssData rfssData : rfssList) {

			if (((rfssData.getX() <= mouseX) && (mouseX <= (rfssData.getX() + rfssData
					.getDimension().width)))
					&& ((rfssData.getY() <= mouseY) && (mouseY <= (rfssData
							.getY() + rfssData.getDimension().height)))) {

				foundRfss = rfssData;
				break;

			}

		}

		return foundRfss;

	}

	private MessageData getMessageByLocation(int mouseX, int mouseY) {

		MessageData foundMessage = null;
		ArrayList messageList = tracePanel.getMessageList();

		for (int i = 0; i < messageList.size(); i++) {

			MessageData messageData = (MessageData) messageList.get(i);

			if (((messageData.getX() <= mouseX) && (mouseX <= (messageData
					.getX() + messageData.getDimension().width)))
					&& ((messageData.getY() <= mouseY) && (mouseY <= (messageData
							.getY() + messageData.getDimension().height)))) {

				foundMessage = messageData;
				break;

			}

		}

		return foundMessage;

	}

	protected void paintComponent(Graphics g) {

		super.paintComponent(g);

		Graphics2D g2d = (Graphics2D) g;
		double zoomValue = currentZoomLevel.value();
		int newFontSize = (int) (FONT_PT_SIZE * zoomValue);
		g2d.setFont(new Font("Tahoma", Font.PLAIN, newFontSize));
		FontMetrics metrics = g2d.getFontMetrics();

		drawNodes(g2d, zoomValue, metrics);
		drawMessages(g2d, zoomValue, metrics);

	}

	public void saveAs(String fileName) {
		logger.debug("path for save: " + fileName + ".jpg");

		try {
			/*
			 * Compute the size of the image first
			 */
			int rfssListSize = tracePanel.getRfssList().size();
			int messageListSize = tracePanel.getMessageList().size();

			int width = NODE_SEPARATOR_PAD + (rfssListSize * NODE_WIDTH)
					+ (rfssListSize * NODE_SEPARATOR_PAD);

			int height = TOP_PAD + NODE_HEIGHT + MESSAGE_SEPARATOR_PAD
					+ (messageListSize * MESSAGE_SEPARATOR_PAD)
					+ MESSAGE_SEPARATOR_PAD;

			BufferedImage bufferedImage = new BufferedImage(width, height,
					BufferedImage.TYPE_INT_RGB);
			Graphics2D g2d = bufferedImage.createGraphics();
			double zoomValue = currentZoomLevel.value();
			int newFontSize = (int) (FONT_PT_SIZE * zoomValue);
			g2d.setFont(new Font("Tahoma", Font.PLAIN, newFontSize));
			FontMetrics metrics = g2d.getFontMetrics();

			g2d.setColor(Color.white); // Color image background white first
			g2d.fillRect(0, 0, width, height);

			drawNodes(g2d, zoomValue, metrics);
			drawMessages(g2d, zoomValue, metrics);

			OutputStream out = new FileOutputStream(fileName);
			ImageIO.write(bufferedImage, "jpg", out);
			out.close();
		} catch (Exception e) {
			logger.error("An exception occured while trying to save file", e);
			JOptionPane.showMessageDialog(null,
					"An Error Occured while saving the file",
					"Save File Error", JOptionPane.ERROR_MESSAGE);
		}
	}

	private void saveAs() {

		JFileChooser fileChooser = new JFileChooser();
		fileChooser.setDialogTitle("Save As");
		TraceFileFilter traceFilter = new TraceFileFilter();
		traceFilter.addExtension("jpg");
		traceFilter.setDescription("JPG Files");
		fileChooser.setFileFilter(traceFilter);

		int returnVal = fileChooser.showSaveDialog(tracePanel.getTabbedPane());

		File file = null;

		if (returnVal == JFileChooser.APPROVE_OPTION) {
			file = fileChooser.getSelectedFile();
			// this is where a real application would save the file.
			logger.debug("Saving: " + file.getName() + "." + newline);
		} else {
			logger.debug("Save command cancelled by user." + newline);
			return;
		}

		this.saveAs(file.getAbsolutePath() + ".jpg");

	}

	private void drawNodes(Graphics2D g2d, double zoomValue, FontMetrics metrics) {

		Collection<RfssData> rfssList = tracePanel.getRfssList();

		g2d.setColor(Color.black);

		int i = 0;
		for (RfssData rfssData : rfssList) {

			String idString = rfssData.getRfssConfig() != null ? rfssData
					.getRfssConfig().getDomainName()  : 
						rfssData.getId();
			Color color = rfssData.getRfssConfig() != null ? rfssData.getRfssConfig().getColor() :
						Color.black;
			String nameString = rfssData.getRfssConfig() != null ? rfssData
					.getRfssConfig().getRfssName()  : 
						null;
			// String idString = rfssData.getId();
			int idStringWidth = metrics.stringWidth(idString);
			int idStringHeight = metrics.getHeight();
			
			int nameStringWidth = metrics.stringWidth(nameString);

			String portNumberString = "";

			for (int j = 0; j < rfssData.getPorts().size(); j++) {

				portNumberString += rfssData.getPorts().get(j);
				if (j < (rfssData.getPorts().size() - 1)) {
					portNumberString += ",";
				}
			}

			int x = (int) ((NODE_SEPARATOR_PAD * zoomValue)
					+ (i * NODE_WIDTH * zoomValue) + (i * NODE_SEPARATOR_PAD * zoomValue));
			int y = (int) (TOP_PAD * zoomValue);

			g2d.setPaint(new Color(255, 235, 205));
			g2d.fill(new Rectangle2D.Double(x, y, NODE_WIDTH * zoomValue,
					NODE_HEIGHT * zoomValue));

			if (rfssData.isSelected())
				g2d.setPaint(Color.red);
			else
				g2d.setPaint(color);

			g2d.setStroke(stroke);
			g2d.draw(new Rectangle2D.Double(x, y, NODE_WIDTH * zoomValue,
					NODE_HEIGHT * zoomValue));

			
			g2d
			.drawString(
					idString,
					x
							+ ((int) (NODE_WIDTH * zoomValue) / 2 - idStringWidth / 2),
					y
							+ ((int) (NODE_HEIGHT * zoomValue) / 2 - idStringHeight * 1));
			g2d
			.drawString(
					nameString,
					x
							+ ((int) (NODE_WIDTH * zoomValue) / 2 - nameStringWidth / 2),
					y
							+ ((int) (NODE_HEIGHT * zoomValue) / 2 ));
	

			/*
			 * g2d .drawString( ipAddrString, x + ((int) (NODE_WIDTH *
			 * zoomValue) / 2 - ipAddrStringWidth / 2), y + ((int) (NODE_HEIGHT *
			 * zoomValue) / 2));
			 * 
			 * 
			 * g2d .drawString( portNumberString, x + ((int) (NODE_WIDTH *
			 * zoomValue) / 2 - portNumberStringWidth / 2), y + ((int)
			 * (NODE_HEIGHT * zoomValue) / 2 + portNumberStringHeight * 1));
			 */

			ArrayList messageList = tracePanel.getMessageList();
			int messageListSize = messageList.size();
			g2d.drawLine(x + (int) ((NODE_WIDTH * zoomValue) / 2),
					(int) (y + (NODE_HEIGHT * zoomValue)), x
							+ (int) ((NODE_WIDTH * zoomValue) / 2),
					(int) ((y + (NODE_HEIGHT * zoomValue))
							+ MESSAGE_SEPARATOR_PAD + (MESSAGE_SEPARATOR_PAD
							* messageListSize * zoomValue)));

			rfssData.setX(x);
			rfssData.setY(y);
			rfssData.setDimension(new Dimension((int) (NODE_WIDTH * zoomValue),
					(int) (NODE_HEIGHT * zoomValue)));
			i++;

		}
	}

	private void printHeadersAndFooters(Graphics2D g2d, FontMetrics metrics,
			int page) {

		// Set footer
		Calendar cal = Calendar.getInstance(TimeZone.getDefault());

		String DATE_FORMAT = "HH:mm:ss MM/dd/yyyy";
		java.text.SimpleDateFormat sdf = new java.text.SimpleDateFormat(
				DATE_FORMAT);

		sdf.setTimeZone(TimeZone.getDefault());

		String header = "DIETS v0.1a: " + tracePanel.getFileName() + "    "
				+ sdf.format(cal.getTime()) + "    " + "Page " + (page + 1)
				+ " of " + (numPrintPages + 1);

		g2d.setStroke(stroke);

		g2d.drawString(header, 0, metrics.getHeight());

	}

	private void printNodes(Graphics2D g2d, double zoomValue,
			FontMetrics metrics, int page) {

		Collection<RfssData> rfssList = tracePanel.getRfssList();
		int rfssListSize = rfssList.size();

		// g2d.setColor(Color.black);

		int i = 0;
		for (RfssData rfssData : rfssList) {

			int x = (int) ((NODE_SEPARATOR_PAD * zoomValue)
					+ (i * NODE_WIDTH * zoomValue) + (i * NODE_SEPARATOR_PAD * zoomValue));
			int y = (int) (TOP_PAD * zoomValue);

			// Only print RFSS nodes on first page

			String idString = rfssData.getRfssConfig() != null ? rfssData
					.getRfssConfig().getDomainName(): 
						rfssData.getId();
					
			Color color = rfssData.getRfssConfig() != null ? rfssData.getRfssConfig().getColor() :
				Color.black;

			// String idString = rfssData.getId();
			int idStringWidth = metrics.stringWidth(idString);
			int idStringHeight = metrics.getHeight();
			String ipAddrString = rfssData.getAddress() + ":";
			int ipAddrStringWidth = metrics.stringWidth(ipAddrString);

			String portNumberString = "";

			for (int j = 0; j < rfssData.getPorts().size(); j++) {
				portNumberString += rfssData.getPorts().get(j);
				if (j < (rfssData.getPorts().size() - 1)) {
					portNumberString += ", ";
				}
			}

			int portNumberStringWidth = metrics.stringWidth(portNumberString);
			int portNumberStringHeight = metrics.getHeight();

			g2d.setPaint(new Color(255, 235, 205));
			g2d.fill(new Rectangle2D.Double(x, y, NODE_WIDTH * zoomValue,
					NODE_HEIGHT * zoomValue));

			if (rfssData.isSelected())
				g2d.setPaint(Color.red);
			else
				g2d.setPaint(color);

			g2d.setStroke(stroke);
			g2d.draw(new Rectangle2D.Double(x, y, NODE_WIDTH * zoomValue,
					NODE_HEIGHT * zoomValue));

			g2d
					.drawString(
							idString,
							x
									+ ((int) (NODE_WIDTH * zoomValue) / 2 - idStringWidth / 2),
							y
									+ ((int) (NODE_HEIGHT * zoomValue) / 2 - idStringHeight * 1));

			g2d
					.drawString(
							ipAddrString,
							x
									+ ((int) (NODE_WIDTH * zoomValue) / 2 - ipAddrStringWidth / 2),
							y + ((int) (NODE_HEIGHT * zoomValue) / 2));

			g2d
					.drawString(
							portNumberString,
							x
									+ ((int) (NODE_WIDTH * zoomValue) / 2 - portNumberStringWidth / 2),
							y
									+ ((int) (NODE_HEIGHT * zoomValue) / 2 + portNumberStringHeight * 1));

			rfssData.setX(x);
			rfssData.setY(y);
			rfssData.setDimension(new Dimension((int) (NODE_WIDTH * zoomValue),
					(int) (NODE_HEIGHT * zoomValue)));

			g2d.drawLine(x + (int) ((NODE_WIDTH * zoomValue) / 2),
					(int) (y + (NODE_HEIGHT * zoomValue)), x
							+ (int) ((NODE_WIDTH * zoomValue) / 2),
					(int) ((y + (NODE_HEIGHT * zoomValue))
							+ MESSAGE_SEPARATOR_PAD + (MESSAGE_SEPARATOR_PAD
							* NUM_PRINTED_MESSAGES_PER_PAGE * zoomValue)));
			i++;

		}
	}

	private void drawMessages(Graphics2D g2d, double zoomValue,
			FontMetrics metrics) {

		ArrayList messageList = tracePanel.getMessageList();
		int messageListSize = messageList.size();
		int startX = 0;
		int startY = (int) ((TOP_PAD * zoomValue) + (NODE_HEIGHT * zoomValue));
		int endX = 0;
		int endY = startY;

		for (int i = 0; i < messageListSize; i++) {

			MessageData messageData = (MessageData) messageList.get(i);

			RfssData fromRfss = messageData.getFromRfssId() != null ? getRfssById(messageData
					.getFromRfssId())
					: getRfssByPort(messageData.getFromPort());
			if (fromRfss == null) {
				logger.error("Could not find an RFSS (fromRfss) "
						+ messageData.getFromRfssId() + ":"
						+ messageData.getFromPort());
				return;
			}
			RfssData toRfss = messageData.getToRfssId() != null ? getRfssById(messageData
					.getToRfssId())
					: getRfssByPort(messageData.getToPort());
			if (toRfss == null) {
				logger.error("Could not find an RFSS "
						+ messageData.getFromRfssId() + ":"
						+ messageData.getToPort());
				return;
			}
			startX = fromRfss.getX() + (fromRfss.getDimension().width / 2);
			startY += (MESSAGE_SEPARATOR_PAD * zoomValue);
			endX = toRfss.getX() + (toRfss.getDimension().width / 2);
			endY = startY;

			// Set message color
			Color color = null;

			if (messageData.isSelected()) {

				color = Color.red;

			} else {

				Hashtable colorMap = (Hashtable) tracePanel.getColorMap();
				color = (Color) colorMap.get(messageData.getColorMapKey());

			}

			g2d.setColor(color);

			if (startX < endX) {

				g2d.drawLine(startX, startY, endX, endY);

				int topX = endX - (int) (ARROW_WIDTH * zoomValue);
				int topY = endY - (int) (ARROW_TOP_HEIGHT * zoomValue);
				int bottomX = endX - (int) (ARROW_WIDTH * zoomValue);
				int bottomY = endY + (int) (ARROW_TOP_HEIGHT * zoomValue);
				int pointX = endX;
				int pointY = endY;

				int[] xPoints = { topX, bottomX, pointX };
				int[] yPoints = { topY, bottomY, pointY };
				int numPoints = 3;

				Polygon arrowHead = new Polygon(xPoints, yPoints, numPoints);
				g2d.fillPolygon(arrowHead);

				messageData.setX(startX);
				messageData.setY(startY
						- (int) (MESSAGE_SELECT_HEIGHT * zoomValue));
				messageData.setDimension(new Dimension((int) (endX - startX),
						(int) (MESSAGE_SELECT_HEIGHT * 2 * zoomValue)));

				String msg = messageData.getMessageType();
				String message = "F" + messageData.getId() + " : " + msg;
				int messageStringWidth = metrics.stringWidth(message);

				int midX = endX - ((endX - startX) / 2);
				int stringX = midX - (messageStringWidth / 2);
				int stringY = startY
						- (int) (MESSAGE_LABEL_SEPARATOR_PAD * zoomValue);

				g2d.drawString(message, stringX, stringY);

				/*
				 * int fontHeight = metrics.getHeight();
				 * 
				 * g2d.drawString(new Integer(i).toString(), (int)
				 * (NODE_SEPARATOR_PAD * zoomValue) + (int) (.3 * NODE_WIDTH *
				 * zoomValue), startY + (fontHeight / 2));
				 */

			} else if (startX > endX) {

				g2d.drawLine(startX, startY, endX, endY);

				int topX = endX + (int) (ARROW_WIDTH * zoomValue);
				int topY = endY - (int) (ARROW_TOP_HEIGHT * zoomValue);
				int bottomX = endX + (int) (ARROW_WIDTH * zoomValue);
				int bottomY = endY + (int) (ARROW_TOP_HEIGHT * zoomValue);
				int pointX = endX;
				int pointY = endY;

				int[] xPoints = { topX, bottomX, pointX };
				int[] yPoints = { topY, bottomY, pointY };
				int numPoints = 3;

				Polygon arrowHead = new Polygon(xPoints, yPoints, numPoints);
				g2d.fillPolygon(arrowHead);

				messageData.setX(endX);
				messageData.setY(endY
						- (int) (MESSAGE_SELECT_HEIGHT * zoomValue));
				messageData.setDimension(new Dimension((int) (startX - endX),
						(int) (MESSAGE_SELECT_HEIGHT * 2 * zoomValue)));

				String message = "F" + messageData.getId() + " : "
						+ messageData.getMessageType();
				int messageStringWidth = metrics.stringWidth(message);

				int midX = endX - ((endX - startX) / 2);
				int stringX = midX - (messageStringWidth / 2);
				int stringY = startY
						- (int) (MESSAGE_LABEL_SEPARATOR_PAD * zoomValue);

				g2d.drawString(message, stringX, stringY);

				/*
				 * int fontHeight = metrics.getHeight();
				 * 
				 * g2d.drawString(new Integer(i).toString(), (int)
				 * (NODE_SEPARATOR_PAD * zoomValue) + (int) (.3 * NODE_WIDTH *
				 * zoomValue), startY + (fontHeight / 2));
				 */

			} else if (startX == endX) {

				g2d.draw(new Arc2D.Double((double) startX
						- (.5 * ARC_WIDTH * zoomValue), (double) startY,
						(double) ARC_WIDTH * zoomValue, (double) ARC_HEIGHT
								* zoomValue, (double) 270, (double) 180,
						Arc2D.OPEN));

				int topX = endX + (int) (ARROW_WIDTH * zoomValue);
				int topY = (int) (MESSAGE_SEPARATOR_PAD * zoomValue) + endY
						- (int) (ARROW_TOP_HEIGHT * zoomValue);
				int bottomX = endX + (int) (ARROW_WIDTH * zoomValue);
				int bottomY = (int) (MESSAGE_SEPARATOR_PAD * zoomValue) + endY
						+ (int) (ARROW_TOP_HEIGHT * zoomValue);
				int pointX = endX;
				int pointY = (int) (MESSAGE_SEPARATOR_PAD * zoomValue) + endY;

				int[] xPoints = { topX, bottomX, pointX };
				int[] yPoints = { topY, bottomY, pointY };
				int numPoints = 3;

				Polygon arrowHead = new Polygon(xPoints, yPoints, numPoints);
				g2d.fillPolygon(arrowHead);

				String message = messageData.getMessageType();
				int messageStringWidth = metrics.stringWidth(message);

				int stringX = (int) (startX - (messageStringWidth / 2));
				messageData.setY(endY
						- (int) (MESSAGE_SELECT_HEIGHT * zoomValue));

				messageData
						.setDimension(new Dimension(
								(int) (messageStringWidth),
								(int) ((ARC_HEIGHT * zoomValue) + (MESSAGE_SELECT_HEIGHT * 2 * zoomValue))));

				int stringY = startY
						- (int) (MESSAGE_LABEL_SEPARATOR_PAD * zoomValue);

				g2d.drawString(message, stringX, stringY);

				int fontHeight = metrics.getHeight();

				g2d.drawString(new Integer(i).toString(),
						(int) (NODE_SEPARATOR_PAD * zoomValue)
								+ (int) (.3 * NODE_WIDTH * zoomValue), startY
								+ (fontHeight / 2));
			}
		}
	}

	private void printMessages(Graphics2D g2d, double zoomValue,
			FontMetrics metrics, int startMsgIndex, int finishMsgIndex) {

		ArrayList messageList = tracePanel.getMessageList();
		int startX = 0;
		int startY = (int) ((TOP_PAD * zoomValue) + (NODE_HEIGHT * zoomValue));
		int endX = 0;
		int endY = startY;

		for (int i = startMsgIndex; i < finishMsgIndex; i++) {

			MessageData messageData = (MessageData) messageList.get(i);

			RfssData fromRfss = getRfssById(messageData.getFromRfssId() + ":"
					+ messageData.getFromPort());
			RfssData toRfss = getRfssById(messageData.getToRfssId() + ":"
					+ messageData.getToPort());

			startX = fromRfss.getX() + (fromRfss.getDimension().width / 2);
			startY += (MESSAGE_SEPARATOR_PAD * zoomValue);
			endX = toRfss.getX() + (toRfss.getDimension().width / 2);
			endY = startY;

			// Set message color
			Color color = null;

			if (messageData.isSelected()) {

				color = Color.red;

			} else {

				Hashtable colorMap = (Hashtable) tracePanel.getColorMap();
				color = (Color) colorMap.get(messageData.getColorMapKey());

			}

			g2d.setColor(color);

			if (startX < endX) {

				g2d.drawLine(startX, startY, endX, endY);

				int topX = endX - (int) (ARROW_WIDTH * zoomValue);
				int topY = endY - (int) (ARROW_TOP_HEIGHT * zoomValue);
				int bottomX = endX - (int) (ARROW_WIDTH * zoomValue);
				int bottomY = endY + (int) (ARROW_TOP_HEIGHT * zoomValue);
				int pointX = endX;
				int pointY = endY;

				int[] xPoints = { topX, bottomX, pointX };
				int[] yPoints = { topY, bottomY, pointY };
				int numPoints = 3;

				Polygon arrowHead = new Polygon(xPoints, yPoints, numPoints);
				g2d.fillPolygon(arrowHead);

				messageData.setX(startX);
				messageData.setY(startY
						- (int) (MESSAGE_SELECT_HEIGHT * zoomValue));
				messageData.setDimension(new Dimension((int) (endX - startX),
						(int) (MESSAGE_SELECT_HEIGHT * 2 * zoomValue)));

				String msg = messageData.getMessageType();
				String message = "F" + messageData.getId() + " : " + msg;
				int messageStringWidth = metrics.stringWidth(message);

				int midX = endX - ((endX - startX) / 2);
				int stringX = midX - (messageStringWidth / 2);
				int stringY = startY
						- (int) (MESSAGE_LABEL_SEPARATOR_PAD * zoomValue);

				g2d.drawString(message, stringX, stringY);

				int fontHeight = metrics.getHeight();

				g2d.drawString(new Integer(i).toString(),
						(int) (NODE_SEPARATOR_PAD * zoomValue)
								+ (int) (.3 * NODE_WIDTH * zoomValue), startY
								+ (fontHeight / 2));

			} else if (startX > endX) {

				g2d.drawLine(startX, startY, endX, endY);

				int topX = endX + (int) (ARROW_WIDTH * zoomValue);
				int topY = endY - (int) (ARROW_TOP_HEIGHT * zoomValue);
				int bottomX = endX + (int) (ARROW_WIDTH * zoomValue);
				int bottomY = endY + (int) (ARROW_TOP_HEIGHT * zoomValue);
				int pointX = endX;
				int pointY = endY;

				int[] xPoints = { topX, bottomX, pointX };
				int[] yPoints = { topY, bottomY, pointY };
				int numPoints = 3;

				Polygon arrowHead = new Polygon(xPoints, yPoints, numPoints);
				g2d.fillPolygon(arrowHead);

				messageData.setX(endX);
				messageData.setY(endY
						- (int) (MESSAGE_SELECT_HEIGHT * zoomValue));
				messageData.setDimension(new Dimension((int) (startX - endX),
						(int) (MESSAGE_SELECT_HEIGHT * 2 * zoomValue)));

				String message = "F" + messageData.getId() + " : "
						+ messageData.getMessageType();
				int messageStringWidth = metrics.stringWidth(message);

				int midX = endX - ((endX - startX) / 2);
				int stringX = midX - (messageStringWidth / 2);
				int stringY = startY
						- (int) (MESSAGE_LABEL_SEPARATOR_PAD * zoomValue);

				g2d.drawString(message, stringX, stringY);

				int fontHeight = metrics.getHeight();

				g2d.drawString(new Integer(i).toString(),
						(int) (NODE_SEPARATOR_PAD * zoomValue)
								+ (int) (.3 * NODE_WIDTH * zoomValue), startY
								+ (fontHeight / 2));

			} else if (startX == endX) {

				g2d.draw(new Arc2D.Double((double) startX
						- (.5 * ARC_WIDTH * zoomValue), (double) startY,
						(double) ARC_WIDTH * zoomValue, (double) ARC_HEIGHT
								* zoomValue, (double) 270, (double) 180,
						Arc2D.OPEN));

				int topX = endX + (int) (ARROW_WIDTH * zoomValue);
				int topY = (int) (MESSAGE_SEPARATOR_PAD * zoomValue) + endY
						- (int) (ARROW_TOP_HEIGHT * zoomValue);
				int bottomX = endX + (int) (ARROW_WIDTH * zoomValue);
				int bottomY = (int) (MESSAGE_SEPARATOR_PAD * zoomValue) + endY
						+ (int) (ARROW_TOP_HEIGHT * zoomValue);
				int pointX = endX;
				int pointY = (int) (MESSAGE_SEPARATOR_PAD * zoomValue) + endY;

				int[] xPoints = { topX, bottomX, pointX };
				int[] yPoints = { topY, bottomY, pointY };
				int numPoints = 3;

				Polygon arrowHead = new Polygon(xPoints, yPoints, numPoints);
				g2d.fillPolygon(arrowHead);

				String message = messageData.getMessageType();
				int messageStringWidth = metrics.stringWidth(message);

				int stringX = (int) (startX - (messageStringWidth / 2));
				messageData.setY(endY
						- (int) (MESSAGE_SELECT_HEIGHT * zoomValue));

				messageData
						.setDimension(new Dimension(
								(int) (messageStringWidth),
								(int) ((ARC_HEIGHT * zoomValue) + (MESSAGE_SELECT_HEIGHT * 2 * zoomValue))));

				int stringY = startY
						- (int) (MESSAGE_LABEL_SEPARATOR_PAD * zoomValue);

				g2d.drawString(message, stringX, stringY);

				int fontHeight = metrics.getHeight();

				g2d.drawString(new Integer(i).toString(),
						(int) (NODE_SEPARATOR_PAD * zoomValue)
								+ (int) (.3 * NODE_WIDTH * zoomValue), startY
								+ (fontHeight / 2));
			}
		}
	}

	public void mouseClicked(MouseEvent e) {

		if (e.getClickCount() == 2) {

			createNewWindow();

		}

		RfssData rfssData = getRfssByLocation(e.getX(), e.getY());

		if (rfssData != null) {

			disableMessageSelected();
			disableRfssSelected();
			rfssData.setSelected(true);
			this.repaint();

			JTextArea textArea = tracePanel.getTextArea();
			StringBuffer sbuf = new StringBuffer();
			sbuf.append("ID: " + rfssData.getId() + "\nIP Address: "
					+ rfssData.getAddress());
			for (int i = 0; i < rfssData.getPorts().size(); i++) {
				if (i == 0)
					sbuf.append("\nSIP Recv Port: "
							+ rfssData.getPorts().get(i));
				else
					sbuf.append("\nRTP Recv Port: "
							+ rfssData.getPorts().get(i));

			}
			if (rfssData.getRfssConfig() != null) {
				sbuf.append("\nRfss Symbolic name = "
						+ rfssData.getRfssConfig().getRfssName());
				sbuf.append("\nDomain name = "
						+ rfssData.getRfssConfig().getDomainName());
				if (rfssData.getRfssConfig().getHomeSubsciberUnits().hasNext()) {
					sbuf.append("\nSu IDs homed at this RFSS = [ ");

					for (Iterator<SuConfig> it = rfssData.getRfssConfig()
							.getHomeSubsciberUnits(); it.hasNext();) {
						SuConfig suConfig = it.next();
						sbuf.append(" ( Symbolic name = ");
						sbuf.append(suConfig.getSuName() + " , ");
						sbuf.append("SU ID = ");
						sbuf.append(suConfig.getSuId() + " ) ");
					}
					sbuf.append("]");
				}

				if (rfssData.getRfssConfig().getInitialServedSubscriberUnits()
						.size() != 0) {
					sbuf.append("\nSu IDs initially served at this RFSS = [ ");
					for (SuConfig suConfig : rfssData.getRfssConfig()
							.getInitialServedSubscriberUnits()) {

						sbuf.append(" ( SU Symbolic name = ");
						sbuf.append(suConfig.getSuName() + " , ");
						sbuf.append("SU ID = %x");
						sbuf.append(Integer.toHexString(suConfig.getSuId()) + " ) ");
					}
					sbuf.append("]");
				}

				if (rfssData.getRfssConfig().getAssignedGroups().hasNext()) {
					sbuf.append("\nGroup IDs homed at this RFSS = [ ");

					for (Iterator<GroupConfig> it = rfssData.getRfssConfig()
							.getAssignedGroups(); it.hasNext();) {
						GroupConfig groupConfig = it.next();
						sbuf.append("( Group Symbolic Name = "
								+ groupConfig.getGroupName());
						sbuf.append(" , Group ID = %x" + Integer.toHexString(groupConfig.getGroupId())
								+ " ) ");
					}
					sbuf.append("]");
				}
				for (PttSessionInfo pttSessionInfo : rfssData
						.getPttSessionInfo()) {
					sbuf.append("\n[").append(pttSessionInfo.toString())
							.append("]\n");
				}
				textArea.setText(sbuf.toString());

				// JOptionPane.showMessageDialog(null, sbuf.toString(),
				// "RFSS Info", JOptionPane.INFORMATION_MESSAGE);
			}

		} else {

			disableRfssSelected();

			// Try messages
			MessageData messageData = getMessageByLocation(e.getX(), e.getY());

			if (messageData != null) {

				disableMessageSelected();
				messageData.setSelected(true);
				this.repaint();

				JTextArea textArea = tracePanel.getTextArea();
				textArea.setText(messageData.getData().trim());
				textArea.setCaretPosition(0);

			} else {

				disableMessageSelected();
				repaint();

			}
		}

	}

	public void mousePressed(MouseEvent e) {

		mousePressedXPos = e.getX();
		mousePressedYPos = e.getY();

		int buttonNumber = e.getButton();

		if (buttonNumber == 3) {

			rightClickPopupMenu.show(this, e.getX(), e.getY());

		}
	}

	public void mouseReleased(MouseEvent e) {
	}

	public void mouseEntered(MouseEvent e) {
	}

	public void mouseExited(MouseEvent e) {
	}

	public void mouseDragged(MouseEvent e) {

		JViewport jv = tracePanel.getScrollPane().getViewport();
		Point p = jv.getViewPosition();
		int newX = p.x - (e.getX() - mousePressedXPos);
		int newY = p.y - (e.getY() - mousePressedYPos);

		int maxX = this.getWidth() - jv.getWidth();
		int maxY = this.getHeight() - jv.getHeight();

		if (newX < 0)
			newX = 0;
		if (newX > maxX)
			newX = maxX;
		if (newY < 0)
			newY = 0;
		if (newY > maxY)
			newY = maxY;

		jv.setViewPosition(new Point(newX, newY));

	}

	public void mouseMoved(MouseEvent arg0) {
	}

	public void actionPerformed(ActionEvent ae) {

		if (ae.getActionCommand().equals("New Window")) {

			createNewWindow();

		} else if (ae.getActionCommand().equals("Save As")) {

			saveAs();

		} else if (ae.getActionCommand().equals("Close")) {

			int selectedIndex = tracePanel.getTabbedPane().getSelectedIndex();
			tracePanel.getTabbedPane().remove(selectedIndex);

		}
	}

	private void zoom(ZoomLevel toZoomLevel) {

		tracePanel.setZoomLabel(toZoomLevel);

		double viewProportion = toZoomLevel.value() / currentZoomLevel.value();
		currentZoomLevel = toZoomLevel;

		double newCursorXPos = (mousePressedXPos * viewProportion);
		double newCursorYPos = (mousePressedYPos * viewProportion);

		JViewport jv = tracePanel.getScrollPane().getViewport();
		Dimension dimension = jv.getExtentSize();
		int viewWidth = dimension.width;
		int viewHeight = dimension.height;

		int newViewX = (int) (newCursorXPos - viewWidth / 2);
		if (newViewX < 0)
			newViewX = 0;

		int newViewY = (int) (newCursorYPos - viewHeight / 2);
		if (newViewY < 0)
			newViewY = 0;

		jv.setViewPosition(new Point((int) (newViewX), (int) (newViewY)));

		this.repaint();

	}

	private void createNewWindow() {

		JFrame frame = new JFrame();
		frame.setLayout(new BorderLayout());
		frame.setMinimumSize(new Dimension(750, 600));
		frame.setPreferredSize(new Dimension(800, 600));
		frame.setMaximumSize(new Dimension(850, 600));
		frame.setSize(800, 600);
		frame.setIconImage(new ImageIcon("gfx/nist.gif").getImage());

		frame.setTitle(this.tracePanel.getFileName());
		TracePanel newTracePanel = new TracePanel(this.tracePanel.getFrame(),
				this.tracePanel.getTabbedPane(), this.tracePanel.getFileName(),
				this.tracePanel.getRfssList(),
				this.tracePanel.getMessageList(), this.tracePanel.getColorMap());
		frame.add(newTracePanel, BorderLayout.CENTER);
		frame.setLocationRelativeTo(this.tracePanel.getFrame());
		frame.setVisible(true);

	}
	
}
