/* The software was developed by employees of the National Institute of
 * Standards and Technology (NIST), an agency of the Federal Government.
 * Pursuant to title 15 United States Code Section 105, works of NIST
 * employees are not subject to copyright protection in the United States
 * and are considered to be in the public domain.  As a result, a formal
 * license is not needed to use the software.
 * 
 * The ISSI Emulator software is provided by NIST as a service and is expressly
 * provided "AS IS".  NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED
 * OR STATUTORY, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT
 * AND DATA ACCURACY.  NIST does not warrant or make any representations
 * regarding the use of the software or the results thereof including, but
 * not limited to, the correctness, accuracy, reliability or usefulness of
 * the software.
 * 
 * Permission to use this software is contingent upon your acceptance
 * of the terms of this agreement and upon your providing appropriate
 * acknowledgments of NIST's ownership of the software.
 */
package gov.nist.p25.issi.traceviewer;

import org.apache.log4j.Logger;
import org.xml.sax.*;
import org.xml.sax.helpers.DefaultHandler;

import javax.xml.parsers.SAXParserFactory; // import
											// javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import java.util.*;
import java.awt.Color;
import java.io.*;

/**
 * This class implements a PTT XML trace file loader.
 * 
 * @author steveq@nist.gov
 * @author mranga@nist.gov
 * @version $Revision: 1.6 $, $Date: 2007/10/27 03:23:33 $
 * @since 1.5
 */
public class PttTraceLoader extends DefaultHandler implements TraceLoader {

	private static Logger logger = Logger.getLogger(PttTraceLoader.class);

	// /////////////////////////////////////////////////////////////////////
	// Variables
	// /////////////////////////////////////////////////////////////////////

	private Properties props = new Properties();

	private TreeSet<MessageData> messageList = null;

	private Hashtable<String, Color> colorMap = null;

	private String myIpAddress = "";

	private String myRtpRecvPort = "";

	private String myRfssId = "";

	private String remoteRtpRecvPort = "";

	private String remoteRfssId = "";

	private String messageType = "";

	private String pttMessageData = "";

	private String rtfData = "";

	private String time = "";

	private int messageColorIndex = 0;

	private static final int MAX_COLOR_INDEX = 11;

	int[][] messageColors = { { 117, 4, 0 }, { 96, 93, 0 }, { 96, 0, 87 },
			{ 31, 104, 0 }, { 27, 0, 96 }, { 0, 159, 109 }, { 0, 145, 159 },
			{ 96, 0, 38 }, { 0, 0, 255 }, { 95, 0, 127 }, { 255, 0, 201 },
			{ 191, 0, 40 } };

	private boolean isSender;

	private String sequenceNumber;

	private boolean inRtfFormat;

	// /////////////////////////////////////////////////////////////////////
	// Constructors
	// /////////////////////////////////////////////////////////////////////

	/**
	 * This constructor gets called when we are loading the traces at the end of
	 * a test case.
	 * 
	 * @param inputStream --
	 *            input stream from which to read the messages.
	 * @param colorMap --
	 *            color map to color the horizontal arcs.
	 */
	public PttTraceLoader(InputStream inputStream,
			Hashtable<String, Color> colorMap) {
		this.messageList = new TreeSet<MessageData>();
		this.colorMap = colorMap;

		SAXParserFactory factory = SAXParserFactory.newInstance();

		try {

			SAXParser saxParser = factory.newSAXParser();
			saxParser.parse(inputStream, this);

		} catch (Exception ex) {

			ex.printStackTrace();

		}

	}

	// /////////////////////////////////////////////////////////////////////
	// Methods
	// /////////////////////////////////////////////////////////////////////

	public void startElement(String namespaceURI, String localName,
			String qualifiedName, Attributes attrs) throws SAXException {

		String elementName = localName;

		if ("".equals(elementName)) {

			elementName = qualifiedName;
		}

		if (!elementName.equals("rtf-format") && !elementName.equals("trace")
				&& !elementName.equals("pttmessages"))
			pttMessageData += elementName + "\n";

		if (elementName.equals("rtf-format"))
			this.inRtfFormat = true;

		if (attrs != null && !elementName.equals("rtf-format")) {

			for (int i = 0; i < attrs.getLength(); i++) {

				String attributeName = attrs.getLocalName(i);

				if ("".equals(attributeName)) {

					attributeName = attrs.getQName(i);
				}

				pttMessageData += "\t" + attributeName + " = "
						+ attrs.getValue(i) + "\n";
				String propname = elementName + "." + attributeName;

				// Capture the fields for pattern matching.
				if (propname.equals("ptt-packet.receivingRfssId")
						|| propname.equals("ptt-packet.sessionType")
						|| propname.equals("rtp-header.version")
						|| propname.equals("rtp-header.marker")
						|| propname.equals("rtp-header.headerExtension")
						|| propname.equals("rtp-header.payloadType")
						|| propname.equals("rtp-header.SSRC")
						|| propname.equals("control-octet.signalBit")
						|| propname.equals("control-octet.compactBit")
						|| propname.equals("block-header.payloadType")
						|| propname.equals("block-header.blockType")
						|| propname.equals("issi-packet-type.muteStatus")
						|| propname.equals("issi-packet-type.losingAudio")
						|| propname.equals("issi-packet-type.serviceOptions")
						|| propname.equals("issi-packet-type.packetType")) {
					this.props.setProperty(elementName + "." + attributeName,
							attrs.getValue(i));
				}

				handleAttribute(attributeName, attrs.getValue(i));

			}
		}

	}

	public void handleAttribute(String attrName, String attrValue) {

		if (attrName.equals("myIpAddress")) {

			myIpAddress = attrValue;

		} else if (attrName.equals("receivingRfssId")) {

			myRfssId = attrValue;

		} else if (attrName.equals("isSender")) {
			this.isSender = attrValue.equals("true");

		} else if (attrName.equals("myRtpRecvPort")) {

			myRtpRecvPort = attrValue;

		} else if (attrName.equals("sendingRfssId")) {
			remoteRfssId = attrValue;
		} else if (attrName.equals("remoteRtpRecvPort")) {

			remoteRtpRecvPort = attrValue;

		} else if (attrName.equals("packetType")) {

			messageType = attrValue;

			if (colorMap != null) {
				if (!colorMap.containsKey(attrValue)) {

					if (messageColorIndex > MAX_COLOR_INDEX)
						messageColorIndex = 0;

					int[] rgb = messageColors[messageColorIndex];
					colorMap.put(attrValue, new Color(rgb[0], rgb[1], rgb[2]));
					messageColorIndex++;

				}
			}
		} else if (attrName.equals("packetNumber")) {
			sequenceNumber = attrValue;

		} else if (attrName.equals("receptionTime")) {

			this.time = attrValue;
			// System.out.println("receptionTime: " + time);
		}
	}

	public void endElement(String namespaceURI, String simpleName,
			String qualifiedName) throws SAXException {

		if (qualifiedName.equals("rtf-format")) {
			this.inRtfFormat = false;
		}
		if (qualifiedName.equals("ptt-packet")) {

			if (!pttMessageData.equals("")) {

				PttMessageData messageData = new PttMessageData(remoteRfssId,
						myRfssId, myRtpRecvPort, remoteRtpRecvPort,
						messageType, pttMessageData, time, sequenceNumber,
						this.rtfData, false, isSender, this.props);

				messageList.add(messageData);

				// Reset data
				this.props = new Properties();
				this.isSender = false;
				myIpAddress = "";
				myRfssId = "";
				myRtpRecvPort = "";
				remoteRtpRecvPort = "";
				messageType = "";
				pttMessageData = "";
				time = "";
				sequenceNumber = "";
				remoteRfssId = "";
				rtfData = "";

			}

		}

	}

	public void characters(char buf[], int offset, int len) throws SAXException {

		String s = new String(buf, offset, len);

		if (!inRtfFormat) {
			if (s.length() > 8)
				pttMessageData += "\t" + s + "\n";
		} else {
			this.rtfData += s;
		}

	}

	public ArrayList<MessageData> getRecords(String rfssDomainName) {
		ArrayList<MessageData> alist = new ArrayList<MessageData>();
		for (MessageData mdata : this.messageList) {
			if ( mdata.toRfssId.equals(rfssDomainName))
				alist.add(mdata);

		}
		return alist;
	}
	
	
	public ArrayList<MessageData> getRecords() {
		int counter = 0;
		ArrayList<MessageData> alist = new ArrayList<MessageData>();
		for (MessageData mdata : this.messageList) {
			mdata.setId(++counter);
			alist.add(mdata);

		}
		return alist;
	}

}
