package gov.nist.p25.issi.traceviewer;

import gov.nist.javax.sip.message.SIPMessage;
import gov.nist.p25.issi.p25body.ContentList;
import gov.nist.p25.issi.utils.RtfConstants;

import java.util.Iterator;

import javax.sip.header.CSeqHeader;
import javax.sip.header.CallIdHeader;
import javax.sip.header.ContentDispositionHeader;
import javax.sip.header.ContentLengthHeader;
import javax.sip.header.ContentTypeHeader;
import javax.sip.header.ExpiresHeader;
import javax.sip.header.FromHeader;
import javax.sip.header.MaxForwardsHeader;
import javax.sip.header.RouteHeader;
import javax.sip.header.TimeStampHeader;
import javax.sip.header.ToHeader;
import javax.sip.header.ViaHeader;
import javax.sip.header.WarningHeader;
import javax.sip.message.Message;
import javax.sip.message.Request;
import javax.sip.message.Response;

import org.apache.log4j.Logger;

public class SipMessageData extends MessageData implements
		Comparable<SipMessageData>, RtfConstants {
	private String timeStamp;

	private static Logger logger = Logger.getLogger(SipMessageData.class);

	private String transactionId;

	private Message sipMessage;
	
	
	@Override
	public String toString() {
		return new StringBuffer()
		.append("timeStamp = \"" + timeStamp + "\"\n")
		.append("fromRfssId = \"" + super.fromRfssId + "\"\n")
		.append("toRfssId = \"" + super.toRfssId + "\"\n" )
		.append("transactionId = \"" + transactionId + "\"\n")
		.append("<sip-message>\n")
		.append(sipMessage == null ? null: sipMessage.toString())
		.append("</sip-message>\n").toString();
	}

	public SipMessageData(String fromRfssId, String toRfssId, String fromPort,
			String toPort, String messageType, String data, String time,
			String timeStamp, String transactionId, boolean selected) {
		super(fromRfssId, toRfssId, fromPort, toPort, messageType, data, time,
				transactionId, selected, false);
		if ( data == null ) throw new NullPointerException("null bad data!");
		this.timeStamp = timeStamp;
		this.transactionId = transactionId;

	}

	private String getKey() {
		if (timeStamp != null) {
			return messageType + transactionId + ":" + timeStamp;
		} else {
			// Generate a random time stamp
			return messageType + transactionId + ":"
					+ (int) (Math.random() * 1000);
		}
	}

	/**
	 * To check for insertion into the hash table.
	 * 
	 */
	@Override
	public int hashCode() {
		return super.hashCode();
	}

	/**
	 * @return the transactionId
	 */
	public String getTransactionId() {
		return transactionId;
	}

	/**
	 * @param sipMessage --
	 *            the message to set.
	 */
	public void setSipMessage(Message sipMessage) {
		if (sipMessage == null)
			throw new NullPointerException(
					"attempt to set a null message in SipMessageData");

		this.sipMessage = sipMessage;

	}
	
	public Message getSipMessage() {
		return this.sipMessage;
		
	}

	/**
	 * Format a via header as RTF. Note that we only store the last four
	 * significant digits of the branch ( should be random enough for a single
	 * call flow).
	 * 
	 */
	private static String formatViaHeader(ViaHeader viaHeader) {
		StringBuffer sbuf = new StringBuffer();

		String branch = viaHeader.getBranch();
		String restOfBranch = branch.substring("z9hG4bK".length());
		sbuf.append(HEADER_START
				+ ViaHeader.NAME
				+ ": SIP/2.0/UDP "
				+ viaHeader.getHost()
				+ ";branch=z9hG4bK"
				+ IGNORED_FIELD
				+ restOfBranch.substring(restOfBranch.length() - 4,
						restOfBranch.length()) + HEADER_END);
		return sbuf.toString();

	}

	private static String formatMaxForwardsHeader(
			MaxForwardsHeader maxForwardsHeader) {
		return new StringBuffer().append(HEADER_START).append(
				MaxForwardsHeader.NAME).append(": ").append(IGNORED_FIELD)
				.append(maxForwardsHeader.getMaxForwards()).append(HEADER_END)
				.toString();
	}

	/**
	 * Format Expires Header in RTF
	 */
	private static String formatExpiresHeader(ExpiresHeader expiresHeader) {
		StringBuffer sbuf = new StringBuffer();
		if (expiresHeader.getExpires() == 0) {
			return sbuf.append(HEADER_START).append(expiresHeader.toString())
					.append(HEADER_END).toString();

		} else {
			return sbuf.append(HEADER_START).append(ExpiresHeader.NAME).append(
					": ").append(IGNORED_FIELD).append(
					expiresHeader.getExpires()).append(HEADER_END).toString();
		}
	}

	private static String formatFromHeader(FromHeader fromHeader) {
		StringBuffer sbuf = new StringBuffer();

		return sbuf.append(HEADER_START).append(FromHeader.NAME).append(": ")
				.append(fromHeader.getAddress().toString()).append(";tag=")
				.append(IGNORED_FIELD).append(fromHeader.getTag()).append(
						HEADER_END).toString();
	}

	private static String formatToHeader(ToHeader toHeader) {
		StringBuffer sbuf = new StringBuffer();
		if (toHeader.getTag() != null) {
			return sbuf.append(HEADER_START).append(ToHeader.NAME).append(": ")
					.append(toHeader.getAddress().toString()).append(";tag=")
					.append(IGNORED_FIELD).append(toHeader.getTag()).append(
							HEADER_END).toString();
		} else {
			return sbuf.append(HEADER_START).append(ToHeader.NAME).append(": ")
					.append(toHeader.getAddress().toString())
					.append(HEADER_END).toString();
		}
	}

	private static String formatCSeqHeader(CSeqHeader cseqHeader) {
		StringBuffer sbuf = new StringBuffer();
		return sbuf.append(HEADER_START).append(CSeqHeader.NAME + ": ").append(
				IGNORED_FIELD).append((int) cseqHeader.getSeqNumber()).append(
				HEADER_RESTART).append(cseqHeader.getMethod()).append(
				HEADER_END).toString();
	}

	private static String formatCallIdHeader(CallIdHeader callIdHeader) {
		StringBuffer sbuf = new StringBuffer();
		return sbuf.append(HEADER_START).append(CallIdHeader.NAME).append(": ")
				.append(IGNORED_FIELD).append(callIdHeader.getCallId()).append(
						HEADER_END).toString();
	}

	private static String formatLine(String line) {
		return (HEADER_START + line + HEADER_END);
	}

	private static String formatWarningHeader(WarningHeader warningHeader) {
		return new StringBuffer().append(OPTIONAL_HEADER).append(
				warningHeader.toString()).append(HEADER_END).toString();
	}

	/**
	 * @return the rtfText
	 */
	public String getRtfText() {
		try {
			StringBuffer sbuf = new StringBuffer();

			sbuf.append("{\\par F" + id + "\\par }\n");
			/*
			 * Dunno what the bloody hell this is but I cut and paste from a
			 * word document.
			 */
			sbuf.append(PREAMBLE);
			String firstLine = ((SIPMessage) sipMessage).getFirstLine();
			sbuf.append(formatLine(firstLine));
			// Format the via header
			Iterator iter = sipMessage.getHeaders(ViaHeader.NAME);
			while (iter.hasNext()) {
				ViaHeader via = (ViaHeader) iter.next();
				sbuf.append(formatViaHeader(via));
			}
			sbuf.append(formatFromHeader((FromHeader) sipMessage
					.getHeader(FromHeader.NAME)));
			sbuf.append(formatToHeader((ToHeader) sipMessage
					.getHeader(ToHeader.NAME)));
			sbuf.append(formatCallIdHeader((CallIdHeader) sipMessage
					.getHeader(CallIdHeader.NAME)));
			sbuf.append(formatCSeqHeader((CSeqHeader) sipMessage
					.getHeader(CSeqHeader.NAME)));
			if (sipMessage.getHeader(MaxForwardsHeader.NAME) != null)
				sbuf
						.append(formatMaxForwardsHeader((MaxForwardsHeader) sipMessage
								.getHeader(MaxForwardsHeader.NAME)));
			if (sipMessage.getHeader(ExpiresHeader.NAME) != null) {
				ExpiresHeader expiresHeader = (ExpiresHeader) sipMessage
						.getExpires();
				sbuf.append(formatExpiresHeader(expiresHeader));
			}
			if (sipMessage.getHeader(WarningHeader.NAME) != null) {
				WarningHeader warningHeader = (WarningHeader) sipMessage
						.getHeader(WarningHeader.NAME);
				sbuf.append(formatWarningHeader(warningHeader));
			}
			Iterator routes = sipMessage.getHeaders(RouteHeader.NAME);
			while (routes != null && routes.hasNext()) {
				RouteHeader routeHeader = (RouteHeader) routes.next();
				sbuf.append(formatLine(routeHeader.toString()));
			}
			Iterator it1 = sipMessage.getHeaderNames();
			while (it1.hasNext()) {
				String name = (String) it1.next();
				if (name.equals(ViaHeader.NAME) || name.equals(FromHeader.NAME)
						|| name.equals(RouteHeader.NAME)
						|| name.equals(ToHeader.NAME)
						|| name.equals(CallIdHeader.NAME)
						|| name.equals(CSeqHeader.NAME)
						|| name.equals(ExpiresHeader.NAME)
						|| name.equals(MaxForwardsHeader.NAME)
						|| name.equals(TimeStampHeader.NAME) 
						|| name.equals(ContentLengthHeader.NAME)
						|| name.equals(ContentDispositionHeader.NAME) 
						|| name.equals(ContentTypeHeader.NAME)) { 
					continue;
				}
				String msgLine = ((SIPMessage) sipMessage)
						.getHeaderAsFormattedString(name);
				sbuf.append(formatLine(msgLine));

			}
			if (sipMessage.getHeader(ContentDispositionHeader.NAME) != null) {
				String msgLine = ((SIPMessage) sipMessage)
						.getHeaderAsFormattedString(ContentDispositionHeader.NAME);
				sbuf.append(formatLine(msgLine));
			}
			if (sipMessage.getHeader(ContentTypeHeader.NAME) != null) {
				String msgLine = ((SIPMessage) sipMessage)
						.getHeaderAsFormattedString(ContentTypeHeader.NAME);
				sbuf.append(formatLine(msgLine));
			}
			sbuf.append(formatLine(sipMessage.getContentLength().toString()));
			sbuf.append(LINE_FEED);
			if (sipMessage.getHeader(ContentTypeHeader.NAME) != null) {
				ContentList clist = ContentList
						.getContentListFromMessage(sipMessage);
				String content = clist.toRtfString();
				sbuf.append(content);
			}

			sbuf.append(END_MESSAGE);
			return sbuf.toString();
		} catch (Exception ex) {
			logger.error("Badly formatted message ", ex);
			return null;
		}

	}

	@Override
	public boolean equals(Object other) {
		return this.hashCode() == other.hashCode();
	}

	public int compareTo(SipMessageData m2) {
		SipMessageData m1 = this;
		try {
			long ts1 = Long.parseLong(m1.getTime());
			long ts2 = Long.parseLong(m2.getTime());
			if (ts1 < ts2)
				return -1;
			else if (ts1 > ts2)
				return 1;
			else {
				if (m2.sipMessage instanceof Response && m1 instanceof Request
						&& m2.transactionId.equals(m1.transactionId)) {
					return -1;
				} else if (m1.sipMessage instanceof Response
						&& m2 instanceof Request
						&& !((CSeqHeader) m1.sipMessage
								.getHeader(CSeqHeader.NAME)).getMethod()
								.equals(
										m2.sipMessage
												.getHeader(CSeqHeader.NAME))) {
					return -1;
				} else
					return 1;

			}

		} catch (NumberFormatException ex) {
			ex.printStackTrace();
			//System.exit(0);
			return 0;
		}
	}

}
