/* The software was developed by employees of the National Institute of
 * Standards and Technology (NIST), an agency of the Federal Government.
 * Pursuant to title 15 United States Code Section 105, works of NIST
 * employees are not subject to copyright protection in the United States
 * and are considered to be in the public domain.  As a result, a formal
 * license is not needed to use the software.
 * 
 * The ISSI Emulator software is provided by NIST as a service and is expressly
 * provided "AS IS".  NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED
 * OR STATUTORY, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT
 * AND DATA ACCURACY.  NIST does not warrant or make any representations
 * regarding the use of the software or the results thereof including, but
 * not limited to, the correctness, accuracy, reliability or usefulness of
 * the software.
 * 
 * Permission to use this software is contingent upon your acceptance
 * of the terms of this agreement and upon your providing appropriate
 * acknowledgments of NIST's ownership of the software.
 */
package gov.nist.p25.issi.traceviewer;

import gov.nist.p25.issi.issiconfig.TopologyConfig;

import java.awt.Color;
import java.io.ByteArrayInputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.TreeSet;

import javax.swing.JFrame;

import org.apache.log4j.Logger;

/**
 * This class implements a SIP XML trace file loader.
 * 
 * @author mranga@nist.gov
 * @author steveq@nist.gov
 * @version $Revision: 1.5 $, $Date: 2007/10/25 21:27:54 $
 * @since 1.5
 */
public class SipTraceLoader implements TraceLoader {

	private HashSet<RfssData> rfssList = new HashSet<RfssData>();

	private Collection<RfssData> sortedRfssList = null;

	private String transactionId = "";

	private int messageColorIndex = 0;

	private static final int MAX_COLOR_INDEX = 11;

	int[][] messageColors = { { 117, 4, 0 }, { 96, 93, 0 }, { 96, 0, 87 },
			{ 31, 104, 0 }, { 27, 0, 96 }, { 0, 159, 109 }, { 0, 145, 159 },
			{ 96, 0, 38 }, { 0, 0, 255 }, { 95, 0, 127 }, { 255, 0, 201 },
			{ 191, 0, 40 } };

	private TopologyConfig topologyConfig;

	private SipTraceFileParser sipTracefileParser;

	// private DietsGUI gui;

	private static Logger logger = Logger.getLogger(SipTraceLoader.class);

	private ArrayList<SipMessageData> messageData;

	// /////////////////////////////////////////////////////////////////////
	// Constructor
	// /////////////////////////////////////////////////////////////////////

	public SipTraceLoader(ByteArrayInputStream byteArrayInputStream,
			Hashtable<String, HashSet<PttSessionInfo>> runtimeData,
			Hashtable<String, Color> colorMap, TopologyConfig topologyConfig)
			throws Exception {

		this.sipTracefileParser = new SipTraceFileParser(byteArrayInputStream,
				runtimeData, topologyConfig);

		this.topologyConfig = topologyConfig;
		this.sipTracefileParser.parse();

		this.messageData = this.sipTracefileParser.getRecords();

		this.rfssList = this.sipTracefileParser.getRfssList();
		
		if (topologyConfig.getTraceOrder().size() != this.rfssList.size()) {
			this.sortedRfssList = new TreeSet<RfssData>();
			this.sortedRfssList.addAll(this.rfssList);
		} else {
			// User has specified a sorting order.
			this.sortedRfssList = new LinkedList<RfssData>();
			for (String name : this.topologyConfig.getTraceOrder()) {

				boolean added = false;
				for (RfssData rfssData : this.rfssList) {
					if (rfssData.getRfssConfig().getRfssName().equals(name)) {
						this.sortedRfssList.add(rfssData);
						added = true;
						break;
					}
				}
				if (!added) {
					this.sortedRfssList = new TreeSet<RfssData>();
					this.sortedRfssList.addAll(this.rfssList);
					break;
				}
			}

		}
		
		for (SipMessageData sipMessageData : this.messageData) {
			transactionId = sipMessageData.getTransactionId();

			if (!colorMap.containsKey(transactionId)) {

				if (messageColorIndex > MAX_COLOR_INDEX)
					messageColorIndex = 0;

				int[] rgb = messageColors[messageColorIndex];
				colorMap.put(transactionId, new Color(rgb[0], rgb[1], rgb[2]));
				messageColorIndex++;

			}
		}

	}

	// /////////////////////////////////////////////////////////////////////
	// Methods
	// /////////////////////////////////////////////////////////////////////

	public Collection<RfssData> getSortedRfssList() {
		return this.sortedRfssList;
	}

	@Override
	public ArrayList<MessageData> getRecords() {
		
		ArrayList <MessageData> retval = new ArrayList<MessageData>();
		retval.addAll(this.messageData);
		return retval;
	}

}
