/* This software was developed by employees of the National Institute of
 * Standards and Technology (NIST), an agency of the Federal Government.
 * Pursuant to title 15 United States Code Section 105, works of NIST
 * employees are not subject to copyright protection in the United States
 * and are considered to be in the public domain.  As a result, a formal
 * license is not needed to use the software.
 * 
 * The ISSI Emulator is provided by NIST as a service and is expressly
 * provided "AS IS".  NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED
 * OR STATUTORY, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT
 * AND DATA ACCURACY.  NIST does not warrant or make any representations
 * regarding the use of the software or the results thereof including, but
 * not limited to, the correctness, accuracy, reliability or usefulness of
 * the software.
 * 
 * Permission to use this software is contingent upon your acceptance
 * of the terms of this agreement.
 */
package gov.nist.p25.issi.transctlmgr;

import gov.nist.p25.issi.issiconfig.*;
import gov.nist.p25.issi.rfss.RFSS;
import gov.nist.p25.issi.rfss.tester.TestScript;
import gov.nist.p25.issi.transctlmgr.ptt.*;

import java.io.*;
import java.util.Hashtable;

import javax.sdp.MediaDescription;
import javax.sdp.SessionDescription;

import org.apache.log4j.Logger;

/**
 * The Transmission control manager is the main RTP listener as well as the
 * manager of rtp Sessions. When an RTP packet comes in, it finds the
 * appropriate point to multipoint session to forward the packet on.
 * 
 * @author mranga@nist.gov
 * @version $Revision: 1.27 $, $Date: 2007/08/21 04:06:24 $
 * @since 1.5
 */
public class TransmissionControlManager {

	// /////////////////////////////////////////////////////////////////////
	// Variables
	// /////////////////////////////////////////////////////////////////////

	PttManager pttManager;

	private Hashtable<String, PttPointToMultipointSession> pttSessions;

	private boolean isRunning;

	private static Logger logger = Logger
			.getLogger(TransmissionControlManager.class);

	
	// /////////////////////////////////////////////////////////////////////
	// Constructors
	// /////////////////////////////////////////////////////////////////////

	public TransmissionControlManager(RFSS rfss) throws Exception {

		this.pttSessions = new Hashtable<String, PttPointToMultipointSession>();
		this.pttManager = new PttManager(rfss);
		this.isRunning = true;

	}

	// /////////////////////////////////////////////////////////////////////
	// Methods
	// /////////////////////////////////////////////////////////////////////

	/**
	 * get pt to mpt session for a given (uniques) session id. The session id is
	 * obtained by concatenating from and to address for pt-to-pt calls
	 * 
	 * @param sessionId
	 * @return
	 */

	public PttPointToMultipointSession getPttPointToMultipointSession(
			String sessionId, SessionType sessionType) {

		if (pttSessions.get(sessionId) != null)

			return pttSessions.get(sessionId);

		else {

			PttPointToMultipointSession retval = new PttPointToMultipointSession(
					this.pttManager, sessionType);

			retval.setSessionId(sessionId);
			pttSessions.put(sessionId, retval);
			return retval;

		}

	}

	public void shutDown() {
		if (!this.isRunning) {
			logger.debug("Session is already shut down");
		}
		for (PttPointToMultipointSession session : this.pttSessions.values()) {
			session.shutDown();
		}
		this.isRunning = false;

	}

	@Override
	public String toString() {
		return this.pttSessions.toString();
	}

	/**
	 * Extract parameters from a session description and add a media leg
	 * corresponding to that session description.
	 * 
	 * @param sessionDescription
	 * @return
	 */
	public  PttPointToMultipointSession  addMediaLeg(long groupId, PttSession pttSession,
			SessionDescription sessionDescription, String mediaLegId) {

		try {

			String ipAddress = sessionDescription.getConnection().getAddress();

			// TODO -- we need to sort through different media types here.
			MediaDescription mediaDescription = (MediaDescription) sessionDescription
					.getMediaDescriptions(true).get(0);

			if (mediaDescription.getConnection() != null) {

				ipAddress = mediaDescription.getConnection().getAddress();

			}

			int port = mediaDescription.getMedia().getMediaPort();

			if (logger.isDebugEnabled()) {

				logger.debug("Setting ipAddress : " + ipAddress);
				logger.debug("Setting port " + port);

			}

			// The last four hex digits of the radical name are the
			// group ID.
			pttSession.setGroupId((int) (groupId & 0xFFFF));
			pttSession.setRemoteIpAddress(ipAddress);
			
			pttSession.setRemoteRtpRecvPort(port);
			
			PttPointToMultipointSession pointToMultipointSession =
				 pttSession.getPointToMultipointSession();
			pointToMultipointSession.addPeerSession(pttSession,mediaLegId);
			return pointToMultipointSession;

		} catch (Exception ex) {

			logger.error("Could not add media leg for given session "
					+ "description", ex);
			throw new RuntimeException("Unexpected exception !", ex);

		}

	}

	public void teardownRTPPointToMultipointSession(String sessionId) {

		// TODO -- clse the session.
		pttSessions.remove(sessionId);

	}

	public PttManager getPttManager() {
		return this.pttManager;
	}

}
