/* This software was developed by employees of the National Institute of
 * Standards and Technology (NIST), an agency of the Federal Government.
 * Pursuant to title 15 United States Code Section 105, works of NIST
 * employees are not subject to copyright protection in the United States
 * and are considered to be in the public domain.  As a result, a formal
 * license is not needed to use the software.
 * 
 * The ISSI Emulator is provided by NIST as a service and is expressly
 * provided "AS IS".  NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED
 * OR STATUTORY, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT
 * AND DATA ACCURACY.  NIST does not warrant or make any representations
 * regarding the use of the software or the results thereof including, but
 * not limited to, the correctness, accuracy, reliability or usefulness of
 * the software.
 * 
 * Permission to use this software is contingent upon your acceptance
 * of the terms of this agreement.
 */
package gov.nist.p25.issi.transctlmgr;

import java.io.*;
import java.util.ArrayList;

import org.apache.log4j.Logger;
import gov.nist.p25.issi.issiconfig.*;
import gov.nist.p25.issi.p25payload.*;
import gov.nist.p25.issi.rfss.*;
import gov.nist.p25.issi.rfss.tester.AbstractScenario;
import gov.nist.p25.issi.transctlmgr.ptt.*;
import gov.nist.rtp.*;

/**
 * The Transmission control SAP.
 * 
 * @author mranga@nist.gov
 * @version $Revision: 1.31 $, $Date: 2007/08/22 16:04:45 $
 * @since 1.5
 */
public class TransmissionControlSAP {

	/***************************************************************************
	 * Variables
	 **************************************************************************/

	private PttSession pttSession;

	/* The RFSS that owns me. */

	private RFSS rfss;

	private static Logger logger = Logger
			.getLogger(TransmissionControlSAP.class);

	/***************************************************************************
	 * Constructors
	 **************************************************************************/

	public TransmissionControlSAP(RFSS rfss, SuConfig suConfig) {

		this.rfss = rfss;
	}

	/***************************************************************************
	 * Methods
	 **************************************************************************/

	public void setPttSession(PttSession pttSession) {

		logger.debug("TrasnmissionCOntrolManager: setPttSession " + pttSession);
		this.pttSession = pttSession;

		assert pttSession.getRfss() == this.rfss;

	}

	/**
	 * This is a placeholder method. The PTT state machine will be used to send
	 * out the talk spurt.
	 * 
	 * @param talkSpurt --
	 *            byte array to transmit.
	 * @throws IOException
	 */
	public void sendTalkSpurt(ArrayList<IMBEVoiceBlock> imbeVoices, int systemId, int unitId,
			TransmitPriorityType priorityType, int priorityLevel, int nblocks)
			throws IOException {

		try {
			if (logger.isDebugEnabled())
				logger.debug("TransmissionControlSAP : sendTalkSpurt "
						+ priorityType + " priorityLevel " + priorityLevel);
			SmfSession smfSession = (SmfSession) pttSession;

			smfSession.getSmfTransmitter().sendSpurtRequest(imbeVoices, systemId, unitId,
					nblocks);

			smfSession.setTransmitPriorityType(priorityType);

			smfSession.setTransmitPriorityLevel(priorityLevel);
		} catch (Exception ex) {

			logger.error("Unexpected error sending message", ex);
			throw new IOException(ex.getMessage());

		}

	}

	/**
	 * Force a talk spurt to a peer SU.
	 * 
	 * @param imbeVoiceBlocks --
	 *            blocks to send.
	 * @param suId --
	 *            the su id.
	 * @param priorityType --
	 *            priority type.
	 * @param priorityLevel --
	 *            priority level.
	 * @param nblocks -
	 *            number of blocks to send.
	 * @throws IOException
	 */
	public void forceTalkSpurt(ArrayList<IMBEVoiceBlock> imbeVoiceBlocks,
			int systemId, int suId, TransmitPriorityType priorityType, int priorityLevel,
			int nblocks) throws IOException {
		try {
			SmfSession smfSession = (SmfSession) pttSession;

			assert smfSession != null;
			
			smfSession.getSmfTransmitter().forceAudio(imbeVoiceBlocks, systemId, suId,
					nblocks);
			smfSession.setTransmitPriorityType(priorityType);

			smfSession.setTransmitPriorityLevel(priorityLevel);
		} catch (Exception ex) {

			logger.error("Unexpected error sending message", ex);
			throw new IOException(ex.getMessage());

		}

	}

	/**
	 * Send a talk spurt to the given group Id (given by its radical name )
	 * 
	 * @param talkSpurt -
	 *            a list of voice blocks read from a file.
	 * @param groupId -
	 *            the group radical name.
	 * @param ptype -
	 *            the transmit priority type.
	 * @param plevel -
	 *            the priority level.
	 */

	public void sendTalkSpurt(ArrayList<IMBEVoiceBlock> imbeVoices,
			String groupRadicalName, SuConfig suConfig,
			TransmitPriorityType ptype, int plevel, int nblocks)
			throws Exception {
		// Construct an ISSI header word
		GroupServing groupServing = this.rfss
				.getServedGroupByRadicalName(groupRadicalName);
		if (groupServing == null)
			throw new IOException("Group Not found " + groupRadicalName);
		if (!groupServing.getMultiplexer().isSuSubscribed(
				suConfig.getRadicalName())) {
			this.rfss.subscribeSuToServedGroup(groupRadicalName, suConfig);
		}
		int tsn = groupServing.getMultiplexer().getTsn(
				suConfig.getRadicalName());
		SmfSession smfSession = groupServing.getSmfSession(tsn);
		if (logger.isDebugEnabled()) {

			logger.debug("Sending talk spurt from RFSS "
					+ this.rfss.getRfssConfig().getRfssName());
			logger.debug("TransmissionControlSAP : sendTalkSpurt " + ptype
					+ " priorityLevel " + plevel);
		}
		if (smfSession == null)
			throw new IOException("No SMF session in table!");

		smfSession.setTransmitPriorityLevel(plevel);
		smfSession.setTransmitPriorityType(ptype);
		smfSession.getSmfTransmitter().sendSpurtRequest(imbeVoices,
				suConfig.getSystemId(), suConfig.getSuId(), nblocks);

	}

	
	/**
	 * Force a talk spurt to the group.
	 * 
	 * @param imbeVoices -- canned list of blocks.
	 * @param groupRadicalName -- radical name of group to send the talk spurt to.
	 * @param suConfig -- sending SU.
	 * @param ptype -- priorty type.
	 * @param plevel -- priority level.
	 * @param nblocks -- number of blocks to send.
	 * @throws Exception
	 */
	public void forceTalkSpurt(ArrayList<IMBEVoiceBlock> imbeVoices,
			String groupRadicalName, SuConfig suConfig,
			TransmitPriorityType ptype, int plevel, int nblocks)
			throws Exception {
		GroupServing groupServing = this.rfss
				.getServedGroupByRadicalName(groupRadicalName);
		if (groupServing == null)
			throw new IOException("Group Not found " + groupRadicalName);
		if (!groupServing.getMultiplexer().isSuSubscribed(
				suConfig.getRadicalName())) {
			this.rfss.subscribeSuToServedGroup(groupRadicalName, suConfig);
		}
		int tsn = groupServing.getMultiplexer().getTsn(
				suConfig.getRadicalName());
		SmfSession smfSession = groupServing.getSmfSession(tsn);
		if (logger.isDebugEnabled()) {

			logger.debug("Sending talk spurt from RFSS "
					+ this.rfss.getRfssConfig().getRfssName());
			logger.debug("TransmissionControlSAP : sendTalkSpurt " + ptype
					+ " priorityLevel " + plevel);
		}
		if (smfSession == null)
			throw new IOException("No SMF session in table!");

		smfSession.setTransmitPriorityLevel(plevel);
		smfSession.setTransmitPriorityType(ptype);
		smfSession.getSmfTransmitter().forceAudio(imbeVoices,
				suConfig.getSystemId(), suConfig.getSuId(), nblocks);

	}
	public void teardownRtpSession() {

		logger.debug("teardownRtpSession");

		// this.rtpSession.endSession();
		pttSession.shutDown();

	}

	

	

}
