/* This software was developed by employees of the National Institute of
 * Standards and Technology (NIST), an agency of the Federal Government.
 * Pursuant to title 15 United States Code Section 105, works of NIST
 * employees are not subject to copyright protection in the United States
 * and are considered to be in the public domain.  As a result, a formal
 * license is not needed to use the software.
 * 
 * The ISSI Emulator is provided by NIST as a service and is expressly
 * provided "AS IS".  NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED
 * OR STATUTORY, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT
 * AND DATA ACCURACY.  NIST does not warrant or make any representations
 * regarding the use of the software or the results thereof including, but
 * not limited to, the correctness, accuracy, reliability or usefulness of
 * the software.
 * 
 * Permission to use this software is contingent upon your acceptance
 * of the terms of this agreement.
 */
package gov.nist.p25.issi.transctlmgr.ptt;

import gov.nist.p25.issi.p25payload.IMBEVoiceBlock;
import gov.nist.p25.issi.p25payload.IMBEVoiceDataGenerator;
import org.apache.log4j.ConsoleAppender;
import org.apache.log4j.Logger;
import org.apache.log4j.PropertyConfigurator;
import org.apache.log4j.SimpleLayout;
import junit.framework.TestCase;
import java.util.ArrayList;

/**
 * Tests SMF to MMF audio send.
 * 
 * @author steveq@nist.gov
 * @version $Revision: 1.10 $, $Date: 2007/07/30 04:22:56 $
 * @since 1.5
 */
public class Test_Smf2Mmf_NormalAudioSendTest extends TestCase implements
		SmfTxListener, MmfRxListener {

	/***************************************************************************
	 * Constants
	 **************************************************************************/

	/** SMF RTP receive port (WARNING: port may already be in use!) */
	private static final int SMF_RTP_RECV_PORT = 1842;

	/** MMF RTP receive port (WARNING: port may already be in use!) */
	private static final int MMF_RTP_RECV_PORT = 1844;

	/** The number of audio packets sent. */
	private static final int NUM_AUDIO_PACKETS = 2;

	/***************************************************************************
	 * Variables
	 **************************************************************************/

	private SmfSession smfSession = null;

	private MmfSession mmfSession = null;

	private PttManager pttManager = new PttManager("127.0.0.1");

	private static Logger logger = Logger
			.getLogger("gov.nist.p25.issi.transctlmgr");

	static {
		PropertyConfigurator.configure("log4j.properties");
		logger.addAppender(new ConsoleAppender(new SimpleLayout()));
	}

	// SmfTx flags
	private boolean gotGrant = false;

	private boolean gotDeny = false;

	private boolean gotWait = false;

	// MmfRx flags
	private boolean gotRequest = false;
	
	private boolean gotRequestWithVoice = false;

	private boolean gotAudio = false;

	private int gotAudioPackets = 0;

	private boolean gotSpurtEnd = false;
	
	private boolean audioTimeout = false;

	/***************************************************************************
	 * Tests
	 **************************************************************************/

	public void test() {

		System.out.println("=============================================\n"
				+ "File: Smf2Mmf_NormalAudioSend\n"
				+ "Description: Tests SMF to MMF audio send:\n"
				+ "Reference: Sections 7.2.1 and 7.2.2\n\n" +

				"SMF -> PTT TRANSMIT REQUEST -> MMF\n"
				+ "SMF <- PTT TRANSMIT GRANT <- MMF\n"
				+ "SMF -> PTT PROGRESS -> MMF (audio packets)\n"
				+ "SMF -> PTT TRANSMIT END -> MMF\n" + "MMF DONE\n"
				+ "SMF TERMINATED\n"
				+ "---------------------------------------------\n");

		try {

			smfSession = pttManager.createTestSmfSession(SMF_RTP_RECV_PORT,
					"127.0.0.1", MMF_RTP_RECV_PORT,
					LinkType.UNIT_TO_UNIT_CALLING_SERVING_TO_CALLING_HOME);

			smfSession.smfTransmitter.addListener(this);

			mmfSession = pttManager.createTestMmfSession(MMF_RTP_RECV_PORT,
					"127.0.0.1", SMF_RTP_RECV_PORT,
					LinkType.UNIT_TO_UNIT_CALLING_HOME_TO_CALLING_SERVING,null);

			mmfSession.getMmfReceiver().addListener(this);
			mmfSession.getMmfReceiver().setArbitrateGrant();

		} catch (Exception ex) {

			ex.printStackTrace();
			fail("Unexpected exception");

		}

		// Start SMF normally with a spurt request
		try {

			ArrayList<IMBEVoiceBlock> imbeVoiceBlocks = IMBEVoiceDataGenerator
					.getIMBEVoiceData("voice/imbe-hex-test.txt",
							NUM_AUDIO_PACKETS);

			smfSession.smfTransmitter.sendSpurtRequest(imbeVoiceBlocks, 3,0,imbeVoiceBlocks.size());

		} catch (Exception e) {

			// TODO Auto-generated catch block
			e.printStackTrace();
			fail("Unexpected exception");

		}

		try {

			Thread.sleep(4000); // IMPORTANT! Make sure this is long enough!

		} catch (InterruptedException ie) {

			// ie.printStackTrace();

		}

	}

	public void tearDown() throws Exception {

		smfSession.shutDown();
		mmfSession.shutDown();

		logger.debug("Checking assertions...");
		
		/* Check SmfTx assertions */
		assertTrue(gotGrant == true);
		assertTrue(gotDeny == false);
		assertTrue(gotWait == false);

		/* Check MmfRx assertions */
		assertTrue(gotRequest == true);
		assertTrue(gotRequestWithVoice == false);
		assertTrue(gotAudio == true);
		assertTrue(gotAudioPackets == NUM_AUDIO_PACKETS);
		assertTrue(gotSpurtEnd == true);
		assertTrue(audioTimeout == false);

		logger.debug("DONE: All assertions passed.");

	}

	/***************************************************************************
	 * SmfTxListener Methods
	 **************************************************************************/

	public void receivedPttGrant(SmfPacketEvent event) {

		logger.debug("[SmfTxListener] GOT GRANT");
		gotGrant = true;

	}

	public void receivedPttDeny(SmfPacketEvent event) {

		logger.debug("[SmfTxListener] GOT DENY");
		gotDeny = true;
	}

	public void receivedPttWait(SmfPacketEvent event) {

		logger.debug("[SmfTxListener] GOT WAIT");
		gotWait = true;
	}

	public void requestTimeout() {

		logger.debug("[SmfTxListener] Request Timeout");

	}

	public void waitTimeout() {

		logger.debug("[SmfTxListener] Wait Timeout");

	}

	/***************************************************************************
	 * MmfRxListener Methods
	 **************************************************************************/

	public void receivedPttRequest(MmfPacketEvent event) {

		logger.debug("[MmfRxListener] GOT REQUEST");
		gotRequest = true;
		mmfSession.getMmfReceiver().arbitrate(event);

	}
	
	public void receivedPttRequestWithVoice(MmfPacketEvent event) {

		logger.debug("[MmfRxListener] GOT REQUEST WITH VOICE");
		gotRequestWithVoice = true;
		mmfSession.getMmfReceiver().arbitrate(event);

	}

	public void receivedPttProgress(MmfPacketEvent event) {

		logger.debug("[MmfRxListener] GOT PROGRESS");
		gotAudio = true;
		gotAudioPackets++;

	}

	public void receivedPttEnd(MmfPacketEvent event) {

		logger.debug("[MmfRxListener] GOT SPURT END");
		gotSpurtEnd = true;

	}
	
	public void audioTimeout() {
		
		logger.debug("[MmfRxListener] AUDIO TIMEOUT");
		audioTimeout = true;
		
	}

}
