/* This software was developed by employees of the National Institute of
 * Standards and Technology (NIST), an agency of the Federal Government.
 * Pursuant to title 15 United States Code Section 105, works of NIST
 * employees are not subject to copyright protection in the United States
 * and are considered to be in the public domain.  As a result, a formal
 * license is not needed to use the software.
 * 
 * The ISSI Emulator software is provided by NIST as a service and is expressly
 * provided "AS IS".  NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED
 * OR STATUTORY, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT
 * AND DATA ACCURACY.  NIST does not warrant or make any representations
 * regarding the use of the software or the results thereof including, but
 * not limited to, the correctness, accuracy, reliability or usefulness of
 * the software.
 * 
 * Permission to use this software is contingent upon your acceptance
 * of the terms of this agreement
 * .
 */
package gov.nist.p25.issi.p25payload;

import org.apache.log4j.Logger;
import org.xml.sax.Attributes;

import gov.nist.p25.issi.utils.ByteUtil;
import gov.nist.p25.issi.utils.RtfConstants;

/**
 * This class implements a P25 block header.
 * <p>
 * 
 * <pre>
 * 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
 * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 * |E|      BT     |             TSO           |         BL        |
 * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 * </pre>
 * 
 * @author steveq@nist.gov
 * @author mranga@nist.gov
 * @version $Revision: 1.12 $, $Date: 2007/07/24 03:20:38 $
 * @since 1.5
 */
public class BlockHeader implements RtfConstants {

	/***************************************************************************
	 * Constants
	 **************************************************************************/

	/** Constant that identifies the total byte length of fixed fields. */
	final static int FIXED_LENGTH = 4;

	/***************************************************************************
	 * Variables:  Note that each of the following use data types that are 
	 * larger than what is required to keep two's complement sign bit from 
	 * changing the sign.
	 **************************************************************************/

	/** Payload type (1 bit). */
	int E = 0;

	/** Block type (7 bits). */
	int BT = 0;

	/**
	 * Timestamp offset (14 bits). Here, we use 0 as the default for all packets
	 * that are not IMBEVoiceBlock packets. Only in the case of IMBEVoiceBlock
	 * packets is this field set to the voice sample duration of an IMBE voice
	 * packet (20ms in 125 microsecond resolution).
	 */
	int timeStampOffset = 0;

	/** Length of corresponding block (10 bits). */
	int blockLength = 0;

	
	/** The logger for this class. */
	private static Logger logger = Logger.getLogger(BlockHeader.class);

	/***************************************************************************
	 * Constructors
	 **************************************************************************/

	/**
	 * Construct a P25 block header.
	 */
	BlockHeader() {
	}

	/**
	 * Construct a P25 block header given its byte array representation.
	 * 
	 * @param bytes The byte array representation of this block header.
	 * @throws P25BlockException
	 */
	BlockHeader(byte[] bytes) throws P25BlockException {
		
		int i = ByteUtil.bytesToInt(bytes);

		this.E = (i >>> 0x1F) & 0x01; 		// Extract E
		this.BT = (i >>> 0x18) & 0x07F; 	// Extract BT
		this.timeStampOffset = (i >>> 0x0A) & 0x3FFF; 	// Extract TSO
		this.blockLength = i & 0x3FF; // Extract BL

		if ((BT != BlockType.IMBE_VOICE.intValue())
				&& (BT != BlockType.PACKET_TYPE.intValue())
				&& (BT != BlockType.ISSI_HEADER_INFO.intValue())
				&& (BT != BlockType.MFGSPEC_MIN_INDEX)
				&& (BT != BlockType.PTT_CONTROL_WORD.intValue())) {

			throw new P25BlockException("Unsupported block type: " + BT);

		}
		
	}

	/***************************************************************************
	 * Methods
	 **************************************************************************/

	/**
	 * Get the size of this block header in bytes.
	 * 
	 * @return The size of this block header in bytes.
	 */
	int size() {

		return FIXED_LENGTH;

	}

	/**
	 * Get the block header as a byte array.
	 * 
	 * @return The block header byte array.
	 */
	byte[] getBytes() {

		int i = 0;
		i |= E; 	// Add E
		i <<= 0x07; // Make room for BT
		i |= BT; 	// Add BT
		i <<= 0x0E; // Make room for TSO
		i |= timeStampOffset; 	// Add TSO
		i <<= 0x0A; // Make room for BL
		i |= blockLength; 	// Add BL

		return ByteUtil.intToBytes(i);

	}

	/**
	 * Set the payload type.
	 * 
	 * @param i
	 *            The payload type (1 bit)
	 * @throws IllegalArgumentException
	 */
	void setE(int i) throws IllegalArgumentException {

		if ((0 <= i) && (i <= ByteUtil.getMaxIntValueForNumBits(1)))
			E = i;
		else
			throw new IllegalArgumentException(P25BlockException.OUT_OF_RANGE);

	}


	/**
	 * Get the payload type int value.
	 * 
	 * @return The payload type int value.
	 */
	int getE() {

		return E;

	}
	
	/**
	 * Get the payload type.
	 * 
	 * @return The payload type.
	 */
	private PayloadType getPayloadType() {

		return PayloadType.getInstance(E);

	}

	/**
	 * Set the block type.
	 * 
	 * @param i
	 *            The block type (7 bits)
	 * @throws IllegalArgumentException
	 */
	void setBT(int i) throws IllegalArgumentException {

		if ((0 <= i) && (i <= ByteUtil.getMaxIntValueForNumBits(7)))
			BT = i;
		else
			throw new IllegalArgumentException(P25BlockException.OUT_OF_RANGE);

	}


	/**
	 * Get the block type.
	 * 
	 * @return The block type.
	 */
	int getBT() {

		return BT;

	}

	/**
	 * Get the block type.
	 * 
	 * @return The block type.
	 */
	BlockType getBlockType() {

		return BlockType.getInstance(BT);

	}

	/**
	 * Set the timestamp offset. Each block header SHALL contain a 14-bit
	 * timestamp offset. This SHALL be in the same units as the RTP header
	 * timestamp (125 microseconds). The sum of the RTP timestamp and the offset
	 * is the timestamp of each block.
	 * 
	 * @param i
	 *            The timestamp offset (14 bits)
	 * @throws IllegalArgumentException
	 */
	void setTimeStampOffset(int i) throws IllegalArgumentException {

		if ((0 <= i) && (i <= ByteUtil.getMaxIntValueForNumBits(14)))
			timeStampOffset = i;
		else
			throw new IllegalArgumentException(P25BlockException.OUT_OF_RANGE);

	}

	/**
	 * Get the timestamp offset.
	 * 
	 * @return the timestamp offset.
	 */
	int getTimeStampOffset() {

		return timeStampOffset;

	}

	/**
	 * Set the block length.
	 * 
	 * @param i
	 *            The block length (10 bits)
	 * @throws IllegalArgumentException
	 */
	void setBlockLength(int i) throws IllegalArgumentException {

		if ((0 <= i) && (i <= ByteUtil.getMaxIntValueForNumBits(10)))
			blockLength = i;
		else
			throw new IllegalArgumentException(P25BlockException.OUT_OF_RANGE);

	}


	/**
	 * Get the block length. Each block header SHALL contain a 10-bit length
	 * indicator. This SHALL be the number of octets contained in the
	 * corresponding block.
	 * 
	 * @return The block length.
	 */
	int getBlockLength() {

		return blockLength;

	}
	
	/**
	 * Compare this object with another object.
	 * 
	 * @param other The other object to be compared to.
	 * @return True if this object is equivalent to the other object.
	 */
	@Override
	public boolean equals(Object other) {
		if (!(other instanceof BlockHeader))
			return false;
		BlockHeader bh = (BlockHeader) other;
		return this.E == bh.E && this.BT == bh.BT && this.timeStampOffset == bh.timeStampOffset
				&& this.blockLength == bh.blockLength;
	}

	/**
	 * Get the XML formatted string representation.
	 * 
	 * @return The XML formatted string representation.
	 */
	@Override
	public String toString() {
		return new StringBuffer().append("<block-header").append(
				"\npayloadType = \"" + getPayloadType() + "\"").append(
				"\nblockType = \"" + getBlockType() + "\"").append(
				"\ntimeStampOffset = \"" + getTimeStampOffset() + "\"").append(
				"\nblockLength = \"" + getBlockLength() + "\"").append("\n/>")
				.toString();
	}
	
	/**
	 * Create an instance from XML attributes.
	 * @return
	 */
	public BlockHeader createFromAttributes(Attributes attrs) {
		BlockHeader retval = new BlockHeader();
		retval.E = PayloadType.getValueFromDescription(attrs.getValue("payloadType"));
		retval.BT = BlockType.getValueFromDescription(attrs.getValue("blockType"));
		retval.setTimeStampOffset ( Integer.parseInt(attrs.getValue ( "timeStampOffset") ));
		retval.setBlockLength(Integer.parseInt(attrs.getValue ( "blockLength")));
		return retval;
	}
	
	
	/**
	 * Convert to an Rtf string (for documents).
	 */

	public String toRtfString() {
		StringBuffer sb = new StringBuffer();
		sb.append(HEADER_START).append("Block Header: ").append(HEADER_END);
		sb.append(HEADER_START).append("\tPayload Type (E)= " + getPayloadType() + " ( " + getE() + " )").append(HEADER_END);
		sb.append(HEADER_START).append("\tBlock Type (BT) = " + getBlockType() + " ( " + getBT() +" )").append(HEADER_END);
		sb.append(HEADER_START).append("\tTime Stamp Offset (TSO) = " + getTimeStampOffset() ).append(HEADER_END);
		sb.append(HEADER_START).append("\tBlock Length (BL)= " + getBlockLength()).append(HEADER_END);
		return sb.toString();
	}

}
