/* This software was developed by employees of the National Institute of
 * Standards and Technology (NIST), an agency of the Federal Government.
 * Pursuant to title 15 United States Code Section 105, works of NIST
 * employees are not subject to copyright protection in the United States
 * and are considered to be in the public domain.  As a result, a formal
 * license is not needed to use the software.
 * 
 * The ISSI Emulator software is provided by NIST as a service and is expressly
 * provided "AS IS".  NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED
 * OR STATUTORY, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT
 * AND DATA ACCURACY.  NIST does not warrant or make any representations
 * regarding the use of the software or the results thereof including, but
 * not limited to, the correctness, accuracy, reliability or usefulness of
 * the software.
 * 
 * Permission to use this software is contingent upon your acceptance
 * of the terms of this agreement
 * .
 */
package gov.nist.p25.issi.p25payload;

import java.io.Serializable;

import gov.nist.p25.issi.utils.ByteUtil;
import gov.nist.p25.issi.utils.RtfConstants;
import org.apache.log4j.Logger;
import org.xml.sax.Attributes;

/**
 * This class implements a P25 control octet. The P25 control octet has the
 * following format:
 * <p>
 * 
 * <pre>
 * 0 1 2 3 4 5 6 7 8
 * +-+-+-+-+-+-+-+-+
 * |S|C|   BHC     |
 * +-+-+-+-+-+-+-+-+
 * </pre>
 * 
 * @author steveq@nist.gov
 * @version $Revision: 1.16 $, $Date: 2007/07/24 03:20:38 $
 * @since 1.5
 */
public class ControlOctet  implements Serializable,RtfConstants {

	/***************************************************************************
	 * Constants
	 **************************************************************************/

	/** Constant that identifies the total byte length of fixed fields. */
	protected final static int FIXED_LENGTH = 1;

	/***************************************************************************
	 * Variables:  Note that each of the following use data types that are 
	 * larger than what is required to keep two's complement sign bit from 
	 * changing the sign.
	 **************************************************************************/

	/** Signal bit (1 bit). */
	private int S = 0;

	/** Indicates a compact (1) or verbose (0) block header (1 bit). */
	private int C = 0;

	/** The number of block headers following this control octet (6 bits). */
	private int BHC = 0;

	/** The logger for this class. */
	private static Logger logger = Logger.getLogger(ControlOctet.class);

	/***************************************************************************
	 * Constructors
	 **************************************************************************/

	/**
	 * Construct a P25 control octet.
	 */
	public ControlOctet() {

	}
	
	/**
	 * Construct a P25 block control octet given its byte array representation.
	 * 
	 * @param bytes The byte array representation of this control octet.
	 * @throws P25BlockException
	 */
	public ControlOctet(byte b) {

		this.S = (int) ((b >>> 0x07) & 0x01);	// Extract S
		this.C = (int) ((b >>> 0x06) & 0x01);	// Extract C
		this.BHC = (int) (b & 0x3F);			// Extract BHC

	}

	/***************************************************************************
	 * Methods
	 **************************************************************************/

	/**
	 * Get the size of this control octet in bytes.
	 * 
	 * @return The size of this control octet in bytes.
	 */
	public int size() {

		return FIXED_LENGTH;

	}

	/**
	 * Get the control octet as a byte array.
	 * 
	 * @return The control octet byte array.
	 */
	public byte[] getBytes() {

		byte b = 0;
		b |= S; 	// Set S
		b <<= 0x01; // Make room for C
		b |= C; 	// Set C
		b <<= 0x06; // Make room for BHC
		b |= BHC; 	// Set BHC

		byte[] bytes = new byte[1];
		bytes[0] = b;

		return bytes;

	}

	/**
	 * Set the signal bit (1=on, 0=off).
	 * 
	 * @param i
	 *            The signal bit (1 bit)
	 * @throws IllegalArgumentException
	 */
	public void setS(int i) throws IllegalArgumentException {

		if ((0 <= i) && (i <= ByteUtil.getMaxIntValueForNumBits(1)))
			S = i;
		else
			throw new IllegalArgumentException(P25BlockException.OUT_OF_RANGE);

	}
	
	/**
	 * Get the signal bit.
	 * 
	 * @return The signal bit (1=on, 0=off).
	 */
	int getS() {

		return S;

	}
	
	/**
	 * Set the compact bit (1=compact, 0=verbose).
	 * 
	 * @param i
	 *            The compact bit (1 bit)
	 * @throws IllegalArgumentException
	 */
	public void setC(int i) throws IllegalArgumentException {

		if ((0 <= i) && (i <= ByteUtil.getMaxIntValueForNumBits(1)))
			C = i;
		else
			throw new IllegalArgumentException(P25BlockException.OUT_OF_RANGE);

	}

	/**
	 * Get the compact bit.
	 * 
	 * @return The compact bit (1=compact, 0=verbose).
	 */
	int getC() {

		return C;

	}

	/**
	 * Set the number of block headers following this control octet.
	 * 
	 * @param i
	 *            The number of block headers following this control octet (6 bits).
	 * @throws IllegalArgumentException
	 */
	void setBHC(int i) throws IllegalArgumentException {

		if ((0 <= i) && (i <= ByteUtil.getMaxIntValueForNumBits(6)))
			BHC = i;
		else
			throw new IllegalArgumentException(P25BlockException.OUT_OF_RANGE);

	}

	/**
	 * Get the number of block headers following this control octet.
	 * 
	 * @return The number of block headers following this control octet.
	 */
	int getBHC() {

		return BHC;

	}
	
	/**
	 * Compare this control octet with another object.
	 * 
	 * @param other The other object.
	 * @return True if this control octet is equivalent to the other object.
	 */
	@Override
	public boolean equals(Object other) {
		if (!(other instanceof ControlOctet))
			return false;
		ControlOctet that = (ControlOctet) other;
		return this.C == that.C && this.BHC == that.BHC && this.S == that.S;

	}
	
	/**
	 * Match the significant fileds of the control octet with another control octet.
	 * This is for conformance testing of PTT.
	 */
	public boolean match(ControlOctet other) {
		ControlOctet that = (ControlOctet) other;
		return this.C == that.C && this.S == that.S;
	}

	/**
	 * Get the XML formatted string representation.
	 * 
	 * @return The XML formatted string representation.
	 */
	@Override
	public String toString() {
		return new StringBuffer().append("<control-octet").append(
				"\nsignalBit = \"" + getS() + "\"").append(
				"\ncompactBit = \"" + getC() + "\"").append(
				"\nblockHeaderCount = \"" + getBHC() + "\"").append("\n/>")
				.toString();
	}
	
	/**
	 * Construct this field from a set of attributes.
	 * @param b
	 */
	
	public ControlOctet createFromAttributes(Attributes attrs) {
		
		ControlOctet retval = new ControlOctet();
		retval.S = Integer.parseInt(attrs.getValue("signalBit"));
		retval.C = Integer.parseInt(attrs.getValue("compactBit"));
		retval.BHC = Integer.parseInt(attrs.getValue("blockHeaderCount"));
		return retval;
		
	}
	
	/**
	 * Set the signal bit.
	 */

	public void setSignal(boolean b) {
		this.setS(b ? 1 : 0);
		
	}

	public String toRtfString() {
		StringBuffer sb = new StringBuffer();
		sb.append(HEADER_START).append("Control Octet : ").append(HEADER_END);
		sb.append(HEADER_START).append("\tSignal Bit = " + getS() ).append(HEADER_END);
		sb.append(HEADER_START).append("\tCompact Bit = " + getC() ).append(HEADER_END);
		sb.append(HEADER_START).append("\tBlock Header Count = " + getBHC()).append(HEADER_END);
		return sb.toString();
		
	}

}
