/* This software was developed by employees of the National Institute of
 * Standards and Technology (NIST), an agency of the Federal Government.
 * Pursuant to title 15 United States Code Section 105, works of NIST
 * employees are not subject to copyright protection in the United States
 * and are considered to be in the public domain.  As a result, a formal
 * license is not needed to use the software.
 * 
 * The ISSI Emulator software is provided by NIST as a service and is expressly
 * provided "AS IS".  NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED
 * OR STATUTORY, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT
 * AND DATA ACCURACY.  NIST does not warrant or make any representations
 * regarding the use of the software or the results thereof including, but
 * not limited to, the correctness, accuracy, reliability or usefulness of
 * the software.
 * 
 * Permission to use this software is contingent upon your acceptance
 * of the terms of this agreement
 * .
 */
package gov.nist.p25.issi.p25payload;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;

import org.apache.log4j.Logger;

/**
 * This class handles IMBE voice data generation.
 * 
 * @author steveq@nist.gov
 * @version $Revision: 1.9 $, $Date: 2007/12/20 04:40:07 $
 * @since 1.5
 */
public class IMBEVoiceDataGenerator {
	
	private static Logger logger = Logger.getLogger(IMBEVoiceDataGenerator.class);

	/***************************************************************************
	 * Constants
	 **************************************************************************/

	/** Constant that identifies the total byte length of fixed fields. */
	final static int FIXED_LENGTH = 4;

	/***************************************************************************
	 * Methods
	 **************************************************************************/

	/**
	 * This method opens an IMBE file, parses the file into an array list of
	 * IMBEVoiceBlock objects, and returns the array list.
	 * 
	 * @param filePath
	 *            The path to the IMBE file.
	 * @param numBlocks
	 *            The number of blocks to read from the IMBE file.
	 * @return An array list of IMBE voice blocks.
	 */
	public static ArrayList<IMBEVoiceBlock> getIMBEVoiceData(String filePath,
			int numBlocks) {

		return readIMBEVoiceData(filePath, numBlocks);

	}

	/**
	 * This method opens an IMBE file, parses the file into an array list of
	 * IMBEVoiceBlock objects, and returns the array list.
	 * 
	 * @param filePath
	 *            The path to the IMBE file.
	 * @return An array list of IMBE voice blocks.
	 */
	public static ArrayList<IMBEVoiceBlock> getIMBEVoiceData(String filePath) {

		return readIMBEVoiceData(filePath, -1);

	}

	private static ArrayList<IMBEVoiceBlock> readIMBEVoiceData(String filePath,
			int numBlocks) {

		ArrayList<IMBEVoiceBlock> imbeVoices = new ArrayList<IMBEVoiceBlock>();

		// Open media file
		BufferedReader inputStream = null;

		try {

			/*
			 * TODO (steveq): Hardcode this until we decide where voice files
			 * will go and how they will be retrieved.
			 */
			inputStream = new BufferedReader(new FileReader(
					"voice/imbe-hex-test.txt"));

			String l;
			int i = 0;
			int imbePacketIndex = 0;
			int superFrameIndex = 0; // 18 IMBE packets

			while ((l = inputStream.readLine()) != null) {

				i++;
				imbePacketIndex++;

				IMBEVoiceBlock imbeVoice = new IMBEVoiceBlock();

				// Get first 6 bytes in hex string format
				String firstHexSubStr = l.substring(0, 12);
				// Get next 5 bytes in hex string format (11 bytes total)
				String secondHexSubStr = l.substring(12, 22);

				// The following was verified by Kameron
				int frameType = IMBEVoiceDataGenerator
						.getFrameType(imbePacketIndex);
				imbeVoice.setFT(frameType);

				long U0_U3 = Long.parseLong(firstHexSubStr, 16);
				long U4_U7 = Long.parseLong(secondHexSubStr, 16);

				imbeVoice.setU0((int) (U0_U3 >>> 0x24) & 0xFFF);
				imbeVoice.setU1((int) (U0_U3 >>> 0x18) & 0xFFF);
				imbeVoice.setU2((int) (U0_U3 >>> 0x0C) & 0xFFF);
				imbeVoice.setU3((int) U0_U3 & 0xFFF);
				imbeVoice.setU4((int) (U4_U7 >>> 0x1D) & 0x7FF);
				imbeVoice.setU5((int) (U4_U7 >>> 0x12) & 0x7FF);
				imbeVoice.setU6((int) (U4_U7 >>> 0x07) & 0x7FF);
				imbeVoice.setU7((int) U4_U7 & 0x7F);

				imbeVoice.setEt(0);
				imbeVoice.setEr(0);
				imbeVoice.setM(0);
				imbeVoice.setL(0);
				imbeVoice.setE4(0);
				imbeVoice.setE1(0);
				imbeVoice.setSF(superFrameIndex);
				imbeVoice.setRsvd(0);

				if (frameType == IMBEVoiceBlock.IMBEFRAME9
						|| frameType == IMBEVoiceBlock.IMBEFRAME18) {

					byte[] additionalFrameData = new byte[3];
					imbeVoice.setAdditionalFrameData(additionalFrameData);

				} else if (frameType == IMBEVoiceBlock.IMBEFRAME12
						|| frameType == IMBEVoiceBlock.IMBEFRAME13
						|| frameType == IMBEVoiceBlock.IMBEFRAME14) {

					byte[] additionalFrameData = new byte[4];
					imbeVoice.setAdditionalFrameData(additionalFrameData);

				} else if (frameType == IMBEVoiceBlock.IMBEFRAME15) {

					byte[] additionalFrameData = new byte[4];
					additionalFrameData[0] = (byte) 0x80;
					imbeVoice.setAdditionalFrameData(additionalFrameData);

				} else if (frameType == IMBEVoiceBlock.IMBEFRAME16) {

					byte[] additionalFrameData = new byte[4];
					additionalFrameData[0] = (byte) 0xac;
					additionalFrameData[1] = (byte) 0xb8;
					additionalFrameData[2] = (byte) 0xa4;
					imbeVoice.setAdditionalFrameData(additionalFrameData);

				} else if (frameType == IMBEVoiceBlock.IMBEFRAME17) {

					byte[] additionalFrameData = new byte[4];
					additionalFrameData[0] = (byte) 0x9b;
					additionalFrameData[1] = (byte) 0xdc;
					additionalFrameData[2] = 0x75;
					imbeVoice.setAdditionalFrameData(additionalFrameData);

				}

				// Add to IMBE voice array
				imbeVoices.add(imbeVoice);

				if (imbePacketIndex == 18) {

					imbePacketIndex = 0;

					if (superFrameIndex < 3)
						superFrameIndex++;
					else
						superFrameIndex = 0;

				}

				// We break if i = numBlocks. Note that if i < 0, we
				// only break if we reach EOF.
				if (i == numBlocks)
					break;

			}

			if (inputStream != null) {

				inputStream.close();

			}

		} catch (IOException ioe) {

			ioe.printStackTrace();

		}

		return imbeVoices;

	}

	/**
	 * STEVEQ: Increment the frame type on each outgoing packet, and loop after
	 * each set of 18. Frame types are defined in IMBEVoiceBlock.
	 * 
	 * @param i
	 *            The index of the IMBE voice block.
	 * @return
	 */
	public static int getFrameType(int i) {

		int frameTypeValue = 0;

		switch (i) {
		case 1:
			frameTypeValue = IMBEVoiceBlock.IMBE1;
			break;
		case 2:
			frameTypeValue = IMBEVoiceBlock.IMBE2;
			break;
		case 3:
			frameTypeValue = IMBEVoiceBlock.IMBE3;
			break;
		case 4:
			frameTypeValue = IMBEVoiceBlock.IMBE4;
			break;
		case 5:
			frameTypeValue = IMBEVoiceBlock.IMBE5;
			break;
		case 6:
			frameTypeValue = IMBEVoiceBlock.IMBE6;
			break;
		case 7:
			frameTypeValue = IMBEVoiceBlock.IMBE7;
			break;
		case 8:
			frameTypeValue = IMBEVoiceBlock.IMBE8;
			break;
		case 9:
			frameTypeValue = IMBEVoiceBlock.IMBEFRAME9;
			break;
		case 10:
			frameTypeValue = IMBEVoiceBlock.IMBE10;
			break;
		case 11:
			frameTypeValue = IMBEVoiceBlock.IMBE11;
			break;
		case 12:
			frameTypeValue = IMBEVoiceBlock.IMBEFRAME12;
			break;
		case 13:
			frameTypeValue = IMBEVoiceBlock.IMBEFRAME13;
			break;
		case 14:
			frameTypeValue = IMBEVoiceBlock.IMBEFRAME14;
			break;
		case 15:
			frameTypeValue = IMBEVoiceBlock.IMBEFRAME15;
			break;
		case 16:
			frameTypeValue = IMBEVoiceBlock.IMBEFRAME16;
			break;
		case 17:
			frameTypeValue = IMBEVoiceBlock.IMBEFRAME17;
			break;
		case 18:
			frameTypeValue = IMBEVoiceBlock.IMBEFRAME18;
			break;
		default:
			logger.error("ERROR: Invalid Frame Type!");
			break;
		}

		return frameTypeValue;

	}

}
