/* This software was developed by employees of the National Institute of
 * Standards and Technology (NIST), an agency of the Federal Government.
 * Pursuant to title 15 United States Code Section 105, works of NIST
 * employees are not subject to copyright protection in the United States
 * and are considered to be in the public domain.  As a result, a formal
 * license is not needed to use the software.
 * 
 * The ISSI Emulator software is provided by NIST as a service and is expressly
 * provided "AS IS".  NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED
 * OR STATUTORY, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT
 * AND DATA ACCURACY.  NIST does not warrant or make any representations
 * regarding the use of the software or the results thereof including, but
 * not limited to, the correctness, accuracy, reliability or usefulness of
 * the software.
 * 
 * Permission to use this software is contingent upon your acceptance
 * of the terms of this agreement
 * .
 */
package gov.nist.p25.issi.p25payload;

import java.io.Serializable;

import org.apache.log4j.Logger;
import org.xml.sax.Attributes;

import gov.nist.p25.issi.utils.ByteUtil;
import gov.nist.p25.issi.utils.RtfConstants;

/**
 * This class implements a P25 PTT Control Word block. The P25 Control Word
 * block has the following format:
 * <p>
 * 
 * <pre>
 *     0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
 *     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *     |                  WACN ID              |       System ID       |
 *     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *     |            Unit ID                            |      TP       |
 *     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 * </pre>
 * 
 * @author steveq@nist.gov
 * @author mranga@nist.gov
 * @version $Revision: 1.15 $, $Date: 2007/07/24 03:20:38 $
 * @since 1.5
 */
public class PTTControlWord 
implements Serializable, RtfConstants {

	/***************************************************************************
	 * Constants
	 **************************************************************************/

	/** Constant that identifies the total byte length of fixed fields. */
	protected final static int FIXED_LENGTH = 8;

	/***************************************************************************
	 * Variables: Note that each of the following use data types that are larger
	 * than what is required to keep two's complement sign bit from changing the
	 * sign.
	 **************************************************************************/

	// NOTE: The SU ID is a fully qualified identifier that includes
	// the WACN, System ID, and Unit ID (56 bits).
	/** The Wide Area Communication Network (WACN) ID (20 bits). */
	private int wacnId = 0;

	/** The system ID (12 bits). */
	private int systemId = 0;

	/** The unit ID (24 bits). */
	private int unitId = 0;

	/**
	 * The transmit priority (8 bits). Note that the first four bits in TP is
	 * the priority type (PT) and the last 4 bits are the priority level (PL).
	 */
	private int TP = 0;

	/** The logger for this class. */
	private static Logger logger = Logger.getLogger(PTTControlWord.class);

	/***************************************************************************
	 * Constructors
	 **************************************************************************/

	/**
	 * Construct a PTT control word.
	 */
	public PTTControlWord() {
	}

	/**
	 * Construct a PTT Control Word given its byte array representation.
	 * 
	 * @param bytes
	 *            The byte array representation of this PTT Control Word.
	 * @throws P25BlockException
	 */
	public PTTControlWord(byte[] bytes) {

		long l = ByteUtil.bytesToLong(bytes);

		this.wacnId = (int) ((l >>> 0x2C) & 0x0FFFFF);
		this.systemId = (int) ((l >>> 0x20) & 0xFFF);
		this.unitId = (int) ((l >>> 0x08) & 0xFFFFFF);
		this.TP = (int) (l & 0xFF);

	}

	/***************************************************************************
	 * Methods
	 **************************************************************************/

	/**
	 * Get the size of this block header in bytes.
	 * 
	 * @return the size of this block header in bytes.
	 */
	public int size() {

		return FIXED_LENGTH;

	}

	/**
	 * Get all PTT control word data.
	 * 
	 * @return all PTT control word data.
	 */
	public byte[] getBytes() {

		long l = 0;
		l |= (wacnId & 0xFFFFF);
		l <<= 0x0C;
		l |= (systemId & 0xFFF);
		l <<= 0x18;
		l |= (unitId & 0xFFFFFF);
		l <<= 0x08;
		l |= (TP & 0xFF);

		byte[] bytes = ByteUtil.longToBytes(l);

		return bytes;
	}

	/**
	 * Set the Wide Area Communication Network ID.
	 * 
	 * @param wacnId
	 *            The Wide Area Communication Network ID (20 bits)
	 * @throws IllegalArgumentException
	 */
	public void setWacnId(int wacnId) throws IllegalArgumentException {

		if ((0 <= wacnId) && (wacnId <= ByteUtil.getMaxIntValueForNumBits(20)))
			this.wacnId = wacnId;
		else
			throw new IllegalArgumentException(P25BlockException.OUT_OF_RANGE);

	}

	/**
	 * Get the Wide Area Communication Network ID.
	 * 
	 * @return The Wide Area Communication Network ID.
	 */
	public int getWacnId() {

		return wacnId;

	}

	/**
	 * Set the system ID.
	 * 
	 * @param systemId
	 *            The system ID (12 bits)
	 * @throws IllegalArgumentException
	 */
	public void setSystemId(int systemId) throws IllegalArgumentException {

		if ((0 <= systemId)
				&& (systemId <= ByteUtil.getMaxIntValueForNumBits(12)))
			this.systemId = systemId;
		else
			throw new IllegalArgumentException(P25BlockException.OUT_OF_RANGE);
	}

	/**
	 * Get the system ID.
	 * 
	 * @return The system ID.
	 */
	public int getSystemId() {

		return systemId;

	}

	/**
	 * Set the unit ID.
	 * 
	 * @param i
	 *            The unit ID (24 bits)
	 * @throws IllegalArgumentException
	 */
	public void setUnitId(int i) throws IllegalArgumentException {

		if ((0 <= i) && (i <= ByteUtil.getMaxIntValueForNumBits(24)))
			this.unitId = i;
		else
			throw new IllegalArgumentException(P25BlockException.OUT_OF_RANGE);

	}

	/**
	 * Get the unit ID.
	 * 
	 * @return The unit ID.
	 */
	public int getUnitId() {

		return unitId;
	}

	/**
	 * Set the transmit priority.
	 * 
	 * @param i
	 *            The transmit priority (8 bits)
	 * @throws IllegalArgumentException
	 */
	void setTP(int i) throws IllegalArgumentException {

		if ((0 <= i) && (i < ByteUtil.getMaxIntValueForNumBits(8)))
			TP = i;
		else
			throw new IllegalArgumentException(P25BlockException.OUT_OF_RANGE
					+ "Transmit Priority can only have the values 0..3");

	}

	/**
	 * Get the transmit priority.
	 * 
	 * @return The transmit priority.
	 */
	int getTP() {

		return TP;

	}

	/**
	 * Set the priority type and priority level.
	 * 
	 * @param pt
	 *            The priority type
	 * @param pl
	 *            The priority level
	 */
	void setPTAndPL(int pt, int pl) {

		if (((0 <= pt) && (pt < TransmitPriorityType.size()))
				&& ((0 <= pt) && (pt <= ByteUtil.getMaxIntValueForNumBits(4)))) {

			TP = (int) (pt & 0x0F);
			TP <<= 0x04;
			TP |= (pl & 0x0F);

		} else {

			throw new IllegalArgumentException(P25BlockException.OUT_OF_RANGE);

		}

	}

	/**
	 * Get the priority type.
	 * 
	 * @return The priority type.
	 */
	int getPT() {

		return (int) ((TP >>> 0x04) & 0x0F);

	}

	/**
	 * Get the priority type.
	 * 
	 * @return The priority type.
	 */
	public int getPriorityType() {
		return getPT();
	}

	/**
	 * Get the priority level.
	 * 
	 * @return The priority level.
	 */
	int getPL() {

		return (int) (TP & 0x0F);

	}

	/**
	 * Get the priority level.
	 * 
	 * @return The priority level.
	 */
	public int getPriorityLevel() {
		return getPL();
	}

	/**
	 * Set the transmit priority and priority level.
	 * 
	 * @param transmitPriority
	 *            The transmit priority.
	 * @param priorityLevel
	 *            The priority level.
	 */
	public void setPriorityTypeAndPriorityLevel(TransmitPriorityType tpType,
			int priorityLevel) {

		int pt = tpType.intValue();
		setPTAndPL(pt, priorityLevel);

	}

	/**
	 * Get transmit priority.
	 * 
	 * @return The transmit priority.
	 */
	public TransmitPriorityType getTransmitPriority() {

		return TransmitPriorityType.getInstance(getPT());

	}

	/**
	 * Compare this object with another object.
	 * 
	 * @param other
	 *            The other object to be compared to.
	 * @return True if this object is equivalent to the other object.
	 */
	@Override
	public boolean equals(Object other) {
		PTTControlWord that = (PTTControlWord) other;
		return this.systemId == that.systemId && this.TP == that.TP
				&& this.unitId == that.unitId && this.wacnId == that.wacnId;
	}

	/**
	 * Get the XML formatted string representation.
	 * 
	 * @return The XML formatted string representation.
	 */
	public String toString() {
		return new StringBuffer().append("<ptt-control-word").append(
				"\nwacnId = \"" + getWacnId() + "\"").append(
				"\nsystemId = \"" + getSystemId() + "\"")
				.append(
						"\ntransmitPriorityLevel = \""
								+ this.getPriorityLevel() + "\"").append(
						"\ntransmitPriority = \"" + 
						this.getTransmitPriority()
								+ "\"").append(
						"\nunitId = \"" + getUnitId() + "\"").append("\n/>")
				.toString();
	}
	
	/**
	 * Populate this field from a set of attributes.
	 * 
	 * @return the populated structure ( constructed from attributes)
	 */
	public static PTTControlWord createFromAttributes(Attributes attrs) {
		PTTControlWord retval  = new PTTControlWord();
		retval.wacnId = Integer.parseInt(attrs.getValue("wacnId"));
		retval.systemId = Integer.parseInt(attrs.getValue("systemId"));
		int pl = Integer.parseInt(attrs.getValue("transmitPriorityLevel"));
		int pt = TransmitPriorityType.valueFromString(attrs.getValue("transmitPriority"));
		retval.setPTAndPL(pt, pl);
		retval.setUnitId(Integer.parseInt(attrs.getValue("unitId")));
		return retval;
		
		
		
	}
	
	/**
	 * Format as an RTF string.
	 */

	public String toRtfString() {
		StringBuffer sb = new StringBuffer();
		sb.append(HEADER_START).append("Ptt Control Word : " ).append(HEADER_END);
		sb.append(HEADER_START).append("\tWacnID = " + getWacnId() ).append(HEADER_END);
		sb.append(HEADER_START).append("\tSystemID = " + getSystemId() ).append(HEADER_END);
		sb.append(HEADER_START).append("\tTransmit Priority Level = " + this.getPriorityLevel() ).append(HEADER_END);
		sb.append(HEADER_START).append("\tTransmit Priority = " + this.getTransmitPriority()).append(HEADER_END);
		sb.append(HEADER_START).append("\tUnit ID = " + getUnitId()).append(HEADER_END);
		return sb.toString();
	}
}
