/* This software was developed by employees of the National Institute of
 * Standards and Technology (NIST), an agency of the Federal Government.
 * Pursuant to title 15 United States Code Section 105, works of NIST
 * employees are not subject to copyright protection in the United States
 * and are considered to be in the public domain.  As a result, a formal
 * license is not needed to use the software.
 * 
 * The ISSI Emulator software is provided by NIST as a service and is expressly
 * provided "AS IS".  NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED
 * OR STATUTORY, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT
 * AND DATA ACCURACY.  NIST does not warrant or make any representations
 * regarding the use of the software or the results thereof including, but
 * not limited to, the correctness, accuracy, reliability or usefulness of
 * the software.
 * 
 * Permission to use this software is contingent upon your acceptance
 * of the terms of this agreement
 * .
 */
package gov.nist.p25.issi.p25payload;

/**
 * Block header payload type. Each block header SHALL contain a bit (E) to
 * identify whether the payload type is a standard IANA type (E set to 0) or a
 * profile specific type (E set to 1). For voice services over the ISSI, the E
 * bit SHALL always be set to 1 (profile specific type).
 * 
 * @author mranga@nist.gov
 * @version $Revision: 1.5 $, $Date: 2007/07/24 03:20:38 $
 * @since 1.5
 */
public enum PayloadType {

	IANA_TYPE(0, "IANA Type Payload"), PROFILE_SPECIFIC_TYPE(1,
			"Profile Specific Type Payload");

	/** The payload type as an int. */
	private int type;

	/** A description of the payload type. */
	private String description;
	
	
	/**
	 * Get the appropriate type from a description.
	 * 
	 */
	public static int getValueFromDescription(String description) {
		if ( description.equals(IANA_TYPE.description)) return IANA_TYPE.type;
		else if (description.equals(PROFILE_SPECIFIC_TYPE.description)) return PROFILE_SPECIFIC_TYPE.type;
		else throw new IllegalArgumentException("Illegal description " + description);
	}

	/**
	 * Construct a payload type.
	 * 
	 * @param type
	 *            the payload type in a byte representation
	 * @param description
	 *            a description of the payload type
	 */
	PayloadType(int type, String description) {

		this.type = type;
		this.description = description;

	}

	/**
	 * Get this payload type as an int.
	 * 
	 * @return this payload type as an int.
	 */
	public int intValue() {

		return type;

	}

	/**
	 * Get payload type as a PayloadType object.
	 * 
	 * @param pt
	 *            the payload type as a byte.
	 * @return the payload type as a PayloadTypes object.
	 */
	public static PayloadType getInstance(int pt) {

		if (pt != 0 && pt != 1)
			throw new IllegalArgumentException("bad arg");

		return pt == IANA_TYPE.type ? IANA_TYPE : PROFILE_SPECIFIC_TYPE;

	}

	/**
	 * Get the description of this payload type.
	 * 
	 * @return the description of this payload type.
	 */
	public String toString() {

		return description;

	}
}
