/* This software was developed by employees of the National Institute of
 * Standards and Technology (NIST), an agency of the Federal Government.
 * Pursuant to title 15 United States Code Section 105, works of NIST
 * employees are not subject to copyright protection in the United States
 * and are considered to be in the public domain.  As a result, a formal
 * license is not needed to use the software.
 * 
 * The ISSI Emulator software is provided by NIST as a service and is expressly
 * provided "AS IS".  NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED
 * OR STATUTORY, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT
 * AND DATA ACCURACY.  NIST does not warrant or make any representations
 * regarding the use of the software or the results thereof including, but
 * not limited to, the correctness, accuracy, reliability or usefulness of
 * the software.
 * 
 * Permission to use this software is contingent upon your acceptance
 * of the terms of this agreement
 * .
 */
package gov.nist.p25.issi.p25payload;

import gov.nist.p25.issi.utils.ByteUtil;
import java.util.ArrayList;
import org.apache.log4j.ConsoleAppender;
import org.apache.log4j.Logger;
import org.apache.log4j.PropertyConfigurator;
import org.apache.log4j.SimpleLayout;
import junit.framework.TestCase;

/**
 * This class tests IMBE Voice block data generator.
 * 
 * @author steveq@nist.gov
 * @version $Revision: 1.2 $, $Date: 2007/05/07 03:23:21 $
 * @since 1.5
 */
public class Test_IMBEVoiceDataGeneratorTest extends TestCase {

	private static Logger logger = Logger
			.getLogger("gov.nist.p25.issi.p25payload");
	static {
		PropertyConfigurator.configure("log4j.properties");
		logger.addAppender(new ConsoleAppender(new SimpleLayout()));
	}

	/***************************************************************************
	 * Methods
	 **************************************************************************/

	/**
	 * Test IMBE voice block fields.  Note that we do not test IMBE Voice
	 * Block packing and unpacking; that should exist as a separate test.
	 * This tests only getIMBEVoiceData(String filePath).
	 */
	public void testIMBEDataGenerator() {

		ArrayList<IMBEVoiceBlock> imbeVoiceBlocks = IMBEVoiceDataGenerator
				.getIMBEVoiceData("voice/imbe-hex-test.txt");

		int superFrameIndex = 0;
		int imbePacketIndex = 0;
		
		for (IMBEVoiceBlock imbeVoice : imbeVoiceBlocks) {

			imbePacketIndex++;
			
			// Check if FT is in possible range (we don't check specific val)
			int frameType = imbeVoice.getFT();
			assertTrue((frameType >= 0x6A) && (frameType <= 0xCC));
			
			assertTrue(imbeVoice.getU0() <= 
				ByteUtil.getMaxIntValueForNumBits(12));
			assertTrue(imbeVoice.getU1() <= 
				ByteUtil.getMaxIntValueForNumBits(12));
			assertTrue(imbeVoice.getU2() <= 
				ByteUtil.getMaxIntValueForNumBits(12));
			assertTrue(imbeVoice.getU3() <= 
				ByteUtil.getMaxIntValueForNumBits(12));
			assertTrue(imbeVoice.getU4() <= 
				ByteUtil.getMaxIntValueForNumBits(11));
			assertTrue(imbeVoice.getU5() <= 
				ByteUtil.getMaxIntValueForNumBits(11));
			assertTrue(imbeVoice.getU6() <= 
				ByteUtil.getMaxIntValueForNumBits(11));
			assertTrue(imbeVoice.getU7() <= 
				ByteUtil.getMaxIntValueForNumBits(7));
			
			// Verify that these block types have 3 bytes of additional
			// frame data.
			if (frameType == IMBEVoiceBlock.IMBEFRAME9
					|| frameType == IMBEVoiceBlock.IMBEFRAME18) {
				
				assertTrue(imbeVoice.getAdditionalFrameData().length == 3);
				
			}
			
			if (frameType == IMBEVoiceBlock.IMBEFRAME12
				|| frameType == IMBEVoiceBlock.IMBEFRAME13
				|| frameType == IMBEVoiceBlock.IMBEFRAME14
				|| frameType == IMBEVoiceBlock.IMBEFRAME15
				|| frameType == IMBEVoiceBlock.IMBEFRAME16
				|| frameType == IMBEVoiceBlock.IMBEFRAME17) {
				
				assertTrue(imbeVoice.getAdditionalFrameData().length == 4);
				
			}
			
			// Check values for these (these values might change later)
			assertTrue(imbeVoice.getEt() == 0);
			assertTrue(imbeVoice.getEr() == 0);
			assertTrue(imbeVoice.getM() == 0);
			assertTrue(imbeVoice.getL() == 0);
			assertTrue(imbeVoice.getE4() == 0);
			assertTrue(imbeVoice.getE1() == 0);
			
			// Check superframe
			assertTrue(imbeVoice.getSF() == superFrameIndex);
			
			// Update superFrame
			if (imbePacketIndex == 18) {

				imbePacketIndex = 0;

				if (superFrameIndex < 3)
					superFrameIndex++;
				else
					superFrameIndex = 0;

			}
			
		}

	}
	
	/**
	 * Test IMBE voice block fields.  Note that we do not test IMBE Voice
	 * Block packing and unpacking; that should exist as a separate test.
	 * This tests only getIMBEVoiceData(String filePath, int numBlocks).
	 */
	public void testIMBEDataGenerator2() {

		int numBlocksToRead = 3;
		
		ArrayList<IMBEVoiceBlock> imbeVoiceBlocks = IMBEVoiceDataGenerator
				.getIMBEVoiceData("voice/imbe-hex-test.txt", numBlocksToRead);

		int superFrameIndex = 0;
		int imbePacketIndex = 0;
		
		assertTrue(imbeVoiceBlocks.size() == numBlocksToRead);
		
		for (IMBEVoiceBlock imbeVoice : imbeVoiceBlocks) {

			imbePacketIndex++;
			
			// Check if FT is in possible range (we don't check specific val)
			int frameType = imbeVoice.getFT();
			assertTrue((frameType >= 0x6A) && (frameType <= 0xCC));
			
			assertTrue(imbeVoice.getU0() <= 
				ByteUtil.getMaxIntValueForNumBits(12));
			assertTrue(imbeVoice.getU1() <= 
				ByteUtil.getMaxIntValueForNumBits(12));
			assertTrue(imbeVoice.getU2() <= 
				ByteUtil.getMaxIntValueForNumBits(12));
			assertTrue(imbeVoice.getU3() <= 
				ByteUtil.getMaxIntValueForNumBits(12));
			assertTrue(imbeVoice.getU4() <= 
				ByteUtil.getMaxIntValueForNumBits(11));
			assertTrue(imbeVoice.getU5() <= 
				ByteUtil.getMaxIntValueForNumBits(11));
			assertTrue(imbeVoice.getU6() <= 
				ByteUtil.getMaxIntValueForNumBits(11));
			assertTrue(imbeVoice.getU7() <= 
				ByteUtil.getMaxIntValueForNumBits(7));
			
			// Verify that these block types have 3 bytes of additional
			// frame data.
			if (frameType == IMBEVoiceBlock.IMBEFRAME9
					|| frameType == IMBEVoiceBlock.IMBEFRAME18) {
				
				assertTrue(imbeVoice.getAdditionalFrameData().length == 3);
				
			}
			
			if (frameType == IMBEVoiceBlock.IMBEFRAME12
				|| frameType == IMBEVoiceBlock.IMBEFRAME13
				|| frameType == IMBEVoiceBlock.IMBEFRAME14
				|| frameType == IMBEVoiceBlock.IMBEFRAME15
				|| frameType == IMBEVoiceBlock.IMBEFRAME16
				|| frameType == IMBEVoiceBlock.IMBEFRAME17) {
				
				assertTrue(imbeVoice.getAdditionalFrameData().length == 4);
				
			}
			
			// Check values for these (these values might change later)
			assertTrue(imbeVoice.getEt() == 0);
			assertTrue(imbeVoice.getEr() == 0);
			assertTrue(imbeVoice.getM() == 0);
			assertTrue(imbeVoice.getL() == 0);
			assertTrue(imbeVoice.getE4() == 0);
			assertTrue(imbeVoice.getE1() == 0);
			
			// Check superframe
			assertTrue(imbeVoice.getSF() == superFrameIndex);
			
			// Update superFrame
			if (imbePacketIndex == 18) {

				imbePacketIndex = 0;

				if (superFrameIndex < 3)
					superFrameIndex++;
				else
					superFrameIndex = 0;

			}
			
		}

	}

}
