/* This software was developed by employees of the National Institute of
 * Standards and Technology (NIST), an agency of the Federal Government.
 * Pursuant to title 15 United States Code Section 105, works of NIST
 * employees are not subject to copyright protection in the United States
 * and are considered to be in the public domain.  As a result, a formal
 * license is not needed to use the software.
 * 
 * The ISSI Emulator software is provided by NIST as a service and is expressly
 * provided "AS IS".  NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED
 * OR STATUTORY, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT
 * AND DATA ACCURACY.  NIST does not warrant or make any representations
 * regarding the use of the software or the results thereof including, but
 * not limited to, the correctness, accuracy, reliability or usefulness of
 * the software.
 * 
 * Permission to use this software is contingent upon your acceptance
 * of the terms of this agreement
 * .
 */
package gov.nist.p25.issi.p25payload;

/**
 * An enumeration of priority types. The priority type is set to decimal 2 for
 * preemptive priority transmissions, 1 for priority transmissions, and 0 for
 * all other transmissions.
 * 
 * @author mranga@nist.gov
 * @version $Revision: 1.6 $, $Date: 2007/07/24 03:20:38 $
 * @since 1.5
 */
public enum TransmitPriorityType {

	PREEMPTIVE_PRIORITY(2, "Preemptive Priority"), PRIORITY(1, "Priority"), NORMAL(
			0, "Normal");

	private static TransmitPriorityType[] tpTypes = { NORMAL, PRIORITY,
			PREEMPTIVE_PRIORITY

	};
	
	/** The payload type as an int. */
	private int type;

	/** A description of the payload type. */
	private String description;

	
	/**
	 * Construct a transmit priority type.
	 * 
	 * @param type
	 *            the transmit priority type as a byte.
	 * @param description
	 *            the description of this transmit priority type.
	 */
	TransmitPriorityType(int type, String description) {
		this.description = description;
		this.type = type;
	}

	/**
	 * Get the description of this transmit priority type.
	 * 
	 * @return the description of this transmit priority type.
	 */
	public String toString() {
	
		return description;
	}
	
	/**
	 * Get the type from the given descriptive string.
	 * 
	 * @param description
	 * @return
	 * @throws IllegalArgumentException
	 */
	public static int valueFromString(String description) throws IllegalArgumentException {
		for ( TransmitPriorityType tp : tpTypes) {
			if (description.equals(tp.description)) return tp.type;
		}
		throw new IllegalArgumentException("Illegal arg " + description);
	}

	/**
	 * Get this transmit priority type as a byte.
	 * 
	 * @return this transmit priority type as a byte.
	 */
	public int intValue() {

		return type;

	}

	/**
	 * Get this transmit priority type as a TransmitPriorityTypes object.
	 * 
	 * @param tpType
	 *            the transmit priority type as a byte.
	 * @return the transmit priority type as a TransmitPriorityTypes object.
	 */
	public static TransmitPriorityType getInstance(int tpType) {

		return tpTypes[tpType];

	}
	
	/* steveq:  Used by other classes to determine the number of elements
	 * in this enumeration.
	 */
	public static int size() {
		return tpTypes.length;
	}

}
